/**
 * $Revision: 1.67 $
 * $Date: 2006/02/02 09:58:36 $
 */

#include "../tinyxml/tinyxml.h"
#include "../MLPCommon/Logger.h"
#include "../MLPDataset/DataSetC45.h"
#include "../MLPDataset/AddNoise.h"
#include "../MLPDT_TD/TDIDTInducer.h"
#include "../MLPDT_TD/TDIDTRandInducer.h"
//#include "../MLPDT_TD_OB/TDIDTInducerOB.h"	
//#include "../MLPDT_EA_AP/EAIDTInducerAP.h"
//#include "../MLPDT_EA_OB/EAIDTInducerOB.h"
#include "../MLPClassification/ConstInducer.h"
#include "../MLPClassification/CVClassificationExperiment.h"
#include "../MLPClassification/TestClassificationExperiment.h"
#ifdef BUILD_MLPSVM 
#include "../MLPSVM/SVMMultiClassOAOInducer.h"
#include "../MLPSVM/SVMMultiClassOAAInducer.h"
#endif
#include "ClassificationTask.h"

#include <iostream>
#include <sstream>
#include <fstream>
#include <string>
#include <iomanip>
#include <cstdlib>
#include <ctime>

#define __MLPVERSION__ "1.2"

using namespace std;

/**
 * Wczytuje z pliku konfiguracyjnego informacje o zbiorach danych.
 * @param element - wskaznik do wezla MLPDatasets
 */
void ReadDatasets(TiXmlElement* element, CClassificationTask &task);

/**
 * Wczytuje z pliku konfiguracyjnego informacje o parametrach zbioru danych
 * (informacje o parametrach kroswalidacji, kosztach, etc.).
 * @param element - wskaznik do wezla MLPDataset
 */
void ReadDatasetParams(TiXmlElement* element, CDatasetDescription *dsDesc);

/** 
 * Wczytuje z pliku konfiguracyjnego informacje o zbiorach danych.
 * @param element - wskaznik do wezla MLPClassifiers
 */
void ReadClassifiers(TiXmlElement* element, CClassificationTask &task);

/**
 * Wczytuje z pliku konfiguracyjnego infrormacje na temat jednego klasyfikatora i
 * tworzy jego obiekt inducera do wykorzystania w eksperymetnach. Sprawdza poprawnosc
 * podanych w pliku konfiguracyjnym parametrow.
 * @param classifier - wskaznikd do wezla MLPClassifier
 */
IInducer* ReadClassifier(TiXmlElement* classifier, CClassifierDescription &desc);

/**
 * Drukuje pomoc do cout.
 */
void PrintHelp();

/**
 * Funkcja generuje przykladowy plik konfiguracyjny. Wynik drukowany jest do cout.
 */
void GenerateFile();

/**
 * Funkcja sprawdza czy okreslony eksperyment nie byl juz wykonany. Uwzglednia listy 
 * parametrow i numer uruchomienia.
 * @return true - kiedy eksperymetn byl wykonany
 */
bool IsCarriedOut(TiXmlElement* element, string logval, CClassifierDescription &desc, CDatasetDescription &ds, int run);

/**
 * Wczytanie atrybutu z elementu XML, ze sprawdzeniem poprawnosci.
 * @param element - tag XML
 * @param name - nazwa atrybutu do wczytania
 * @param domain - wektor napisow stanowiacych dziedzine wczytywanego atrybutu
 */
const char* ReadAttribute(TiXmlElement* element, char* name, vector<string> domain);

void ReadDatasets(TiXmlElement* element, CClassificationTask &task){
	vector<string> vstr;
	TiXmlElement* dataset = element->FirstChildElement();
	if(dataset == 0){
		CLogger::LogError("MLPDataset not found",__FILE__,__LINE__);
	}
	do{
		if(strcmp(dataset->Value(),"MLPDataset") != 0){
			CLogger::LogError("MLPDataset expected",__FILE__,__LINE__);
		}
		CDatasetDescription *dsDescr = new CDatasetDescription();
		int iDSID;
		int queryOkay = dataset->QueryIntAttribute("DSID",&iDSID);
		if(queryOkay == TIXML_SUCCESS){
			dsDescr->m_iDSID = iDSID;
		}else{
			CLogger::LogError("DSID not specified or is not integer",__FILE__,__LINE__);
		}
		vstr.clear();
		dsDescr->m_sName = ReadAttribute(dataset,"Name",vstr);
		dsDescr->m_sPath2Stem = ReadAttribute(dataset,"Path2Stem",vstr);
		const char* str = dataset->Attribute("Columns2Rem");
		vector<unsigned> cols2rem;
		if(str){
			char *tempstr = strdup(str);
			char *token = strtok(tempstr,",");
			while(token){
				cols2rem.push_back((unsigned)atoi(token));
				token = strtok(NULL,",");
			}
			delete tempstr;
		}
		dsDescr->m_vsCols2Rem = cols2rem;
		vstr.clear();
		vstr.push_back("C4.5");
		dsDescr->m_sDSFormat = ReadAttribute(dataset,"DSFormat",vstr);
		
		ReadDatasetParams(dataset,dsDescr);
		
		// To ladownie danych ponizej jest po to, zeby na poczatku sprawdzic
		// czy dane wczytuja sie poprawnie (mimo ze moze to dawac spory narzut)
		if(dsDescr->LoadData()){
			dsDescr->ReleaseData();
		}
		task.m_vDatasets.push_back(dsDescr);
		dataset=dataset->NextSiblingElement();
	}while(dataset);
}

void ReadDatasetParams(TiXmlElement* element, CDatasetDescription *dsDesc){
	TiXmlElement* dsParam = element->FirstChildElement();
	while(dsParam){
		if(strcmp(dsParam->Value(),"MLPExperimentType") == 0){
			vector<string> vstr;
			vstr.clear();
			vstr.push_back("test");
			vstr.push_back("crossvalidation");
			string str = ReadAttribute(dsParam,"Type",vstr);
			if(strcmp(str.c_str(),"test") == 0){
				dsDesc->m_iExperimentType = 1;
			}else{
				if(strcmp(str.c_str(),"crossvalidation") == 0){
					dsDesc->m_iExperimentType = 0;
					TiXmlElement* typeParam = dsParam->FirstChildElement();
					while(typeParam){
						if(strcmp(typeParam->Value(),"MLPExpParam") == 0){
							const char* param = typeParam->Attribute("Name");
							const char* value = typeParam->Attribute("Value");
							if(param!=0 && value!=0){
								bool found = false;
								if(strcmp(param,"Stratified") == 0){
									found = true;
									if(strcmp(value,"yes") == 0){
										dsDesc->m_bStratified = true;
									}else{
										dsDesc->m_bStratified = false;
									}
								}
								if(strcmp(param,"MakeSplit") == 0){
									found = true;
									if(strcmp(value,"yes") == 0){
										dsDesc->m_bmakeSplit = true;
									}else{
										dsDesc->m_bmakeSplit = false;
									}
								}
								if(strcmp(param,"Folds") == 0){
									found = true;
									int iFolds;
									int queryOkay = typeParam->QueryIntAttribute("Value",&iFolds);
									if(queryOkay == TIXML_SUCCESS){
										dsDesc->m_iFolds = iFolds;
									}else{
										CLogger::LogError("Value in Folds is not specified or is not integer",__FILE__,__LINE__);
									}
								}
								if(!found){
									CLogger::LogError("Unknown Name in MLPExpParam",__FILE__,__LINE__);
								}
							}else{
								CLogger::LogError("Name and Value expected in MLPExpParam",__FILE__,__LINE__);
							}
						}else{
							CLogger::LogError("MLPExpParam expected",__FILE__,__LINE__);
						}
						
						typeParam = typeParam->NextSiblingElement();
					}
					
				}
			}
		}else{
			if(strcmp(dsParam->Value(),"MLPDSParam") == 0){
				const char* param = dsParam->Attribute("Name");
				const char* value = dsParam->Attribute("Value");
				if(param!=0 && value!=0){
					if(strcmp(param,"costofattributes") == 0){
						if(strcmp(value,"yes") == 0){
							dsDesc->m_bCostOfAttrs = true;
						}else{
							dsDesc->m_bCostOfAttrs = false;
						}
					}
					if(strcmp(param,"costofclassification") == 0){
						if(strcmp(value,"yes") == 0){
							dsDesc->m_bCostOfClassification = true;
						}else{
							dsDesc->m_bCostOfClassification = false;
						}
					}
				}else{
					CLogger::LogError("Name and Value expected in MLPDSParam",__FILE__,__LINE__);
				}
			}else{
				CLogger::LogError("MLPExperimentType or MLPDSParam expected",__FILE__,__LINE__);
			}
		}
		
		dsParam = dsParam->NextSiblingElement();
	}
}

void ReadClassifiers(TiXmlElement* element, CClassificationTask &task){
	vector<string> vstr;
	TiXmlElement* classifier = element->FirstChildElement();
	if(classifier == 0){
		CLogger::LogError("MLPClassifier not found",__FILE__,__LINE__);
	}
	do{
		if(strcmp(classifier->Value(),"MLPClassifier") != 0){
			CLogger::LogError("MLPClassifier expected",__FILE__,__LINE__);
		}
		CClassifierDescription *clDescr = new CClassifierDescription();

		int iCLID;
		int queryOkay = classifier->QueryIntAttribute("CLID",&iCLID);
		if(queryOkay == TIXML_SUCCESS){
			clDescr->m_iCLID = iCLID;
		}else{
			CLogger::LogError("CLID not specified or is not integer",__FILE__,__LINE__);
		}
		vstr.clear();
		clDescr->m_sName = ReadAttribute(classifier,"Name",vstr);
		clDescr->m_pInducer = ReadClassifier(classifier,*clDescr);
		
		// Ustawienie od razu poziomu logowania.
		clDescr->m_pInducer->SetLogLevel(task.m_iLogLevel);
		
		task.m_vClassifiers.push_back(clDescr);
		classifier=classifier->NextSiblingElement();
	}while(classifier);
}

IInducer* ReadClassifier(TiXmlElement* classifier, CClassifierDescription &desc){
	IInducer *inducer = 0;
//	if(desc.m_sName == "MLPDT_EA_AP"){
//		inducer = new CEAIDTInducerAP();
//	}
//	if(desc.m_sName == "MLPDT_EA_OB"){
//		inducer = new CEAIDTInducerOB();
//	}
	if(desc.m_sName == "MAJORITYCLASS"){
		inducer = new CConstInducer();
	}
	if(desc.m_sName == "MLPDT_TD_AP"){
		inducer = new CTDIDTInducer();
	}
//	if(desc.m_sName == "MLPDT_TD_OB"){
//		inducer = new CTDIDTInducerOB();
//	}
	if(desc.m_sName == "MLPDT_TD_AP_RAND"){
		inducer = new CTDIDTRandInducer();
	}
#ifdef BUILD_MLPSVM 
	if(desc.m_sName == "MLPSVM_One-Against-All"){
		inducer = new CSVMMultiClassOAAInducer();
	}
	if(desc.m_sName == "MLPSVM_One-Against-One"){
		inducer = new CSVMMultiClassOAOInducer();
	}
#endif
	if(inducer == 0){
		stringstream ss;
		ss<<"Unknown classifier: "<<desc.m_sName;
		CLogger::LogError(ss.str().c_str(),__FILE__,__LINE__);
		return NULL;
	}else{
		TiXmlElement* MLPParamElement = classifier->FirstChildElement();
		while(MLPParamElement){
			if(strcmp(MLPParamElement->Value(),"MLPParamList") == 0){
				if(desc.m_bParam1 && desc.m_bParam2){
					CLogger::LogError("MLPParamList can't be specified more than twice","",__LINE__);
				}
				const char* param = MLPParamElement->Attribute("Name");
				const char* values = MLPParamElement->Attribute("Value");
				if(param!=0 && values!=0){
					string sParamName = param;
					vector<string> vsParamList;
					char *tempstr = strdup(values);
					char *token = strtok(tempstr,",");
					if(token==0){
						CLogger::LogError("MLPParamList values not specified",__FILE__,__LINE__);
					}
					while(token){
						if(!inducer->SetParam(param,token)){
							stringstream ss;
							ss<<param<<" = "<<token<<" not accepted by "<<desc.m_sName;
							CLogger::LogError(ss.str().c_str(),__FILE__,__LINE__);
						}
						vsParamList.push_back(token);
						token = strtok(NULL,",");
					}
					free(tempstr);
					inducer->SetParam(sParamName,vsParamList[0]);
					if(!desc.m_bParam1){
						desc.m_bParam1 = true;
						desc.m_sParamName1 = sParamName;
						desc.m_vsParamList1 = vsParamList;
					}else{
						if(!desc.m_bParam2){
							desc.m_bParam2 = true;
							desc.m_sParamName2 = sParamName;
							desc.m_vsParamList2 = vsParamList;
						}	
					}
				}else{
					CLogger::LogError("MLPParamList requires exactly two attributes (Name & Value)",__FILE__,__LINE__);
				}
			}else{
				if(strcmp(MLPParamElement->Value(),"MLPParam") == 0){
					const char *param = MLPParamElement->Attribute("Name");
					const char* value = MLPParamElement->Attribute("Value");
					if(param!=0 && value!=0){
						if(!inducer->SetParam(param,value)){
							stringstream ss;
							ss<<"["<<param<<" = "<<value<<"] not accepted by "<<desc.m_sName;
							CLogger::LogError(ss.str().c_str(),__FILE__,__LINE__);
						}
					}else{
						CLogger::LogError("MLPParam requires exactly two attributes (Name & Value)",__FILE__,__LINE__);
					}
					// Wczytanie tagow MLPSubParam
					TiXmlElement* MLPSubParamElement = MLPParamElement->FirstChildElement();
					while(MLPSubParamElement){
						if(strcmp(MLPSubParamElement->Value(),"MLPSubParam") == 0){
							const char* subparam = MLPSubParamElement->Attribute("Name");
							const char* subvalue = MLPSubParamElement->Attribute("Value");
							if(subparam!=0 && subvalue!=0){
								if(!inducer->SetSubParam(param,subparam,subvalue)){
									stringstream ss;
									ss<<"["<<subparam<<" = "<<subvalue<<"] not accepted by "<<desc.m_sName;
									CLogger::LogError(ss.str().c_str(),__FILE__,__LINE__);
								}
							}else{
								CLogger::LogError("MLPSubParam requires exactly two attributes (Name & Value)",__FILE__,__LINE__);
							}
						}else{
							CLogger::LogError("MLPSubParam is expected",__FILE__,__LINE__);
						}
						MLPSubParamElement=MLPSubParamElement->NextSiblingElement();
					}
				}else{
					CLogger::LogError("MLPParam or MLPParamList expected",__FILE__,__LINE__);
				}
			}
			MLPParamElement=MLPParamElement->NextSiblingElement();
		}
		return inducer;
	}
}

void PrintHelp(){
	cout<<"NAME"<<endl;
	cout<<"\tMLP - Machine Learning Platform"<<endl<<endl;
	cout<<"VERSION"<<endl;
	cout<<"\t"<<__MLPVERSION__<<endl<<endl;
	cout<<"USAGE"<<endl;
	cout<<"\tMLPAppMain [options]"<<endl<<endl;
	cout<<"OPTIONS"<<endl;
	cout<<"\t-f, --file file"<<endl;
	cout<<"\t\tUse file as a configuration file. By defaul program"<<endl;
	cout<<"\t\tlooks in working directory for file MLPExperiment.xml."<<endl;
	cout<<"\t-g, --generate-file"<<endl;
	cout<<"\t\tPrint on standard output simple configuration file."<<endl;
	cout<<"\t-h, --help"<<endl;
	cout<<"\t\tSpecify --help to receive this message."<<endl;
	cout<<"\t-v, --version"<<endl;
	cout<<"\t\tSpecify --version to print MLP version number."<<endl;
	cout<<"\t-s, --split stem nsplits"<<endl;
	cout<<"\t\tSpecify --split to split dataset in C4.5 format."<<endl<<endl;
	cout<<"\t-st, --split-test stem"<<endl;
	cout<<"\t\tSpecify --split-test to split dataset in C4.5 format into training (2/3) and test (1/3) sets."<<endl<<endl;
	cout<<"PLUGINS"<<endl;
#ifdef BUILD_MLPSVM
	cout<<"\tMLPSVM plugin .........................: yes"<<endl;
#endif
#ifdef MLPEAOPERATORSTEST 
	cout<<"\tMLPEAOPERATORSTEST ....................: yes"<<endl;
#endif
	cout<<endl;
	cout<<"BUGS"<<endl;
	cout<<"\tReport bugs to <marekg@ii.pb.bialystok.pl>."<<endl<<endl;
	cout.fill(' ');
	cout<<setw(45)<<__DATE__<<endl;
}

void GenerateFile(){
cout<<"<?xml version = \'1.0\' encoding = \'ISO-8859-1\'?>\n"
"<!-- MLP configuration file generated by Machine Learning Platform -->\n"
"<MLPExperiment Description=\"description\" >\n"
"  <!-- everything is logged to OutputFolder: ControlFile, OutputFiles -->\n"
"  <MLPClassification OutputFolder=\"/home/marekg/mlp/experiments/ex1\" Runs=\"10\" LogLevel=\"1\" >\n"
"    <MLPDatasets>\n"
"      <MLPDataset DSFormat=\"C4.5\" DSID=\"1\" Name=\"iris\" Path2Stem=\"/home/marekg/mlp/dane/iris\" > <!-- Columns2Rem=\"0,1\" --> \n"
"      <!-- DSFormat=\"C4.5 | ...\" -->\n"
"        <MLPExperimentType Type=\"test\" />\n"
"        <MLPDSParam Name=\"costofattributes\" Value=\"no\" />\n"
"        <MLPDSParam Name=\"costofclassification\" Value=\"no\" />\n"
"        <!-- Type=\"test\" testing data set -->\n"
"      </MLPDataset>\n"
"      <MLPDataset DSFormat=\"C4.5\" DSID=\"2\" Name=\"iris\" Path2Stem=\"/home/marekg/mlp/dane/iris\" > <!-- Columns2Rem=\"0\" --> \n"
"        <MLPExperimentType Type=\"crossvalidation\" >\n"
"          <MLPExpParam Name=\"Stratified\" Value=\"yes\" />\n"
"          <MLPExpParam Name=\"Folds\" Value=\"10\" />\n"
"          <MLPExpParam Name=\"MakeSplit\" Value=\"yes\" />\n" 
"        </MLPExperimentType>\n"
"        <MLPDSParam Name=\"costofattributes\" Value=\"no\" />\n"
"        <MLPDSParam Name=\"costofclassification\" Value=\"no\" />\n"
"        <!-- MakeSplit=\"yes\" - entire data will be divided authomatically for crossvalidation -->\n"
"        <!-- MakeSplit=\"no\" - cross-validation division is available -->\n"
"      </MLPDataset>\n"
"    </MLPDatasets>\n"
"    <MLPClassifiers>\n";

cout<<"      <MLPClassifier CLID=\"1\" Name=\"MAJORITYCLASS\" >\n";
CConstInducer t1;
t1.PrintParamsXML(cout,8);
cout<<"      </MLPClassifier>\n";
//cout<<"      <MLPClassifier CLID=\"2\" Name=\"MLPDT_EA_AP\" >\n";
//CEAIDTInducerAP t2;
//t2.PrintParamsXML(cout,8);
//cout<<"      </MLPClassifier>\n";
//cout<<"      <MLPClassifier CLID=\"3\" Name=\"MLPDT_EA_OB\" >\n";
//CEAIDTInducerOB t3;
//t3.PrintParamsXML(cout,8);
//cout<<"      </MLPClassifier>\n";
cout<<"      <MLPClassifier CLID=\"4\" Name=\"MLPDT_TD_AP\" >\n";
CTDIDTInducer t4;
t4.PrintParamsXML(cout,8);
cout<<"      </MLPClassifier>\n";
//cout<<"      <MLPClassifier CLID=\"5\" Name=\"MLPDT_TD_OB\" >\n";
//CTDIDTInducerOB t5;
//t5.PrintParamsXML(cout,8);
//cout<<"      </MLPClassifier>\n";
cout<<"      <MLPClassifier CLID=\"6\" Name=\"MLPDT_TD_AP_RAND\" >\n";
CTDIDTRandInducer t6;
t6.PrintParamsXML(cout,8);
cout<<"      </MLPClassifier>\n";
#ifdef BUILD_MLPSVM 
cout<<"      <MLPClassifier CLID=\"7\" Name=\"MLPSVM_One-Against-All\" >\n";
CSVMMultiClassOAAInducer t7;
t7.PrintParamsXML(cout,8);
cout<<"      </MLPClassifier>\n";
cout<<"      <MLPClassifier CLID=\"8\" Name=\"MLPSVM_One-Against-One\" >\n";
CSVMMultiClassOAOInducer t8;
t8.PrintParamsXML(cout,8);
cout<<"      </MLPClassifier>\n";
#endif
cout<<"    </MLPClassifiers>\n"
"  </MLPClassification>\n"
"</MLPExperiment>\n";	
}

const char* ReadAttribute(TiXmlElement* element, char* name, vector<string> domain){
	const char* str = element->Attribute(name);
	if(str){
		string d;
		if(domain.empty()){
			return str;
		}
		for(size_t t=0; t<domain.size(); t++){
			if(t==0){
				d = domain[t];
			}else{
				d += " ,"+domain[t];
			}
			if(strcmp(domain[t].c_str(),str) == 0){
				return str;
			}
		}
		stringstream ss;
		ss<<name<<" can be one of ("<<d<<")"; 
		CLogger::LogError(ss.str().c_str(),"",__LINE__);
	}else{
		stringstream ss;
		ss<<"Attribute "<<name<<" expected in "<<element->Value();
		CLogger::LogError(ss.str().c_str(),"",__LINE__);
	}
	return 0;
}

bool IsCarriedOut(TiXmlElement* element, string logval, string pnames, CClassifierDescription &desc, CDatasetDescription &ds, int run){
	TiXmlElement* MLPClassResultElement = element->FirstChildElement("MLPClassResult");
	int found=-1;
	while(MLPClassResultElement){
		int clid;
		int dsid;
		if(MLPClassResultElement->QueryIntAttribute("CLID",&clid)==TIXML_SUCCESS &&
		MLPClassResultElement->QueryIntAttribute("DSID",&dsid)==TIXML_SUCCESS){
			if(MLPClassResultElement->Attribute("ParamNamesList")==0){
				CLogger::LogError("ParamNamesList required",__FILE__,__LINE__);
			}
			if(MLPClassResultElement->Attribute("ParamListID")==0){
				CLogger::LogError("ParamListID required",__FILE__,__LINE__);
			}
			if(MLPClassResultElement->Attribute("CLName")==0){
				CLogger::LogError("CLName required",__FILE__,__LINE__);
			}
			if(MLPClassResultElement->Attribute("DSName")==0){
				CLogger::LogError("DSName required",__FILE__,__LINE__);
			}
			if(
				pnames == MLPClassResultElement->Attribute("ParamNamesList") &&
				logval == MLPClassResultElement->Attribute("ParamListID") &&
				desc.m_sName == MLPClassResultElement->Attribute("CLName") &&
				desc.m_iCLID == clid &&
				ds.m_sName == MLPClassResultElement->Attribute("DSName") &&
				ds.m_iDSID == dsid
			){
				found++;
				if(found==run){
					return true;
				}
			}
		}
		MLPClassResultElement = MLPClassResultElement->NextSiblingElement("MLPClassResult");
	}
	return false;
}

/**
 * Dodawanie szumu do danych.
 */
int AddNoise(char* stem, int probClass, int prob, double q, int n, double min, double max, string suffix){
	CAddNoise aNoise;
	string data;
	string names;
	string test;
	data = stem;
	names = stem;
	test = stem;
	data += ".data";
	names += ".names";
	test += ".test";
	CDataSetC45 *ds = new CDataSetC45("");
	CDataSetC45 *testds = new CDataSetC45("");
	ds->SetDataSetName(stem);
	testds->SetDataSetName(stem);
	ds->ReadData((char*)(data.c_str()),(char*)(names.c_str()));
	testds->ReadData((char*)(test.c_str()),(char*)(names.c_str()));

	aNoise.ClassLabels(ds,probClass);
	aNoise.ClassLabels(testds,probClass);
	aNoise.Features(ds,prob,q);
	aNoise.Features(testds,prob,q);
	aNoise.AddNoisyFeatures(ds,n,min,max);	
	aNoise.AddNoisyFeatures(testds,n,min,max);	

	ofstream outfile;
	ofstream outnames;
	ofstream outtest;

	data=stem;
	names=stem;
	test=stem;
	data+=suffix;
	data+=".data";
	names+=suffix;
	names+=".names";
	test+=suffix;
	test+=".test";
	outfile.open(data.c_str(),ofstream::out);
	outnames.open(names.c_str(),ofstream::out);
	outtest.open(test.c_str(),ofstream::out);
	ds->PrintDSInC45Format(&outfile,&outnames);
	testds->PrintDSInC45Format(&outtest,&cerr);
	return 0;
}

int main(int argc, char* argv[]){
ofstream	outstream;
char *path=0;
ifstream MLPExpFile;
vector<string> vstr;

	srand( (unsigned)time( NULL ) );
	
	/**
     * Fragment kodu do dodawania szumu do danych.
	 */
	// mlplink /tmp/chess3x3 10(20) 10(20) 1(q) 2(4) 0 3 _NALL
	/*
	cout<<"stem <int probClass> <int prob> <double q> <int n> <double min> <double max> <suffix>"<<endl;
	if(argc==9){
		AddNoise(argv[1],atoi(argv[2]),atoi(argv[3]),atof(argv[4]),atoi(argv[5]),atof(argv[6]),atof(argv[7]),argv[8]);
	}else{
		cerr<<"ups"<<endl;
	}
	return 0;
	*/
	/**
     *
	 */

	if(argc>1){
		if(strcmp("-h",argv[1])==0 || strcmp("--help",argv[1])==0){
			PrintHelp();
			return 0;
		}
		if(strcmp("-g",argv[1])==0 || strcmp("--generate-file",argv[1])==0){
			GenerateFile();
			return 0;
		}
		if(strcmp("-f",argv[1])==0 || strcmp("--file",argv[1])==0){
			if(argc>2){
				path = strdup(argv[2]);
			}else{
				CLogger::LogError("option --file requires an argument","",__LINE__);
			}
		}
		if(strcmp("-v",argv[1])==0 || strcmp("--version",argv[1])==0){
			cout<<"MLP version "<<__MLPVERSION__<<" compiled "<<__DATE__<<endl;
			return 0;
		}
		if(strcmp("-s",argv[1])==0 || strcmp("--split",argv[1])==0){
			if(argc>3){
				char* stem = strdup(argv[2]);
				int n = atoi(argv[3]);
				string data;
				data = stem;
				data += ".data";
				string names;
				names = stem;
				names += ".names";
				CDataSetC45 *ds = new CDataSetC45("");
				ds->SetDataSetName(stem);
				ds->ReadData((char*)(data.c_str()),(char*)(names.c_str()));
				deque<const IDataSet*> spDs = ds->SplitDataSet(n,true);
				for(size_t ids=0; ids<spDs.size(); ids++){
					ofstream outnames;
					string s;
					s = stem;
					s += ".names";
					outnames.open(s.c_str(),ofstream::out);
					ofstream outdata;
					stringstream st;
					st<<stem<<".data."<<ids;
					outdata.open(st.str().c_str(),ofstream::out);
					spDs[ids]->PrintDSInC45Format(&outdata,&outnames);
					delete spDs[ids];
				}
				delete ds;
				free(stem);
				return 0;
			}else{
				CLogger::LogError("option --split requires two arguments","",__LINE__);
			}
		}
		if(strcmp("-st",argv[1])==0 || strcmp("--split-test",argv[1])==0){
			if(argc>2){
				char* stem = strdup(argv[2]);
				string data;
				data = stem;
				data += ".data";
				string names;
				names = stem;
				names += ".names";
				CDataSetC45 *ds = new CDataSetC45("");
				ds->SetDataSetName(stem);
				ds->ReadData((char*)(data.c_str()),(char*)(names.c_str()));
				deque<const IDataSet*> spDs = ds->SplitDataSet23(true);
					cout<<ds->GetSize()<<" instances divided into:"<<endl;
					cout<<spDs[0]->GetSize()<<" training data and "<<spDs[1]->GetSize()<<" test data"<<endl;
					ofstream outnames;
					string s;
					s = stem;
					s += "_23.names";
					outnames.open(s.c_str(),ofstream::out);
					ofstream outdata;
					stringstream st;
					st<<stem<<"_23.data";
					outdata.open(st.str().c_str(),ofstream::out);
					spDs[0]->PrintDSInC45Format(&outdata,&outnames);
					delete spDs[0];
					outdata.close();
					st.str("");
					st<<stem<<"_23.test";
					outdata.open(st.str().c_str(),ofstream::out);
					spDs[1]->PrintDSInC45Format(&outdata,&outnames);
					delete spDs[1];
					outdata.close();
				delete ds;
				free(stem);
				return 0;
			}else{
				CLogger::LogError("option --split-test requires one argument","",__LINE__);
			}
		}

	}

	if(path == 0){
		path=strdup("MLPExperiment.xml");
		MLPExpFile.open(path,ifstream::in);
		if(!MLPExpFile.good()){
			free(path);
			// Tu nastapi wyjscie z programu
			CLogger::LogError("MLPExperiment.xml does not exist!","",__LINE__);
		}
	}else{
		MLPExpFile.open(path,ifstream::in);
		if(!MLPExpFile.good()){
			stringstream ss;
			ss<<path<<" does not exist!";
			free(path);
			CLogger::LogError(ss.str().c_str(),"",__LINE__);
		}
	}
		
	MLPExpFile.close();	
	TiXmlDocument mlpexpdoc(path);
	bool loadOkay = mlpexpdoc.LoadFile();

	if ( !loadOkay ) {
		stringstream ss;
		ss<<"Could not load file "<<path<<"'. Error="<<mlpexpdoc.ErrorDesc();
		free(path);
		CLogger::LogError(ss.str().c_str(),"",__LINE__);
	}else{
		free(path);
	}

/*#ifndef NDEBUG
	cout<<mlpexpdoc<<endl;
#endif*/

	TiXmlElement* MLPExperimentElement = mlpexpdoc.RootElement();
	if(MLPExperimentElement == 0){
		CLogger::LogError("MLPExperiment not found",__FILE__,__LINE__);
	}
	string sDescription;
	if(strcmp(MLPExperimentElement->FirstAttribute()->Name(),"Description")==0){
		sDescription = MLPExperimentElement->FirstAttribute()->Value();
	}else{
		CLogger::LogError("Attribute Description expected",__FILE__,__LINE__);
	}
	
	/**
	 * Jak bedzie wiecej rodzajow eksperymentow, to tu trzeba bedzie sprawdzic jaki
	 * zostal podany (np. MLPRegression, itp).
	 */
	TiXmlElement* MLPClassElement= MLPExperimentElement->FirstChildElement("MLPClassification");
	if(MLPClassElement == 0){
		CLogger::LogError("MLPClassification not found",__FILE__,__LINE__);
	}
	/**
	 * Na razie w tym miejscu jest MLPClassification.
	 */
	CClassificationTask clTask;
	clTask.m_sDescription = sDescription;

	int iRuns;
	int queryOkay = MLPClassElement->QueryIntAttribute("Runs",&iRuns);
	if(queryOkay == TIXML_SUCCESS){
		clTask.m_iRuns = iRuns;
	}else if(queryOkay == TIXML_WRONG_TYPE){
		CLogger::LogError("Runs is not integer",__FILE__,__LINE__);
	}
	
	int iLogLevel;
	queryOkay = MLPClassElement->QueryIntAttribute("LogLevel",&iLogLevel);
	if(queryOkay == TIXML_SUCCESS){
		clTask.m_iLogLevel = iLogLevel;
	}else{
		CLogger::LogError("LogLevel is not specified or is not integer",__FILE__,__LINE__);
	}
	
	vstr.clear();
	clTask.m_sOutputFolder = ReadAttribute(MLPClassElement,"OutputFolder",vstr);

	TiXmlElement* MLPDatasetsElement=MLPClassElement->FirstChildElement("MLPDatasets");
	if(MLPDatasetsElement == 0){
		CLogger::LogError("MLPDatasets not found",__FILE__,__LINE__);
	}
	ReadDatasets(MLPDatasetsElement,clTask);
	TiXmlElement* MLPClassifiersElement=MLPDatasetsElement->NextSiblingElement();
	if(MLPClassifiersElement == 0 ||  strcmp(MLPClassifiersElement->Value(),"MLPClassifiers") != 0){
		CLogger::LogError("MLPClassfiers not found",__FILE__,__LINE__);
	}
	ReadClassifiers(MLPClassifiersElement,clTask);
	TiXmlElement* temp=MLPClassifiersElement->NextSiblingElement();
	if(temp){
		CLogger::LogError("MLPClassification must have exactly two childeren (MLPDatasets and MLPClassfiers)",__FILE__,__LINE__);
	}

/*
 * _____________________________________________________________________________________________________________
 *					Plik konfiguracyjny wczytany
 * _____________________________________________________________________________________________________________
 *
 */ 	
	cout<<clTask<<endl;
	/*
	for(size_t t=0; t<clTask.m_vDatasets.size(); t++){
		clTask.m_vDatasets[t]->m_ptrainDS->PrintDSInHumanReadableFormat(&cout);
		if(clTask.m_vDatasets[t]->m_ptestDS){
			clTask.m_vDatasets[t]->m_ptestDS->PrintDSInHumanReadableFormat(&cout);
		}
	 
	for(size_t t=0; t<clTask.m_vClassifiers.size(); t++){
		cout<<clTask.m_vClassifiers[t]->m_sName<<endl;
	}
	*/
	
	if(clTask.m_sOutputFolder[clTask.m_sOutputFolder.size()-1]!='/'){
		clTask.m_sOutputFolder += "/";
	}
	
	string resfile = clTask.m_sOutputFolder;
	resfile += "MLPResults.xml";
	TiXmlDocument resultxml(resfile.c_str());
	// Sprawdzenie czy nie ma juz pliku wynikowego (z czesciowym wynikiem z poprzedniego uruchomienia)
	MLPExpFile.open(resfile.c_str(),ifstream::in);
	if(!MLPExpFile.good()){
		// Utworzenie pliku z szablonu
		string sMLPResult;
		sMLPResult ="<?xml version = '1.0' encoding = 'ISO-8859-1'?>\n";
		sMLPResult +=
		"<MLPResults Description=\""+clTask.m_sDescription+"\" >\n"+
		"  <MLPClassification>\n"+
		"  </MLPClassification>\n"+
		"</MLPResults>\n";
		resultxml.Parse(sMLPResult.c_str());
		if(resultxml.Error()){
			stringstream ss;
			ss<<"Error in "<<resultxml.Value()<<": "<<resultxml.ErrorDesc();
			CLogger::LogError(ss.str().c_str(),"",__LINE__);
		}
	}else{
		// Wczytanie istniejacego pliku
		MLPExpFile.close();
		bool loadOkay = resultxml.LoadFile();
		if ( !loadOkay ) {
			stringstream ss;
			ss<<"Could not load file "<<resfile.c_str()<<"'. Error="<<resultxml.ErrorDesc();
			CLogger::LogError(ss.str().c_str(),"",__LINE__);
		}
	}
	
	TiXmlElement* MLPResultsElement = resultxml.RootElement();
	assert(MLPResultsElement);
	MLPClassElement = MLPResultsElement->FirstChildElement("MLPClassification");
	assert(MLPClassElement);
	
	for(int runs=0; runs<clTask.m_iRuns; runs++){
		for(size_t ids=0; ids<clTask.m_vDatasets.size(); ids++){
			for(size_t icl=0; icl<clTask.m_vClassifiers.size(); icl++){
				string logval;
				string pnames;
				IInducer* inducer = clTask.m_vClassifiers[icl]->GetFirstInducer(logval,pnames);
				while(inducer){
					if(IsCarriedOut(MLPClassElement,logval,pnames,*(clTask.m_vClassifiers[icl]),*(clTask.m_vDatasets[ids]),runs)){
						cerr<<clTask.m_vClassifiers[icl]->m_sName<<" ";
						cerr<<clTask.m_vClassifiers[icl]->m_iCLID<<" ";
						cerr<<clTask.m_vDatasets[ids]->m_sName<<" ";
						cerr<<clTask.m_vDatasets[ids]->m_iDSID<<" ";
						cerr<<logval<<" skipped in "<<runs<<" run"<<endl;
						inducer = clTask.m_vClassifiers[icl]->GetNextInducer(logval,pnames);
						continue;
					}else{
						cerr<<clTask.m_vClassifiers[icl]->m_sName<<" ";
						cerr<<clTask.m_vClassifiers[icl]->m_iCLID<<" ";
						cerr<<clTask.m_vDatasets[ids]->m_sName<<" ";
						cerr<<clTask.m_vDatasets[ids]->m_iDSID<<" ";
						cerr<<logval<<" run "<<runs<<endl;
					}
					char *strtemp = strdup(logval.c_str());
					char *strtemp2 = strdup(pnames.c_str());
					strtemp = trim(strtemp);
					strtemp2 = trim(strtemp2);
					TiXmlElement MLPClassResultElement("MLPClassResult");
					MLPClassResultElement.SetAttribute("CLName",clTask.m_vClassifiers[icl]->m_sName);
					MLPClassResultElement.SetAttribute("CLID",clTask.m_vClassifiers[icl]->m_iCLID);
					MLPClassResultElement.SetAttribute("DSName",clTask.m_vDatasets[ids]->m_sName);
					MLPClassResultElement.SetAttribute("DSID",clTask.m_vDatasets[ids]->m_iDSID);
					MLPClassResultElement.SetAttribute("ParamListID",strtemp);
					MLPClassResultElement.SetAttribute("ParamNamesList",strtemp2);
					// Tu mozna zrobic sprawdzenie !!!
					stringstream filestem; 
					filestem<<"__"<<clTask.m_vClassifiers[icl]->m_sName<<"_";
					filestem<<clTask.m_vClassifiers[icl]->m_iCLID<<"_"<<clTask.m_vDatasets[ids]->m_sName<<"_";
					filestem<<clTask.m_vDatasets[ids]->m_iDSID<<"__RUN"<<runs<<"_"<<strtemp;
					free(strtemp);
					free(strtemp2);
					// Ladowanie danych dopiero przed wykorzystaniem, zeby
					// niepotrzebnie nie przechowywac ich w pamieci
					clTask.m_vDatasets[ids]->LoadData();
					// To sprawdzenie nizej dopiero po wczytaniu danych
					if(inducer->IsCostOfClassification()){
						// jak jeden zbior
						if(clTask.m_vDatasets[ids]->m_ptrainDS != 0){
							if(!clTask.m_vDatasets[ids]->m_ptrainDS->IsCostOfClassification()){
								CLogger::LogEvent("Dataset without cost of classification",__FILE__,__LINE__);
							}
						}else{
							// tutaj jest zadany podzial (kroswalidacja)
							if(clTask.m_vDatasets[ids]->m_deqDatasets.size() > 0 && clTask.m_vDatasets[ids]->m_deqDatasets[0] != 0 ){
								if(!clTask.m_vDatasets[ids]->m_deqDatasets[0]->IsCostOfClassification()){
									CLogger::LogEvent("Dataset without cost of classification",__FILE__,__LINE__);
								}
							}
						}
					}
					if(inducer->IsCostOfAttrs()){
						// jak jeden zbior
						if(clTask.m_vDatasets[ids]->m_ptrainDS != 0){
							if(!clTask.m_vDatasets[ids]->m_ptrainDS->IsCostOfAttrs()){
								CLogger::LogEvent("Dataset without cost of attributes",__FILE__,__LINE__);
							}
						}else{
							// tutaj jest zadany podzial (kroswalidacja)
							if(clTask.m_vDatasets[ids]->m_deqDatasets.size() > 0 && clTask.m_vDatasets[ids]->m_deqDatasets[0] != 0 ){
								if(!clTask.m_vDatasets[ids]->m_deqDatasets[0]->IsCostOfAttrs()){
									CLogger::LogEvent("Dataset without cost of attributes",__FILE__,__LINE__);
								}
							}
						}
					}
					switch(clTask.m_vDatasets[ids]->GetDSType()){
						case e_TEST:{
							ofstream outfile;
							string soutfile=clTask.m_sOutputFolder+filestem.str()+".txt";
							string sResultsFileStem=clTask.m_sOutputFolder+filestem.str();
							outfile.open(soutfile.c_str(),ofstream::out);
							if(!outfile.good()){
								stringstream ss;
								ss<<"Can't open "<<soutfile;
								CLogger::LogError(ss.str().c_str(),"",__LINE__);
							}
							clTask.m_vClassifiers[icl]->m_pInducer->SetResultsFileStem(sResultsFileStem);
							TiXmlElement MLPTestElement("MLPTest");
							MLPTestElement.SetAttribute("OutFile",filestem.str()+".txt");
							MLPTestElement.SetAttribute("LogFile",filestem.str()+".log");
							CTestClassificationExperiment cds(&outfile);
							cds.RunTestXML(MLPTestElement,"","",clTask.m_vDatasets[ids]->m_ptrainDS,inducer,clTask.m_vDatasets[ids]->m_ptestDS);
							MLPClassResultElement.InsertEndChild(MLPTestElement);
							outfile.close();
							break;
						}
						case e_CV:{
							TiXmlElement MLPCrossElement("MLPCrossvalidation");
							CCVClassificationExperiment cds(&cerr);
							cds.RunTestXML(MLPCrossElement,clTask.m_sOutputFolder,filestem.str(),\
							clTask.m_vDatasets[ids]->m_deqDatasets,inducer);
							MLPClassResultElement.InsertEndChild(MLPCrossElement);
							break;
						}
						case e_CVMAKESPLIT:{
							TiXmlElement MLPCrossElement("MLPCrossvalidation");
							CCVClassificationExperiment cds(&cerr);
							if(clTask.m_vDatasets[ids]->m_bStratified){
								cds.m_bStratified = true;
							}else{
								cds.m_bStratified = false;
							}
							cds.RunTestXML(MLPCrossElement,clTask.m_sOutputFolder,filestem.str(),clTask.m_vDatasets[ids]->m_ptrainDS,inducer);
							MLPClassResultElement.InsertEndChild(MLPCrossElement);
							break;
						}
					}
					// Zwolnienie danych po wykorzystaniu, zeby
					// niepotrzebnie nie przechowywac ich w pamieci
					clTask.m_vDatasets[ids]->ReleaseData();
					
					MLPClassElement->InsertEndChild(MLPClassResultElement);
					resultxml.SaveFile();
					inducer = clTask.m_vClassifiers[icl]->GetNextInducer(logval,pnames);
				}
			}
		}
	}
	return 0;
}

