#include "CVClassificationExperiment.h"
#include <cmath>
#include <algorithm>
#include "../MLPCommon/OutFormat.h"
#include "../MLPCommon/MLPStatistics.h"
#include "../MLPDataset/TempDataSet.h"
#include <sstream>
#include <typeinfo>

CCVClassificationExperiment::CCVClassificationExperiment(ostream *ostr):IClassificationExperiment(ostr){
	m_uNDataSets = 10;
	m_bStratified = true;
}

CCVClassificationExperiment::~CCVClassificationExperiment(void){
}

void CCVClassificationExperiment::PrintCVResultXML(TiXmlElement &crossElement, vector<CSingleTestResult> cvres){
	if(cvres.empty()) return;

	CMLPStatistics stat;

	// *********************************************
	CSingleTestResult min_;
	CSingleTestResult max_;

	// *********************************************
	CSingleTestResult average;
	vector<double> vquality;
	vector<double> vsize;
	vector<double> vtime;
	vector<double> vcost;
	bool bCost = false;
	for(size_t t=0; t<cvres.size();t++){
		vquality.push_back(cvres[t].m_dQuality);
		vsize.push_back(cvres[t].m_dSize);
		vtime.push_back(cvres[t].m_dTime);
		vcost.push_back(cvres[t].m_dCost);
		if(!bCost && cvres[t].m_bCost){
			bCost = true;
		}
	}
	average.m_dQuality = stat.MLPAverage<double>(vquality);
	average.m_dSize = stat.MLPAverage<double>(vsize);
	average.m_dTime = stat.MLPAverage<double>(vtime);
	average.m_dCost = stat.MLPAverage<double>(vcost);
	average.m_bCost = true;
	min_.m_dQuality = *min_element(vquality.begin(),vquality.end());
	min_.m_dSize = *min_element(vsize.begin(),vsize.end());
	min_.m_dTime= *min_element(vtime.begin(),vtime.end());
	min_.m_dCost= *min_element(vcost.begin(),vcost.end());
	min_.m_bCost = true;
	max_.m_dQuality = *max_element(vquality.begin(),vquality.end());
	max_.m_dSize = *max_element(vsize.begin(),vsize.end());
	max_.m_dTime= *max_element(vtime.begin(),vtime.end());
	max_.m_dCost= *max_element(vcost.begin(),vcost.end());
	max_.m_bCost = true;
	// *********************************************
	CSingleTestResult variance;
	variance.m_dQuality = stat.MLPVariance<double>(vquality);
	variance.m_dSize = stat.MLPVariance<double>(vsize);
	variance.m_dTime = stat.MLPVariance<double>(vtime);
	variance.m_dCost = stat.MLPVariance<double>(vcost);
	variance.m_bCost = true;
		
	// *********************************************
	CSingleTestResult stdeviation;
	stdeviation.m_dQuality = sqrt(variance.m_dQuality);
	stdeviation.m_dTime = sqrt(variance.m_dTime);
	stdeviation.m_dSize = sqrt(variance.m_dSize);
	stdeviation.m_dCost = sqrt(variance.m_dCost);
	stdeviation.m_bCost = true;
	/* Drukowanie wyniku.
	 */
	stringstream ss;
	
	// Ustawinie formatowania (zawsze z kropka - zeby analiza byla jednolita)
    // ss.setf(ios::showpoint);
    ss.setf(ios::fixed);
    // ss.precision(8);
	
	TiXmlElement MLPResulElement("MLPResult");
	MLPResulElement.SetAttribute("Name","quality");
	ss.str("");ss<<average.m_dQuality;
	MLPResulElement.SetAttribute("Value",ss.str());
	ss.str("");ss<<min_.m_dQuality;
	MLPResulElement.SetAttribute("min",ss.str());
	ss.str("");ss<<max_.m_dQuality;
	MLPResulElement.SetAttribute("max",ss.str());
	ss.str("");ss<<variance.m_dQuality;
	MLPResulElement.SetAttribute("var",ss.str());
	ss.str("");ss<<stdeviation.m_dQuality;
	MLPResulElement.SetAttribute("stdev",ss.str());
	crossElement.InsertEndChild(MLPResulElement);

	TiXmlElement MLPResulElement1("MLPResult");
	MLPResulElement1.SetAttribute("Name","size");
	ss.str("");ss<<average.m_dSize;
	MLPResulElement1.SetAttribute("Value",ss.str());
	ss.str("");ss<<min_.m_dSize;
	MLPResulElement1.SetAttribute("min",ss.str());
	ss.str("");ss<<max_.m_dSize;
	MLPResulElement1.SetAttribute("max",ss.str());
	ss.str("");ss<<variance.m_dSize;
	MLPResulElement1.SetAttribute("var",ss.str());
	ss.str("");ss<<stdeviation.m_dSize;
	MLPResulElement1.SetAttribute("stdev",ss.str());
	crossElement.InsertEndChild(MLPResulElement1);
	
	TiXmlElement MLPResulElement2("MLPResult");
	MLPResulElement2.SetAttribute("Name","time");
	ss.str("");ss<<average.m_dTime;
	MLPResulElement2.SetAttribute("Value",ss.str());
	ss.str("");ss<<min_.m_dTime;
	MLPResulElement2.SetAttribute("min",ss.str());
	ss.str("");ss<<max_.m_dTime;
	MLPResulElement2.SetAttribute("max",ss.str());
	ss.str("");ss<<variance.m_dTime;
	MLPResulElement2.SetAttribute("var",ss.str());
	ss.str("");ss<<stdeviation.m_dTime;
	MLPResulElement2.SetAttribute("stdev",ss.str());
	crossElement.InsertEndChild(MLPResulElement2);
	
	if(bCost){
		TiXmlElement MLPResulElement3("MLPResult");
		MLPResulElement3.SetAttribute("Name","cost");
		ss.str("");ss<<average.m_dCost;
		MLPResulElement3.SetAttribute("Value",ss.str());
		ss.str("");ss<<min_.m_dCost;
		MLPResulElement3.SetAttribute("min",ss.str());
		ss.str("");ss<<max_.m_dCost;
		MLPResulElement3.SetAttribute("max",ss.str());
		ss.str("");ss<<variance.m_dCost;
		MLPResulElement3.SetAttribute("var",ss.str());
		ss.str("");ss<<stdeviation.m_dCost;
		MLPResulElement3.SetAttribute("stdev",ss.str());
		crossElement.InsertEndChild(MLPResulElement3);	
	}
}


bool CCVClassificationExperiment::RunTestXML(TiXmlElement &crossElement, string outputFolder, string filestem, const IDataSet *trainDS, IInducer* inducer, const IDataSet *testDS){
	deque<const IDataSet*> cvdatasets = trainDS->SplitDataSet(m_uNDataSets,m_bStratified);
	
	bool res = RunTestXML(crossElement,outputFolder,filestem,cvdatasets,inducer,testDS);
	
	while(!cvdatasets.empty()){
		delete cvdatasets.front();
		cvdatasets.pop_front();
	}
	return res;
}

bool CCVClassificationExperiment::RunTestXML(TiXmlElement &crossElement, string outputFolder, string filestem, deque<const IDataSet*> deqtrainDS, IInducer* inducer, const IDataSet *testDS){
	// Pierwszy wymiar wektora jest po vinducer.size.
	vector<CSingleTestResult> cvres;

	for(size_t t=0; t<deqtrainDS.size(); t++){
		const IDataSet* test = deqtrainDS.front();
		deqtrainDS.pop_front();
		CTempDataSet* train = new CTempDataSet(*(deqtrainDS.front()));
		for(size_t i=0; i<deqtrainDS.size(); i++){
			*train += *deqtrainDS[i];
		}
		ofstream outfile;
		stringstream ssfilestem;
		ssfilestem<<outputFolder<<filestem<<"_FOLD"<<(unsigned)t;
		string soutfile=ssfilestem.str()+".txt";
		outfile.open(soutfile.c_str(),ofstream::out);
		if(!outfile.good()){
			stringstream ss;
			ss<<"Can't open "<<soutfile;
			CLogger::LogError(ss.str().c_str(),"",__LINE__);
		}
		stringstream ss;
		ss<<filestem<<"_FOLD"<<(unsigned)t;
		TiXmlElement MLPFoldElement("MLPFold");
		MLPFoldElement.SetAttribute("OutFile",ss.str()+".txt");
		MLPFoldElement.SetAttribute("LogFile",ss.str()+".log");
		m_pOutStream=&outfile;
		inducer->SetResultsFileStem(ssfilestem.str());
		
		CSingleTestResult temp = RunSingleExp((const IDataSet*)train,inducer,test);
		if(temp.m_dQuality == -1){
			CLogger::LogError("Single Experiment Failed",__FILE__,__LINE__);
			return false;
		}
		cvres.push_back(temp);
		delete train;
		deqtrainDS.push_back(test);
		
		outfile.close();
		
		TiXmlElement MLPResulElement1("MLPResult");
		MLPResulElement1.SetAttribute("Name","size");
		ss.str("");ss<<temp.m_dSize;
		MLPResulElement1.SetAttribute("Value",ss.str());
		MLPFoldElement.InsertEndChild(MLPResulElement1);
		TiXmlElement MLPResulElement2("MLPResult");
		MLPResulElement2.SetAttribute("Name","time");
		MLPResulElement2.SetAttribute("Value",(int)(temp.m_dTime));
		MLPFoldElement.InsertEndChild(MLPResulElement2);
		TiXmlElement MLPResulElement3("MLPResult");
		MLPResulElement3.SetAttribute("Name","quality");
		ss.str("");ss<<temp.m_dQuality;
		MLPResulElement3.SetAttribute("Value",ss.str());
		MLPFoldElement.InsertEndChild(MLPResulElement3);
		if(temp.m_bCost){
			TiXmlElement MLPResulElement4("MLPResult");
			MLPResulElement4.SetAttribute("Name","cost");
			ss.str("");ss<<temp.m_dCost;
			MLPResulElement4.SetAttribute("Value",ss.str());
			MLPFoldElement.InsertEndChild(MLPResulElement4);
		}

		crossElement.InsertEndChild(MLPFoldElement);
	}

	PrintCVResultXML(crossElement,cvres);
	AddXMLParams(crossElement,inducer);
	
	return true;
}

void CCVClassificationExperiment::SetDataSetsCount(unsigned n){
	m_uNDataSets = n;
}

CSingleTestResult CCVClassificationExperiment::RunSingleExp(const IDataSet *trainDS, IInducer* inducer, const IDataSet* testDS){
CSingleTestResult res;
vector<vector<sConfusionMatrixElement> > ConfiusionMatrix;

	time_t t1 = time(NULL);
	IClassifier *iClassifier = inducer->Train(trainDS);
	time_t t2 = time(NULL);

	res.m_dTime = t2 - t1;

	if(iClassifier){
		// 0 - czyli zwracana bedzie liczba lisci
		res.m_dSize = iClassifier->GetClassifierSize();
	
		*m_pOutStream<<"_______________________________________________________________"<<endl;
		*m_pOutStream << endl;
		CLogger::TimeStamp("Crossvalidation:\t",*m_pOutStream);
		*m_pOutStream << "Training data set:\t" << trainDS->m_sDataSetName.c_str() << endl;
		*m_pOutStream << "Test data set:\t\t" << testDS->m_sDataSetName.c_str() << endl;
		*m_pOutStream << "Classifier:\t\t" << iClassifier->GetClassifierName().c_str() << endl;
	
		iClassifier->DisplayStruct(*m_pOutStream,0);
		*m_pOutStream<<"Time:\t"<<(long)(t2-t1)<<endl;
		*m_pOutStream<<"\nEvaluation on training data ("<<trainDS->GetSize()<<" items):"<<endl<<endl;

		vector<IAttrValue*> classvalues = trainDS->GetClassValues();
		for(unsigned i=0; i<classvalues.size(); i++){
			for(unsigned j=0; j<classvalues.size(); j++){
				sConfusionMatrixElement sCME;
				sCME.s_pIAGoodClass = classvalues[i];
				sCME.s_pIAClassifiedAs = classvalues[j];
				sCME.s_uNum = 0;
				if(j==0){
					vector<sConfusionMatrixElement> temp;
					temp.push_back(sCME);
					ConfiusionMatrix.push_back(temp);
				}else{
					ConfiusionMatrix[i].push_back(sCME);
				}
			}
		}

		int all=0;
		int ok=0;
		double costOfClassification = 0.0;
		double costOfAttrs = 0.0;
		pair<double,double> costs;
		for(unsigned int i=0; i<trainDS->GetSize(); i++){
			CDataObject *inst = (*trainDS)[i];
			IAttrValue* pIA=0;
			if(inducer->IsCostOfClassification() || inducer->IsCostOfAttrs()){
				pIA=iClassifier->Classify(inst,costs);
				if(inducer->IsCostOfClassification()){
					costOfClassification+=costs.first;
				}
				if(inducer->IsCostOfAttrs()){
					costOfAttrs+=costs.second;
				}
			}else{
				pIA=iClassifier->Classify(inst);
			}
			all++;
			IAttrValue *ia=inst->GetClass();
			if((*pIA)==(*ia))ok++;
			// Zapis testu do ConfiusionMatrix
			for(unsigned i=0; i<ConfiusionMatrix.size(); i++){
				bool tempb=false;
				for(unsigned j=0; j<ConfiusionMatrix[i].size(); j++){
					if( (*(ConfiusionMatrix[i][j].s_pIAGoodClass) == *ia) &&
						(*(ConfiusionMatrix[i][j].s_pIAClassifiedAs) == *pIA) ){
							ConfiusionMatrix[i][j].s_uNum++;
							tempb=true;
							break;
					}
				}
				if(tempb) break;
			}
		}
		// Kosz trzeba podzielic przez liczbe przykladow w zbiorze uczacym.
		assert(trainDS->GetSize() > 0);
		costOfClassification /= (double)(trainDS->GetSize());
		costOfAttrs /= (double)(trainDS->GetSize());

		//______________________________________________________________________
		char a='a';
		for(unsigned i=0; i<ConfiusionMatrix.size(); i++){
			*m_pOutStream<<"("<<a++<<")\t";
		}
		*m_pOutStream<<"<-classified as"<<endl<<endl;
		a='a';
		for(unsigned i=0; i<ConfiusionMatrix.size(); i++){
			for(unsigned j=0; j<ConfiusionMatrix[i].size(); j++){
				*m_pOutStream<<ConfiusionMatrix[i][j].s_uNum<<"\t";
			}
			*m_pOutStream<<"("<<a++<<"): ";
			ConfiusionMatrix[i][0].s_pIAGoodClass->PrintValue(*m_pOutStream);
			*m_pOutStream<<endl;
		}
		//______________________________________________________________________
		*m_pOutStream << "\nResult on training data:\t" << 100.0*(double)ok/(double)all << "%" << endl;
		if(inducer->IsCostOfClassification()){
			*m_pOutStream << "Cost of classification on training data:\t $" << costOfClassification << endl;
		}
		if(inducer->IsCostOfAttrs()){
			*m_pOutStream << "Cost of attributes:\t $" << costOfAttrs << endl;
		}
		if(inducer->IsCostOfAttrs() && inducer->IsCostOfClassification()){
			*m_pOutStream << "Total cost:\t $" << costOfClassification + costOfAttrs << endl;
		}
		*m_pOutStream << "\nEvaluation on test data ("<<testDS->GetSize()<<" items):"<<endl<<endl;
		
		for(unsigned i=0; i<ConfiusionMatrix.size(); i++){
			for(unsigned j=0; j<ConfiusionMatrix[i].size(); j++){
				ConfiusionMatrix[i][j].s_uNum=0;
			}
		}

		all=0;
		ok=0;
		costOfClassification = 0.0;
		costOfAttrs = 0.0;
		for(unsigned int i=0; i<testDS->GetSize(); i++){
			CDataObject *inst = (*testDS)[i];
			IAttrValue* pIA=0;
			if(inducer->IsCostOfClassification() || inducer->IsCostOfAttrs()){
				pIA=iClassifier->Classify(inst,costs);
				if(inducer->IsCostOfClassification()){
					costOfClassification+=costs.first;
				}
				if(inducer->IsCostOfAttrs()){
					costOfAttrs+=costs.second;
				}
			}else{
				pIA=iClassifier->Classify(inst);
			}
			all++;
			IAttrValue *ia=inst->GetClass();
			if((*pIA)==(*ia))ok++;
			// Zapis testu do ConfiusionMatrix
			for(unsigned i=0; i<ConfiusionMatrix.size(); i++){
				bool tempb=false;
				for(unsigned j=0; j<ConfiusionMatrix[i].size(); j++){
					if( (*(ConfiusionMatrix[i][j].s_pIAGoodClass) == *ia) &&
						(*(ConfiusionMatrix[i][j].s_pIAClassifiedAs) == *pIA) ){
							ConfiusionMatrix[i][j].s_uNum++;
							tempb=true;
							break;
					}
				}
				if(tempb) break;
			}
		}
		// Kosz trzeba podzielic przez liczbe przykladow w zbiorze testowym.
		assert(testDS->GetSize() > 0);
		costOfClassification /= (double)(testDS->GetSize());
		costOfAttrs /= (double)(testDS->GetSize());
		
		//______________________________________________________________________
		a='a';
		for(unsigned i=0; i<ConfiusionMatrix.size(); i++){
			*m_pOutStream<<"("<<a++<<")\t";
		}
		*m_pOutStream<<"<-classified as"<<endl<<endl;
		a='a';
		for(unsigned i=0; i<ConfiusionMatrix.size(); i++){
			for(unsigned j=0; j<ConfiusionMatrix[i].size(); j++){
				*m_pOutStream<<ConfiusionMatrix[i][j].s_uNum<<"\t";
			}
			*m_pOutStream<<"("<<a++<<"): ";
			ConfiusionMatrix[i][0].s_pIAGoodClass->PrintValue(*m_pOutStream);
			*m_pOutStream<<endl;
		}
		//______________________________________________________________________
		*m_pOutStream << "\nResult on test data:\t" << 100.0*(double)ok/(double)all << "%" << endl;
		double cost = 0;
		if(inducer->IsCostOfClassification()){
			*m_pOutStream << "Cost of classification on test data:\t $" << costOfClassification << endl;
			cost = costOfClassification;
		}
		if(inducer->IsCostOfAttrs()){
			*m_pOutStream << "Cost of attributes:\t $" << costOfAttrs << endl;
			cost += costOfAttrs;
		}
		if(inducer->IsCostOfAttrs() && inducer->IsCostOfClassification()){
			*m_pOutStream << "Total cost:\t $" << costOfClassification + costOfAttrs << endl;
		}
		if(inducer->IsCostOfAttrs() || inducer->IsCostOfClassification()){
			res.m_dCost = cost;
			res.m_bCost = true;
		}
		*m_pOutStream<<"_______________________________________________________________"<<endl;
		
		delete iClassifier;
		if(all > 0){
			res.m_dQuality = 100.0 * (double)ok / (double)all;
			//- stringstream ss;
			//- ss<<"Result on test data: "<<res.m_dQuality;
			//- CLogger::LogEvent(ss.str().c_str(),trainDS->m_sDataSetName.c_str(),__LINE__,cerr);
			return res;
		}
		else{
			CLogger::LogError("Empty Data Set",__FILE__,__LINE__);
			res.m_dQuality = -1;
			return res;
		}
	}else{
		CLogger::LogError("NULL Classifier returned by IInducer::Train()",__FILE__,__LINE__);
		res.m_dQuality = -1;
		return res;
	}
}
