#ifndef __CCVCLASSIFICATIONEXPERIMENT__
#define __CCVCLASSIFICATIONEXPERIMENT__

#include "../tinyxml/tinyxml.h"
#include "ClassificationExperiment.h"
#include "SingleTestResult.h"
#include <deque>

/**
 * Kroswalidacja dla jednego lub kilku klasyfikatorw.
 * @author Marek Grzes'
 */
class CCVClassificationExperiment : public IClassificationExperiment
{
public:
	/**
	 * Sposob dzielenia danych.
	 */
	bool m_bStratified;

	/**
	 * Kroswalidacja na jednym zbiorze danych dla jednego ucznia. Metoda sama dzieli dane do
	 * kroswalidacji. Pliki wynikowe dla poszczegolnych wywolan
	 * algorytmu uczenia tworzone sa w tej funkcji.
	 * @param crossElement - tag XML kroswalidacji w pliku wynikowym
	 * @param outputFolder - sciezka do katalogu wynikow
	 * @param filestem - prefix dla nazw plikow wyjsciowych i plikow z informacjami diagnostycznymi
	 * @param trainDS - zbior uczacy
	 * @param inducer - uczen
	 * @param testDS - nie jest barany pod uwage
	 * @see IClassificationExperiment#m_pOutStream
	 */
	virtual bool RunTestXML(TiXmlElement &crossElement, string outputFolder, string filestem, const IDataSet *trainDS, IInducer* inducer, const IDataSet *testDS=NULL);

	/**
	 * Kroswalidacja na jednym zbiorze danych dla jednego ucznia. Metoda nie dzieli danych. 
	 * Zadany podzial danych jest przekazywany w parametrze. Pliki wynikowe dla poszczegolnych wywolan
	 * algorytmu uczenia tworzone sa w tej funkcji.
	 * @param crossElement - tag XML kroswalidacji w pliku wynikowym
	 * @param outputFolder - sciezka do katalogu wynikow
	 * @param filestem - prefix dla nazw plikow wyjsciowych i plikow z informacjami diagnostycznymi
	 * @param vtrainDS - zadany podzial zbioru uczacego
	 * @param inducer - uczen
	 * @param testDS - nie jest barany pod uwage
	 * @see IClassificationExperiment#m_pOutStream
	 */
	virtual bool RunTestXML(TiXmlElement &crossElement, string outputFolder, string filestem, deque<const IDataSet*> deqtrainDS, IInducer* inducer, const IDataSet *testDS=NULL);
	
protected:
	/**
	 * Przeprowadzenie pojedynczego eksperymentu.
	 * @param trainDS - zbir danych do uczenia klasyfikatora
	 * @param inducer - ucze, ktry ma stworzy klasyfikator do testowania
	 * @param testdatasets - kolejka zbiorw danych do testowania klasyfikatora
	 * utworzonego z danych przekazanych w pierwszym parametrze
	 * @return Wynik zapisany w obiekcie typu CSingleTestResult zawiera jako
	 * klasyfikacji (w procentach), czas uczenia klasyfikatora oraz rozmiar
	 * klasyfikatora. W przypadku bdu pole m_dQuality zwrconego obiektu
	 * CSingleTestResult ma wartosc -1.
	 * @see CSingleTestResult
	 */
	virtual CSingleTestResult RunSingleExp(const IDataSet *trainDS, IInducer* inducer, const IDataSet* testDS);

	/**
	 * Zapisuje do pliku XML wynik kroswalidacji.
	 * @param crossElement - tag XML wyniku kroswalidacji
	 * cvres - wektor wynikow czastkowych
	 */
	virtual void PrintCVResultXML(TiXmlElement &crossElement, vector<CSingleTestResult> cvres);
	
	/**
	 * Parametr wskazujcy na ile zbiorw ma by dzielony zbir danych, na ktrym
	 * ma by przeprowadzona kroswalidacja.
	 */
	unsigned m_uNDataSets;

public:

	/**
	 * Okresla na ile zbiorw ma byc podzielony zbir danych do kroswalidacji.
	 * @see m_uNDataSets
	 */
	virtual void SetDataSetsCount(unsigned n);

	/**
	 * Konstruktor pobierajacy strumien, ktry przekazywany jest do konstruktora
	 * IClassificationExperiment.
	 * @see IClassificationExperiment#m_pOutStream
	 */
	CCVClassificationExperiment(ostream *ostr=&cout);

	virtual ~CCVClassificationExperiment(void);
};

#endif

