#ifndef __CLASSIFIERH__
#define __CLASSIFIERH__

#include "../MLPDataset/DataObject.h"
#include "../MLPDataset/IDataSet.h"

/**
 * Interfejs klasyfikatora.
 * @author Marek Grzes
 */
class IClassifier{
public:
	/**
	 * Dla podanego przykladu zwraca jego klase.
	 */
	virtual IAttrValue* Classify(CDataObject *instance)=0;

	/**
	 * Dla podanego przykladu zwraca jego klase. W zmiennej dCost jest zwracany koszt.
	 * cost.first koszt blednej klasyfikacji
	 * cost.second koszt atrybutow
	 */
	virtual IAttrValue* Classify(CDataObject *instance, pair<double,double> &cost)=0;

	/**
	 * Drukuje do strumienia - ostr - strukture klasyfikatora, w
	 * formacie okreslonym parametrem dispformat.
	 */
	virtual void DisplayStruct(ostream &ostr, int dispformat)=0;

	virtual string GetClassifierName()=0;
	
	/**
	 * Zwraca rozmiar klasyfikatora. Interpretacja tej wartosci zalezy od
	 * rodzaju klasyfikatora.
	 *
	 * @paam iResType - rodzaj zwracanego wyniku, interpretacja tego 
	 * parametru zalezy od klasyfikatora. Np. w przypadku drzew 1 
	 * oznacza, ze ma byc zwrocony rozmiar calego drzewa, a 0 to liczba 
	 * lisci drzewa.
	 */
	virtual unsigned GetClassifierSize(int iResType=0) const=0;

	IClassifier(const IDataSet *dataset, bool bcostOfClass, bool bcostOfAttrs){
		assert(dataset);
		m_pIDataset = dataset;
		m_bCostOfClassification = bcostOfClass;
		m_bCostOfAttrs = bcostOfAttrs;
	}
	
	virtual ~IClassifier(){
	}

	/**
	 * Czy klasyfikator byl nauczany z uwzglednieniem kosztow.
	 */
	bool m_bCostOfClassification;
	bool m_bCostOfAttrs;
	
protected:
	const IDataSet* m_pIDataset;
};

#endif


