/**
 * $Revision: 1.10 $
 * $Date: 2005/04/11 22:44:31 $
 */

#include "Inducer.h"
#include <iomanip>

IInducer::IInducer(){
	m_bCostOfAttrs = false;
	m_bCostOfClassification = false;
	m_pDataset = 0;
}

IInducer::~IInducer(){
	if(m_ofsLogFile.is_open()){
		m_ofsLogFile.close();
	}
}

vector<string> IInducer::GetParams() const{
vector<string> res;
stringstream ss;

	res.push_back("costofattributes");
	if(m_bCostOfAttrs) res.push_back("yes");
		else res.push_back("no");
	res.push_back("costofclassification");
	if(m_bCostOfClassification) res.push_back("yes");
		else res.push_back("no");
	return res;
}

void IInducer::OpenLogFile(){
	// Dodawanego tutaj rozszerzenia lepiej nie zmieniac, bo tak jest zapisywane
	// tez wynikowych plikach w xml'u.
	OpenLogFile(m_sResultsFileStem+".log");
}

void IInducer::OpenLogFile(string fname){
	if(m_ofsLogFile.is_open()){
		m_ofsLogFile.close();
	}
	m_ofsLogFile.open(fname.c_str(),ofstream::out);
	if(!m_ofsLogFile.good()){
		stringstream ss;
		ss<<"Can't open "<<fname;
		CLogger::LogEvent(ss.str().c_str(),"",__LINE__);
	}
}

void IInducer::PrintParamsXML(ostream &ostr, int nsp){
	ostr.fill(' ');
	ostr<<setw(nsp)<<" "<<"<MLPParam Name=\"costofattributes\" Value=\"no\" />"<<endl;	
	ostr<<setw(nsp)<<" "<<"<MLPParam Name=\"costofclassification\" Value=\"no\" />"<<endl;
	ostr.fill(' ');
}

bool IInducer::SetParam(string sParamName, string sParamValue){
	string::iterator i = sParamName.begin();
	while(i != sParamName.end()){
		*i = tolower(*i);
		++i;
	}
	i = sParamValue.begin();
	while(i != sParamValue.end()){
		*i = tolower(*i);
		++i;
	}
	if(sParamName=="cost of attributes" || sParamName=="costofattributes"){
		if(sParamValue=="yes"){
			m_bCostOfAttrs = true;
			return true;
		}
		if(sParamValue=="no"){
			m_bCostOfAttrs = false;
			return true;
		}
		return false;
	}
	if(sParamName=="cost of classification" || sParamName=="costofclassification"){
		if(sParamValue=="yes"){
			m_bCostOfClassification = true;
			return true;
		}
		if(sParamValue=="no"){
			m_bCostOfClassification = false;
			return true;
		}
		return false;
	}
	
	return false;
}

IClassifier *IInducer::Train(const IDataSet *trainingDataSet){
	PreTrain(trainingDataSet);
	IClassifier *res = ActualTrain(trainingDataSet);
	PostTrain(trainingDataSet);
	return res;
}

void IInducer::PreTrain(const IDataSet *trainingDataSet){
	m_pDataset = trainingDataSet;
	if(m_sResultsFileStem.length()==0){
#ifdef __linux__
		OpenLogFile("/dev/null");
#endif
	}else{
		// Otwarcie pliku na informacje diagnostyczne
		OpenLogFile();
	}
}
