#include "TestClassificationExperiment.h"
#include <ctime>
#include <sstream>
#include <typeinfo>

bool CTestClassificationExperiment::RunTestXML(TiXmlElement &testElement, string outputFolder, string filestem, const IDataSet *trainDS, IInducer* inducer, const IDataSet *testDS){
vector<vector<sConfusionMatrixElement> > ConfiusionMatrix;

	time_t t1 = time(NULL);
	IClassifier *iClassifier = inducer->Train(trainDS);
	time_t t2 = time(NULL);

	TiXmlElement MLPResulElement1("MLPResult");
	MLPResulElement1.SetAttribute("Name","size");
	// 0 - czyli zwracan liczna lisci
	MLPResulElement1.SetAttribute("Value",iClassifier->GetClassifierSize());
	testElement.InsertEndChild(MLPResulElement1);
	TiXmlElement MLPResulElement2("MLPResult");
	MLPResulElement2.SetAttribute("Name","time");
	MLPResulElement2.SetAttribute("Value",(int)(t2-t1));
	testElement.InsertEndChild(MLPResulElement2);
	
	if(iClassifier){
		*m_pOutStream<<"_______________________________________________________________"<<endl;
		*m_pOutStream << endl;
		CLogger::TimeStamp("Classification test:\t",*m_pOutStream);
		*m_pOutStream << "Training data set:\t" << trainDS->m_sDataSetName.c_str() << endl;
		*m_pOutStream << "Test data set:\t\t" << testDS->m_sDataSetName.c_str() << endl;
		*m_pOutStream << "Classifier:\t\t" << iClassifier->GetClassifierName().c_str() << endl;

		iClassifier->DisplayStruct(*m_pOutStream,0);
		*m_pOutStream<<"Time:\t"<<(long)(t2-t1)<<endl;
		*m_pOutStream<<"\nEvaluation on training data ("<<trainDS->GetSize()<<" items):"<<endl<<endl;

		vector<IAttrValue*> classvalues = trainDS->GetClassValues();
		for(unsigned i=0; i<classvalues.size(); i++){
			for(unsigned j=0; j<classvalues.size(); j++){
				sConfusionMatrixElement sCME;
				sCME.s_pIAGoodClass = classvalues[i];
				sCME.s_pIAClassifiedAs = classvalues[j];
				sCME.s_uNum = 0;
				if(j==0){
					vector<sConfusionMatrixElement> temp;
					temp.push_back(sCME);
					ConfiusionMatrix.push_back(temp);
				}else{
					ConfiusionMatrix[i].push_back(sCME);
				}
			}
		}

		int all=0;
		int ok=0;
		double costOfClassification = 0.0;
		double costOfAttrs = 0.0;
		pair<double,double> costs;
		for(unsigned int i=0; i<trainDS->GetSize(); i++){
			CDataObject *inst = (*trainDS)[i];
			IAttrValue* pIA=0;
			if(inducer->IsCostOfClassification() || inducer->IsCostOfAttrs()){
				pIA=iClassifier->Classify(inst,costs);
				if(inducer->IsCostOfClassification()){
					costOfClassification+=costs.first;
				}
				if(inducer->IsCostOfAttrs()){
					costOfAttrs+=costs.second;
				}
			}else{
				pIA=iClassifier->Classify(inst);
			}
			all++;
			IAttrValue *ia=inst->GetClass();
			if((*pIA)==(*ia))ok++;
			// Zapis testu do ConfiusionMatrix
			for(unsigned i=0; i<ConfiusionMatrix.size(); i++){
				bool tempb=false;
				for(unsigned j=0; j<ConfiusionMatrix[i].size(); j++){
					if( (*(ConfiusionMatrix[i][j].s_pIAGoodClass) == *ia) &&
						(*(ConfiusionMatrix[i][j].s_pIAClassifiedAs) == *pIA) ){
							ConfiusionMatrix[i][j].s_uNum++;
							tempb=true;
							break;
					}
				}
				if(tempb) break;
			}
		}
		// Kosz trzeba podzielic przez liczbe przykladow w zbiorze uczacym.
		assert(trainDS->GetSize() > 0);
		costOfClassification /= (double)(trainDS->GetSize());
		costOfAttrs /= (double)(trainDS->GetSize());

		//______________________________________________________________________
		char a='a';
		for(unsigned i=0; i<ConfiusionMatrix.size(); i++){
			*m_pOutStream<<"("<<a++<<")\t";
		}
		*m_pOutStream<<"<-classified as"<<endl<<endl;
		a='a';
		for(unsigned i=0; i<ConfiusionMatrix.size(); i++){
			for(unsigned j=0; j<ConfiusionMatrix[i].size(); j++){
				*m_pOutStream<<ConfiusionMatrix[i][j].s_uNum<<"\t";
			}
			*m_pOutStream<<"("<<a++<<"): ";
			ConfiusionMatrix[i][0].s_pIAGoodClass->PrintValue(*m_pOutStream);
			*m_pOutStream<<endl;
		}
		//______________________________________________________________________
		*m_pOutStream << "\nResult on training data:\t" << 100.0*(double)ok/(double)all << "%" << endl;
		if(inducer->IsCostOfClassification()){
			*m_pOutStream << "Cost of classification on training data:\t $" << costOfClassification << endl;
		}
		if(inducer->IsCostOfAttrs()){
			*m_pOutStream << "Cost of attributes:\t $" << costOfAttrs << endl;
		}
		if(inducer->IsCostOfAttrs() && inducer->IsCostOfClassification()){
			*m_pOutStream << "Total cost:\t $" << costOfClassification + costOfAttrs << endl;
		}
		*m_pOutStream << "\nEvaluation on test data ("<<testDS->GetSize()<<" items):"<<endl<<endl;

		for(unsigned i=0; i<ConfiusionMatrix.size(); i++){
			for(unsigned j=0; j<ConfiusionMatrix[i].size(); j++){
				ConfiusionMatrix[i][j].s_uNum=0;
			}
		}

		all=0;
		ok=0;
		costOfClassification = 0.0;
		costOfAttrs = 0.0;
		for(unsigned int i=0; i<testDS->GetSize(); i++){
			CDataObject *inst = (*testDS)[i];
			IAttrValue* pIA=0;
			if(inducer->IsCostOfClassification() || inducer->IsCostOfAttrs()){
				pIA=iClassifier->Classify(inst,costs);
				if(inducer->IsCostOfClassification()){
					costOfClassification+=costs.first;
				}
				if(inducer->IsCostOfAttrs()){
					costOfAttrs+=costs.second;
				}
			}else{
				pIA=iClassifier->Classify(inst);
			}
			all++;
			IAttrValue *ia=inst->GetClass();
			if((*pIA)==(*ia))ok++;
			// Zapis testu do ConfiusionMatrix
			for(unsigned i=0; i<ConfiusionMatrix.size(); i++){
				bool tempb=false;
				for(unsigned j=0; j<ConfiusionMatrix[i].size(); j++){
					if( (*(ConfiusionMatrix[i][j].s_pIAGoodClass) == *ia) &&
						(*(ConfiusionMatrix[i][j].s_pIAClassifiedAs) == *pIA) ){
							ConfiusionMatrix[i][j].s_uNum++;
							tempb=true;
							break;
					}
				}
				if(tempb) break;
			}
		}
		// Kosz trzeba podzielic przez liczbe przykladow w zbiorze testowym.
		assert(testDS->GetSize() > 0);
		costOfClassification /= (double)(testDS->GetSize());
		costOfAttrs /= (double)(testDS->GetSize());

		//______________________________________________________________________
		a='a';
		for(unsigned i=0; i<ConfiusionMatrix.size(); i++){
			*m_pOutStream<<"("<<a++<<")\t";
		}
		*m_pOutStream<<"<-classified as"<<endl<<endl;
		a='a';
		for(unsigned i=0; i<ConfiusionMatrix.size(); i++){
			for(unsigned j=0; j<ConfiusionMatrix[i].size(); j++){
				*m_pOutStream<<ConfiusionMatrix[i][j].s_uNum<<"\t";
			}
			*m_pOutStream<<"("<<a++<<"): ";
			ConfiusionMatrix[i][0].s_pIAGoodClass->PrintValue(*m_pOutStream);
			*m_pOutStream<<endl;
		}
		//______________________________________________________________________
		TiXmlElement MLPResulElement3("MLPResult");
		MLPResulElement3.SetAttribute("Name","quality");
		stringstream ss;
		
		// Ustawinie formatowania (zawsze z kropka - zeby analiza byla jednolita)
		// ss.setf(ios::showpoint);
		ss.setf(ios::fixed);
		// ss.precision(8);
	
		ss<<100.0*(double)ok/(double)all;
		MLPResulElement3.SetAttribute("Value",ss.str());
		testElement.InsertEndChild(MLPResulElement3);

		double cost = 0;
		*m_pOutStream << "\nResult on test data:\t" << 100.0*(double)ok/(double)all << "%" << endl;
		if(inducer->IsCostOfClassification()){
			*m_pOutStream << "Cost of classification on test data:\t $" << costOfClassification << endl;
			cost = costOfClassification;
		}
		if(inducer->IsCostOfAttrs()){
			*m_pOutStream << "Cost of attributes:\t $" << costOfAttrs << endl;
			cost += costOfAttrs;
		}
		if(inducer->IsCostOfAttrs() && inducer->IsCostOfClassification()){
			*m_pOutStream << "Total cost:\t $" << costOfClassification + costOfAttrs << endl;
		}
		
		if(inducer->IsCostOfClassification() || inducer->IsCostOfAttrs()){
			TiXmlElement MLPResulElement4("MLPResult");
			MLPResulElement4.SetAttribute("Name","cost");
			ss.str("");
			ss<<cost;
			MLPResulElement4.SetAttribute("Value",ss.str());
			testElement.InsertEndChild(MLPResulElement4);
		}
		*m_pOutStream<<"_______________________________________________________________"<<endl;

		// Zwolnienie klasyfikatora
		delete iClassifier;
		
		AddXMLParams(testElement,inducer);

	}else{
		CLogger::LogError("NULL Classifier returned by IInducer::Train()",__FILE__,__LINE__);
		return false;
	}

	
return true;	
}

bool CTestClassificationExperiment::RunTestXML(TiXmlElement &testElement, string outputFolder, string filestem, deque<const IDataSet*> deqtrainDS, IInducer* inducer, const IDataSet *testDS){
	CLogger::LogError("This method cannot be called from this class",__FILE__,__LINE__);
	return false;
}

