/**
 * $Revision: 1.38 $
 * $Date: 2006/04/19 23:53:46 $
 */
 
#include "common.h" 
#include "IEvolutionaryAlg.h"
#include "Logger.h"
#include "MLPStatistics.h"
#include <sstream>
#include <iomanip>
#include <algorithm>
#include <cmath>
#include <fstream>
#include <sys/types.h>
#include <sys/stat.h>
#include <ctime>

#ifdef __linux__
#include <sys/time.h>
#include <unistd.h>
#endif

IEvolutionaryAlg::IEvolutionaryAlg(void){
	m_bFitnessLog	= false;
	m_uSizeOfPop = 100;
	m_uMaxIterations = 1000;
	m_uMaxIdleIterations		=	200;
	m_iSelection = 1;
	f_ptSelection = &IEvolutionaryAlg::RankingLinearSelection;
	m_dSelectivePressure = 1.5;
	m_uMinIterations = 1;
	m_iProbabilityOfMutation	=	5;
	m_iProbabilityOfCrossOver	=	80;
	m_BestIndividual = 0;
	m_vdDist.resize(m_uSizeOfPop);
	m_vtMulti.resize(m_uSizeOfPop);
}

IEvolutionaryAlg::~IEvolutionaryAlg(void){
	if(m_BestIndividual) delete m_BestIndividual;
}

bool IEvolutionaryAlg::SetParam(string sParamName, string sParamValue){
	sParamName=trimS(sParamName);
	sParamName=tolowerS(sParamName);
	sParamValue=trimS(sParamValue);
	sParamValue=tolowerS(sParamValue);

	if(sParamName=="fitnesslog"){
		if(sParamValue=="yes"){
			m_bFitnessLog = true;
			return true;
		}
		if(sParamValue=="no"){
			m_bFitnessLog = false;
			return true;
		}
		return false;
	}
	if(sParamName=="sizeofpopulation"){
		m_uSizeOfPop = (unsigned)atoi(sParamValue.c_str());
		if(m_uSizeOfPop<1){
			m_uSizeOfPop = 1;
			return false;
		}else{
			m_vdDist.resize(m_uSizeOfPop);
			m_vtMulti.resize(m_uSizeOfPop);
			return true;
		}
	}
	if(sParamName=="maximumiterations"){
		m_uMaxIterations = (unsigned)atoi(sParamValue.c_str());
		if(m_uMaxIterations<1){
			m_uMaxIterations = 1;
			return false;
		}else{
			return true;
		}
	}
	if(sParamName=="minimumiterations"){
		m_uMinIterations = (unsigned)atoi(sParamValue.c_str());
		if(m_uMinIterations<1){
			m_uMinIterations = 1;
			return false;
		}else{
			return true;
		}
	}
	if(sParamName=="selection"){
		if(sParamValue=="roulettewheelselection"){
			m_iSelection = 0;
			f_ptSelection = &IEvolutionaryAlg::RouletteWheelSelection;
			return true;
		}
		if(sParamValue=="rankinglinearselection"){
			m_iSelection = 1;
			f_ptSelection = &IEvolutionaryAlg::RankingLinearSelection;
			return true;
		}
		return false;
	}
	if(sParamName=="probabilityofmutation"){
		m_iProbabilityOfMutation = (int)atoi(sParamValue.c_str());
		if(m_iProbabilityOfMutation < 0 || m_iProbabilityOfMutation > 100){
			m_iProbabilityOfMutation = 5;
			return false;
		}else{
			return true;
		}
	}
	if(sParamName=="selectionpressure"){
		m_dSelectivePressure = (double)atof(sParamValue.c_str());
		if(m_dSelectivePressure < 1.0 || m_dSelectivePressure > 2.0){
			m_dSelectivePressure = 1.5;
			return false;
		}else{
			return true;
		}
	}
	if(sParamName=="probabilityofcrossover"){
		m_iProbabilityOfCrossOver = (int)atoi(sParamValue.c_str());
		if(m_iProbabilityOfCrossOver < 0 || m_iProbabilityOfCrossOver > 100){
			m_iProbabilityOfCrossOver = 25;
			return false;
		}else{
			return true;
		}
	}
	if(sParamName=="maximumofidleiterations"){
		m_uMaxIdleIterations = (unsigned)atoi(sParamValue.c_str());
		if(m_uMaxIdleIterations < 0){
			m_uMaxIdleIterations = 10000;
			return false;
		}else{
			return true;
		}
	}
	return false;
}

vector<string> IEvolutionaryAlg::GetParams() const{
vector<string> res;
stringstream ss;

	res.push_back("fitnesslog");
	if(m_bFitnessLog){
		res.push_back("yes");
	}else{
		res.push_back("no");
	}
	res.push_back("sizeofpopulation");
	ss.str(""); ss<<m_uSizeOfPop;
	res.push_back(ss.str());
	res.push_back("maximumiterations");
	ss.str(""); ss<<m_uMaxIterations;
	res.push_back(ss.str());
	res.push_back("minimumiterations");
	ss.str(""); ss<<m_uMinIterations;
	res.push_back(ss.str());
	res.push_back("selection");
	if(m_iSelection==0){
		res.push_back("roulette wheel selection");
	}else{
		if(m_iSelection==1){
			res.push_back("ranking linear selection");
		}
	}
	res.push_back("probabilityofmutation");
	ss.str(""); ss<<m_iProbabilityOfMutation;
	res.push_back(ss.str());
	res.push_back("selectionpressure");
	ss.str(""); ss<<m_dSelectivePressure;
	res.push_back(ss.str());
	res.push_back("probabilityofcrossover");
	ss.str(""); ss<<m_iProbabilityOfCrossOver;
	res.push_back(ss.str());
	res.push_back("maximumofidleiterations");
	ss.str(""); ss<<m_uMaxIdleIterations;
	res.push_back(ss.str());
	
	return res;
}

void IEvolutionaryAlg::PrintXML(ostream &ostr, int nsp, string sParamName){
	ostr.fill(' ');
	ostr<<setw(nsp)<<" "<<"<"<<sParamName<<" Name=\"fitnesslog\" Value=\"no\" />"<<endl;
	ostr<<setw(nsp)<<" "<<"<"<<sParamName<<" Name=\"sizeofpopulation\" Value=\"50\" />"<<endl;
	ostr<<setw(nsp)<<" "<<"<"<<sParamName<<" Name=\"maximumiterations\" Value=\"1000\" />"<<endl;
	ostr<<setw(nsp)<<" "<<"<"<<sParamName<<" Name=\"minimumiterations\" Value=\"1\" />"<<endl;
	ostr<<setw(nsp)<<" "<<"<"<<sParamName<<" Name=\"selection\" Value=\"ranking linear selection\" />"<<endl;
	ostr<<setw(nsp)<<" "<<"<"<<sParamName<<" Name=\"probabilityofmutation\" Value=\"5\" />"<<endl;
	ostr<<setw(nsp)<<" "<<"<"<<sParamName<<" Name=\"selectionpressure\" Value=\"1.2\" />"<<endl;
	ostr<<setw(nsp)<<" "<<"<"<<sParamName<<" Name=\"probabilityofcrossover\" Value=\"20\" />"<<endl;
	ostr<<setw(nsp)<<" "<<"<"<<sParamName<<" Name=\"maximumofidleiterations\" Value=\"200\" />"<<endl;
	ostr.fill(' ');
}

void IEvolutionaryAlg::PrintParamsXML(ostream &ostr, int nsp){
	// Inaczej wywola sie metoda z klasy podrzednej.
	IEvolutionaryAlg::PrintXML(ostr,nsp,"MLPParam");
}

void IEvolutionaryAlg::PrintSubParamsXML(ostream &ostr, int nsp){
	// Inaczej wywola sie metoda z klasy podrzednej.
	IEvolutionaryAlg::PrintXML(ostr,nsp,"MLPSubParam");
}

void IEvolutionaryAlg::RouletteWheelSelection(deque<IIndividual*> &p){

	if(p.empty()){
		CLogger::LogError("Population is empty.",__FILE__,__LINE__);
		return;
	}

	deque<IIndividual*>::iterator min_ = min_element(p.begin(),p.end(),IIndividualLess);

	// Czynnik, ktory przesunie wartosci dopasowania, tak aby nie mialy wartosci
	// ujemnych.
	double fp = (*min_)->m_dFitness;
	if(fp>0){
		fp=0;
	}
	size_t pop_size = p.size();
		
	m_vdDist[0] = p[0]->m_dFitness + fp;
	double F = m_vdDist[0];
	for(size_t i=1; i<pop_size; i++){
		F += p[i]->m_dFitness + fp;
		m_vdDist[i] = (m_vdDist[i-1] + p[i]->m_dFitness + fp);
	}
	for(size_t i=0; i<pop_size; i++){
		m_vdDist[i] /= F;
	}

	if(fabs(m_vdDist.back()-1.0)>0.02){
		CLogger::LogError("Error in probability values",__FILE__,__LINE__);
	}
	m_vdDist.back()=1.0;
	
	SelectIndividuals(p);
}

void IEvolutionaryAlg::RankingLinearSelection(deque<IIndividual*> &p){

	if(p.size()<2){
		CLogger::LogError("Population is empty.",__FILE__,__LINE__);
		return;
	}

	// sortujemy populacje malejaco (najlepsze osobniki sa na poczatku)
	// takie sortowanie dla przypadku, kiedy m_dFitness jest maksymalizowana
	sort(p.begin(),p.end(),IIndividualGreater);
	
	/*
		for(size_t x=0; x<p.size(); x++)
			cerr<<p[x]->m_dFitness<<" ";
		}
		cerr<<endl;
	*/
	
	size_t pop_size = p.size();
	double q = m_dSelectivePressure * 1.0/(double)pop_size;
	double r = 2.0* q / ((double)pop_size-1.0) - 2.0/((double)pop_size*((double)pop_size-1.0));
	m_vdDist[0] = q;
	for(size_t i=1; i<pop_size; i++){
		m_vdDist[i] = (m_vdDist[i-1] + q - i * r);
	}

	
	if(abs(m_vdDist.back()-1.0)>0.02){
		CLogger::LogError("Error in probability values",__FILE__,__LINE__);
	}
	m_vdDist.back()=1.0;

	/*
		for(size_t k=0; k<pop_size; k++)
			cerr<<m_vdDist[k]<<endl;
		}
		cerr<<endl<<endl;
	*/
	
	SelectIndividuals(p);
}

void IEvolutionaryAlg::SelectIndividuals(deque<IIndividual*> &p){
	for(size_t t=0; t<m_vtMulti.size(); t++){
		m_vtMulti[t] = 0;
	}
	for(unsigned i=0; i<m_uSizeOfPop - 1; i++){
		size_t ind = 0;
		double rnd = CMLPStatistics::Rand_01();
		if(rnd <= m_vdDist[0]){
			ind = 0;
		}else{
			size_t t=0;
			while(t<m_vdDist.size() && rnd>m_vdDist[t]){
				t++;
			}
			ind = t;
		}
		if(ind>=p.size()){
			ind = p.size() - 1;
		}
		m_vtMulti[ind]++;
	}

	// Powielenie wylosowanych i wstawienie NULL w pozycjach nie wylosowanych
	// Rozmiar populacji trzeba zapamietac, bo bedzie ulegal on zmianie w ponizszej petli
	size_t tSize = p.size();
	for(size_t i=0; i<tSize; i++){
		if(m_vtMulti[i] == 0){
			delete p[i];
			p[i] = 0;
		}else{
			for(size_t m=1; m<m_vtMulti[i]; m++ ){
				IIndividual* temp = p[i]->Clone();
				p.push_back(temp);
			}
		}
	}
	
	// Usuniecie z listy tych pozycji, ktore nie przechodza do kolejnej populacji
	deque<IIndividual*>::iterator newend = remove(p.begin(),p.end(),(IIndividual*)0);
	p.erase(newend,p.end());

	
}

void IEvolutionaryAlg::SelectIndividualsToReproduce(){
	(this->*f_ptSelection)(m_dqPopulation);
	if(m_BestIndividual){
		IIndividual *temp = m_BestIndividual->Clone();
		m_dqPopulation.push_front(temp);
	}
	random_shuffle(m_dqPopulation.begin(),m_dqPopulation.end(), Rand);
	return;
}

void IEvolutionaryAlg::EvaluateFitness(deque<IIndividual*> &population){
	/*
	 * evaluate each individual's fitness
	 */
	for(size_t k=0; k<population.size(); k++){
		EvaluateIndividualsFitness(population[k]);
	}
}

IIndividual* IEvolutionaryAlg::Run(const string sResultsFileStem, int iDebugLevel, ofstream &ofLogStream){
#ifdef __linux__
	long double	loopTime = 0.0;
#else
	time_t	loopTime=0;
#endif
	/**
	 * Strumien, do ktorego sa drukowane informacje o zmianach funkcji fitness i wartosci, od
	 * ktorych ona zalezy
	 */
	ofstream	ofsFitnessLog;
	long		n = m_uMaxIterations;
	unsigned	idle = 0;
	unsigned	maxit = 0;

// ------------------------------- OP TEST BEGIN ----------------------------
// Utworzenie plikow danych do przeprowadzenia testu operatorow. Tworzenie projektow .plt dla gnuplota.
#ifdef MLPEAOPERATORSTEST
	OPTEST_OpenFiles(sResultsFileStem);
#endif
// ------------------------------- OP TEST END ----------------------------

	// Jesli jest logowanie dopasowania to tworzony jest plik danych na te informacje oraz 
	// plik .plt dla gnuplota
	if(m_bFitnessLog){
		string fname = sResultsFileStem+"_FLOG"+".dat";
		ofsFitnessLog.open(fname.c_str(),ofstream::out);
		PrintFitnessLogHeader(ofsFitnessLog);
		CreateGnuplotProjects(fname,sResultsFileStem);
	}
	
	stringstream msg;
	
	// Utworzenie populacji poczatkowej.
	CreateInitialPopulation();
	
	// Obliczenie dopasowania populacji poczatkowej.
	EvaluateFitness(m_dqPopulation);
	
	// Znajdowanie osobnika z maksymalnym dopasowaniem.
	deque<IIndividual*>::iterator maxind = max_element(m_dqPopulation.begin(),m_dqPopulation.end(),IIndividualLess);
	m_BestIndividual = (*maxind)->Clone();
	
// ------------------------------- OP TEST BEGIN ----------------------------
#ifdef MLPEAOPERATORSTEST
	m_dMaxFitness = m_BestIndividual->m_dFitness;
#endif
// ------------------------------- OP TEST END ------------------------------

	// Zapisanie do logow informacji o najlepszym osobniku z populacji poczatkowej
	if(iDebugLevel>=1){
		PrintIndividualToLog("Initial fitness",m_BestIndividual,true);
	}

	// Poczatkowa populacja do FitnessLog
	if(ofsFitnessLog.is_open()){
		PrintFitnessLog(m_dqPopulation,m_BestIndividual,0,ofsFitnessLog);
	}

// ____________________________________________________________________________ 
// Glowna petla algorytmu ewolucyjnego
	do{
#ifdef __linux__
		timeval tim;
		gettimeofday(&tim,NULL);
		long double t1=tim.tv_sec+(tim.tv_usec/1000000.0);
#else
		time_t start = time(NULL);
#endif
		SelectIndividualsToReproduce();
		CrossoverOperator();
		MutationOperator(m_uMaxIterations-n+1);
		PostEAOperations(m_dqPopulation);
		if(!m_dqPopulation.empty()){
			deque<IIndividual*>::iterator maxind = max_element(m_dqPopulation.begin(),m_dqPopulation.end(),IIndividualLess);
			if(m_BestIndividual->m_dFitness < (*maxind)->m_dFitness){
				delete m_BestIndividual;
				m_BestIndividual = (*maxind)->Clone();
				maxit = m_uMaxIterations-n+1;
				if(iDebugLevel>=1){
					msg.str("");
					msg<<"Better fitness found in "<<m_uMaxIterations-n+1<<" iteration";
					PrintIndividualToLog(msg.str(),m_BestIndividual,true);
				}
				idle=0;
			}
		}
// ------------------------------- OP TEST BEGIN ----------------------------
#ifdef MLPEAOPERATORSTEST
		m_dMaxFitness = m_BestIndividual->m_dFitness;
#endif	
// ------------------------------- OP TEST END ------------------------------

#ifdef __linux__
		gettimeofday(&tim, NULL);
		long double t2=tim.tv_sec+(tim.tv_usec/1000000.0);
		loopTime += t2-t1;
#else
		time_t stop = time(NULL);
		loopTime += stop - start;
#endif
		// Biezaca populacja do FitnessLog
		if(ofsFitnessLog.is_open()){
			PrintFitnessLog(m_dqPopulation,m_BestIndividual,m_uMaxIterations-n+1,ofsFitnessLog);
		}
		
		if(iDebugLevel>=2){
			stringstream msg;
			msg<<"Iteration number: "<<m_uMaxIterations-n+1<<" finished";
			CLogger::LogEvent(msg.str(),"",__LINE__,ofLogStream);
		}
		
		if(m_uMinIterations < m_uMaxIterations-n+1){//0MIN
			if(++idle>=m_uMaxIdleIterations){
				if(iDebugLevel>=1){
					stringstream msg;
					msg<<"Training process stopped with "<<idle<<" idle iterations.";
					CLogger::LogEvent(msg.str(),"",__LINE__,ofLogStream);
				}
// ------------------------------- OP TEST BEGIN ----------------------------
#ifdef MLPEAOPERATORSTEST
					OPTEST_PrintOperatorsTestLog(m_uMaxIterations-n+1);
#endif
// ------------------------------- OP TEST END ------------------------------
				break;
			}
		}//0MIN

// ------------------------------- OP TEST BEGIN ----------------------------
#ifdef MLPEAOPERATORSTEST
	OPTEST_PrintOperatorsTestLog(m_uMaxIterations-n+1);
#endif
// ------------------------------- OP TEST END ------------------------------

	}while(--n>0);
// ____________________________________________________________________________ 
// Koniec petli glownej algorytmu ewolucyjnego
	
	if(ofsFitnessLog.is_open()){
		ofsFitnessLog.close();
	}

// ------------------------------- OP TEST BEGIN ----------------------------
#ifdef MLPEAOPERATORSTEST
	OPTEST_PrintOperatorsNames();
	m_ofsOperatorTest.close();
	m_ofsOperatorTestBestIndv.close();
	m_ofsMutationCounting.close();
#endif
// ------------------------------- OP TEST END ------------------------------

	// Average loop time
	msg.str("");
	msg<<"Average loop time ";
#ifdef __linux__
	msg<<"[gettimeofday()]";
	long double avrgLoopTime = loopTime / (long double)(m_uMaxIterations-n+1);
#else
	msg<<"[time()]";
	double avrgLoopTime = (double)loopTime / (double)(m_uMaxIterations-n+1);
#endif
	msg<<": "<<avrgLoopTime<<" [sec]";
	CLogger::LogEvent(msg.str(),"",__LINE__,ofLogStream);

	// Wypisanie informacji o najlepszym znalezionym osobniku
	if(iDebugLevel>=1){
		stringstream msg;
		msg<<"The best individual found in "<<maxit<<" iteration";
		PrintIndividualToLog(msg.str(),m_BestIndividual);
		msg.str("");
		msg<<"The best fitness: "<<m_BestIndividual->m_dFitness;
		CLogger::LogEvent(msg.str(),"",__LINE__,ofLogStream);
	}
	
	// Zwolnienie populacji
	for(size_t i=0; i<m_dqPopulation.size(); i++){
		delete m_dqPopulation[i];
	}
	m_dqPopulation.clear();
	
	// Zwrocenie najlepszego osobnika
	if(m_BestIndividual){
		ImproveTheBestIndividual(m_BestIndividual);
		IIndividual* res=m_BestIndividual;
		m_BestIndividual=0;
		// zwracany jest wskaznik, ktory funkcja wolajaca musi usunac
		return res;
	}
	return 0;
}

// ------------------------------- OP TEST BEGIN ----------------------------
#ifdef MLPEAOPERATORSTEST
void IEvolutionaryAlg::OPTEST_PrintOperatorsTestLog(unsigned iterNum){
	// Drukowanie testu operatorow na biezacaej populacji - dla calej populacji
	m_ofsOperatorTest<<iterNum;
	for(size_t t = 0; t<m_vOpNames.size(); t++){
		// Drukowanie wynikow w porzadku operatorow zadanym przez wektro m_vOpNames
		map<string,pair<int,int> >::iterator iter = m_mOpEfficiency.find(m_vOpNames[t]);
		if(iter != m_mOpEfficiency.end()){
			m_ofsOperatorTest<<"\t"<<iter->second.first<<"\t"<<iter->second.second;
			// Od razu zerowanie licznikow.
			iter->second.first = 0;
			iter->second.second = 0;
		}else{
			CLogger::LogError("OPERATOR TEST ERROR",__FILE__,__LINE__);
		}
	}
	m_ofsOperatorTest<<endl;
	// Drukowanie testu operatorow na biezacaej populacji - dla najlepszego osobnika
	m_ofsOperatorTestBestIndv<<iterNum;
	for(size_t t = 0; t<m_vOpNames.size(); t++){
		// Drukowanie wynikow w porzadku operatorow zadanym przez wektro m_vOpNames	
		map<string,pair<int,int> >::iterator iter2 = m_mOpEfficiencyBestIndv.find(m_vOpNames[t]);
		if(iter2 != m_mOpEfficiencyBestIndv.end()){
			m_ofsOperatorTestBestIndv<<"\t"<<iter2->second.first<<"\t"<<iter2->second.second;
			// Od razu zerowanie licznikow.
			iter2->second.first = 0;
			iter2->second.second = 0;
		}else{
			CLogger::LogError("OPERATOR TEST ERROR",__FILE__,__LINE__);
		}
	}
	m_ofsOperatorTestBestIndv<<endl;
}

void IEvolutionaryAlg::OPTEST_PrintOperatorsNames(){
	m_ofsOperatorTest<<"# Iter.";
	for(size_t t = 0; t<m_vOpNames.size(); t++){
		// Drukowanie wynikow w porzadku operatorow zadanym przez wektro m_vOpNames
		map<string,pair<int,int> >::iterator iter = m_mOpEfficiency.find(m_vOpNames[t]);
		if(iter != m_mOpEfficiency.end()){
			m_ofsOperatorTest<<"\t"<<iter->first<<"\t"<<iter->first<<" - ALL";
		}else{
			CLogger::LogError("OPERATOR TEST ERROR",__FILE__,__LINE__);
		}
	}
	m_ofsOperatorTest<<endl;
	//--------------------------------------------------------------------------------------------------------
	m_ofsOperatorTestBestIndv<<"# Iter.";
	for(size_t t = 0; t<m_vOpNames.size(); t++){
		// Drukowanie wynikow w porzadku operatorow zadanym przez wektro m_vOpNames	
		map<string,pair<int,int> >::iterator iter2 = m_mOpEfficiencyBestIndv.find(m_vOpNames[t]);
		if(iter2 != m_mOpEfficiencyBestIndv.end()){
			m_ofsOperatorTestBestIndv<<"\t"<<iter2->first<<"\t"<<iter2->first<<" - ALL";
		}else{
			CLogger::LogError("OPERATOR TEST ERROR",__FILE__,__LINE__);
		}
	}
	m_ofsOperatorTestBestIndv<<endl;
}

void IEvolutionaryAlg::OPTEST_OpenFiles(string sFilestem){
	string fname = sFilestem+"_OPTEST"+".dat";
	m_ofsOperatorTest.open(fname.c_str(),ofstream::out);
	CLogger::TimeStamp("# This file is generated by Machine Learning Platform ",m_ofsOperatorTest);
	string fname2 = sFilestem+"_THEBEST"+".dat";
	m_ofsOperatorTestBestIndv.open(fname2.c_str(),ofstream::out);
	CLogger::TimeStamp("# This file is generated by Machine Learning Platform ",m_ofsOperatorTestBestIndv);
	string fname3 = sFilestem+"_MUTATION"+".dat";
	m_ofsMutationCounting.open(fname3.c_str(),ofstream::out);
	CLogger::TimeStamp("# This file is generated by Machine Learning Platform ",m_ofsMutationCounting);
	OPTEST_CreateGnuplotProjects(fname,fname2,fname3,sFilestem+"_OPTEST",sFilestem+"_THEBEST",sFilestem+"_MUTATION");
}

void IEvolutionaryAlg::OPTEST_WholePopulationGnuplotProject(string sOpTestsFileData, string sOpTestsPltFileStem, int opIndex,int iFirstIndex){
	vector<string> params = GetParams();
	stringstream ssParams;
	for(size_t i=0; i<params.size()-1; i+=2){
		ssParams<<"["<<params[i]<<"="<<params[i+1]<<"]";
		if(i%3==0){
			ssParams<<"\\n";
		}
	}

	ofstream pltFile;
	stringstream ssfName;
	ssfName<<sOpTestsPltFileStem<<"_"<<opIndex<<".plt";
	pltFile.open(ssfName.str().c_str(),ofstream::out);
	pltFile<<"set terminal png"<<endl;
	pltFile<<"set title \"Whole population - "<<ssParams.str()<<"\""<<endl;
	pltFile<<"set y2label \" \""<<endl;
	pltFile<<"set ytics mirror"<<endl;
	pltFile<<"set noy2tics"<<endl;
	string pngName = sOpTestsFileData;
	string::size_type pos = pngName.find(".dat");
	if(pos != string::npos){
		pngName = pngName.substr(0,pos);
	}
	pltFile<<"set output '"<<pngName<<"_"<<opIndex<<".png'"<<endl;
	pltFile<<"plot \""<<sOpTestsFileData<<"\" using 1:"<<iFirstIndex+1<<" title '"<<m_vOpNames[opIndex]<<"' with lines, \\"<<endl;
	pltFile<<"\t\""<<sOpTestsFileData<<"\" using 1:"<<iFirstIndex<<" title '"<<m_vOpNames[opIndex]<<" - ALL' with lines"<<endl<<endl;
	pltFile.close();
}

void IEvolutionaryAlg::OPTEST_TheBestGnuplotProject(string sBestFileData, string sBestPltFileStem, int opIndex,int iFirstIndex){
	vector<string> params = GetParams();
	stringstream ssParams;
	for(size_t i=0; i<params.size()-1; i+=2){
		ssParams<<"["<<params[i]<<"="<<params[i+1]<<"]";
		if(i%3==0){
			ssParams<<"\\n";
		}
	}

	ofstream pltFile;
	stringstream ssfName;
	ssfName<<sBestPltFileStem<<"_"<<opIndex<<".plt";
	pltFile.open(ssfName.str().c_str(),ofstream::out);
	pltFile<<"set terminal png"<<endl;
	pltFile<<"set title \"The best individual - "<<ssParams.str()<<"\""<<endl;
	pltFile<<"set y2label \" \""<<endl;
	pltFile<<"set ytics mirror"<<endl;
	pltFile<<"set noy2tics"<<endl;
	string pngName = sBestFileData;
	string::size_type pos = pngName.find(".dat");
	if(pos != string::npos){
		pngName = pngName.substr(0,pos);
	}
	pltFile<<"set output '"<<pngName<<"_"<<opIndex<<".png'"<<endl;
	pltFile<<"plot \""<<sBestFileData<<"\" using 1:"<<iFirstIndex+1<<" title '"<<m_vOpNames[opIndex]<<"' with lines, \\"<<endl;
	pltFile<<"\t\""<<sBestFileData<<"\" using 1:"<<iFirstIndex<<" title '"<<m_vOpNames[opIndex]<<" - ALL' with lines"<<endl<<endl;
	pltFile.close();
}

void IEvolutionaryAlg::OPTEST_MutationGnuplotProject(string sMutationFileData, string sMutationPltFileStem){
	vector<string> params = GetParams();
	stringstream ssParams;
	for(size_t i=0; i<params.size()-1; i+=2){
		ssParams<<"["<<params[i]<<"="<<params[i+1]<<"]";
		if(i%3==0){
			ssParams<<"\\n";
		}
	}

	ofstream pltFile;
	stringstream ssfName;
	ssfName<<sMutationPltFileStem<<".plt";
	pltFile.open(ssfName.str().c_str(),ofstream::out);
	pltFile<<"set terminal png"<<endl;
	pltFile<<"set title \"Mutation - "<<ssParams.str()<<"\""<<endl;
	pltFile<<"set y2label \" \""<<endl;
	pltFile<<"set ytics mirror"<<endl;
	pltFile<<"set noy2tics"<<endl;
	string pngName = sMutationFileData;
	string::size_type pos = pngName.find(".dat");
	if(pos != string::npos){
		pngName = pngName.substr(0,pos);
	}
	pltFile<<"set output '"<<pngName<<".png'"<<endl;
	pltFile<<"plot \""<<sMutationFileData<<"\" using 1:2 title 'mutated' with lines, \\"<<endl;
	pltFile<<"\t\""<<sMutationFileData<<"\" using 1:3 title 'successfully' with lines"<<endl<<endl;
	pltFile.close();
}

void IEvolutionaryAlg::OPTEST_CreateGnuplotProjects(string sOpTestsFileData, string sBestFileData, string sMutationFileData, string sOpTestsPltFileStem, string sBestPltFileStem, string sMutationPltFileStem){
	for(size_t t=0; t<m_vOpNames.size(); t++){
		OPTEST_WholePopulationGnuplotProject(sOpTestsFileData,sOpTestsPltFileStem,t,(t+1)*2);
	}
	for(size_t t=0; t<m_vOpNames.size(); t++){
		OPTEST_TheBestGnuplotProject(sBestFileData,sBestPltFileStem,t,(t+1)*2);
	}
	OPTEST_MutationGnuplotProject(sMutationFileData,sMutationPltFileStem);
}

#endif
// ------------------------------- OP TEST END ------------------------------

void IEvolutionaryAlg::CreateInitialPopulation(){
	PreCreateInitialPopulation();
	ActualCreateInitialPopulation();
	PostCreateInitialPopulation();
}

