#ifndef __CMLPSTATISTICS__
#define __CMLPSTATISTICS__

#include <ctime>
#include <vector>

using namespace std;

/** Klasa z podstawowymi funkcjami statystyki.
 * @author Marek Grze
 */
class CMLPStatistics
{
public:
	/** Metoda zwraca dla współczynnika ufnoci 25% górny próg przeciału ufnoci
	 * wskanika struktury
	 * rozkładu Bernouliego.
	 * @param E - zaobserwowana liczba sukcesów
	 * @param N - liczba obserwacji
	 * @return górny próg przedziału ufnoci
	 */
	static double UCF(double E, double N);

	/** Decyzja o zajciu zdarzenia, którego p-o wynosi percent.
	 * @param percent - prawdopodobienstwo w procentach
	 * @return true kiedy zdarzenie zachodzi
	 */
	static bool Event(int percent);

	/** Zwraca wartoć losowš z przedzialu obustronnie domkniętego [0,max-1].
	 */
	static int RandVal(int max);

	/** Zwraca losowš liczbę w przedziale [0,1]. */
	static double Rand_01();

	/** Wyliczenie wartoci redniej elementów wektora. */
	template<class T> T MLPAverage(vector<T> &v){
		if(v.empty()) return (T)-1;
		T res = 0;
		for(size_t t=0; t<v.size(); t++)
			res += v[t];
		res /= (T)(v.size());
		return res;
	}

	/** Wyliczenie wariancji elementów wektora. */
	template<class T> T MLPVariance(vector<T> &v){
		if(v.empty()) return (T)-1;
		T res = 0;
		T average = MLPAverage<T>(v);
		for(size_t t=0; t<v.size(); t++)
			res += (v[t]-average)*(v[t]-average);
		res /= (T)(v.size());
		return res;
	}

	CMLPStatistics(void);
	virtual ~CMLPStatistics(void);
};

#endif
