#include "DTClassifier.h"
#include "../MLPDT/ObliqueTest.h"
#include <typeinfo>

IAttrValue* CDTClassifier::Classify(CDataObject *instance){
	return RecursiveSearch(m_pCTreeRootNode,instance);
}

IAttrValue* CDTClassifier::RecursiveSearch(CDTreeNode *node, CDataObject *instance){
	assert(node);
	if(node->m_bLeaf){
		if(node->m_pIAClass == NULL){
			CLogger::LogError("Class atrribute in CTDIDTTree leaf is NULL",__FILE__,__LINE__);
		}
		return node->m_pIAClass;
	}
	else{
		assert(node->m_pITest);
		unsigned id = node->m_pITest->GetBranchID(instance);
		node=node->m_vBranch[id];
		return RecursiveSearch(node,instance);
	}		
}


void CDTClassifier::PrintSpaces(int n, ostream &ostr){
	for(int i=0; i<n; i++){ 
		if(i%3==0)
			ostr<<"|";
		else
			ostr << " ";
	}
}

void CDTClassifier::DisplayStruct(ostream &ostr, int dispformat){
	ostr<<"_______________________________________________________________"<<endl;
	ostr<<"Decision Tree:"<<endl;
	ostr<<"_______________________________________________________________"<<endl;
	PrintTree(m_pCTreeRootNode,0,ostr,dispformat);
	ostr<<"_______________________________________________________________"<<endl;
	ostr<<"Estimate error:\t"<<m_pCTreeRootNode->GetEstimateError()<<"%"<<endl;
	if(m_bCostOfAttrs){
		ostr<<"Cost of attributes:\t$"<<m_pCTreeRootNode->GetCostOfAttributesOnTrainDS()<<endl;
	}
	if(m_bCostOfClassification){
		ostr<<"Cost of classification on training data:\t$"<<m_pCTreeRootNode->GetCostOfClassificationOnTrainDS()<<endl;
	}
	if(m_bCostOfAttrs && m_bCostOfClassification){
		ostr<<"Total cost on training data:\t$"<<m_pCTreeRootNode->GetCostOfClassificationOnTrainDS()+m_pCTreeRootNode->GetCostOfAttributesOnTrainDS()<<endl;
	}
	ostr<<"Tree Size:\t"<<m_pCTreeRootNode->GetNodeCount()<<endl;
	ostr<<"_______________________________________________________________"<<endl;
}

void CDTClassifier::PrintTree(CDTreeNode *node, int level,ostream &ostr, int dispformat){
	if(node->m_bLeaf){
		node->m_pIAClass->PrintValue(ostr);
		ostr<<" ("<<node->m_dN<<"/"<<node->m_dE<<"/"<<node->m_dEstE<<")";
		ostr<<endl;
		return;
	}

	for(unsigned i=0; i<node->m_vBranch.size(); i++){
		PrintSpaces(level,ostr);
		node->m_pITest->PrintTest(i,m_pIDataset,ostr);
		if(node->m_vBranch[i]->m_bLeaf){
			ostr<<" : ";
			node->m_vBranch[i]->m_pIAClass->PrintValue(ostr);
			ostr<<" ("<<node->m_vBranch[i]->m_dN<<"/"<<node->m_vBranch[i]->m_dE<<"/"<<node->m_vBranch[i]->m_dEstE<<")";
			ostr<<endl;
		}
		else{
			//- *ostr<<" ("<<node->m_vBranch[i]->m_dN<<"/"<<node->m_vBranch[i]->m_dE<<"/"<<node->m_vBranch[i]->m_dEstE<<")";
			ostr<<endl;
			PrintTree(node->m_vBranch[i],level+3,ostr,dispformat);
		}
	}
}

string CDTClassifier::GetClassifierName(){
	return "Decision Tree Classifier";
}

CDTClassifier::CDTClassifier(CDTreeNode *root, bool bcostOfClass, bool bcostOfAttrs, const IDataSet* dataset):IClassifier(dataset,bcostOfClass,bcostOfAttrs){
	assert(dataset);
	m_pCTreeRootNode=root;
}

CDTClassifier::~CDTClassifier(){
	delete m_pCTreeRootNode;
}

CDTreeNode* CDTClassifier::GetRoot(){
	return m_pCTreeRootNode;
}

unsigned CDTClassifier::GetClassifierSize(int iResType)const{
	if(iResType==1){
		if(m_pCTreeRootNode){
			return m_pCTreeRootNode->GetNodeCount();
		}else{
			return 0;
		}
	}
	if(iResType==0){
		if(m_pCTreeRootNode){
			return m_pCTreeRootNode->GetLeafsCount();
		}else{
			return 0;
		}
	}
	CLogger::LogError("unknown size type",__FILE__,__LINE__);
	return 0;
}

void CDTClassifier::PrintGnuplotHiperplanes(ostream &ostr){
	if(m_pCTreeRootNode){
		unsigned k=0;
		PrintGnuplotHiperplanes(ostr,m_pCTreeRootNode,k);
		ostr<<endl;
	}
}

void CDTClassifier::PrintGnuplotHiperplanes(ostream &ostr, CDTreeNode* node, unsigned &k){
	if(m_pCTreeRootNode && !m_pCTreeRootNode->m_bLeaf){
		CObliqueTest* oblTest = dynamic_cast<CObliqueTest*>(node->m_pITest);
        	if(oblTest){
			vector<double> vdW = oblTest->GetHiperplaneFactors();
			if(vdW.size()==3){
				double d = vdW[1];
				if(k>0){
					ostr<<", \\"<<endl;
				}
				if(d){
					ostr<<"test"<<k<<"(x)=";
					ostr<<"x*"<<-vdW[0]/d<<"+"<<-vdW[2]/d<<", test"<<k<<"(x)";
				}else{
					if(vdW[0]){
						ostr<<"x=";
						ostr<<-vdW[2]/vdW[0];
					}
				}
			}
			k++;
		}
		for(size_t t=0; t<node->m_vBranch.size(); t++){
			PrintGnuplotHiperplanes(ostr,node->m_vBranch[t],k);
		}
	}
}

IAttrValue* CDTClassifier::Classify(CDataObject *instance, pair<double,double> &cost){
	cost.first=0.0;
	cost.second=0.0;
	vector<bool> vUsed(m_pIDataset->GetCurrentAttrCount(),false);
	return RecursiveSearch(m_pCTreeRootNode,instance,cost,vUsed);
}

IAttrValue* CDTClassifier::RecursiveSearch(CDTreeNode *node, CDataObject *instance, pair<double,double> &cost, vector<bool> vUsed){
	assert(node);
	if(node->m_bLeaf){
		if(node->m_pIAClass == NULL){
			CLogger::LogError("Class atrribute in CTDIDTTree leaf is NULL",__FILE__,__LINE__);
		}
		vUsed.clear();
		try{
			CNomAttr& predC = dynamic_cast<CNomAttr&>(*node->m_pIAClass);
			CNomAttr& trueC = dynamic_cast<CNomAttr&>(*(instance->GetClass()));
			cost.first += m_pIDataset->CostOfClassification(predC,trueC);
		}
		catch(bad_cast){
		}
		return node->m_pIAClass;
	}
	else{
		assert(node->m_pITest);
		// Metoda GetCostOfAttributes uaktualni wektor vUsed.
		cost.second += node->m_pITest->GetCostOfAttributes(m_pIDataset,vUsed);
		unsigned id = node->m_pITest->GetBranchID(instance);
		node=node->m_vBranch[id];
		return RecursiveSearch(node,instance,cost,vUsed);
	}
}
