#ifndef __CDTCLASSIFIERH__
#define __CDTCLASSIFIERH__

#include "../MLPClassification/Classifier.h"
#include "DTreeNode.h"

/**
 * Drzewo decyzyjne.
 * @author Marek Grzes'
 */
class CDTClassifier: public IClassifier{
public:
	/**
	 * Wykorzystywana do wypisania do logow wzrow na hiperplaszczyzny stanowiace
	 * testy skosne. Wypisuje tylko testy majace 2 atrybuty.
	 */
	virtual void PrintGnuplotHiperplanes(ostream &ostr);
	virtual IAttrValue* Classify(CDataObject *instance);
	virtual IAttrValue* Classify(CDataObject *instance, pair<double,double> &cost);
	virtual void DisplayStruct(ostream &ostr, int dispformat);
	virtual string GetClassifierName();
	virtual CDTreeNode* GetRoot();
	
	/**
	 * @param iResType rodzaj zwracanego wyniku. 1 - zwracany jest calkowity
	 * rozmiar dzewa. 0 - zwracana jest liczba lisci.
	 */
	virtual unsigned GetClassifierSize(int iResType=0)const;

	/**
	 * @param root - wskaznik do korzenia drzewa
	 * @param bcostOfClass - czy uczenie uwzglednialo koszt klasyfikacji
	 * @param bcostOfAttrs - czy uczenie uwzglednialo koszt testow
	 * @param dataset - zbior danych, na ktorym bylo nauczanie klasyfikatora
	 */
	CDTClassifier(CDTreeNode *root, bool bcostOfClass, bool bcostOfAttrs, const IDataSet* dataset);
	virtual ~CDTClassifier();

private:
	CDTreeNode *m_pCTreeRootNode;
	IAttrValue* RecursiveSearch(CDTreeNode *node, CDataObject *instance);
	IAttrValue* RecursiveSearch(CDTreeNode *node, CDataObject *instance, pair<double,double> &cost, vector<bool> vUsed);
	virtual void PrintTree(CDTreeNode *node, int level,ostream &ostr, int dispformat);
	virtual void PrintSpaces(int n,ostream &ostr);
	virtual void PrintGnuplotHiperplanes(ostream &ostr, CDTreeNode *node, unsigned &k);
};

#endif

