#include "DTreeNode.h"
#include "ObliqueTest.h"
#include "../MLPDataset/NomAttr.h"
#include "../MLPDataset/RealAttr.h"
#include "../MLPCommon/MLPStatistics.h"
#include "../MLPDataset/IDataSet.h"
#include <iostream>
#include <cassert>
#include <typeinfo>

using namespace std;

CDTreeNode::CDTreeNode(){
	m_bLeaf = false;
	m_pIAClass = NULL;
	m_pITest = NULL;
	m_iSubtreeSize = -1;
	m_fCostOfClassification = 0.0;
	m_fCostOfAttrs = 0.0;
	m_uNExamples = 1;
}

CDTreeNode::CDTreeNode(const CDTreeNode& templ){
	this->m_bLeaf = templ.m_bLeaf;
	this->m_dE = templ.m_dE;
	this->m_dEstE = templ.m_dEstE;
	this->m_dN = templ.m_dN;
	this->m_iSubtreeSize = templ.m_iSubtreeSize;
	this->m_fCostOfClassification = templ.m_fCostOfClassification;
	this->m_fCostOfAttrs = templ.m_fCostOfAttrs;
	this->m_uNExamples = templ.m_uNExamples;
	assert(templ.m_pIAClass);
	IAttrValue* ia = templ.m_pIAClass;
	if(ia->GetAttrType() == e_NominalAttr){
		this->m_pIAClass = new CNomAttr((CNomAttr&)*ia);
	}
	if(ia->GetAttrType() == e_RealAttr){
		this->m_pIAClass = new CRealAttr((CRealAttr&)*ia);
	}
	for(unsigned i=0; i<templ.m_vBranch.size(); i++){
		// Utworzenie kopii poddrzewa.
		CDTreeNode* temp = new CDTreeNode(*templ.m_vBranch[i]);
		this->m_vBranch.push_back(temp);
	}
	if(!templ.m_bLeaf && templ.m_pITest){
		this->m_pITest = templ.m_pITest->CloneTest();
	}else{
		this->m_pITest = NULL;
	}
	this->m_deqCI = templ.m_deqCI;
}

CDTreeNode::~CDTreeNode(){
	if(m_pITest) delete m_pITest;
	for(unsigned i=0;i<m_vBranch.size();i++)
		delete m_vBranch[i];
	m_vBranch.clear();
	if(m_pIAClass) delete m_pIAClass;
}

void CDTreeNode::ClearDeqCI(){
	m_deqCI.clear();
	for(unsigned i=0;i<m_vBranch.size();i++){
		m_vBranch[i]->ClearDeqCI();
	}
}

double CDTreeNode::GetErrorOnTestDS(){
	if(m_dN > 0)
		return m_dE/m_dN*100.0;
	else return -1;
}

double CDTreeNode::GetEstimateError(){
	if(m_dN > 0)
		return m_dEstE/m_dN*100.0;
	else return -1;
}

unsigned CDTreeNode::GetNodeCount(){
	// Wliczenie wezla biezacego
	unsigned n=1;
	// Policzenie synow wezla
	for(unsigned i=0; i<m_vBranch.size(); i++){
		n+=m_vBranch[i]->GetNodeCount();
	}
	return n;
}

unsigned CDTreeNode::GetLeafsCount()const{
	unsigned n=0;
	// Wliczenie wezla biezacego, tylko wtedy kiedy jest on lisciem
	if(m_bLeaf){
		n=1;
	}
	// Policzenie synow wezla
	for(unsigned i=0; i<m_vBranch.size(); i++){
		n+=m_vBranch[i]->GetLeafsCount();
	}
	return n;
}

unsigned CDTreeNode::GetDepth(){
	unsigned n=1;
	unsigned max=0;
	if(!m_vBranch.empty()){
		max = m_vBranch.front()->GetDepth();
	}
	for(unsigned i=1; i<m_vBranch.size(); i++){
		unsigned t = m_vBranch[i]->GetDepth();
		if(t > max) max = t;
	}
	n+=max;
	return n;
}

CDTreeNode* CDTreeNode::FindRandomLeaf(){
	if(m_bLeaf) return this;

	int n = (int)m_vBranch.size();
	// Losowanie galezi, go ktorej nastapi zejscie.
	n = CMLPStatistics::RandVal(n);
	return m_vBranch[n]->FindRandomLeaf();
}

CDTreeNode* CDTreeNode::FindRandomNodeOrLeaf(){
	vector<CDTreeNode*> v;
	return FindRandomNodeOrLeaf(v);
}

CDTreeNode* CDTreeNode::FindRandomNodeOrLeaf(vector<CDTreeNode*> &vCDT){
	if(m_bLeaf){
		int n = (int)vCDT.size();
		if(n == 0) return this;
		n = CMLPStatistics::RandVal(n+1);
		CDTreeNode* temp;
		if((size_t)n < vCDT.size()){
			temp = vCDT[n];
		}else{
			temp = this;
		}
		vCDT.clear();
		return temp;
	}else{
		vCDT.push_back(this);
		int n = (int)m_vBranch.size();
		n = CMLPStatistics::RandVal(n);
		return m_vBranch[n]->FindRandomNode(vCDT);
	}
}

CDTreeNode* CDTreeNode::FindRandomNode(){
	vector<CDTreeNode*> v;
	return FindRandomNode(v);
}

CDTreeNode* CDTreeNode::FindRandomNode(vector<CDTreeNode*> &vCDT){
	if(m_bLeaf){
		int n = (int)vCDT.size();
		if(n == 0) return NULL;
		n = CMLPStatistics::RandVal(n);
		CDTreeNode* temp = vCDT[n];
		vCDT.clear();
		return temp;
	}else{
		vCDT.push_back(this);
		int n = (int)m_vBranch.size();
		n = CMLPStatistics::RandVal(n);
		return m_vBranch[n]->FindRandomNode(vCDT);
	}
}

CDTreeNode* CDTreeNode::FindRandomNode(unsigned t){
	vector<CDTreeNode*> v;
	return FindRandomNode(t,v);
}

CDTreeNode* CDTreeNode::FindRandomNode(unsigned t, vector<CDTreeNode*> &vCDT){
	if(m_bLeaf){
		int n = (int)vCDT.size();
		if(n == 0) return NULL;
		n = CMLPStatistics::RandVal(n);
		CDTreeNode* temp = vCDT[n];
		vCDT.clear();
		return temp;
	}else{
		if(m_vBranch.size() == t){
			vCDT.push_back(this);
		}
		int n = (int)m_vBranch.size();
		n = CMLPStatistics::RandVal(n);
		return m_vBranch[n]->FindRandomNode(t,vCDT);
	}

}

CDTreeNode* CDTreeNode::FindNodeCuttingDipol(pair<CDataObject*,CDataObject*> p){
	CObliqueTest* oblTest = dynamic_cast<CObliqueTest*>(m_pITest);
	// Jesli to nie jest test skosny to zwracamy NULL
	if(oblTest==NULL){
		return NULL;
	}
	// Jesli dipol jest przecinany w tym wezle to zwracamy jego wskaznik.
	if ( oblTest->GetInnerProduct(p.first) * oblTest->GetInnerProduct(p.second) < 0 ){
		return this;
	}
	// Rekurencyjne schodzenie nizej.
	for(size_t i=0; i<m_vBranch.size(); i++){
		CDTreeNode* temp = m_vBranch[i]->FindNodeCuttingDipol(p);
		if(temp!=NULL){
			return temp;
		}
	}
	return NULL;
}	

CDTreeNode* CDTreeNode::GetNodeOrLeafEvenProb(){
	vector<CDTreeNode*> v;
	GetNodeOrLeafEvenProb(v);
	if(v.empty()){
		return NULL;
	}else{
		return v[rand()%v.size()];
	}
}

void CDTreeNode::GetNodeOrLeafEvenProb(vector<CDTreeNode*> &vCDT){
	vCDT.push_back(this);
	for(size_t t=0; t<m_vBranch.size(); t++){
		m_vBranch[t]->GetNodeOrLeafEvenProb(vCDT);
	}
}

CDTreeNode* CDTreeNode::GetLeafEvenProb(){
	vector<CDTreeNode*> v;
	GetLeafEvenProb(v);
	if(v.empty()){
		return NULL;
	}else{
		return v[rand()%v.size()];
	}
}

void CDTreeNode::GetLeafEvenProb(vector<CDTreeNode*> &vCDT){
	if(m_bLeaf){
		vCDT.push_back(this);
	}
	for(size_t t=0; t<m_vBranch.size(); t++){
		m_vBranch[t]->GetLeafEvenProb(vCDT);
	}
}

CDTreeNode* CDTreeNode::GetNodeEvenProb(){
	vector<CDTreeNode*> v;
	GetNodeEvenProb(v);
	if(v.empty()){
		return NULL;
	}else{
		return v[rand()%v.size()];
	}
}

void CDTreeNode::GetNodeEvenProb(vector<CDTreeNode*> &vCDT){
	if(!m_bLeaf){
		vCDT.push_back(this);
	}
	for(size_t t=0; t<m_vBranch.size(); t++){
		m_vBranch[t]->GetNodeEvenProb(vCDT);
	}
}

CDTreeNode* CDTreeNode::GetNodeEvenProb(unsigned uSize){
	vector<CDTreeNode*> v;
	GetNodeEvenProb(v,uSize);
	if(v.empty()){
		return NULL;
	}else{
		return v[rand()%v.size()];
	}
}

void CDTreeNode::GetNodeEvenProb(vector<CDTreeNode*> &vCDT, unsigned uSize){
	if(!m_bLeaf){
		if(m_vBranch.size() == uSize){
			vCDT.push_back(this);
		}
	}
	for(size_t t=0; t<m_vBranch.size(); t++){
		m_vBranch[t]->GetNodeEvenProb(vCDT,uSize);
	}
}
