#include "Entropy.h"
#include <map>
#include <cmath>

CEntropy::CEntropy(void){
}

CEntropy::~CEntropy(void){
}

long double CEntropy::Entropy(deque<CDataObject*> &deqCI){
map<IAttrValue*,unsigned> mapIAu;
map<IAttrValue*,unsigned>::iterator iter;
long double P;
long double info_P;

	P = deqCI.size();

	if(P == 0){
		// CLogger::LogEvent("|P|=0",__FILE__,__LINE__);
		return 0;
	}

	/* Obliczenie licznosci w ramach kazdej z klas.
	 */
	for(unsigned i=0; i < deqCI.size(); i++){
		int n=0;
		IAttrValue *pIA = deqCI[i]->GetClass();

		for(iter=mapIAu.begin(); iter!=mapIAu.end(); iter++){
			if( *(iter->first) == *pIA ){
				iter->second++;
				n=1;
				break;
			}
		}
		if(n==0){
			pair<IAttrValue*,unsigned> p;
			p.first = pIA;
			p.second = 1;
			mapIAu.insert(p);
		}
	}
	
	/* Obliczenie info(P)
	 *
	 *				  	
	 *		info(P) = sum(d in C) [ - |Pd| / |P| * log ( |Pd| / |P|  ) ]
	 */
	info_P = 0;
	for(iter=mapIAu.begin(); iter!=mapIAu.end(); iter++){
		long double wsp = (long double)iter->second/P;
		if(wsp != 0)
			info_P += - wsp * log(wsp);
	}
	return info_P;
}

long double CEntropy::Gain(deque<CDataObject*> &deqCI, ITest *pIT){
long double P;
long double info_P;
long double Et;


	P = deqCI.size();

	if(P == 0){
		//CLogger::LogEvent("|P|=0",__FILE__,__LINE__);
		return 0;
	}

	/* Entropia calego zbioru przed wykonaniem testu.
	 */ 	 
	info_P = Entropy(deqCI);

	/* Pobranie podzialu jaki wyznacza pITT.
	 */
	vector<vector<unsigned> > vvDiv = pIT->GetSubSets(deqCI);

	/* Obliczenie info_test(P) czyli entropii zbioru przykladow deqCI, ze 
	 * wzgledu na test pITT;
	 */
	Et = 0;
	for(unsigned i=0; i<vvDiv.size(); i++){
		deque<CDataObject*> subset;
		vector<unsigned> subIDs = vvDiv[i];
		for(unsigned j=0; j<subIDs.size(); j++){
			subset.push_back(deqCI[subIDs[j]]);
		}
		/* Obliczenie entropii dla i-tego przedzialu wyznaczonego przez
		 * i-ty wynik testu.
		 */
		long double Etr = Entropy(subset);

		Et += (long double) subset.size() / P * Etr;
	}

	return info_P - Et;
}

long double CEntropy::GainRatio(deque<CDataObject*> &deqCI, ITest *pIT){
long double P;
long double info_P;
long double Et;
long double split_info;


	P = deqCI.size();

	if(P == 0){
		//CLogger::LogEvent("|P|=0",__FILE__,__LINE__);
		return 0;
	}

	/* Entropia calego zbioru przed wykonaniem testu.
	 */ 	 
	info_P = Entropy(deqCI);

	/* Pobranie podzialu jaki wyznacza pITT.
	 */
	vector<vector<unsigned> > vvDiv = pIT->GetSubSets(deqCI);

	/* Obliczenie info_test(P) czyli entropii zbioru przykladow deqCI, ze 
	 * wzgledu na test pITT;
	 */
	Et = 0;
	split_info = 0;
	for(unsigned i=0; i<vvDiv.size(); i++){
		deque<CDataObject*> subset;
		vector<unsigned> subIDs = vvDiv[i];
		for(unsigned j=0; j<subIDs.size(); j++){
			subset.push_back(deqCI[subIDs[j]]);
		}
		/* Obliczenie entropii dla i-tego przedzialu wyznaczonego przez
		 * i-ty wynik testu.
		 */
		long double Etr = Entropy(subset);

		Et += (long double) subset.size() / P * Etr;

		long double wsp = (long double) subset.size() / P;

		if(wsp != 0)
			split_info += - wsp * log(wsp);
	}
	
	if(split_info == 0){
		// CLogger::LogEvent("split_info=0",__FILE__,__LINE__);
		// Jak split_info == 0 to info_p == Et. Zwracane jest 0, bo ten
		// test nie powinien byc brany pod uwage.
		return 0;
	}
	return (info_P - Et) / split_info;
}
