#include "IDTInducer.h"
#include "DTClassifier.h"
#include "RealTest.h"
#include "NominalTest.h"
#include "TestContainer2.h"
#include "../MLPCommon/MLPStatistics.h"
#include "../MLPDataset/NomAttr.h"
#include "../MLPDataset/RealAttr.h"

#include <iomanip>
#include <typeinfo>

IDTInducer::IDTInducer(void){
	m_iError	= 1;
}

IDTInducer::~IDTInducer(void){
}

IAttrValue* IDTInducer::Category(const IDataSet* dataset, deque<CDataObject*> &deqCI, IAttrValue *pIA){
// Ta zmienna mozna raz zainicjalizowac, bo ona moze miescic maksymalnie tyle elementow
// ile jest klas
static vector<sCounter*> vs(m_pDataset->GetClassCount(),(sCounter*)0);
// Liczba elementow w wektorze vs
unsigned vsSize = 0;

	if(m_bCostOfClassification){
		if(deqCI.size()>0){
			// Wektor par (klasa, koszt)
			vector<pair<IAttrValue*, double> > classCost;
			// Wyszukanie wszystkich klas
			for(unsigned int i=0;i<deqCI.size();i++){
				CDataObject *inst = deqCI[i];
				int x=0;
				for(unsigned int j=0;j<classCost.size();j++){
					IAttrValue *ia1=classCost[j].first;
					IAttrValue *ia2=inst->GetClass();
					if (ia1==NULL || ia2==NULL){
						return pIA;
					}
					if((*ia1)==(*ia2)){
						x=1;
						break;
					}
				}
				if(x==0){
					pair<IAttrValue*, double> p(inst->GetClass(),0);
					classCost.push_back(p);
				}
			}
			// Sprawdzenie jaki jest koszt klasyfikacji przy wyborze kazdej z klas
			// od razu znalezenie najlepszej klasy (czyli mininalny koszt)
			int minIndex = -1;
			double minCost = -1;
			for(size_t clID=0; clID<classCost.size(); clID++){
				IAttrValue *ia1=classCost[clID].first;
				for(unsigned int i=0;i<deqCI.size();i++){
					IAttrValue *ia2=deqCI[i]->GetClass();
					try{
						CNomAttr& predC = dynamic_cast<CNomAttr&>(*ia1);
						CNomAttr& trueC = dynamic_cast<CNomAttr&>(*ia2);
						classCost[clID].second += dataset->CostOfClassification(predC,trueC);
					}
					catch(bad_cast){
					}
					// Jezeli juz mamy wiekszy koszt niz min to nie ma co isc dalej
					// przerywamy
					if(minIndex != -1 && classCost[clID].second > minCost){
						break;
					}
				}
				if(minIndex == -1){
					minIndex = clID;
					minCost = classCost[clID].second;
				}else{
					if(classCost[clID].second < minCost){
						minIndex = clID;
						minCost = classCost[clID].second;
					}
				}
			}
			return classCost[minIndex].first;
		}else{
			return pIA;
		}
	}else{
		if(deqCI.size()>0){
			for(unsigned int i=0;i<deqCI.size();i++){
				CDataObject *inst = deqCI[i];
				int x=0;
				for(unsigned int j=0;j<vsSize;j++){
					IAttrValue *ia1=vs[j]->s_pIAttrValue;
					IAttrValue *ia2=inst->GetClass();
					if((*ia1)==(*ia2)){
						vs[j]->s_Counter++;
						x=1;
						break;
					}
				}
				if(x==0){
					sCounter *t=new sCounter;
					t->s_Counter=1;
					t->s_pIAttrValue=inst->GetClass();
					// Zakladamy ze wektor vs byl dobrze zainicjowany i nie wyjdziemy poza zakres
					vs[vsSize++]=t;
				}
			}
		
			int max = 0;
			IAttrValue* pmaxClass=NULL;
			for(unsigned int i=0;i<vsSize;i++){
				if(vs[i]->s_Counter>max){
					max = vs[i]->s_Counter;
					pmaxClass = vs[i]->s_pIAttrValue;
				}
			}
			for(unsigned int i=0; i<vsSize;i++){
				delete vs[i];
			}
			return pmaxClass;
		}else{
			return pIA;
		}
	}
}

bool IDTInducer::SetParam(string sParamName, string sParamValue){
	string::iterator i = sParamName.begin();
	while(i != sParamName.end()){
		*i = tolower(*i);
		++i;
	}
	i = sParamValue.begin();
	while(i != sParamValue.end()){
		*i = tolower(*i);
		++i;
	}

	if(sParamName=="error"){
		if(sParamValue=="pessimistic"){
			m_iError = 0;
			return true;
		}
		if(sParamValue=="retest"){
			m_iError = 1;
			return true;
		}
		return false;
	}

	if(IInducer::SetParam(sParamName,sParamValue)) return true;
	return false;
}

CDTreeNode* IDTInducer::MakeTreeLeaf(const IDataSet* dataset, deque<CDataObject*> &set, IAttrValue* ia){
	CDTreeNode * pnode = new CDTreeNode();
	IAttrValue* temp = Category(dataset,set,ia);
	if(temp->GetAttrType() == e_NominalAttr){
		pnode->m_pIAClass = new CNomAttr((CNomAttr&)*temp);
	}
	if(temp->GetAttrType() == e_RealAttr){
		pnode->m_pIAClass = new CRealAttr((CRealAttr&)*temp);
	}
	pnode->m_deqCI = set;
	pnode->m_bLeaf = true;
	return pnode;
}

void IDTInducer::PrintParamsXML(ostream &ostr, int nsp){
	IInducer::PrintParamsXML(ostr,nsp);
	ostr.fill(' ');
	ostr<<setw(nsp)<<" "<<"<MLPParam Name=\"error\" Value=\"retest\" />"<<endl;
	ostr.fill(' ');
}

vector<string> IDTInducer::GetParams() const{
vector<string> res;
stringstream ss;

	res = IInducer::GetParams();

	res.push_back("error");
	if(m_iError==0) res.push_back("pessimistic");
		else res.push_back("retest");
	
	return res;
}

