/**
 * $Revision: 1.8 $
 * $Date: 2006/05/05 06:11:15 $
 */

#include "../MLPCommon/common.h"
#include "../MLPDataset/RealAttr.h"
#include "IDTInducerAP.h"
#include "IDTInducer.h"
#include "RealTest.h"
#include <iomanip>
#include <sstream>
#include <algorithm>

IDTInducerAP::IDTInducerAP(){
	m_bInnerDisjunction = false;
}

bool IDTInducerAP::AddBranch(CNominalTestInnerDisj* pTest, CDTreeNode* pNode){
deque<size_t> dq;

	// Wrzucamy do kolejki wszystkie wartosci, zeby preferowac grupy wieksze aby to z nich losowac z wiekszym p-em 
	// wartosc do wydzielenia z oddzielna galezia
	for(size_t t=0; t<pTest->m_vtInnerDisIndex.size(); t++){
		if(search_n(pTest->m_vtInnerDisIndex.begin(),pTest->m_vtInnerDisIndex.end(),2,pTest->m_vtInnerDisIndex[t]) != pTest->m_vtInnerDisIndex.end()){
			//if(find(dq.begin(),dq.end(),pTest->m_vtInnerDisIndex[t]) == dq.end()){
			dq.push_back(t);
			//}
		}
	}
	
	// Musi wystapic co najmniej jedna grupa
	if(dq.size() > 1){
		// losujemy pozycje, dla ktorej powstanie nowa galaz (mozna od razu tu losowac, bo
		// w petli wyzej byly dodane wszystkie pozycje
		size_t pos = dq[rand()%dq.size()];
		// Galaz odpowiadajacy pozycji pos z wektora m_vtInnerDisIndex jest powielana
		CDTreeNode* newNode = new CDTreeNode(*(pNode->m_vBranch[pTest->m_vtInnerDisIndex[pos]]));
		// Powielona galaz na koniec
		pNode->m_vBranch.push_back(newNode);
		// Uaktualnienie wektora m_vtInnerDisIndex aby na pozycji pos bylo wskazanie na nowa galaz
		pTest->m_vtInnerDisIndex[pos] = pNode->m_vBranch.size()-1;
		return true;
	}

	return false;
}

bool IDTInducerAP::CutMixedDipolAP(CDataObject* ob1, CDataObject* ob2, ITest* pITest){
	CRealTest* rTest = dynamic_cast<CRealTest*>(pITest);
	assert(rTest);

	// Pobranie wartosci atrybutow, na ktorych jest ciecie
	// Tu pobieramy przez operator a nie przez metode GetRealAttrValue, bo
	// testy CRealTest maja m_iAttrNum wzgledem numeracji globalnej a nie
	// numeracji tylko dla atrybutow rzeczywistych jak w CObliqueTest
	IAttrValue* pIA1 = (*ob1)[rTest->GetAttrNum()];
	IAttrValue* pIA2 = (*ob2)[rTest->GetAttrNum()];
	assert(pIA1);
	assert(pIA2);
	CRealAttr* ref_ra1 = dynamic_cast<CRealAttr*>(pIA1);
	CRealAttr* ref_ra2 = dynamic_cast<CRealAttr*>(pIA2);
	assert(ref_ra1!=NULL);
	assert(ref_ra2!=NULL);

	// Jak sa rowne wartosci to nie szukamy innego atrybutu.
	if(ref_ra2->m_AttrValue.u_dReal == ref_ra1->m_AttrValue.u_dReal){
		return false;
	}

	// Obliczenie wsp. punktu przez, ktory trzeba poprowadzic hiperplane.
	double newThr;
	if(ref_ra2->m_AttrValue.u_dReal > ref_ra1->m_AttrValue.u_dReal){
		newThr = (ref_ra2->m_AttrValue.u_dReal - ref_ra1->m_AttrValue.u_dReal) / 2.0 + ref_ra1->m_AttrValue.u_dReal;
	}else{
		newThr = (ref_ra1->m_AttrValue.u_dReal - ref_ra2->m_AttrValue.u_dReal) / 2.0 + ref_ra2->m_AttrValue.u_dReal;
	}

	// Aktualizacja punktu ciecia
	rTest->m_dThreshold = newThr;

	//cerr<<ref_ra1->m_AttrValue.u_dReal<<" "<<ref_ra2->m_AttrValue.u_dReal<<" "<<newThr<<endl;

	return true;
}

bool IDTInducerAP::FindMixedDipolAP(deque<CDataObject*> &deqCI, CDataObject* &ob1, CDataObject* &ob2, ITest* pITest){
	CRealTest* rTest = dynamic_cast<CRealTest*>(pITest);
	assert(rTest);

	bool found=false;
	if(deqCI.size()>0){
		IAttrValue *pIA1 = deqCI[0]->GetClass();
		for(unsigned i=1; i<deqCI.size(); i++){
			IAttrValue *pIA2 = deqCI[i]->GetClass();
			if( ! ((*pIA1) == (*pIA2)) ){
				found=true;
				break;
			}
		}
	}else{
		return false;
	}
	if(!found){
		return false;
	}

	// Pierwszy losujemy
	size_t ind1 = rand()%deqCI.size();
	unsigned bid1 = rTest->GetBranchID(deqCI[ind1]);

	// Wektor indeksow obiektow z klas innych niz atrind (inizjalizujemy na maksa
	// ze wzgledow wydajnosciowych)
	vector<unsigned> vInd(deqCI.size());
	// Liczba obiektow w tym wektorze
	unsigned nVIND = 0;

	// wrzucamy do vInd indeksy tych, ktore leza po tej samej stronie threshold
	// i sa z innych klas (czyli, te ktore maja ten sam branchID)
	// zeby z nich wylosowac
	for(size_t c=0; c<deqCI.size(); c++){
		// Dla klas roznych niz atrind
		if( *(deqCI[ind1]->GetClass()) != *(deqCI[c]->GetClass()) ){
			if(rTest->GetBranchID(deqCI[c]) == bid1){
				vInd[nVIND++]=c;
			}
		}
	}
	if(nVIND == 0){
		return false;
	}
	// I tu juz losowanie z tej samej klasy lezacego po drugiej stronie
	// marginesu threshold
	size_t ind2 = vInd[rand()%nVIND];

	ob1 = deqCI[ind1];
	ob2 = deqCI[ind2];
	//cerr<<rTest->GetBranchID(deqCI[ind1])<<" "<<rTest->GetBranchID(deqCI[ind2])<<endl;
	return true;
}

bool IDTInducerAP::FindPureDipolAP(deque<CDataObject*> &deqCI, CDataObject* &ob1, CDataObject* &ob2, ITest* pITest){
	CRealTest* rTest = dynamic_cast<CRealTest*>(pITest);
	assert(rTest);

	bool found=false;
	if(deqCI.size()>0){
		IAttrValue *pIA1 = deqCI[0]->GetClass();
		for(unsigned i=1; i<deqCI.size(); i++){
			IAttrValue *pIA2 = deqCI[i]->GetClass();
			if( ! ((*pIA1) == (*pIA2)) ){
				found=true;
				break;
			}
		}
	}else{
		return false;
	}
	if(!found){
		return false;
	}
	vector<vector<CDataObject*> > splitCI(1);
	splitCI[0].push_back(deqCI[0]);
	for(size_t t=1; t<deqCI.size(); t++){
		bool found = false;
		for(size_t k=0; k<splitCI.size(); k++){
			if( *(splitCI[k].front()->GetClass()) == *(deqCI[t]->GetClass()) ){
				splitCI[k].push_back(deqCI[t]);
				found=true;
				break;
			}
		}
		if(!found){
			splitCI.resize(splitCI.size()+1);
			splitCI[splitCI.size()-1].push_back(deqCI[t]);
		}
	}
	if(!(splitCI.size()>(size_t)1)){
		return false;
	}
	// Losowanie indeksu klasy.
	size_t atrind = rand()%splitCI.size();
	if(splitCI[atrind].size()<2){
		return false;
	}
	// Pierwszy losujemy
	size_t ind1 = rand()%splitCI[atrind].size();
	unsigned bid1 = rTest->GetBranchID(splitCI[atrind][ind1]);

	// Wektor indeksow obiektow z klasy z atrind (inizjalizujemy na maksa
	// ze wzgledow wydajnosciowych)
	vector<unsigned> vInd(splitCI[atrind].size());
	// Liczba obiektow w tym wektorze
	unsigned nVIND = 0;

	// wrzucamy do vInd indeksy tych, ktore leza po drugiej stronie threshold
	// (czyli, te ktore maja inny branchID) zeby z nich wylosowac
	for(size_t i=0; i<splitCI[atrind].size(); i++){
		if(rTest->GetBranchID(splitCI[atrind][i]) != bid1){
			vInd[nVIND++]=i;
		}
	}
	if(nVIND == 0){
		return false;
	}
	// I tu juz losowanie z tej samej klasy lezacego po drugiej stronie
	// marginesu threshold
	size_t ind2 = vInd[rand()%nVIND];

	ob1 = splitCI[atrind][ind1];
	ob2 = splitCI[atrind][ind2];
	//cerr<<rTest->GetBranchID(splitCI[atrind][ind1])<<" "<<rTest->GetBranchID(splitCI[atrind][ind2])<<endl;
	return true;
}

vector<string> IDTInducerAP::GetParams() const{
vector<string> res;
stringstream ss;

	res.push_back("innerdisjunction");
	if(m_bInnerDisjunction) res.push_back("yes");
		else res.push_back("no");
	return res;
}

CDTreeNode* IDTInducerAP::JoinBranches(IDTInducer *inducer, CNominalTestInnerDisj* pTest, CDTreeNode* pNode){
deque<size_t> dq;

	// Wrzucamy do kolejki po jednej wartosci, bo inaczej bedziemy preferowac grupy
	// wieksze aby loczyly sie w jeszcze wieksze
	for(size_t t=0; t<pTest->m_vtInnerDisIndex.size(); t++){
		if(find(dq.begin(),dq.end(),pTest->m_vtInnerDisIndex[t]) == dq.end()){
			dq.push_back(pTest->m_vtInnerDisIndex[t]);
		}
	}

	// Musza pozostac co najmniej 2 galezie
	if(dq.size() > 2){
		size_t branch1 = dq[rand()%dq.size()];
		deque<size_t>::iterator it = find(dq.begin(),dq.end(),branch1);
		dq.erase(it);
		size_t branch2 = dq[rand()%dq.size()];
		CDTreeNode* node1 = pNode->m_vBranch[branch1];
		CDTreeNode* node2 = pNode->m_vBranch[branch2];
		deque<CDataObject*> subset = node1->m_deqCI;
		subset.resize(node1->m_deqCI.size()+node2->m_deqCI.size());
		for(size_t i=0; i<node2->m_deqCI.size(); i++){
			subset[node1->m_deqCI.size()+i] = node2->m_deqCI[i];
		}
		CDTreeNode* pResNode = inducer->MakeTreeLeaf(inducer->m_pDataset,subset,pNode->m_pIAClass);
		delete node1;
		delete node2;
		// Wiekszy numer laczonych galezi
		size_t bBranch;
		// Mniejszy numer laczonych galezi
		size_t lBranch;
		if(branch1<branch2){
			pNode->m_vBranch[branch1] = pResNode;
			bBranch = branch2;
			lBranch = branch1;
		}else{
			pNode->m_vBranch[branch2] = pResNode;
			bBranch = branch1;
			lBranch = branch2;
		}
		// Uaktualnie wektora podwojnego indeksowania dla wewn. dyzjunkcji
		for(size_t t=0; t<pTest->m_vtInnerDisIndex.size(); t++){
			if(pTest->m_vtInnerDisIndex[t] == bBranch){
				pTest->m_vtInnerDisIndex[t] = lBranch;
			}
		}
		
		// Skrocenie wektora galezi m_vBranch
		if(bBranch == pNode->m_vBranch.size() - 1){
			pNode->m_vBranch.pop_back();
		}else{
			// Jesli usuwany indeks nie jest ostanim to robimy przesuniecie.
			for(size_t t=bBranch; t<pNode->m_vBranch.size()-1; t++){
				pNode->m_vBranch[t] = pNode->m_vBranch[t+1];
			}
			// Zmniejszenie o 1 wartosci w indeksie, ktore sa wieksze od bBranch
			for(size_t t=0; t<pTest->m_vtInnerDisIndex.size(); t++){
				if(pTest->m_vtInnerDisIndex[t] > bBranch){
					pTest->m_vtInnerDisIndex[t]--;
				}
			}
			pNode->m_vBranch.pop_back();
		}
		return pResNode;
	}
	
	return 0;
}

bool IDTInducerAP::JoinPureDipolAP(CDataObject* ob1, CDataObject* ob2, ITest* pITest){
	CRealTest* rTest = dynamic_cast<CRealTest*>(pITest);
	assert(rTest);

	// Pobranie wartosci atrybutow, na ktorych jest ciecie
	// Tu pobieramy przez operator a nie przez metode GetRealAttrValue, bo
	// testy CRealTest maja m_iAttrNum wzgledem numeracji globalnej a nie
	// numeracji tylko dla atrybutow rzeczywistych jak w CObliqueTest
	IAttrValue* pIA1 = (*ob1)[rTest->GetAttrNum()];
	IAttrValue* pIA2 = (*ob2)[rTest->GetAttrNum()];
	assert(pIA1);
	assert(pIA2);
	CRealAttr* ref_ra1 = dynamic_cast<CRealAttr*>(pIA1);
	CRealAttr* ref_ra2 = dynamic_cast<CRealAttr*>(pIA2);
	assert(ref_ra1!=NULL);
	assert(ref_ra2!=NULL);

	// Obliczenie wsp. punktu przez, ktory trzeba poprowadzic hiperplane.
	double newThr;
	if(ref_ra2->m_AttrValue.u_dReal > ref_ra1->m_AttrValue.u_dReal){
		newThr = (ref_ra2->m_AttrValue.u_dReal - ref_ra1->m_AttrValue.u_dReal) / 2.0;
		if(rand()%2){
			newThr = ref_ra2->m_AttrValue.u_dReal + newThr;
		}else{
			newThr = ref_ra1->m_AttrValue.u_dReal - newThr;
		}
	}else{
		newThr = (ref_ra1->m_AttrValue.u_dReal - ref_ra2->m_AttrValue.u_dReal) / 2.0;
		if(rand()%2){
			newThr = ref_ra1->m_AttrValue.u_dReal + newThr;
		}else{
			newThr = ref_ra2->m_AttrValue.u_dReal - newThr;
		}
	}

	// Aktualizacja punktu ciecia
	rTest->m_dThreshold = newThr;

	//cerr<<ref_ra1->m_AttrValue.u_dReal<<" "<<ref_ra2->m_AttrValue.u_dReal<<" "<<newThr<<endl;

	return true;
}

bool IDTInducerAP::MoveValueBetweenBranches(CNominalTestInnerDisj* pTest){
deque<size_t> dq;

	// Wrzucamy do kolejki wszystkie wartosci, zeby preferowac grupy wieksze aby to z nich losowac z wiekszym p-em 
	// element do przerzucenia do inneg grupy.
	for(size_t t=0; t<pTest->m_vtInnerDisIndex.size(); t++){
		if(search_n(pTest->m_vtInnerDisIndex.begin(),pTest->m_vtInnerDisIndex.end(),2,pTest->m_vtInnerDisIndex[t]) != pTest->m_vtInnerDisIndex.end()){
			//if(find(dq.begin(),dq.end(),pTest->m_vtInnerDisIndex[t]) == dq.end()){
			dq.push_back(t);
			//}
		}
	}
	
	// Musi wystapic co najmniej jedna grupa
	if(dq.size() > 1){
		// Numer wartosci z wektora m_vtInnerDisIndex, ktora bedzie zmieniona
		size_t valPos = dq[rand()%dq.size()];
		
		vector<size_t> vt = pTest->m_vtInnerDisIndex;
		// Usuniece tych pozycji, na ktorych wartosc jest taka sama jak na pozycji valPos
		vector<size_t>::iterator newend = remove(vt.begin(),vt.end(),vt[valPos]);
		vt.erase(newend,vt.end());
	
		// Losowanie pozycji
		size_t valPos2 = vt[rand()%vt.size()];
		
		// Zamiana wartosci na pozycjaci valPos, pozycja valPos wchodzi do grupy wartosci
		// w jakiej jest pozycja valPos2
		pTest->m_vtInnerDisIndex[valPos] = pTest->m_vtInnerDisIndex[valPos2];
		return true;
	}
	
	return false;
}

void IDTInducerAP::PrintParamsXML(ostream &ostr, int nsp){
	ostr.fill(' ');
	ostr<<setw(nsp)<<" "<<"<MLPParam Name=\"innerdisjunction\" Value=\"no\" />"<<endl;
	ostr.fill(' ');
}

bool IDTInducerAP::SetParam(string sParamName, string sParamValue){
	sParamName=trimS(sParamName);
	sParamName=tolowerS(sParamName);
	sParamValue=trimS(sParamValue);
	sParamValue=tolowerS(sParamValue);

	if(sParamName=="innerdisjunction"){
		if(sParamValue=="yes"){
			m_bInnerDisjunction = true;
			return true;
		}
		if(sParamValue=="no"){
			m_bInnerDisjunction = false;
			return true;
		}
		return false;
	}
	return false;
}
