/**
 * $Revision: 1.4 $
 * $Date: 2006/04/19 07:16:15 $
 */

#ifndef __IDTINDUCERAPH__
#define __IDTINDUCERAPH__

#include <string>
#include <vector>
#include <iostream>

#include "DTreeNode.h"
#include "NominalTestInnerDisj.h"
#include "IDTInducer.h"

using namespace std;

/**
 * Klasa zawiera elementy wspolne dla inducerow tworzacych proste drzewa decyzyjne.
 * @author Marek Grzes'
 */
class IDTInducerAP{
public:

	vector<string> GetParams() const;
	void PrintParamsXML(ostream &ostr, int nsp);
	bool SetParam(string sParamName, string sParamValue);

	IDTInducerAP();
	virtual ~IDTInducerAP(){
	}
protected:
	/**
	 * Parametr wskazuje czy ma byc przeprowadzana wewnetrzna dyzjunkcja.
	 */
	bool m_bInnerDisjunction;

	/**
	 * Dodaje galaz do testu z wewnetrzna dyzjunkcja.
	 *
	 * @return true - kiedy nastapilo dodanie testu
	 */
	bool AddBranch(CNominalTestInnerDisj* pTest, CDTreeNode* pNode);

	/**
	 * Funkcja laczy o ile to mozliwe dwie galezie. Zloczone poddrzewa sa usuwane, a
	 * w ich miejcu po zlaczeniu wstawiany jest nowy lisc, ktory jest zwracany
	 * przez ta funkcje.
	 * @param pTest - test z wewnetrzna dyzjunkcja, ten parametr po to, zeby dwa razy
	 * nie wywolywac dynamic_cast.
	 * @param pNode - wezle drzewa
	 * @return zwrocony wezel ma ustawione pokrywane przyklady.
	 */	
	CDTreeNode* JoinBranches(IDTInducer *inducer, CNominalTestInnerDisj* pTest, CDTreeNode* pNode);

	/**
	 * Funkcja zmienia przyporzadkowanie wartosci nominalnej do gelezi. Zmiana moze nastapic tylko dla takiej wartosci,
	 * ktorej galezi przyporzodkowana jest co najmniej jedna inna wartosc.
	 * @param pTest - test do modyfikacji
	 * @return true - kiedy nastapila modyfikacja
	 */
	bool MoveValueBetweenBranches(CNominalTestInnerDisj* pTest);

	/**
	 * Znajduje mieszany dipol, ktory jest nie jest przecinany przez hiperplaszczyzne testu.
	 * @return false kiedy dane sa tylko z jednej klasy.
	 *
	 * UWAGA: Traci to sens kiedy sa wiecej niz dwie klasy??
	 */
	virtual bool FindMixedDipolAP(deque<CDataObject*> &deqCI, CDataObject* &ob1, CDataObject* &ob2, ITest* pITest);

	/**
	 * Przecina mieszany dipol.
	 */
	virtual bool CutMixedDipolAP(CDataObject* ob1, CDataObject* ob2, ITest* pITest);

	/**
	 * Znajduje czysty dipol, ktory jest przecinany przez hiperplaszczyzne testu.
	 * @return false kiedy dane sa tylko z jednej klasy.
	 */
	virtual bool FindPureDipolAP(deque<CDataObject*> &deqCI, CDataObject* &ob1, CDataObject* &ob2, ITest* pITest);

	/**
	 * Laczy przeciety czysty dipol.
	 */
	virtual bool JoinPureDipolAP(CDataObject* ob1, CDataObject* ob2, ITest* pITest);


};

#endif

