#include "ITDIDTInducer.h"
#include "DTClassifier.h"
#include "RealTest.h"
#include "NominalTest.h"
#include "TestContainer2.h"
#include "../MLPCommon/MLPStatistics.h"
#include "../MLPDataset/NomAttr.h"
#include "../MLPDataset/RealAttr.h"

#include <iomanip>

ITDIDTInducer::ITDIDTInducer(void){
	m_sThreshold = "midpoint";
	m_uMinimunItems = 2;
	m_sPruning = "no";
	m_sClearCoverd = "yes";
	m_dRandomStop = 0.0;
	m_iMaxDepth = -1;
	m_iCurrentDepth = -1;
}

ITDIDTInducer::~ITDIDTInducer(void){
}

IClassifier *ITDIDTInducer::ActualTrain(const IDataSet *trainingDataSet){
CTestContainer2 tests(trainingDataSet->GetCurrentAttrCount());

	// Utworzenie tablicy wskaznikow do danych (przykladow).
	deque<CDataObject*> tempCDataObject;
	for(unsigned i=0; i<trainingDataSet->GetSize(); i++){
		tempCDataObject.insert(tempCDataObject.end(),(*trainingDataSet)[i]);
	}

	// Wyznaczenie kategorii domyslnej.	
	IAttrValue *d = Category(trainingDataSet,tempCDataObject,NULL);

	PreBuildTree(tests,trainingDataSet);

	// Rekurencyja funkcja budowania drzewa.
	CDTreeNode* pCTree = BuildTree(tempCDataObject,tests,d,trainingDataSet);
	
	PostBuildTree(tests);

	if(pCTree){
		// Lokalna estymacja parametrow.
		SubTreeErrors(pCTree);
		if(m_sPruning == "yes")
			// Pruning. Po jego wykonaniu estymacja parametrow jest zaktualizowana 
			// poprawnie.
			PruneTree(pCTree);
		else
			// Globalna estymacja parametrow.
			SubTreeDepErrors(pCTree);
		if(m_sClearCoverd=="yes"){
			// Usuniecie z wezlow drzewa informacji o przykladach pokrywanych przez 
			// dany wezel.
			pCTree->ClearDeqCI();
		}
		CDTClassifier *pClass = new CDTClassifier(pCTree,false,false,trainingDataSet);
		return pClass;
	}
	else 
		return NULL;
}

CDTreeNode* ITDIDTInducer::BuildTree(deque<CDataObject*> &deqCI, ITestContainer &tests, IAttrValue *pIA, const IDataSet *trainingDataSet){
// CTDIDTInducerAPDepthK
	m_iCurrentDepth++;
	if(m_iLogLevel>=2){
		stringstream ss;
		ss<<"Current depth: "<<m_iCurrentDepth<<" ("<<deqCI.size()<<") observations"<<endl;
 		CLogger::LogEvent(ss.str(),"",__LINE__,m_ofsLogFile);
	}
// CTDIDTInducerAPDepthK
	CDTreeNode * pnode = new CDTreeNode();
	IAttrValue* temp = Category(trainingDataSet,deqCI,pIA);
	if(temp->GetAttrType() == e_NominalAttr){
		pnode->m_pIAClass = new CNomAttr((CNomAttr&)*temp);
	}
	if(temp->GetAttrType() == e_RealAttr){
		pnode->m_pIAClass = new CRealAttr((CRealAttr&)*temp);
	}
	pnode->m_deqCI = deqCI;

	if(StopCondition(deqCI,tests.size())){
		pnode->m_bLeaf = true;
// CTDIDTInducerAPDepthK
		m_iCurrentDepth--;
// CTDIDTInducerAPDepthK
		return pnode;
	}else{
		// Wybierz test, zwrocony test jest usuwany z vITraintest.
		ITest* test = PickOutTest(deqCI,tests,trainingDataSet);
	
		// NULL bedzie zwrocony jak juz zadan test nie daje przyrostu
		// informacji, czyli warunek stopu i utworzenie liscia.
		if(test == NULL){
			pnode->m_bLeaf = true;
// CTDIDTInducerAPDepthK
			m_iCurrentDepth--;
// CTDIDTInducerAPDepthK
			return pnode;
		}

		// Utworz tablice podzialow
		vector<vector<unsigned> > vvDiv = test->GetSubSets(deqCI);

		// Sprawdzamy czy sa chociaz 2 niepuste podzialy
		int div=0;
		for(unsigned i=0; i<vvDiv.size(); i++){
			if(vvDiv[i].size()>0){
				div++;
			}
			if(div==2){
				break;
			}
		}
		// Jesli nie ma co namjmniej dwoch nie pustych podzialow to tworzymy lisc w tym miejscu
		if(div<2){
			delete test;
			pnode->m_bLeaf = true;
			m_iCurrentDepth--;
			return pnode;
		}

		//- cerr<<"Podzialow="<<vvDiv.size()<<endl;
		// Dla kazego podzbioru wynikajacego z podzialu: BuildTree
		for(unsigned i=0; i<vvDiv.size(); i++){
			deque<CDataObject*> subset;
			vector<unsigned> subIDs = vvDiv[i];
			for(unsigned j=0; j<subIDs.size(); j++){
				subset.insert(subset.end(),deqCI[subIDs[j]]);
			}
			//-cerr<<"Node="<<deqCI.size()<<"  Part="<<subset.size()<<" depht="<<m_iCurrentDepth<<endl;
			CDTreeNode* node = BuildTree(subset,tests,pnode->m_pIAClass, trainingDataSet);
			pnode->m_vBranch.push_back(node);

		}

		//memcheck pnode->m_pITest = test->CloneTest();
		pnode->m_pITest = test;
// CTDIDTInducerAPDepthK
		m_iCurrentDepth--;
// CTDIDTInducerAPDepthK
		return pnode;
		
	}
// CTDIDTInducerAPDepthK
	m_iCurrentDepth--;
// CTDIDTInducerAPDepthK
	return NULL;	
}

bool ITDIDTInducer::StopCondition(deque<CDataObject*> &deqCI, size_t iTestNum, void *pVoid){
	if(deqCI.size() >= m_uMinimunItems && iTestNum>0){
		IAttrValue *pIA1 = deqCI[0]->GetClass();
		for(unsigned i=1; i<deqCI.size(); i++){
			IAttrValue *pIA2 = deqCI[i]->GetClass();
			if( ! ((*pIA1) == (*pIA2)) ){
				if(CMLPStatistics::Event((int)(100.0*m_dRandomStop))){
					return true;
				}
				return false;
			}
		}
		return true;
	}else
		return true;
}

void ITDIDTInducer::PruneTree(CDTreeNode* pDTree){
	if(pDTree->m_bLeaf) return;

	for(unsigned i=0; i<pDTree->m_vBranch.size(); i++)
		PruneTree(pDTree->m_vBranch[i]);

	double temp = 0;
	double tempN = 0;
	double tempE = 0;
	for(unsigned i=0; i<pDTree->m_vBranch.size(); i++){
		temp += pDTree->m_vBranch[i]->m_dEstE;
		tempN += pDTree->m_vBranch[i]->m_dN;
		tempE += pDTree->m_vBranch[i]->m_dE;
	}
	
	if(temp >= pDTree->m_dEstE){
		pDTree->m_bLeaf = true;
		for(unsigned i=0; i<pDTree->m_vBranch.size(); i++)
			delete pDTree->m_vBranch[i];
		pDTree->m_vBranch.clear();
	}

	if(!pDTree->m_bLeaf){
		pDTree->m_dEstE = temp;
		pDTree->m_dN = tempN;
		pDTree->m_dE = tempE;
	}

}

void ITDIDTInducer::SubTreeErrors(CDTreeNode* pDTree){
	pDTree->m_dN = pDTree->m_deqCI.size();
	pDTree->m_dE=0;
	pDTree->m_dEstE=0;

	if(pDTree->m_dN>0){
		for(unsigned i=0; i<pDTree->m_deqCI.size(); i++){
			if( *(pDTree->m_deqCI[i]->GetClass()) != *(pDTree->m_pIAClass) )
				pDTree->m_dE++;
		}
		pDTree->m_dEstE = pDTree->m_dN*CMLPStatistics::UCF(pDTree->m_dE,pDTree->m_dN);
	}
	for(unsigned i=0; i<pDTree->m_vBranch.size(); i++)
		SubTreeErrors(pDTree->m_vBranch[i]);
}

void ITDIDTInducer::SubTreeDepErrors(CDTreeNode* pDTree){
	if(pDTree->m_bLeaf) return;

	for(unsigned i=0; i<pDTree->m_vBranch.size(); i++)
		SubTreeDepErrors(pDTree->m_vBranch[i]);

	pDTree->m_dEstE = 0;
	pDTree->m_dN = 0;
	pDTree->m_dE = 0;
	for(unsigned i=0; i<pDTree->m_vBranch.size(); i++){
		pDTree->m_dEstE += pDTree->m_vBranch[i]->m_dEstE;
		pDTree->m_dN += pDTree->m_vBranch[i]->m_dN;
		pDTree->m_dE += pDTree->m_vBranch[i]->m_dE;
	}
}

bool ITDIDTInducer::SetParam(string sParamName, string sParamValue){
	sParamName=trimS(sParamName);
	sParamName=tolowerS(sParamName);
	sParamValue=trimS(sParamValue);
	sParamValue=tolowerS(sParamValue);

	if(sParamName=="maxdepth"){
		m_iMaxDepth = atoi(sParamValue.c_str());
		if(m_iMaxDepth<-1){
			m_iMaxDepth = -1;
			return false;
		}
		return true;
	}
	
	if(sParamName=="randomstop"){
		m_dRandomStop = atof(sParamValue.c_str());
		if(m_dRandomStop<0.0 || m_dRandomStop>1.0){
			m_dRandomStop = 0.0;
			return false;
		}
		return true;
	}
	if(sParamName=="minimumitems"){
		m_uMinimunItems = (unsigned)atoi(sParamValue.c_str());
		return true;
	}
	if(sParamName=="threshold"){
		if(sParamValue=="midpoint" ||
			sParamValue=="largestvalue"){
				m_sThreshold=sParamValue;
				return true;
		}
		else
			return false;
	}
	if(sParamName=="pruning"){
		if(sParamValue=="yes" ||
			sParamValue=="no"){
				m_sPruning=sParamValue;
				return true;
		}
		else
			return false;
	}
	if(sParamName=="clearcoverd"){
		if(sParamValue=="yes" ||
			sParamValue=="no"){
				m_sClearCoverd=sParamValue;
				return true;
		}
		else
			return false;
	}
	if(IDTInducer::SetParam(sParamName,sParamValue)) return true;
	return false;
}

void ITDIDTInducer::PrintParamsXML(ostream &ostr, int nsp){
	IDTInducer::PrintParamsXML(ostr,nsp);
	ostr.fill(' ');
	ostr<<setw(nsp)<<" "<<"<MLPParam Name=\"maxdepth\" Value=\"-1\" />"<<endl;
	ostr<<setw(nsp)<<" "<<"<MLPParam Name=\"randomstop\" Value=\"0\" />"<<endl;
	ostr<<setw(nsp)<<" "<<"<MLPParam Name=\"minimumitems\" Value=\"5\" />"<<endl;
	ostr<<setw(nsp)<<" "<<"<MLPParam Name=\"threshold\" Value=\"midpoint\" />"<<endl;
	ostr<<setw(nsp)<<" "<<"<MLPParam Name=\"pruning\" Value=\"yes\" />"<<endl;
	ostr<<setw(nsp)<<" "<<"<MLPParam Name=\"clearcoverd\" Value=\"yes\" />"<<endl;
	ostr.fill(' ');
}

vector<string> ITDIDTInducer::GetParams() const{
vector<string> res;
stringstream ss;

	res = IDTInducer::GetParams();

	res.push_back("maxdepth");
	ss<<m_iMaxDepth;
	res.push_back(ss.str());
	res.push_back("randomstop");
	ss.str(""); ss<<m_dRandomStop;
	res.push_back(ss.str());
	res.push_back("minimumitems");
	ss.str(""); ss<<m_uMinimunItems;
	res.push_back(ss.str());
	res.push_back("threshold");
	res.push_back(m_sThreshold);
	res.push_back("pruning");
	res.push_back(m_sPruning);
	res.push_back("clearcoverd");
	res.push_back(m_sClearCoverd);
	
	return res;
}

