/**
 * $Revision: 1.14 $
 * $Date: 2005/04/20 21:43:05 $
 */

#ifndef __ITDIDTINDUCERH__
#define __ITDIDTINDUCERH__

#include "../MLPDT/IDTInducer.h"
#include "DTreeNode.h"
#include "ITestContainer.h"

/**
 * Interfejs, ktorego implementacje sluza do tworzenia drzew decyzyjnych
 * metoda Top Down Induction.
 * @author Marek Grzes'
 */
class ITDIDTInducer: public IDTInducer{
public:
	virtual bool SetParam(string sParamName, string sParamValue);
	virtual void PrintParamsXML(ostream &ostr, int nsp);
	virtual vector<string> GetParams() const;

	ITDIDTInducer(void);
	virtual ~ITDIDTInducer(void);

protected:
	/**
	 * Maksymalna glebokosc generowanych drzew.
	 * -1 - oznacza ignorowanie tego parametru
	 */
	int m_iMaxDepth;
	
	/**
	 * Biezaca glebokosc generowanego drzewa. Jak jest tylko jeden lisc to takie drzewo 
	 * ma glebokosc 0.
	 */
	int m_iCurrentDepth;

	/**
	 * Parametr wskazuje jak ma byc tworzony podzial dla wartosci ciaglych.
	 *  midpoint
	 *  largest value
	 */
	string m_sThreshold;

	/**
	 * yes
	 *  no
	 */
	string m_sPruning;

	/** 
	 * yes - tzn. ze bedzie skasowana informacja o tym jakie przyklady
	 *		 pokrywane sa przez dany wezel.
	 * no
	 */
	string m_sClearCoverd;

	/**
	 * Minimalna liczba przykladow dla ktorej bedzie tworzony wezel drzewa.
	 * DEFAULT: 2
	 */
	unsigned m_uMinimunItems;

	/**
	 * Prawodopobienstwo stopu przy schodzeniu w dol drzewa podczas jego
	 * budowy. Z tym p-em funkcja StopCondition zwraca true niezaleznie
	 * od innych czynnikow.
	 */
	double	m_dRandomStop;

	/**
	 * Funkcja sluzaca do rekurencyjnego budowania drzewa.
	 */
	virtual CDTreeNode* BuildTree(deque<CDataObject*> &deqCI, ITestContainer &tests, IAttrValue *pIA, const IDataSet *trainingDataSet);
	
	/**
	 * Metoda wywolywana przed uruchomieniem funkcji rekurenyjnej majaca
	 * na celu zainicjowanie parametrow algorytmu.
	 */
	virtual void PreBuildTree(ITestContainer &tests, const IDataSet *trainingDataSet)=0;

	/**
	 * Metoda wywolywana po zakonczeniu rekurencyjnego budowania drzewa.
	 */
	virtual void PostBuildTree(ITestContainer &tests)=0;

	/**
	 * Pruning drzewa.
	 */
	virtual void PruneTree(CDTreeNode* pDTree);

	/**
	 * Obliczenie parametrow oceny wezlow. <b>Wartosci w wezlach liczone
	 * wzgledem klasyfikatora z tych wezlow a nie z poddrzew.</b>
	 *
	 * Powinna byc wywolana przed PruneTree.
	 */
	virtual void SubTreeErrors(CDTreeNode* pDTree);

	/**
	 * Obliczenie parametrow oceny wezlow. Wartosci w wezlach liczone
	 * zgodnie z wartosciami w lisciach i poddrzewach.
	 *
	 * Powinna byc wywolania kiedy nie jest wywolywana PruneTree
	 * i po wywolaniu SubTreeErrors
	 */
	virtual void SubTreeDepErrors(CDTreeNode* pDTree);

	/** 
	 * Funkcja sprawdzajaca czy zostal spelniony warunek stopu.
	 */
	virtual bool StopCondition(deque<CDataObject*> &deqCI, size_t iTestNum, void *pVoid=NULL);

	/**
	 * Parametry przez referencje, zeby nie kopiowac danych, i zmienic kolekcje
	 * testow. Funkcja ta przed zwroceniem testu tworzy na podstawie wybranego testu
	 * jego kopie i zwraca ta kopie. Dlatego obiekt, ktory go otrzyma musi
	 * zadbac o jego usuniecie. 
	 */ 
	virtual ITest* PickOutTest(deque<CDataObject*> &deqCI, ITestContainer &tests, const IDataSet *trainingDataSet)=0;

	virtual void PostTrain(const IDataSet *trainingDataSet){
	}
	
private:
	virtual IClassifier *ActualTrain(const IDataSet *trainingDataSet);
};

#endif

