#ifndef __ITESTCONTAINERH__
#define __ITESTCONTAINERH__

#include "ITest.h"

/**
 * Interfejs kolekcji do przechowywania testow do naczuczania drzew
 * decyzyjnych algorytemami top-down i ewolucyjnymi.
 * @author Marek Grzes' marekg@ii.pb.bialystok.pl, (C) 2004
 */
class ITestContainer{
public:
	virtual ~ITestContainer(){}

	/**
	 * Zwraca test o numerze t odliczajac po kolei wszystkie testy. 
	 * UWAGA: tu jest zwracany wskaznik z kolekcji a nie do kopii testu. Kopie do wykorzystania
	 * dalej trzeba zrobic samemu.
	 * @return NULL kiedy t poza zakresem, albo nie ma zadnego testu.
	 */
	virtual ITest* operator[](size_t t)=0;
	
	/**
	 * Zwraca test o numerze t dla atrybuty o indeksie attrId.
	 * UWAGA: tu jest zwracany wskaznik z kolekcji a nie do kopii testu. Kopie do wykorzystania
	 * dalej trzeba zrobic samemu.
	 * @return NULL kiedy t lub attrId poza zakresem, albo nie ma zadnego testu.
	 */
	//virtual ITest* operator[][](size_t attrId, size_t t) const =0;
	virtual ITest* GetAttrsTest(size_t attrId, size_t t) const =0;

	/**
	 * Zwraca losowe ciecie na atrybucie attrId.
	 */
	virtual double GetRandomThreshold(size_t attrId) const = 0;

	/**
	 * Zwraca losowe ciecie na atrybucie attrId, ktore jest pomiedzy wartosciami
	 * val1 i val2. Wartosci val1 i val2 przekazane do tej funkcji nie musza byc
	 * uporzadkowane. Funkcja sama je porzadkuje.
	 * @return zwracany jest wskaznik do nowego testu. Funkcja, ktora go
	 * otrzyma musi nie robi juz kopii.
	 */
	virtual ITest* GetTestWithRandomThreshold(size_t attrId, double val1, double val2) const = 0;

	/**
	 * Zwraca liczbe wszystkich testow.
	 */
	virtual size_t size()=0;
	
	/**
	 * Zwraca liczbe wszystkich testow dla atrybutu o indeksie attrId.
	 * @return 0 kiedy nie ma atrybutu o indeksie arrtId albo nie ma
	 * zadnego testu na tym atrybucie.
	 */
	virtual size_t size(size_t attrId) const =0;
	
	/**
	 * Dodaje test dla atrybutu attrId. Przekazany wskaznik do testu jest zapamietywany
	 * w strukturach tej klasy. Dlatego ewentualna kopie trzeba zrobic przed wywolaniem
	 * tej metody.
	 * Jesli chcemy pamietac testy w sposob liniowy to mozemy zawsze podawac zero
	 * jako attrId. Wtedy wszystkie testy beda w jednej liscie.
	 * @return true kiedy wstawianie OK
	 */
	virtual bool push(size_t attrId, ITest* test)=0;
	
	/**
	 * Usuwa test atrybutu attrId o indeksie t.
	 * @return true kiedy usuwanie powiodlo sie
	 */
	virtual bool erase(size_t attrId, size_t t)=0;
};

#endif

