/**
 * $Revision: 1.4 $
 * $Date: 2005/11/02 00:38:07 $
 */

#include "NominalTestInnerDisj.h"

#include <sstream>
#include <algorithm>
#include <typeinfo>

CNominalTestInnerDisj::CNominalTestInnerDisj(CAttrDesc* pAttrDesc, int attrNum):CNominalTest(pAttrDesc, attrNum){
	m_vtInnerDisIndex.resize(pAttrDesc->m_vAttrDomain.size());
	for(size_t t=0; t<m_vtInnerDisIndex.size(); t++){
		m_vtInnerDisIndex[t] = t;
	}
}

CNominalTestInnerDisj::~CNominalTestInnerDisj(void){
}

CNominalTestInnerDisj::CNominalTestInnerDisj(const CNominalTestInnerDisj& templ):CNominalTest(templ){
	this->m_vtInnerDisIndex = templ.m_vtInnerDisIndex;
}

unsigned CNominalTestInnerDisj::GetBranchID(CDataObject *pCInst)const{
	assert(pCInst);
	IAttrValue* pIA1 = (*pCInst)[m_iAttrNum];
	assert(pIA1);
	CNomAttr* ref_nomattr = dynamic_cast<CNomAttr*>(pIA1);
	assert(ref_nomattr);
	size_t ret = ref_nomattr->m_AttrValue.u_iNom;
	if(ret<0 || ret>=m_pAttrDesc->m_vAttrDomain.size()){
		CLogger::LogError("Index out of scope",__FILE__,__LINE__);
		return 0;
	}
	assert(m_vtInnerDisIndex.size() <= ret);
	// Numer galezi jest w wektorze m_vtInnerDisIndex na pozycji odpowiadajacej numerowi wartoci
	// nominalnej
	return m_vtInnerDisIndex[ret];
}

vector<vector<unsigned> > CNominalTestInnerDisj::GetSubSets(deque<CDataObject*> deqCI)const{
// Liczba galezi zlezy od m_vtInnerDisIndex dlatego trzeba wywolac funkcje GetSubSetCount()
vector<vector<unsigned> > res(GetSubSetCount());

	for(unsigned i=0; i<deqCI.size(); i++){
		unsigned uSubSetId = GetBranchID(deqCI[i]);
		res[uSubSetId].push_back(i);
	}
	return res;
}

unsigned CNominalTestInnerDisj::GetSubSetCount()const{
deque<size_t> dq;

	for(size_t t=0; t<m_vtInnerDisIndex.size(); t++){
		if(find(dq.begin(),dq.end(),m_vtInnerDisIndex[t]) == dq.end()){
			dq.push_back(m_vtInnerDisIndex[t]);
		}
	}
	return dq.size();
}

bool CNominalTestInnerDisj::operator==(const ITest& it)const{
   try {
		CNominalTestInnerDisj& ref_nomout = dynamic_cast<CNominalTestInnerDisj&>((ITest&)it);
		if(ref_nomout.m_iAttrNum == m_iAttrNum){
			return true;
		}
		return false;
   }
   catch (bad_cast) {
      return false;
   }
}

void CNominalTestInnerDisj::PrintTest(unsigned branchid, const IDataSet* dataset, ostream &ostr)const{
	assert(m_pAttrDesc);
	assert(m_pAttrDesc->m_vAttrDomain.size()>branchid);
	stringstream values;
	bool found = false;
	for(size_t t=0; t<m_vtInnerDisIndex.size(); t++){
		if(m_vtInnerDisIndex[t] == branchid){
			if(found){
				values<<", ";
			}else{
				found = true;
			}
			values<<"\'"<<t<<"\'";
		}
	}	
	ostr<<m_pAttrDesc->m_sAtrrName<<" in {"<<values.str()<<"} ("<<ValueToName(branchid)<<")";
}

bool CNominalTestInnerDisj::SetSubSets(size_t nSubSets){
	if(m_vtInnerDisIndex.size()<nSubSets || nSubSets<2){
		return false;
	}
	for(size_t t=0; t<m_vtInnerDisIndex.size(); t++){
		m_vtInnerDisIndex[t] = t;
	}
	random_shuffle(m_vtInnerDisIndex.begin(),m_vtInnerDisIndex.end(),Rand);
	for(size_t t=0; t<m_vtInnerDisIndex.size(); t++){
		// Tam gdzie wartosc w wektorze jest wieksza rowna nSubSets losujemy nowa wartosc
		if(m_vtInnerDisIndex[t] >= nSubSets){
			m_vtInnerDisIndex[t] = rand()%nSubSets;
		}
	}
	return true;
}

string CNominalTestInnerDisj::ValueToName(unsigned uIndex)const{
string res;

	assert(m_pAttrDesc);
	assert(m_pAttrDesc->m_vAttrDomain.size()>uIndex);
	assert(GetSubSetCount()>uIndex);
	for(size_t t=0; t<m_vtInnerDisIndex.size(); t++){
		if(m_vtInnerDisIndex[t] == uIndex){
			if(res.length()>0){
				res += ", ";
			}
			res += "\'";
			res += m_pAttrDesc->m_vAttrDomain[t];
			res += "\'";
		}
	}
	return res;
}
