#include "ObliqueTest.h"
#include "../MLPDataset/RealAttr.h"
#include <numeric>
#include <typeinfo>
#include <cmath>

CObliqueTest::CObliqueTest(vector<double> vdW):ITest(-1){
	m_vdWReal=vdW;
}

CObliqueTest::~CObliqueTest(void){
}

CObliqueTest::CObliqueTest(const CObliqueTest& templ):ITest(templ.m_iAttrNum){
	m_vdWReal=templ.m_vdWReal;
}

inline float CObliqueTest::GetCostOfAttributes(const IDataSet* dataset)const{
	float cost = 0;
	for(size_t t=0; t<m_vdWReal.size(); t++){
		if(m_vdWReal[t] != 0){
			// Popbrenie pozycji, pod ktora pamietany jest dany atrybut rzeczywisty.
			int id = dataset->GetRealAttrPos(t);
			cost += dataset->GetCostOfAttribute(id);
		}
	}
	return cost;
}

float CObliqueTest::GetCostOfAttributes(const IDataSet* dataset, vector<bool> &vUsed)const{
	float cost = 0;
	for(size_t t=0; t<m_vdWReal.size(); t++){
		if(m_vdWReal[t] != 0){
			if(!vUsed[t]){
				// Popbrenie pozycji, pod ktora pamietany jest dany atrybut rzeczywisty.
				int id = dataset->GetRealAttrPos(t);
				cost += dataset->GetCostOfAttribute(id);
				vUsed[id]=true;
			}
		}
	}
	return cost;
}

double CObliqueTest::GetMI(){
	double N=0;
	size_t t;
	for(t=0; t<m_vdWReal.size()-1; t++){
		N+=(m_vdWReal[t]*m_vdWReal[t]);
	}
	N=sqrt(N);
	if(m_vdWReal[t]<=0){
		return 1.0/N;
	}else{
		return -1.0/N;
	}
	return 0;
}

bool CObliqueTest::operator==(const ITest& it)const{
   try {
		CObliqueTest& ref_ob = dynamic_cast<CObliqueTest&>((CObliqueTest&)it);
		if(this->m_vdWReal.size() == ref_ob.m_vdWReal.size()){
			for(size_t t=0; t<this->m_vdWReal.size(); t++){
				if(this->m_vdWReal[t] != ref_ob.m_vdWReal[t]) return false;
			}
			return true;
		}
		return false;
   }
   catch (bad_cast) {
      return false;
   }
}

double CObliqueTest::GetInnerProduct(CDataObject *pCInst)const{
	unsigned size = (unsigned)(m_vdWReal.size());
	assert(pCInst);
	assert(size>1);
	vector<double> vdX(size);
	for(unsigned t=0; t<size-1; t++){
		IAttrValue* pIA1 = pCInst->GetRealAttrValue(t);
		assert(pIA1);
		CRealAttr* ref_ra = dynamic_cast<CRealAttr*>(pIA1);
		assert(ref_ra!=NULL);
		//vdX.push_back(ref_ra->m_AttrValue.u_dReal);
		vdX[t]=ref_ra->m_AttrValue.u_dReal;
	}
	//vdX.push_back(1.0);
	vdX[size-1]=1.0;
	return inner_product(m_vdWReal.begin(),m_vdWReal.end(),vdX.begin(),0.0);
}

unsigned CObliqueTest::GetBranchID(CDataObject *pCInst)const{
	double iprod = GetInnerProduct(pCInst);
	if(iprod <= 0){
		return 0;
	}else{
		return 1;
	}
}

void CObliqueTest::PrintTest(unsigned branchid, const IDataSet* dataset, ostream &ostr)const{
	assert(dataset);
	assert(branchid==0 || branchid==1);
	bool first=true;
	for(unsigned t=0; t<m_vdWReal.size(); t++){
		if(t<m_vdWReal.size()-1){
			if(m_vdWReal[t] != 0.0){
				if(first){
					//- ostr<<m_vdW[t]<<"*("<<dataset->GetAttrDesc(t)->m_sAtrrName.c_str()<<")";
					ostr<<m_vdWReal[t]<<"*(x"<<t<<")";
					first=false;
				}else{
					//- ostr<<"+"<<m_vdW[t]<<"*("<<dataset->GetAttrDesc(t)->m_sAtrrName.c_str()<<")";
					ostr<<"+"<<m_vdWReal[t]<<"*(x"<<t<<")";
				}
			}
		}else{
			ostr<<"+"<<m_vdWReal[t];
		}
	}
	if(branchid==0){
		ostr<<" <= ";
	}else{
		ostr<<" > ";
	}
	ostr<<" 0";
}

double CObliqueTest::GetDistance(CDataObject *pCInst, double mi){
	double res=0;
	size_t t;
	for(t=0; t<m_vdWReal.size()-1; t++){
		IAttrValue* pIA1 = pCInst->GetRealAttrValue(t);
		assert(pIA1);
		CRealAttr* ref_ra = dynamic_cast<CRealAttr*>(pIA1);
		assert(ref_ra!=NULL);
		res+=m_vdWReal[t]*mi*ref_ra->m_AttrValue.u_dReal;
	}
	res  += m_vdWReal[t] * mi;
	return res;
}
