#ifndef __COBLIQUETEST__
#define __COBLIQUETEST__

#include "ITest.h"

/**
 * Test wielowymiarowy.
 * @author Marek Grze
 */
class CObliqueTest : public ITest
{
friend class CEAObliqueTestCreator;
friend class CObliqueTestCreator;
friend class CEAIDTIndividual;

public:
	/**
	 * Koszt atrybutow jest mierzony w taki sposob, ze sumowany jest koszt 
	 * tych atrybutow, dla ktorych wspolczynniki hiperplaszczyzny sa rozne od 0.
	 */
	virtual float GetCostOfAttributes(const IDataSet* dataset)const;
	
	virtual float GetCostOfAttributes(const IDataSet* dataset, vector<bool> &vUsed)const;

	virtual vector<double> GetHiperplaneFactors(){
		return m_vdWReal;
	}
	virtual unsigned GetSubSetCount()const{
		return (unsigned)2;
	}
	vector<vector<unsigned> > GetSubSets(deque<CDataObject*> deqCI)const{
		vector<vector<unsigned> > res(2);
		for(unsigned i=0; i<deqCI.size(); i++){
			unsigned uSubSetId = GetBranchID(deqCI[i]);
			res[uSubSetId].push_back(i);
		}
		return res;	
	}
	virtual bool operator==(const ITest& it)const;
	virtual bool operator<(const ITest& it)const{
		return false;
	}
	virtual bool operator>(const ITest& it)const{
		return false;
	}
	virtual unsigned GetBranchID(CDataObject *pCInst)const;
	virtual double GetInnerProduct(CDataObject *pCInst)const;
	virtual void PrintTest(unsigned branchid, const IDataSet* dataset, ostream &ostr)const;
	virtual ITest* CloneTest()const{
		CObliqueTest* temp = new CObliqueTest(*this);
		return (ITest*) temp;
	}
	virtual double GetDistance(CDataObject *pCInst, double mi);

	/**
	 * Zwraca czynnik normujacy wykorzystywany do przejscia z postaci ogolnej plaszczyzny
	 * do postaci normalnej.
	 */
	double GetMI();

	CObliqueTest(const CObliqueTest& templ);
	CObliqueTest(vector<double> vdW);
	virtual ~CObliqueTest(void);

protected:
	/**
	 * Wektor wszystkich wspolczynikow hiperplaszczyzny. Dla atrybutow, ktore 
	 * sa do pominiecia odpowiadajace wartosci wspolczynikow sa rowne 0. Liczba
	 * wspolczynnikow obejmuje wyraz wolny.
	 * <b>UWAGA:</b> W wektorze tym pamietane sa tylko wspolczynniki odpowiadajace
	 * atrybutom rzeczywistym. Dla cech nominalnych nie ma tu wspolczynnikow.
	 * Aby odnalezc z ktorego atrybutu jest dana pozycja w tym wektorze, trzeba skorzystac
	 * z podwojnego indeksowania, ktore wyklucza cechy nominalne (przeskakuje je).
	 */
	vector<double> m_vdWReal;

};

#endif
