#include "RealTest.h"
#include <algorithm>
#include <cmath>
#include <typeinfo>

CRealTest::CRealTest(double threshold, int attrNum):ITest(attrNum){
	m_dThreshold = threshold;
}

CRealTest::~CRealTest(void){
}

CRealTest::CRealTest(const CRealTest& templ):ITest(templ.m_iAttrNum){
	this->m_dThreshold = templ.m_dThreshold;
}

inline float CRealTest::GetCostOfAttributes(const IDataSet* dataset)const{
	return dataset->GetCostOfAttribute(m_iAttrNum);
}

float CRealTest::GetCostOfAttributes(const IDataSet* dataset, vector<bool> &vUsed)const{
	if(vUsed[m_iAttrNum]){
		return 0;
	}else{
		vUsed[m_iAttrNum]=true;
		return dataset->GetCostOfAttribute(m_iAttrNum);
	}
}

vector<vector<unsigned> > CRealTest::GetSubSets(deque<CDataObject*> deqCI)const{
vector<vector<unsigned> > res(2);

	for(unsigned i=0; i<deqCI.size(); i++){
		unsigned uSubSetId = GetBranchID(deqCI[i]);
		res[uSubSetId].push_back(i);
	}

	return res;	
}

unsigned CRealTest::GetBranchID(CDataObject *pCInst)const{
	assert(pCInst);
	IAttrValue* pIA1 = (*pCInst)[m_iAttrNum];
	assert(pIA1);
	if(pIA1->GetAttrType() != e_RealAttr){
		CLogger::LogError("e_RealAttr expected",__FILE__,__LINE__);
		return 0;
	}
	if(pIA1->m_AttrValue.u_dReal <= m_dThreshold){
		return 0;
	}else{
		return 1;
	}
	return 0;
}

bool CRealTest::operator==(const ITest& it)const{
   try {
		CRealTest& ref_nomout = dynamic_cast<CRealTest&>((ITest&)it);
		if(this->m_iAttrNum == ref_nomout.m_iAttrNum){
			if(this->m_dThreshold == ref_nomout.m_dThreshold){
				return true;
			}
		}
		return false;
   }
   catch (bad_cast) {
      return false;
   }
}

bool CRealTest::operator<(const ITest& it)const{
   try {
		CRealTest& ref_nomout = dynamic_cast<CRealTest&>((ITest&)it);
		if(this->m_iAttrNum == ref_nomout.m_iAttrNum){
			if(this->m_dThreshold < ref_nomout.m_dThreshold){
				return true;
			}
		}
		return false;
   }
   catch (bad_cast) {
      return false;
   }
}

bool CRealTest::operator>(const ITest& it)const{
   try {
		CRealTest& ref_nomout = dynamic_cast<CRealTest&>((ITest&)it);
		if(this->m_iAttrNum == ref_nomout.m_iAttrNum){
			if(this->m_dThreshold > ref_nomout.m_dThreshold){
				return true;
			}
		}
		return false;
   }
   catch (bad_cast) {
      return false;
   }
}

void CRealTest::PrintTest(unsigned branchid, const IDataSet* dataset, ostream &ostr)const{
	assert(dataset);
	assert(branchid==0 || branchid==1);
	ostr<<dataset->GetAttrDesc(m_iAttrNum)->m_sAtrrName.c_str();
	if(branchid==0){
		ostr<<" <= ";
	}else{
		ostr<<" > ";
	}
	ostr<<m_dThreshold;
}
