#include "TestContainer2.h"
#include "RealTest.h"
#include <algorithm>


CTestContainer2::CTestContainer2(){
	m_bChanged = true;
}

CTestContainer2::CTestContainer2(size_t attrCount){
	deque<ITest*> temp;
	m_ddITest.resize(attrCount,temp);
	m_vAttrsNumTests.resize(attrCount,0);
}

size_t CTestContainer2::divideANDconquer(deque<ITest*> dqtests, int p, int k, ITest* test) const{
	int s;
	do{
		s = (k - p) / 2 + p;
		//cerr<<p<<" "<<s<<" "<<k<<endl;
		if(s==p || s==k){
			if(dqtests[k]->operator==(*test)){
				return k;
			}
			if(dqtests[p]->operator==(*test)){
				return p;
			}
			// Jak nie ma wartosci szukanej to wtedy zwracamy wartosc najblizsza czyli pozycje s
			return s;
		}
		if(dqtests[s]->operator<(*test)){
			p = s;
			continue;
		}
		if(dqtests[s]->operator>(*test)){
			k = s;
			continue;
		}
		if(dqtests[s]->operator==(*test)){
			return s;
		}
	}while(p<=s && s<=k);

	return p;
}

CTestContainer2::~CTestContainer2(){
	for(size_t t=0; t<m_ddITest.size(); t++){
		for(size_t i=0; i<m_ddITest[t].size(); i++){
			/*
			CRealTest* test = dynamic_cast<CRealTest*>(m_ddITest[t][i]);
			if(test){
				cerr<<test->m_dThreshold<<" ";
			}
			*/
			delete m_ddITest[t][i];
		}
		//cerr<<endl;
	}
}

ITest* CTestContainer2::operator[](size_t t){
	if(m_bChanged){
		UpdateAttrsVector();
	}
	size_t attr = 0;
	size_t lastSum = 0;
	while(attr < m_vAttrsNumTests.size() && m_vAttrsNumTests[attr] <= t){
		lastSum = m_vAttrsNumTests[attr];
		attr++;
	}
	if(attr < m_vAttrsNumTests.size()){
		return m_ddITest[attr][t-lastSum];
	}
	return NULL;
}

ITest* CTestContainer2::GetAttrsTest(size_t attrId, size_t t) const{
	// Tutaj sa asserty zamiast sprawdzania "if", bo ta funkcja wywoluje sie bardzo czesto.
	// Quantify pokazuje ze zabiera one 5% czasy calosci na EAAP.
	assert(attrId < m_ddITest.size());
	assert(t < m_ddITest[attrId].size());
	return m_ddITest[attrId][t];
}

double CTestContainer2::GetRandomThreshold(size_t attrId) const{
	assert(attrId < m_ddITest.size());
	CRealTest *rtest = dynamic_cast<CRealTest*>(m_ddITest[attrId][rand()%m_ddITest[attrId].size()]);
	if(rtest){
		return rtest->m_dThreshold;
	}else{
		CLogger::LogError("ups ... test inny niz rzeczywisty",__FILE__,__LINE__);
	}
	return 0;
}

ITest* CTestContainer2::GetTestWithRandomThreshold(size_t attrId, double val1, double val2)const{
	if(val1 > val2){
		double temp = val1;
		val1 = val2;
		val2 = temp;
	}
	deque<ITest*> aTests = m_ddITest[attrId];
	CRealTest rt1(val1,aTests[0]->GetAttrNum());
	CRealTest rt2(val2,aTests[0]->GetAttrNum());
	size_t first = divideANDconquer(aTests,0,aTests.size()-1,&rt1);
	size_t second = divideANDconquer(aTests,0,aTests.size()-1,&rt2);
	size_t r;
	if(second - first < 2){
		r = first;
	}else{
		r = first + 1 + rand()%(second - first - 1);
	}
	CRealTest* rt = dynamic_cast<CRealTest*>(aTests[r]);
	// Jezeli wartosci sa poza zakresem val1 i val2 to wtedy bierzemy wartosc ze srodka
	// Bo inaczej taki test moze nic nie przecinac i wtedy test jest bezwartosciowy
	if(rt->m_dThreshold < val1 || rt->m_dThreshold > val2){
		CRealTest* newtest = new CRealTest( (val1+val2)/2.0 + val1 , aTests[0]->GetAttrNum());
		return newtest;
	}
	//cerr<<first<<" "<<r<<" "<<second<<endl;
	//cerr<<val1<<" "<<rt->m_dThreshold<<" "<<val2<<endl;
	return aTests[r]->CloneTest();
}

void CTestContainer2::MakeRealTests(deque<sColAndClass> &sCAC, ITestContainer &tests, int attrNum, const IDataSet* dataset, string sthreshold){
	if(sCAC.empty()) return;
    
	// Sortowanie niemalejcego, od tego zalezy ponizszy kod, szczegolnie
	// sposob wyliczania wartosci threshold.
	sort(sCAC.begin(),sCAC.end(),sColAndClassGreater);
	
	vector<IAttrValue*> v1;
	vector<IAttrValue*> v2;
	double d1 = sCAC[0].s_dVal;
	v1.push_back(sCAC[0].s_pIAClass);
	size_t i=1;
	while(i<sCAC.size() && sCAC[i].s_dVal == d1){
		bool found=false;
		for(size_t t=0; t<v1.size(); t++){
			if( *(v1[t]) == *(sCAC[i].s_pIAClass) ){
				found = true;
				break;
			}
		}
		if(!found){
			v1.push_back(sCAC[i].s_pIAClass);
		}
		i++;
	}
	
	while(i<sCAC.size()){
		v2.clear();
		double d2 = sCAC[i].s_dVal;
		v2.push_back(sCAC[i].s_pIAClass);
		i++;
		while(i<sCAC.size() && sCAC[i].s_dVal == d2){
			bool found=false;
			for(size_t t=0; t<v2.size(); t++){
				if( *(v2[t]) == *(sCAC[i].s_pIAClass) ){
					found = true;
					break;
				}
			}
			if(!found){
				v2.push_back(sCAC[i].s_pIAClass);
			}
			i++;
		}
		if( v1.size() > 1 || v2.size() > 1 || (v1.size() == 1 && v2.size() == 1 && *(v1[0]) != *(v2[0])) ){
			double threshold;
			if(sthreshold=="midpoint"){
				threshold = d2 + (d1 - d2) / 2.0;
				//cout<<"d1="<<d1<<" d2="<<d2<<" threshold="<<threshold<<endl;
			}
			else{
				threshold = d1;
			}
			CRealTest *temp = new CRealTest(threshold,attrNum);
			tests.push(attrNum,temp);
		}

		v1.clear();
		v1=v2;
		d1=d2;
	}
}

size_t CTestContainer2::size(){
	if(m_bChanged){
		UpdateAttrsVector();
	}
	// Na ostatniej pozycji wektora m_vAttrsNumTests jest liczba wszystkich testow
	return m_vAttrsNumTests[m_vAttrsNumTests.size()-1];
}

size_t CTestContainer2::size(size_t attrId) const{
	if(attrId < m_ddITest.size()){
		return m_ddITest[attrId].size();
	}
	return 0;
}

bool CTestContainer2::push(size_t attrId, ITest* test){
	if(attrId != (size_t)test->GetAttrNum()){
		CLogger::LogError("Niespojnosc danych - numery atrybutow musza sie zgadac",__FILE__,__LINE__);
	}

	// Jak jeszcze nie ma danego atrybutu do trzea go dodac
	while(attrId >= m_ddITest.size()){
		deque<ITest*> temp;
		m_ddITest.push_back(temp);
	}
	// kiedy pierwszy test to wstawiamy od razu
	if(m_ddITest[attrId].size() == 0){
		m_ddITest[attrId].push_back(test);
		m_bChanged = true;
		return true;
	}else{
		CRealTest* p_test = dynamic_cast<CRealTest*>(test);
		// Kiedy test rzeczywisty to wstawienie z zachowaniem kolejnosci.
		if(p_test != 0){
			deque<ITest*>::iterator iter = m_ddITest[attrId].begin();
			while( iter != m_ddITest[attrId].end() ){
				CRealTest* p_test2 = dynamic_cast<CRealTest*>(*iter);
				if( p_test->m_dThreshold <= p_test2->m_dThreshold){
					m_ddITest[attrId].insert(iter,test);
					m_bChanged = true;
					return true;
				}
				iter++;
			}
			// jezlie do tad nie wyszlismy to trzeba wstawic na koncu
			m_ddITest[attrId].push_back(test);
			m_bChanged = true;
			return true;
		}else{
			// Tak na prawede to kiedy dla atrybutow nominalnych bedzie tylko po jednym tescie
			// to to sie nie wykona nigdy.
			m_ddITest[attrId].push_back(test);
			m_bChanged = true;
			return true;
		}
	}
}

bool CTestContainer2::erase(size_t attrId, size_t t){
	if(attrId < m_ddITest.size()){
		deque<ITest*>::iterator iter = m_ddITest[attrId].begin();
		size_t counter = 0;
		while(iter != m_ddITest[attrId].end()){
			if(counter == t){
				m_ddITest[attrId].erase(iter);
				// Bardzo wazne ustawienie tego nizej na true;
				m_bChanged = true;
				return true;
			}
			counter++;
			iter++;
		}
	}
	return false;
}

void CTestContainer2::UpdateAttrsVector(){
	if(m_vAttrsNumTests.size() != m_ddITest.size()){
		m_vAttrsNumTests.resize(m_ddITest.size(),0);
	}
	m_duManyOutcomes.clear();
	size_t sum = 0;
	for(size_t t=0; t<m_ddITest.size(); t++){
		sum += m_ddITest[t].size();
		m_vAttrsNumTests[t] = (unsigned)sum;

		// Przy okazji aktualizujemy m_duManyOutcomes
		if(m_ddITest[t][0]->GetSubSetCount() > 2){
			m_duManyOutcomes.push_back(t);
		}
	}
	m_bChanged = false;
}
