#ifndef __CTESTCONTAINER2H__
#define __CTESTCONTAINER2H__

#include "ITestContainer.h"

/**
 * Implementacja kolekcji do przechowywania testow do naczuczania drzew
 * decyzyjnych algorytemami top-down i ewolucyjnymi.
 * @author Marek Grzes' marekg@ii.pb.bialystok.pl, (C) 2004
 */
class CTestContainer2: public ITestContainer{
public:
	CTestContainer2();
	
	/**
	 * Konstruktor, ktory od razu ustawia rozmiar kolejki m_ddITest oraz wektora m_vAttrsNumTests.
	 * W przypadku duzej liczby atrybutow optymalnie bedzie wykorzystac ten drugi konstruktor.
	 */
	CTestContainer2(size_t attrCount);
	
	virtual ~CTestContainer2();

	virtual ITest* operator[](size_t t);
	//virtual ITest* operator[][](size_t attrId, size_t t) const;
	virtual ITest* GetAttrsTest(size_t attrId, size_t t) const;
	
	/**
	 * Zwraca losowe ciecie na atrybucie attrId.
	 */
	virtual double GetRandomThreshold(size_t attrId) const;

	virtual ITest* GetTestWithRandomThreshold(size_t attrId, double val1, double val2) const;

	static void MakeRealTests(deque<sColAndClass> &sCAC, ITestContainer &tests, int attrNum, const IDataSet* dataset, string sthreshold);

	/**
	 * Zwraca liczbe wszystkich testow w kontenerze.
	 */
	virtual size_t size();

	/**
	 * Zwraca liczbe ciec na atrybucie o indeksie attrId.
	 */
	virtual size_t size(size_t attrId) const;

	/**
	 * Dodaje test na pozycji odpowiadajacej atrybutowi attrId. Testy na atrybutach rzeczywistych
	 * sa dodawane w sposob uporzadkowany (niemalejaca).
	 */
	virtual bool push(size_t attrId, ITest* test);

	virtual bool erase(size_t attrId, size_t t);
	
protected:
	/**
	 * Kolejka kolejek z testami.
	 * <b> Numery w pierwszej kolejce musza isc wzgledem CurrentAttrsCount. </b>
	 */
	deque<deque<ITest*> > m_ddITest;
	
	/**
	 * Kolejka indeksow tych testow (pierwszy wymiar tablicy m_ddITest), ktore maja
	 * wiecej niz 2 dwyjscia. Testy, ktore maja 2 wyjscia nie moga sie znalesc w tej
	 * kolejce. Atrybut jest aktualizowany w metodzie: UpdateAttrsVector
	 */
	deque<unsigned> m_duManyOutcomes;

	/**
	 * Zmienna wskazujaca czy zostala zmieniona zawartosc kolekcji testow.
	 */
	bool	m_bChanged;
	
	/**
	 * Skumulowana suma testow dla poszczegolnych atrybutow.
	 */
	vector<size_t>	m_vAttrsNumTests;

	/**
	 * To nie jest typowa implementacja funkcji dziel i zwyciezaj, bo jak wartosc szukana
	 * nie jest znajdowana to zwracana jest wartsc najblizsza (wartosc srodka). Takie wartosci
	 * moga wynikac z dosuwania marginesu w roznych operacjach ewolucyjnych.
	 */
	size_t divideANDconquer(deque<ITest*> dqtests, int p, int k, ITest* test)const;

	void UpdateAttrsVector();
};

#endif
