#include "TDIDTInducer.h"
#include "../MLPDT/DTClassifier.h"
#include "../MLPDT/NominalTest.h"
#include "../MLPDT/NominalTestInnerDisj.h"
#include "../MLPDT/RealTest.h"
#include "../MLPDT/Entropy.h"
#include "../MLPDT/TestCreator.h"
#include "../MLPDT/TestContainer2.h"
#include "../MLPDataset/NomAttr.h"
#include "../MLPDataset/RealAttr.h"
#include <algorithm>
#include <cstdlib>
#include <cmath>
#include <ctime>
#include "../MLPCommon/MLPStatistics.h"
#include <iomanip>

CTDIDTInducer::~CTDIDTInducer(){
}

CTDIDTInducer::CTDIDTInducer(){
	m_sSelectionOfATest = "none";
}

ITest* CTDIDTInducer::PickOutTest(deque<CDataObject*> &deqCI, ITestContainer &tests, const IDataSet *trainingDataSet){
long double maxgain;
size_t maxidx;

	// Jak jest osiagnieta maksymalna glebokosc drzewa to zwracamy NULL i robudowywanie
	// drzewa zostanie zatrzymane.
	if(m_iMaxDepth > -1 && m_iCurrentDepth >= m_iMaxDepth){
		return 0;
	}

	size_t testsSize = tests.size();

	if(m_sSelectionOfATest=="ganzrandom"){
		if(testsSize > 0){
			ITest* res = tests[rand()%testsSize]->CloneTest();
			if(m_bInnerDisjunction){
				RandomInnerDisjuntion(res);
			}
			return res;
		}else{
			return NULL;
		}
	}
	
	if(m_sSelectionOfATest=="dipolar"){
		if(testsSize > 0){
			static CTestCreator tc;
			CDataObject* pob1;
			CDataObject* pob2;
			if(tc.FindMixedDipol(deqCI,pob1,pob2)){
				// zeby nie dodawac przyrostowo do wektora to od razo tworze o maksymalnym rozmiarze
				vector<int> vAttrs(trainingDataSet->GetCurrentAttrCount(),-1);
				// liczba elementow w wektorze atrybutow
				size_t vmax = 0;
				for(size_t t=0; t<trainingDataSet->GetCurrentAttrCount(); t++){
					if( ! ( *((*pob1)[t]) == *((*pob2)[t]) ) ){
						vAttrs[vmax] = (int)t;
						vmax++;
					}
				}
				// attrId jest to losowany index atrybutu do uzycia przy wyszukiwaniu testu
				if(vmax == 0){
					return NULL;
				}
				size_t attrId = vAttrs[rand()%vmax];
				if(tests.size(attrId) == 0){
					return NULL;
				}
				
				if((*(pob1->operator [ ](attrId))).GetAttrType() == e_NominalAttr){
					// Jezeli to jest atrybut nomilany to losujemy jeden test, tu na razie
					// jest wylko jeden test to losujemy jeden z jednego
					ITest* res = tests.GetAttrsTest(attrId, rand()%tests.size(attrId))->CloneTest();
					if(m_bInnerDisjunction){
						RandomInnerDisjuntion(res);
					}
					return res;
				}
				
				if((*(pob1->operator[ ](attrId))).GetAttrType() == e_RealAttr){
					double val1 = pob1->operator [ ](attrId)->m_AttrValue.u_dReal;
					double val2 = pob2->operator [ ](attrId)->m_AttrValue.u_dReal;
					return tests.GetTestWithRandomThreshold(attrId,val1,val2);
				}
				
				return tests[rand()%testsSize]->CloneTest();
			}else{
				return NULL;
			}
		}else{
			return NULL;
		}
	}

	if(m_sSelectionOfATest=="gain criterion"){
		if(testsSize>0){
			maxidx = 0;
			maxgain = CEntropy::Gain(deqCI,tests[0]);
			for(size_t t=1; t<testsSize; t++){
				long double gain = CEntropy::Gain(deqCI,tests[t]);
				if(gain > maxgain){
					maxgain = gain;
					maxidx = t;
				}
			}
			// Jan nie ma przyrostu informacji (maxgain == 0) to warunek stopu -
			// zwrocenie NULL;
			if(maxgain == 0) return NULL;
			ITest* res = tests[maxidx]->CloneTest();
			if(m_bInnerDisjunction){
				RandomInnerDisjuntion(res);
			}
			return res;
		}else{
			return NULL;
		}
	}

	if(m_sSelectionOfATest=="gain ratio criterion"){
		if(testsSize>0){
			maxidx = 0;
			maxgain = CEntropy::GainRatio(deqCI,tests[0]);
//- cout<<maxgain<<" "<<(*maxiter)->GetDescription()<<" "<<((CRealTrainTest*)*maxiter)->m_dThreshold<<endl;
			for(size_t t=1; t<testsSize; t++){
				long double gain = CEntropy::GainRatio(deqCI,tests[t]);
//- cout<<gain<<" "<<(*triter)->GetDescription()<<" "<<((CRealTrainTest*)*triter)->m_dThreshold<<endl;
				if(gain >= maxgain){
					maxgain = gain;
					maxidx = t;
				}
			}
			// Jan nie ma przyrostu informacji (maxgain == 0) to warunek stopu -
			// zwrocenie NULL;
			if(maxgain == 0) return NULL;
			ITest* res = tests[maxidx]->CloneTest();
			if(m_bInnerDisjunction){
				RandomInnerDisjuntion(res);
			}
			return res;
		}else{
			return 0;
		}
	}
	
	/**
	 * Test wybierany jest w sposob losowy. Dla wybranego testu spawdzane jest
	 * gain ratio. Jesli wynosi ono 0 to nastepuje ponowne losowanie. Liczba
	 * losowan jest rowna liczbie testow. Kiedy w zadnym z losowan nie uzyska
	 * sie testu dajacego gain ratio wieksze od zera to zwracany jest NULL
	 * co stanowi warunek stopu.
	 */
	if(m_sSelectionOfATest=="random"){
		if(testsSize>0){
			size_t repeat=testsSize;
			while(repeat-->0){
				size_t n = rand()%testsSize;
				if(CEntropy::GainRatio(deqCI,tests[n])>0){
					ITest* res = tests[n]->CloneTest();
					if(m_bInnerDisjunction){
						RandomInnerDisjuntion(res);
					}
					return res;
				}
			}
		}else{
			return NULL;
		}
	}

	/**
	 * Kiedy parametr m_sSelectionOfATest nie jest ustawiony
	 * zwracany jest test zupelnie losowy.
	 */
	if(testsSize>0){
		ITest* res = tests[rand()%testsSize]->CloneTest();
		if(m_bInnerDisjunction){
			RandomInnerDisjuntion(res);
		}
		return res;
	}
	return 0;
}

bool CTDIDTInducer::SetParam(string sParamName, string sParamValue){
	string::iterator i = sParamName.begin();
	while(i != sParamName.end()){
		*i = tolower(*i);
		++i;
	}
	i = sParamValue.begin();
	while(i != sParamValue.end()){
		*i = tolower(*i);
		++i;
	}

	if(sParamName=="selection of a test" || sParamName=="selectionofatest"){
		if(sParamValue=="gain criterion" ||
			sParamValue=="gain ratio criterion" ||
			sParamValue=="none" ||
			sParamValue=="random" ||
			sParamValue=="dipolar" ||
			sParamValue=="ganzrandom"){
				m_sSelectionOfATest=sParamValue;
				return true;
		}
		else
			return false;
	}

	if(IDTInducerAP::SetParam(sParamName,sParamValue)) return true;
	if(ITDIDTInducer::SetParam(sParamName,sParamValue)) return true;

	return false;
}

void CTDIDTInducer::PrintParamsXML(ostream &ostr, int nsp){
	IDTInducerAP::PrintParamsXML(ostr,nsp);
	ITDIDTInducer::PrintParamsXML(ostr,nsp);
	ostr.fill(' ');
	ostr<<setw(nsp)<<" "<<"<MLPParam Name=\"selectionofatest\" Value=\"gain ratio criterion\" />"<<endl;
	ostr.fill(' ');
}

vector<string> CTDIDTInducer::GetParams() const{
vector<string> res;
stringstream ss;

	res = ITDIDTInducer::GetParams();

	vector<string> temp = IDTInducerAP::GetParams();
	for(size_t t=0; t<temp.size(); t++){
		res.push_back(temp[t]);
	}
	
	res.push_back("selectionofatest");
	res.push_back(m_sSelectionOfATest);
		
	return res;	
}

void CTDIDTInducer::RandomInnerDisjuntion(ITest* pTest){
	CNominalTestInnerDisj* pITest = dynamic_cast<CNominalTestInnerDisj*>(pTest);
	if(pITest!=0){
		size_t n = rand()%pITest->GetSubSetCount();
		if(n<2){
			n = 2;
		}
		pITest->SetSubSets(n);
	}
}

string CTDIDTInducer::GetName(){
	string ret;
	ret="TDIDT";
	return ret;
}

void CTDIDTInducer::PreBuildTree(ITestContainer &tests, const IDataSet *trainingDataSet){
	// Tworzenie zbioru testow.	
	for(unsigned i=0; i<trainingDataSet->GetCurrentAttrCount(); i++){
		CAttrDesc *pCAD = trainingDataSet->GetAttrDesc(i);
		// Test dla atrybutu nominalnego.
		if(pCAD->m_iContinuous == 0){
			CNominalTest* pnomtest = 0;
			if(m_bInnerDisjunction){
				pnomtest = new CNominalTestInnerDisj(pCAD,i);
			}else{
				pnomtest = new CNominalTest(pCAD,i);
			}
			tests.push(i,(ITest*)pnomtest);
		}
		// Test dla atrybutu ciaglego.
		else if(pCAD->m_iContinuous == 1){
			deque<sColAndClass> sCAC = trainingDataSet->GetColAndClass(i);
			CTestContainer2::MakeRealTests(sCAC,tests,i,trainingDataSet,m_sThreshold);
		}
	}

	if(tests.size() == 0){
		CLogger::LogError("No training tests",__FILE__,__LINE__);
	}
}

void CTDIDTInducer::PostBuildTree(ITestContainer &tests){
}
