#include "TDIDTRandInducer.h"
#include "../MLPDT/DTClassifier.h"
#include "../MLPDT/NominalTest.h"
#include "../MLPDT/RealTest.h"
#include "../MLPDT/TestContainer2.h"
#include "../MLPDataset/NomAttr.h"
#include "../MLPDataset/RealAttr.h"
#include <algorithm>
#include <cstdlib>
#include <cmath>
#include <ctime>
#include <iomanip>

CTDIDTRandInducer::CTDIDTRandInducer(void){
	m_uNumberOfIterations = 50;
	m_sSelectionOfATest = "random";
	m_uMinimunItems = 2;
	m_sThreshold = "midpoint";
	m_sPruning = "yes";
}

CTDIDTRandInducer::~CTDIDTRandInducer(void){
}

IClassifier *CTDIDTRandInducer::ActualTrain(const IDataSet *trainingDataSet){
CTestContainer2 tests(trainingDataSet->GetCurrentAttrCount());

	// Utworzenie tablicy wskaznikow do danych (przykladow).
	deque<CDataObject*> tempCDataObject;
	for(unsigned i=0; i<trainingDataSet->GetSize(); i++){
		tempCDataObject.insert(tempCDataObject.end(),(*trainingDataSet)[i]);
	}

	// Wyznaczenie kategorii domyslnej.	
	IAttrValue *d = Category(trainingDataSet,tempCDataObject,NULL);

	CDTreeNode* mintree = NULL;

	PreBuildTree(tests,trainingDataSet);

	for(unsigned i=0; i<m_uNumberOfIterations; i++){
		// Rekurencyja funkcja budowania drzewa.
		CDTreeNode* pCTree = BuildTree(tempCDataObject,tests,d,trainingDataSet);
		
		if(pCTree){
			// Lokalna estymacja parametrow.
			SubTreeErrors(pCTree);
			if(m_sPruning == "yes")
				// Pruning. Po jego wykonaniu estymacja parametrow jest zaktualizowana 
				// poprawnie.
				PruneTree(pCTree);
			else
				// Globalna estymacja parametrow.
				SubTreeDepErrors(pCTree);
			// Usuniecie z wezlow drzewa informacji o przykladach pokrywanych przez 
			// dany wezel.
			pCTree->ClearDeqCI();
			cout<<"["<<i<<"]\tActual error:\t" <<pCTree->GetEstimateError();
			if(mintree)
				cout<<"\tMinimal error:\t"<<mintree->GetEstimateError()<<endl;
			else
				cout<<endl;
			if(mintree == NULL){
				mintree = pCTree;
			}else{
				if(mintree->GetEstimateError() > pCTree->GetEstimateError()){
					CDTreeNode* temp = mintree;
					mintree = pCTree;
					delete temp;
				}else{
					delete pCTree;
				}
			}
		}
	}

	PostBuildTree(tests);

	if(mintree){
		CDTClassifier *pClass = new CDTClassifier(mintree,false,false,trainingDataSet);
		return pClass;
	}else
		return NULL;
}

bool CTDIDTRandInducer::SetParam(string sParamName, string sParamValue){
	string::iterator i = sParamName.begin();
	while(i != sParamName.end()){
		*i = tolower(*i);
		++i;
	}
	i = sParamValue.begin();
	while(i != sParamValue.end()){
		*i = tolower(*i);
		++i;
	}

	if(sParamName=="selection of a test" || sParamName=="selectionofatest"){
		if( sParamValue=="gain criterion" || sParamValue=="gain ratio criterion" )
			return false;
	}

	if(CTDIDTInducer::SetParam(sParamName,sParamValue)) return true;

	if(sParamName=="number of iterations" || sParamName=="numberofiterations"){
		m_uNumberOfIterations = (unsigned)atoi(sParamValue.c_str());
		return true;
	}

	return false;
}

void CTDIDTRandInducer::PrintParamsXML(ostream &ostr, int nsp){
	CTDIDTInducer::PrintParamsXML(ostr,nsp);
	ostr.fill(' ');
	ostr<<setw(nsp)<<" "<<"<!-- Only \"random\" value can by specified for selectionofatest param in MLPDT_TD_AP_RAND classifier -->"<<endl;
	ostr<<setw(nsp)<<" "<<"<MLPParam Name=\"numberofiterations\" Value=\"10\" />"<<endl;
	ostr.fill(' ');
}

vector<string> CTDIDTRandInducer::GetParams() const{
vector<string> res;
stringstream ss;

	res = CTDIDTInducer::GetParams();

	res.push_back("numberofiterations");
	ss.str(""); ss<<m_uNumberOfIterations;
	res.push_back(ss.str());

	return res;
}

string CTDIDTRandInducer::GetName(){
	string ret;
	ret="TDIDT Rand";
	return ret;
}
