/**
 * Author: Marek Grzes <marekg@ii.pb.bialystok.pl>, (C) 2006
 * $Revision: 1.5 $
 * $Date: 2006/04/04 23:15:31 $
 */

#include <sstream>
#include "AddNoise.h"
#include "RealAttr.h"

int CAddNoise::AddNoisyFeatures(IDataSet* data, int n, double min, double max){
	// Lecimy po wszystkich przykladach
	for(size_t i=0; i<data->GetSize(); i++){
		CDataObject *o = (*data)[i];
		// ta petla po dodawanych cechach
		for(int k=0; k<n; k++){
			CRealAttr *nra = new CRealAttr();
			nra->m_bMissedValue = false;
			// Losujemy wartosc w przedziale [min-max]
			nra->m_AttrValue.u_dReal = (max - min)*rand()/((double)RAND_MAX) + min;
			o->InsertAttr(nra);
			
		}
	}
	// Uaktualnienie pol pamietajacych ilosc cech
	data->m_uAttrCount += n;
	data->m_uCurrentAttrCount = data->m_uAttrCount;
	// Anulowanie startych ustawien.
	for(unsigned i=0;i<data->m_vMapVector.size();i++){
		data->m_vMapVector[i]=i;
	}
	// Dodanie kolejnych pozycji do wektora map
	for(int k=0; k<n; k++){
		data->m_vMapVector.push_back(data->m_vMapVector.size());
		// Dodanie tez opisu dodanych cech do datasetu
		CAttrDesc *ad = new CAttrDesc();
		ad->m_iContinuous=1;
		stringstream ss;
		ss<<"noise"<<data->m_vMapVector.size();
		ad->m_sAtrrName=ss.str();
		data->m_vAttrDesc.push_back(ad);
	}

	// Po zmianach wektora map trzeba uaktualnic wektor, ktory od niego zalezy.
	data->SetRealAttrsMap();

	return 0;
}

int CAddNoise::ClassLabels(IDataSet* data, int prob){
	// Zwracanny jest wektor, w ktorym sa wskazniki prosto do danych w dataset
	vector<IAttrValue*> cV = data->GetClassValues();
	// Robienie kopii danych, bo jak sie jakis przyklad zmutuje i jego klasa sie zmini
	// to wtedy w cV pozostanie stary niezmieniony wskaznik.
	for(size_t k=0; k<cV.size(); k++){
		IAttrValue* ia = NULL;
		CNomAttr* na = dynamic_cast<CNomAttr*>(cV[k]);
		if(na){
			ia = new CNomAttr(*na);
		}else{
			CRealAttr* ra = dynamic_cast<CRealAttr*>(cV[k]);
			if(ra){
				ia = new CRealAttr(*ra);
			}else{
				CLogger::LogError("unknown attr type",__FILE__,__LINE__);
				return 1;
			}
		}
		// Odwolanie pamietane w cV[k] zastepujemy wskaznikiem do nowego obiektu.
		cV[k] = ia;
	}
	
	for(size_t t=0; t<data->GetSize(); t++){
		if(rand()%100 <= prob){
			CDataObject *dO = (*data)[t];
			// Losowanie indeksu nowej klasy
			size_t cid = rand()%cV.size();
			// Jesli wylosowalismy ta sama klase to biezemy wartosc kolejna.
			if( *(cV[cid]) == *(dO->GetClass()) ){
				cid++;
				cid %= cV.size();
			}
			IAttrValue* ia = NULL;
			CNomAttr* na = dynamic_cast<CNomAttr*>(cV[cid]);
			if(na){
				ia = new CNomAttr(*na);
			}else{
				CRealAttr* ra = dynamic_cast<CRealAttr*>(cV[cid]);
				if(ra){
					ia = new CRealAttr(*ra);
				}else{
					CLogger::LogError("unknown attr type",__FILE__,__LINE__);
					return 1;
				}
			}
			delete dO->m_pIAClass;
			dO->m_pIAClass = ia;
		}
	}
	// Trzeba zwonic dane zdupliowane w liscie cV
	for(size_t k=0; k<cV.size(); k++){
		delete cV[k];
	}
	return 0;
}

int CAddNoise::Features(IDataSet* data, int prob, double q){
	for(size_t t=0; t<data->GetSize(); t++){
		CDataObject *o = (*data)[t];
		// Tu berzemy liczbe wszystkich atrybutow (czyli podwojne indeksowanie jest ignorowane)
		// Liczbe cech biezemy wprost z wektora w CDataObject
		for(size_t i=0; i<o->m_vIAttr.size(); i++){
			CRealAttr* ra = dynamic_cast<CRealAttr*>(o->m_vIAttr[i]);
			// mieszamy tylko atrybuty rzeczywiste
			if(ra!=0 && rand()%100 <= prob){
				if(!ra->m_bMissedValue){
					double f = ra->m_AttrValue.u_dReal*q*(rand()+1.0)/(RAND_MAX+1.0);
					rand()%2 ? ra->m_AttrValue.u_dReal+=f : ra->m_AttrValue.u_dReal-=f;
				}
			}
		}
	}
	return 0;
}
