/**
 * Author: Marek Grzes <marekg@ii.pb.bialystok.pl>, (C) 2006
 * $Revision: 1.3 $
 * $Date: 2006/01/24 23:49:44 $
 */

#ifndef __CADDNOISE__
#define __CADDNOISE__

#include "IDataSet.h"

/**
 * Klasa typu "control class" zawierajaca operacje dodawania szumu do zbioru danych.
 * Klasa ta jest zaprzyjazniona z CDataObject, bo bezposrennio modyfikuje dane.
 */
class CAddNoise{
public:
	/**
	 * Metoda dodaje n cech (szumu) do zbioru danych, wartosci kazdej dodanej cechy losowane sa
	 * zgodnie z rozkladem jednowstajnym w przeciale min, max. Metoda ta resetuje wektor
	 * podwojnego indeksowania pamietany w zbiorze danych.
	 * @see IDataSet#m_vMapVector
	 * @param data - zbior danych
	 * @param n - n cech do dodania
	 * @param min
	 * @param min
	 */
	int AddNoisyFeatures(IDataSet* data, int n, double min, double max);

	/**
	 * Metoda przeprowadza losowe (z p-em prob) zmiany wartosci klasy.
	 * @param data - zbior danych
	 * @param prob - p-o z jakim kazdy przyklad bedzie podlegal losowemu zaburzeniu wartosci
	 * klasy. Wartosc tego p-a podawany jest w procentach [0-100].
	 */
	int ClassLabels(IDataSet* data, int prob);

	/**
	 * Metoda przeprowadza losowe zmiany wartosci wszystkich atrybutow z p-em prob.
	 * @param data - zbior danych
	 * @param prob - p-o z jakim kazdy atrybut bedzie podlegal losowemu zaburzeniu wartosci.
	 * Wartosc tego p-a podawany jest w procentach [0-100].
	 * @param q - wartosci beda modyfikowane o wartosc_poczatkowa*(0-q]
	 */
	int Features(IDataSet* data, int prob, double q);

};

#endif
