#ifndef __ATTRVALUEH__
#define __ATTRVALUEH__

#include <iostream>
#include <cassert>
#include "../MLPCommon/Logger.h"
#include "../MLPCommon/common.h"

/** Typ danych atrybutu.
 */
enum AttrType
{
	e_NominalAttr,
	e_RealAttr
};

/** 
 * Interfejs wartosci atrybutu.
 * @author Marek Grzes'
 */
class IAttrValue {
	public:
		/**
		 * Wartosc atrybutu.
		 */
		union {
			int u_iNom;
			double u_dReal;
		} m_AttrValue;

		/**
		 * Pole wskazujace czy wartosc atrybutu jest okreslona.
		 * false - nieokreslona
	 	 * true - okreslona
		 */
		bool m_bMissedValue;

		/** 
		 * Zwraca typ atrubutu.
		 * @see AttrType
		 */
		virtual AttrType GetAttrType()const=0;

		virtual bool operator==(const IAttrValue &ia)const=0;
		virtual bool operator!=(const IAttrValue &ia)const=0;
		virtual bool operator>(const IAttrValue &ia)const=0;
		virtual bool operator<(const IAttrValue &ia)const=0;

		/**
		 * Funkcja drukuje odpowiednio sformatowana wartosc do strumienia ostr.
		 * @param ostr - wskaznik do strumienia
		 */
		virtual void PrintValue(ostream &ostr)const=0;

		/**
		 * Konstruktor.
		 */
		IAttrValue(void);

		/**
		 * Konstruktor ustawiajacy pole m_bMissedValue.
		 * @see m_bMissedValue
		 */
		IAttrValue(bool missed);

		/**
		 * Destruktor.
		 */
		virtual ~IAttrValue(void);
};

/**
 * Struktura wykorzystywana do zliczania przykladow o danej wartosci atrybutu.
 */
typedef struct sCounter{
	IAttrValue*	s_pIAttrValue;
	int			s_Counter;
} sCounter;

/**
 * Para wartosc atrybutu w kolumnie i klasa przykladu. Wykorzystywana
 * przy tworzeniu testow dla atrybutow ciaglych.
 */
typedef struct sColAndClass{
	double		s_dVal;
	IAttrValue*	s_pIAClass; 
} sColAndClass;

/**
 * Struktura do reprezentacji pól w macierzy Confiusion Matrix.
 */
typedef struct sConfusionMatrixElement{
	unsigned	s_uNum;
	IAttrValue	*s_pIAGoodClass;
	IAttrValue	*s_pIAClassifiedAs;
} sConfusionMatrixElement;

/**
 * Return whether first element is greater than the second.
 */
bool sColAndClassGreater( sColAndClass elem1, sColAndClass elem2 );

/**
 * Return whether first element is greater than the second.
 */
bool IAgrater(IAttrValue *pIA1, IAttrValue *pIA2);

#endif
