#ifndef __CDATAOBJECTH__
#define __CDATAOBJECTH__

#include "AttrValue.h"
#include <vector>

class IDataSet;

/** 
 * Klasa do reprezentacji pojedynczego przykladu ze zbioru danych.
 * @author Marek Grzes'
 */
class CDataObject{

friend class CAddNoise;

protected:
		/**
		 * Wektor wskaznikow do atrybutow.
		 */	
		vector<IAttrValue*> m_vIAttr;

		/**
		 * Wskaznik na atrybut, ktory jest klasa.
  		 */
		IAttrValue* m_pIAClass;

		/**
		 * Wskaznik do datasetu, do ktorego nalezy instancja.
		 */
		IDataSet *m_pIDataSet;

public:
		/**
		 * Dodaje atrybut na koniec wektora atrybutow.
		 * @param pIA - atrybut do dodania
		 */
		void InsertAttr(IAttrValue *pIA);
		
		/**
		 * Zwraca atrybut, ktory okresla podzial na klasy.
		 * @return NULL - kiedy dataset nie ma okreslonej klasy
		 */
		IAttrValue *GetClass()const;
		
		const IDataSet* GetDataset()const{
			return m_pIDataSet;
		}

		/**
		 * Atrybut o indeksie uIndex przypisuje do m_pIAClass i
		 * usuwa go z listy pozostałych atrybutow. Jesli wczesniej do 
		 * m_pIAClass byla przypisana klasa to zostanie wstawiona na koncu
		 * wektora m_vIAttr. Czyli biezaca klasa staje sie ostatnim atrybutem.
		 */
		void SetClass(unsigned uIndex);
		
		/**
		 * Ustawia atrybut klasy wartoscia podana jako parametr. Moze byc 
		 * wywolywana jedynie dla obiektu, ktory nie ma jeszcze ustawionej klasy.
		 * Zalecane wykorzystywanie metody jedynie przy wczytywaniu danych.
		 * @param pIA - wskaznik do atrybutu, ktory ma byc klasa
		 */
		void SetClass(IAttrValue* pIA);

		/**
		 * Zwraca wskaznik do atrybutu.
		 * @param uIndex - indeks atrybutu
		 * @return NULL - kiedy podany index jest poza zakresem
		 */		
		IAttrValue* operator[](unsigned uIndex)const;

		/**
		 * Zwraca wskaznik do atrybutu rzeczywistego.
		 * @param uIndex - indeks atrybutu w zakresie atybutow rzeczywistych
		 * @return NULL - kiedy podany index jest poza zakresem
		 */
		IAttrValue* GetRealAttrValue(unsigned uIndex)const;

		/**
		* Zaraca wektor mozliwych wartosci atrybutu bedacego klasa.
		* Wektor ten jest posortowany nierosnaco. Kiedy klasa nie jest
		* zdefiniowana w danym zbiorze danych zwracany jest wektor pusty.
		*/
		virtual vector<IAttrValue*> GetClassValues()const;
		
		/**
		 * Konstruktor ustawiajacy wskaznik do IDataSet. Wskazenie to jest potrzebne,
		 * aby przez ten wskaznik można było odwolac sie do wektora realizujacego podwojne
		 * indeksowanie.
		 * @see m_pIDataSet
		 * @see IDataSet#m_vMapVector
		 */
		CDataObject(IDataSet *pIDS);

		virtual ~CDataObject(void);
};


#endif
