#include "DataSetC45.h"
#include "RealAttr.h"
#include "NomAttr.h"
#include <fstream>
#include <cstring>
#include <cstdlib>
#include <sstream>

#define __MLPMAXLINE__ 300000

CDataSetC45::CDataSetC45(char* name):IDataSet(name){
}

CDataSetC45::~CDataSetC45(void){
}

int CDataSetC45::ReadCostData(char *costofclassification, char *costofattrs){
ifstream infile;
char *line;
char *token;

	if(!this->IsClass()){
		CLogger::LogError("can not read cost data when class is not set",__FILE__,__LINE__);
		return 1;
	}
	assert(m_pCAttrDescClass);
	
	if(costofclassification){
		infile.open (costofclassification, ifstream::in);
		if(!infile.good()){
			string str="ifstream::open(";
			str += costofclassification;
			str += ")";
			CLogger::LogError(str.c_str(),__FILE__,__LINE__);
			return 1;
		}
		
		line = (char*)malloc(sizeof(char) * 1024);
		
		while (infile.good()){
			infile.getline(line, 1023 );
			string logstring(line);
			if(line[0] == '|' || line[0]==' ' || strlen(line)==0){
#ifndef NDEBUG
			cerr<<"line: "<<line<<" skipped\n";
#endif
				continue;
			}

			// obciecie wszyskiego co jest komentarzem
			TrimFrom(line,'|',true);
			// obcina od konca wyszstko (znaki pasujace do isspace()) do wystapienia kropki
			TrimFromBack(line,'.',true);
			
			token = strtok(line,",");
			CNomAttr *predC = 0;
			CNomAttr *trueC = 0;
			float cost = 0;
			if(token){
				token = trimOutside(token);
				predC = new CNomAttr(m_pCAttrDescClass);
				predC->m_AttrValue.u_iNom = m_pCAttrDescClass->GetPos(token);
			}else{
				string s(line);
				CLogger::LogError("Error in: '"+logstring+"' ["+s+"]",__FILE__,__LINE__);
				return 1;
			}
			token = strtok(NULL,":");
			if(token){
				token = trimOutside(token);
				trueC = new CNomAttr(m_pCAttrDescClass);
				trueC->m_AttrValue.u_iNom = m_pCAttrDescClass->GetPos(token);
			}else{
				CLogger::LogError("Error in: "+logstring,__FILE__,__LINE__);
				return 1;
			}
			token = strtok(NULL,"\0");
			if(token){
				token = trim(token);
				cost = atof(token);
				if( (*trueC) == (*predC) && cost!=0){
					CLogger::LogError("Cost value should be 0 for correct classification: "+logstring,__FILE__,__LINE__);
				}
				pair<CNomAttr, float> p(*trueC,cost);
				pair<CNomAttr, pair<CNomAttr, float> > mp(*predC,p);
				m_mpCostOfClassification.insert(mp);
			}else{
				CLogger::LogError("Error in: "+logstring,__FILE__,__LINE__);
				return 1;
			}
			delete trueC;
			delete predC;
		}
		m_bCostOfClassification = true;
		free(line);
		infile.close();
	}
	if(costofattrs){
		infile.open (costofattrs, ifstream::in);
		if(!infile.good()){
			string str="ifstream::open(";
			str += costofattrs;
			str += ")";
			CLogger::LogError(str.c_str(),__FILE__,__LINE__);
			return 1;
		}
		
		line = (char*)malloc(sizeof(char) * 1024);
		
		while (infile.good()){
			infile.getline(line, 1023 );
			string logstring(line);
			if(line[0] == '|' || line[0]==' ' || strlen(line)==0){
#ifndef NDEBUG
			cerr<<"line: "<<line<<" skipped\n";
#endif
				continue;
			}

			// obciecie wszyskiego co jest komentarzem
			TrimFrom(line,'|',true);
			// obcina od konca wyszstko (znaki pasujace do isspace()) do wystapienia kropki
			TrimFromBack(line,'.',true);
			token = strtok(line,":");
			if(token){
				token = trimOutside(token);
				char* aCost = strtok(NULL,"\0");
				if(aCost){
					aCost = trim(aCost);
					for(size_t t=0; t<m_uAttrCount; t++){
						if(strcasecmp_(m_vAttrDesc[t]->m_sAtrrName.c_str(),token) == 0 ){
							m_vAttrDesc[t]->m_fCost = atof(aCost);
							break;
						}
					}
				}else{
					CLogger::LogError("Error in: "+logstring,__FILE__,__LINE__);
					return 1;
				}
			}else{
				CLogger::LogError("Error in: "+logstring,__FILE__,__LINE__);
				return 1;
			}
		}
		m_bCostOfAttrs = true;
		free(line);
		infile.close();
	}
	return 0;
}

int CDataSetC45::ReadData(char *data, char *names){
ifstream infile;
char *line;
char *token;
unsigned int atr_num;
	
	if(ReadMetaData(names)!=0) return 1;
	
	infile.open (data, ifstream::in);
	if(!infile.good()){
		string str="ifstream::open(";
		str += data;
		str += ")";
		CLogger::LogError(str.c_str(),__FILE__,__LINE__);
		return 1;
	}
	
	line = (char*)malloc(sizeof(char) * __MLPMAXLINE__);
	
	while (infile.good()){
		infile.getline(line,__MLPMAXLINE__ -1 );
		if(__MLPMAXLINE__ * 0.8 < strlen(line) ){
			CLogger::LogEvent("__MLPMAXLINE__ seems to be too short",__FILE__,__LINE__);
		}
		if(__MLPMAXLINE__ < strlen(line) ){
			CLogger::LogError("__MLPMAXLINE__ is to too short",__FILE__,__LINE__);
		}
		if(line[0] == '|' || line[0]==' ' || strlen(line)==0){
#ifndef NDEBUG
			cerr<<"line: "<<line<<" skipped\n";
#endif
			continue;
		}

		// obciecie wszyskiego co jest komentarzem
		TrimFrom(line,'|',true);
		// obcina od konca wyszstko (znaki pasujace do isspace()) do wystapienia kropki
		TrimFromBack(line,'.',true);
		
		CDataObject *pCDataObject = new CDataObject(this);
		atr_num=0;
		token = strtok(line,",");
		if(m_vAttrDesc[atr_num]->m_iContinuous == 1){
			CRealAttr *pCRealAttr;		
			if(token){
				token = trim(token);
				pCRealAttr = new CRealAttr();
				if(strcmp(token,"?")==0)
					pCRealAttr->m_bMissedValue=true;
				else				
					pCRealAttr->m_AttrValue.u_dReal = atof(token);
				pCDataObject->InsertAttr((IAttrValue*)pCRealAttr);
			}
		}
		else{
			CNomAttr *pCNomAttr;
			if(token){
				token = trimOutside(token);
				pCNomAttr = new CNomAttr(m_vAttrDesc[atr_num]);
				if(strcmp(token,"?")==0)
					pCNomAttr->m_bMissedValue=true;
				else{
					m_vAttrDesc[atr_num]->AddDomainElement(token);
					pCNomAttr->m_AttrValue.u_iNom = m_vAttrDesc[atr_num]->GetPos(token);
				}
				pCDataObject->InsertAttr((IAttrValue*)pCNomAttr);
			}
		}
		while((token=strtok(NULL,","))!=NULL){
			atr_num++;
			if(atr_num>=m_vAttrDesc.size()){
				token = trimOutside(token);
				// Wczytanie klasy jako atrybut nominalny.
				CNomAttr *pCNomAttr;
				pCNomAttr = new CNomAttr(m_pCAttrDescClass);
				if(strcmp(token,"?")==0)
					pCNomAttr->m_bMissedValue=true;
				else{
					m_pCAttrDescClass->AddDomainElement(token);
					pCNomAttr->m_AttrValue.u_iNom = m_pCAttrDescClass->GetPos(token);
				}
				pCDataObject->SetClass((IAttrValue*)pCNomAttr);
				break;
			}
	
			if(m_vAttrDesc[atr_num]->m_iContinuous == 1){
				CRealAttr *pCRealAttr;
				token = trim(token);
				pCRealAttr = new CRealAttr();
				if(strcmp(token,"?")==0)
					pCRealAttr->m_bMissedValue=true;
				else
					pCRealAttr->m_AttrValue.u_dReal=atof(token);
				pCDataObject->InsertAttr((IAttrValue*)pCRealAttr);
			}else{
				CNomAttr *pCNomAttr;
				token = trimOutside(token);
				pCNomAttr = new CNomAttr(m_vAttrDesc[atr_num]);
				if(strcmp(token,"?")==0)
					pCNomAttr->m_bMissedValue=true;
				else{
					m_vAttrDesc[atr_num]->AddDomainElement(token);
					pCNomAttr->m_AttrValue.u_iNom = m_vAttrDesc[atr_num]->GetPos(token);
				}					
				pCDataObject->InsertAttr((IAttrValue*)pCNomAttr);
			}				
		}
		m_dCDataObject.insert(m_dCDataObject.end(),pCDataObject);
	}
	
	infile.close();
	
	m_uAttrCount=(unsigned)m_vAttrDesc.size();
	m_uCurrentAttrCount=m_uAttrCount;
	// Po wczytaniu danych trzeba zainicjowac wektor m_vMapVector
	for(unsigned uNum=0;uNum<m_uCurrentAttrCount;uNum++){
		m_vMapVector.insert(m_vMapVector.end(),uNum);
	}

	// Po zmianach wektora map trzeba uaktualnic wektor, ktory od niego zalezy.
	SetRealAttrsMap();
	
	free(line);
	return 0;
}

int CDataSetC45::ReadMetaData(char *fname){
ifstream infile;
char line[2048];
char *token;
int cl = 0;

	infile.open (fname, ifstream::in);
	if(!infile.good()){
		string str="ifstream::open(";
		str += fname;
		str += ")";
		CLogger::LogError(str.c_str(),__FILE__,__LINE__);
		return 1;
	}
	while (infile.good()){
		// getline zwraca linie bez znakow konca linii
		infile.getline(line,2048);
		if(line[0] == '|' || line[0]==' ' || line[0]=='\n' || line[0]=='\r'\
			|| line[0] == '\t' || strlen(line)==0){
#ifndef NDEBUG
			cerr<<"line: "<<line<<" skipped\n";
#endif
				continue;
		}

		// obciecie wszyskiego co jest komentarzem
		TrimFrom(line,'|',true);
		// obcina od konca wyszstko (znaki pasujace do isspace()) do wystapienia kropki
		TrimFromBack(line,'.',true);
		
		if(cl==0){
			m_bClass=true;
			m_pCAttrDescClass = new CAttrDesc();
			m_pCAttrDescClass->m_iContinuous=0;
			m_pCAttrDescClass->m_sAtrrName="class";
			token = strtok(line,",");
			if(token){
				token = trimOutside(token);
				m_pCAttrDescClass->m_vAttrDomain.insert(m_pCAttrDescClass->m_vAttrDomain.end(),token);
			}
			while((token=strtok(NULL,","))!=NULL){
				token = trimOutside(token);
				m_pCAttrDescClass->m_vAttrDomain.insert(m_pCAttrDescClass->m_vAttrDomain.end(),token);
			}
			cl = 1;
		}else{
			CAttrDesc *pCAttrDesc = new CAttrDesc();
			m_vAttrDesc.insert(m_vAttrDesc.end(),pCAttrDesc);
			token = strtok(line,":");
			if(token){
				token=trimOutside(token);
				pCAttrDesc->m_sAtrrName = token;
			}
			if((token=strtok(NULL,","))!=NULL){
				token = trimOutside(token);
				if(strcasecmp_("continuous",token)==0){
					pCAttrDesc->m_iContinuous = 1;
				} else if(strcasecmp_("discrete",token)==0){
					pCAttrDesc->m_iContinuous = 0;
				}
				else{
					pCAttrDesc->m_iContinuous = 0;
					pCAttrDesc->m_vAttrDomain.insert(pCAttrDesc->m_vAttrDomain.end(),token);
					while((token=strtok(NULL,","))!=NULL){
						token = trimOutside(token);
						pCAttrDesc->m_vAttrDomain.insert(pCAttrDesc->m_vAttrDomain.end(),token);
					}
				}
			}
		}
	}

	infile.close();	
	return 0;
}
