#include "IDataSet.h"
#include <ctime>
#include <cstring>
#include "NomAttr.h"
#include "RealAttr.h"
#include <algorithm>
#include "TempDataSet.h"
#include "../MLPCommon/MLPStatistics.h"

IDataSet::IDataSet(char *name){
	m_sDataSetName = name;
	m_bClass=false;
	m_uAttrCount=0;
	m_uCurrentAttrCount=0;
	m_pCAttrDescClass=NULL;
	m_bTempDataSet = false;
	m_bCostOfClassification = false;
	m_bCostOfAttrs = false;
}

IDataSet::~IDataSet(void){
	// Kiedy jest do obiekt typu CTempDataSet to nie usuwamy danych, bo 
	// naleza one do innego datasetu.
	if(m_bTempDataSet) return;

	for(unsigned i=0; i<m_dCDataObject.size(); i++)
		delete m_dCDataObject[i];
	for(unsigned i=0; i<m_vAttrDesc.size(); i++)
		delete m_vAttrDesc[i];
	if(m_pCAttrDescClass) delete m_pCAttrDescClass;
	
}

inline float IDataSet::CostOfClassification(CNomAttr& predictedClass, CNomAttr& trueClass)const{
	// Jesli klasyfikacja jest bezbledna to koszt wynosi 0.0
	if(predictedClass == trueClass){
		return 0.0;
	}
	// Jesli jest blad klasyfikacji to szukamy kosztu dla tego bledu
	pair<cfcmap_type::const_iterator, cfcmap_type::const_iterator> p = 
		m_mpCostOfClassification.equal_range(predictedClass);
	for (cfcmap_type::const_iterator i = p.first; i != p.second; ++i){
		pair<CNomAttr,float> cost = (*i).second;
		if( cost.first == trueClass){
			return cost.second;
		}
	}
	// Jesli nie zostal znaleziony koszt biezacego bledu to zwracana jest wartosc
	// domyslna kosztu blednej klasyfikacji, czyli 1
	return 1.0;
}

inline float IDataSet::GetCostOfAttribute(size_t currentID)const{
	assert(currentID<m_uCurrentAttrCount);
	return GetAttrDesc(currentID)->m_fCost;
}

void IDataSet::PrintDSInC45Format(ostream *ostr1, ostream *ostr2)const{
time_t	now = time(NULL);
struct	tm   *currDate;
char	dateString[64];	
unsigned int i=0;
unsigned int j=0;

	currDate = localtime(&now);
	strftime(dateString,sizeof dateString,"%m/%d/%Y-%H:%M:%S",currDate);

	if(ostr2 != NULL){
		*ostr2 <<"| Names file in C4.5 format" << endl;
		*ostr2 <<"| Data set name: "<< m_sDataSetName.c_str()<< endl;
		*ostr2 <<"| Saved on "<<dateString<<" by Machine Learning Platform" << endl;
		*ostr2 <<"| It is a comment. I love comments." << endl << endl;
		for(i=0;i<m_pCAttrDescClass->m_vAttrDomain.size();i++){
			*ostr2<<m_pCAttrDescClass->m_vAttrDomain[i].c_str();
			if(i<m_pCAttrDescClass->m_vAttrDomain.size()-1)
				*ostr2<<",";
		}
		*ostr2<<endl<<endl;
		
		// while - GetAttrDesc nie zwroci NULL mozna czytac 
		// opisy atrubutow.
		i=0;
		CAttrDesc *pCAttrDesc;
		while((pCAttrDesc=GetAttrDesc(i))!=NULL){
			*ostr2 << pCAttrDesc->m_sAtrrName.c_str() << ":";
			if(pCAttrDesc->m_iContinuous == 1)
				*ostr2 << " continuous." << endl;
			else{
				for(unsigned int j=0;j<pCAttrDesc->m_vAttrDomain.size();j++){
					*ostr2<<" "<<pCAttrDesc->m_vAttrDomain[j].c_str(); 
					if(j<pCAttrDesc->m_vAttrDomain.size()-1) 
						*ostr2<<",";
					else
						*ostr2<<".";
				}
				*ostr2 << endl;
			}
			i++;
		}
		*ostr2<<endl;
	}

	*ostr1 <<"| Data file in C4.5 format" << endl;
	*ostr1 <<"| Data set name: "<< m_sDataSetName.c_str()<< endl;
	*ostr1 <<"| Saved on "<<dateString<<" by Machine Learning Platform" << endl;
	*ostr1 <<"| It is a comment. I love comments." << endl << endl;
	i=0;
	while(i<m_dCDataObject.size()){
		j=0;
		while(j<m_uCurrentAttrCount){
			IAttrValue *ia = (*m_dCDataObject[i])[j];
			if(ia->m_bMissedValue){
				*ostr1<<"?, ";
			}
			else{
				if(ia->GetAttrType() == e_NominalAttr){
					*ostr1<<((CNomAttr*)ia)->ValueToName()<<", ";
				}
				if(ia->GetAttrType() == e_RealAttr){
					*ostr1 << ia->m_AttrValue.u_dReal<<", ";
				}
			}
			j++;
		}
		IAttrValue *pCNA = m_dCDataObject[i]->GetClass();
		if(pCNA)
			*ostr1<<((CNomAttr*)pCNA)->ValueToName()<<".\n";
		else
			*ostr1<<".\n";
		i++;
	}
	
}
void IDataSet::PrintDSInHumanReadableFormat(ostream *ostr)const{
unsigned int i=0;
unsigned int j=0;
	
	*ostr <<"Data set name: "<< m_sDataSetName.c_str()<< endl << "Classes: ";
	for(i=0;i<m_pCAttrDescClass->m_vAttrDomain.size();i++){
		*ostr << m_pCAttrDescClass->m_vAttrDomain[i].c_str() << " ";
	}
	*ostr << endl << "Attributes:" << endl;
	
	// while - GetAttrDesc nie zwroci NULL mozna czytac 
	// opisy atrubutow.
	i=0;
	CAttrDesc *pCAttrDesc;
	while((pCAttrDesc=GetAttrDesc(i))!=NULL){
		*ostr<<"["<<i<<"] " << pCAttrDesc->m_sAtrrName.c_str() << ":";
		if(pCAttrDesc->m_iContinuous == 1)
			*ostr << " continuous" << endl;
		else{
			for(unsigned int j=0;j<pCAttrDesc->m_vAttrDomain.size();j++){
				*ostr << " " << pCAttrDesc->m_vAttrDomain[j].c_str(); 
			}
			*ostr << endl;
		}
		i++;
	}
	*ostr << endl;
	
	i=0;
	while(i<m_dCDataObject.size()){
		j=0;
		while(j<m_uCurrentAttrCount){
			IAttrValue *ia = (*m_dCDataObject[i])[j];
			if(ia->m_bMissedValue){
				*ostr<<"?,\t";
			}
			else{
				if(ia->GetAttrType() == e_NominalAttr){
					*ostr << ia->m_AttrValue.u_iNom << " (" <<((CNomAttr*)ia)->ValueToName() <<")"<<"\t";
				}
				if(ia->GetAttrType() == e_RealAttr){
					*ostr << ia->m_AttrValue.u_dReal << "\t";
				}
			}
			j++;
		}
		IAttrValue *pCNA = m_dCDataObject[i]->GetClass();
		if(pCNA)
			*ostr<<pCNA->m_AttrValue.u_iNom<<" ("<<((CNomAttr*)pCNA)->ValueToName()<<")\n";
		else
			*ostr << endl;
		i++;
	}
}

void IDataSet::RemoveAttr(vector<unsigned> attr2rem){
	// Anulowanie startych ustawien.
	for(unsigned i=0;i<m_vMapVector.size();i++){
		m_vMapVector[i]=i;
	}
	m_uCurrentAttrCount=m_uAttrCount;
	// Pominiecie atrybutow wystepujacych w attr2rem.
	for(unsigned j=0;j<attr2rem.size();j++){
		for(unsigned k=attr2rem[j] ;k<m_vMapVector.size()-1 ;k++){
			m_vMapVector[k]=m_vMapVector[k+1];
		}
		m_uCurrentAttrCount--;
	}

	// Po zmianach wektora map trzeba uaktualnic wektor, ktory od niego zalezy.
	SetRealAttrsMap();
}

CDataObject *IDataSet::operator[](unsigned uIndex)const{
	if(uIndex >= m_dCDataObject.size()) return NULL;
	return m_dCDataObject[uIndex];
}

bool IDataSet::OnlyRealAttributes()const{
	for(size_t t=0; t<m_uCurrentAttrCount; t++){
		CAttrDesc* d=GetAttrDesc(t);
		if(d->m_iContinuous != 1){
			return false;
		}
	}
	return true;
}

void IDataSet::SetClass(unsigned uIndex){
vector<CAttrDesc*>::iterator desc_iter;
unsigned i;	
	
	if(uIndex >= m_vAttrDesc.size()) return;
	CAttrDesc *pCAD=m_vAttrDesc[uIndex];

	for (i=0,desc_iter=m_vAttrDesc.begin(); desc_iter != m_vAttrDesc.end(); desc_iter++,i++){
		if(i==uIndex){
			m_vAttrDesc.erase(desc_iter);
			break;
		}
	}
	
	if(m_pCAttrDescClass){
		m_vAttrDesc.insert(m_vAttrDesc.end(),m_pCAttrDescClass);
		// Jezeli wczesniej klasa nie bya ustawiona to nalezy
		// zmniejszyc liczb m_uAttrCount oraz zresetowac ustawienia
		// nieaktywnych kolumn.
		m_uAttrCount--;
		vector<unsigned> v;
		RemoveAttr(v);
	}
	m_pCAttrDescClass=pCAD;
	
	// Ustawienie klasy we wszystkich instancjach wartosci.
	for(i=0;i<m_dCDataObject.size();i++){
		m_dCDataObject[i]->SetClass(uIndex);
	}
	
}

CAttrDesc* IDataSet::GetAttrDesc(unsigned uIndex)const{
	if(uIndex >= m_uCurrentAttrCount) return NULL;
	unsigned virtualIndex = m_vMapVector[uIndex];	
	if(virtualIndex >= m_vAttrDesc.size()) return NULL;
	return m_vAttrDesc[virtualIndex];
}

deque<sColAndClass> IDataSet::GetColAndClass(unsigned uIndex)const{
deque<sColAndClass> res;

	if(uIndex >= m_uCurrentAttrCount){
		CLogger::LogError("Index out of scope",__FILE__,__LINE__);
		return res;	
	}

	sColAndClass sCAC;
	for(unsigned i=0; i<m_dCDataObject.size(); i++){
		CDataObject *pCI = m_dCDataObject[i];
		IAttrValue *pIA = (*pCI)[uIndex];
		if(pIA->GetAttrType() != e_RealAttr){
			CLogger::LogError("e_RealAttr expected",__FILE__,__LINE__);
			return res;
		}
		if(pIA->m_bMissedValue){
			CLogger::LogEvent("Missed value skipped",__FILE__,__LINE__);
		}else{
			sCAC.s_dVal = pIA->m_AttrValue.u_dReal;
			sCAC.s_pIAClass = pCI->GetClass();
			res.push_back(sCAC);
		}
	}

	return res;
}

vector<IAttrValue*> IDataSet::GetClassValues()const{
vector<IAttrValue*> res;
	if(!m_bClass) return res;

	for(unsigned i=0; i<GetSize(); i++){
		IAttrValue *pClass = (*this)[i]->GetClass();
		int n=0;
		for(unsigned j=0; j<res.size(); j++){
			if( *(res[j]) == *pClass){
				n=1;
				break;
			}
		}
		if(n==0){
			res.push_back(pClass);
		}
	}

	sort(res.begin(),res.end(),IAgrater);

	return res;
}

unsigned IDataSet::GetCurrentClassCount()const{
	vector<int> vClassIndexes;
	size_t iObsCount = m_dCDataObject.size();

	for (size_t t=0; t<iObsCount; t++){
		CDataObject* dataobj = m_dCDataObject[t];
		IAttrValue* cl = dataobj->GetClass();
		int in = cl->m_AttrValue.u_iNom;
		if (!IsInIndexes(vClassIndexes, in))
			vClassIndexes.push_back(in);	
	}

	return vClassIndexes.size();
}

int IDataSet::GetClassCount() const{
	if(m_bClass){
		return m_pCAttrDescClass->m_vAttrDomain.size();
	}else{
		return -1;
	}
}

int IDataSet::GetRealAttrPos(unsigned uIndex)const{
	if(uIndex >= (unsigned)m_inRealAttrs) return -1;
	unsigned virtualIndex = m_vRealAttrsMapVector[uIndex];
	if(virtualIndex >= m_vAttrDesc.size()) return -1;
	return virtualIndex;
}

deque<const IDataSet*> IDataSet::SplitDataSet(unsigned n, bool stratified)const{
deque<const IDataSet*> ret;
deque<IDataSet*> res;
	for(unsigned i=0; i<n; i++){
		CTempDataSet *temp = new CTempDataSet(*this);
		res.push_back(temp);
	}

	deque<CDataObject*> tempque(this->m_dCDataObject);

	size_t set=0;

	// Podzial stratyfikowany z rownomiernym rozlozeniem przykladow wzgledem klas.
	if(stratified){//003
		while(!tempque.empty()){
			IAttrValue* cl = tempque.front()->GetClass();
			// W tej kolejce zbieramy przyklady z tej samej klasy (co przyklad pierwszy).
			deque<CDataObject*> singleclass;
			singleclass.push_back(tempque.front());
			tempque.pop_front();
			deque<CDataObject*>::iterator it;
			for(it=tempque.begin();it<tempque.end();it++){
				IAttrValue* ia = (*it)->GetClass();
				if( *cl == *ia ){
					singleclass.push_back(*it);
					tempque.erase(it);
				}
			}
			// Rozrzucenie przykladow z jedenej klasy po tworzonych zbiorach.
			while(!singleclass.empty()){
				deque<CDataObject*>::iterator sit;
				sit = singleclass.begin();
				sit += CMLPStatistics::RandVal((int)(singleclass.size())-1);
				res[set]->m_dCDataObject.push_back(*sit);
				singleclass.erase(sit);
				set++;
				if(set >= res.size()) set=0;
			}
		}
	}//003
	// Podzial losowy.
	else{//005
		while(!tempque.empty()){
			deque<CDataObject*>::iterator it;
			it = tempque.begin();
			it += CMLPStatistics::RandVal( (int)(tempque.size()) - 1 );
			res[set]->m_dCDataObject.push_back(*it);
			tempque.erase(it);
			set++;
			if(set >= res.size()) set=0;
		}
	}//005

	while(!res.empty()){
		ret.push_back((const IDataSet*)(res.front()));
		res.pop_front();
	}
	return ret;
}

void IDataSet::SetRealAttrsMap(){
	m_vRealAttrsMapVector.clear();
	m_inRealAttrs = 0;

	for(size_t t=0; t<m_uCurrentAttrCount; t++){
		CAttrDesc* d=GetAttrDesc(t);
		if(d->m_iContinuous == 1){
			m_vRealAttrsMapVector.push_back(t);
			m_inRealAttrs++;
		}
	}
}

deque<const IDataSet*> IDataSet::SplitDataSet23(bool stratified)const{
deque<const IDataSet*> ret;
deque<IDataSet*> res;
	for(unsigned i=0; i<2; i++){
		CTempDataSet *temp = new CTempDataSet(*this);
		res.push_back(temp);
	}

	deque<CDataObject*> tempque(this->m_dCDataObject);

	// Podzial stratyfikowany z rownomiernym rozlozeniem przykladow wzgledem klas.
	if(stratified){//003
		while(!tempque.empty()){
			IAttrValue* cl = tempque.front()->GetClass();
			// W tej kolejce zbieramy przyklady z tej samej klasy (co przyklad pierwszy).
			deque<CDataObject*> singleclass;
			singleclass.push_back(tempque.front());
			tempque.pop_front();
			deque<CDataObject*>::iterator it;
			for(it=tempque.begin();it<tempque.end();it++){
				IAttrValue* ia = (*it)->GetClass();
				if( *cl == *ia ){
					singleclass.push_back(*it);
					tempque.erase(it);
				}
			}
			// Rozrzucenie przykladow z jedenej klasy po tworzonych zbiorach.
			while(!singleclass.empty()){
				deque<CDataObject*>::iterator sit;
				sit = singleclass.begin();
				sit += CMLPStatistics::RandVal((int)(singleclass.size())-1);
				
				double ratio = 1.0;
				if(res[0]->GetSize() > 0){
					ratio = (double)(res[1]->GetSize()) / (double)(res[0]->GetSize());
				}	
				
				if( ratio > 0.5){
					res[0]->m_dCDataObject.push_back(*sit);
				}else{
					res[1]->m_dCDataObject.push_back(*sit);
				}
				singleclass.erase(sit);
			}
		}
	}//003
	// Podzial losowy.
	else{//005
		while(!tempque.empty()){
			deque<CDataObject*>::iterator it;
			it = tempque.begin();
			it += CMLPStatistics::RandVal( (int)(tempque.size()) - 1 );

			double ratio = 1.0;
			if(res[0]->GetSize() > 0){
				ratio = (double)(res[1]->GetSize()) / (double)(res[0]->GetSize());
			}	
				
			if( ratio > 0.5){
				res[0]->m_dCDataObject.push_back(*it);
			}else{
				res[1]->m_dCDataObject.push_back(*it);
			}
			tempque.erase(it);
		}
	}//005

	while(!res.empty()){
		ret.push_back((const IDataSet*)(res.front()));
		res.pop_front();
	}
	return ret;
}

void IDataSet::SetDataSetName(string name){
	m_sDataSetName=name;
}

char* IDataSet::TrimFrom(char *str, char c, bool inclusion){
	unsigned len = strlen(str);
	for(unsigned i=0;i<len;i++){
		if(str[i]==c){
			if(inclusion){
				str[i]='\0';
				break;
			}else{
				if(i+1 < len){
					str[i+1]='\0';
					break;
				}
			}
		}
	}
	return str;
}

char* IDataSet::TrimFromBack(char *str, char c, bool inclusion){
	unsigned len = strlen(str);
	for(unsigned i=len-1;i>=0;i--){
		// Jak spotkamy znak c to obcinamy do niego (inclusion or not)
		if(str[i]==c){
			if(inclusion){
				str[i] = '\0';
			}else{
				// Uzyta tu wartosc len, moze byc slugoscia calosci (bo sprawdzamy czy nie wyjdziemy poza zakres calego stringa).
				if(i<len-1){
					str[i+1]='\0';
				}
			}
			break;
		}
		// Jak spotkamy znak inny niz isspace to za nim wstawiamy znak konca (czyli biale znaki mamy przeskoczone).
		if(isspace(str[i]) == 0){
			str[i+1] = '\0';
			break;
		}
	}
	return str;
}

// DORDU BEGIN
bool IDataSet::IsInIndexes(const vector<int> &vClassIndexes, int iClassIndex)const
{
	size_t iClassCount=vClassIndexes.size();

	for (size_t i=0; i<iClassCount; i++)
		if (vClassIndexes[i] == iClassIndex)
			return true;

	return false;
}
//DORDU END
