#ifndef __IDATASETH__
#define __IDATASETH__

#include "DataObject.h"
#include "AttrDesc.h"
#include "NomAttr.h"
#include <deque>
#include <map>

/** 
 * Interfejs zbioru danych.
 * @author Marek Grzes'
 */
class IDataSet{
friend class CDataObject;
friend class CTempDataSet;
friend class CAddNoise;

protected:

	bool m_bCostOfClassification;
	bool m_bCostOfAttrs;
	
	/**
	 * True jezli jest to zbior tymczasowy majacy odwolania do danych z innego
	 * danych zbioru.
	 * @see CTempDataSet
	 */
	bool m_bTempDataSet;

	/**
	 * Wektor wszystkich przykladow danego zbioru danych.
	 */
	deque<CDataObject*> m_dCDataObject;

	/**
	 * Wektor obiektow opisujacych kazdy z atrybutow danych (bez
	 * opisu klasy jesli taka została okreslona).
	 * @see m_pCAttrDescClass
	 */
	vector<CAttrDesc*> m_vAttrDesc;

	/**
	 * Opis atrybutu klasy.
	 * @see m_bClass
	 */
	CAttrDesc* m_pCAttrDescClass;

	/**
	 * Odwzorowanie indeksow, które pozwolaja realizowac eliminowanie
	 * okreslonych kolumn, bez fizycznego usuwania ich ze zbioru danych.
	 */
	vector<unsigned> m_vMapVector;

	/**
	 * Wektor potrojnego indeksowania (indeksuje wzgledem m_vMapVector)
	 * w celu zapamietania pozycji, na ktorych sa cechy rzeczywiste.
	 */
	vector<unsigned> m_vRealAttrsMapVector;

	/**
	 * Liczba atrybutow rzeczywistych z uwzglednieniem wektorow
	 * podwojnego i potrojnego indeksowania.
	 */
	int m_inRealAttrs;

	/**
	 * Wskazuje czy zostala zdefiniowana klasa dla tego zbioru danych.
	 * true - jest klasa
	 * false - brak klasy
	 * @see m_pCAttrDescClass
	 */
	bool m_bClass;

	/**
	 * Liczba wszystkich atrybutow (bez klasy).
	 */
	unsigned m_uAttrCount; 

	/**
	 * Liczba aktualnych atrybutow (bez klasy).
	 * @see m_vMapVector
	 */
	unsigned m_uCurrentAttrCount;

	typedef multimap<CNomAttr, pair<CNomAttr,float> > cfcmap_type;
	/**
	 * Struktura do pamietamia kosztu klasyfikacji. Pierwszy klucz CNomAttr
	 * to klasa przewidywana. Drugi klucz CNomAttr to klasa prawdziwa. float to
	 * koszt klasyfikacji. 
	 * Jesli jakas para nie wystapi w tej strukturze, to oznacza
	 * ze koszt wynosi zero dla poprawnej klasyfikacji i jeden dla bledu.
	 * @see CostOfClassification
	 */
	cfcmap_type m_mpCostOfClassification;
	
	/**
	 * Funkcja bierze string str i za pierwszym wystapieniem znaku c wstawia
	 * znak konca linii '\0'.
	 * @param inclusion true to znak c jest rowniez obcinany, false to obciecie
	 * nastepuje za znakiem c.
	 * @return Zwraca przekazany w parametrze string str.
	 */
	virtual char* TrimFrom(char *str, char c, bool inclusion);
	
	/**
	 * Podobnie jak funkcja wyzej obcina do pierwszego wystapienia, tylko ze tym razem
	 * idzie od konca. Szukajac znaku c i idac od konca pomija znaki, ktore pasuja do
	 * isspace().
	 * @param inclusion true to znak c jest rowniez obcinany, false to obciecie
	 * nastepuje za znakiem c (liczac od poczatku napisu).
	 * @return Zwraca przekazany w parametrze string str.
	 */
	char* TrimFromBack(char *str, char c, bool inclusion);
	
	// DORDU BEGIN
	/**
	 * Sprawdzenie, czy iClassIndex jest jednym z elementow wektora vClassIndexes.
	 */
	bool IsInIndexes(const vector<int> &vClassIndexes, int iClassIndex)const;
	// DORDU END

public:
	/**
	 * Nazwa zbioru danych.
	 */
	string m_sDataSetName;

	/**
	 * Funkcja zwraca koszt klasyfikacji przykladu jako predictedClass kiedy w 
	 * rzeczywistosci jest on klasy trueClass.
	 * Domyslnie zwracane jest zero dla poprawnej klasyfikacji i jeden dla bledu.
	 * @return cost of classification
	 * @see m_mpCostOfClassification
	 */
	virtual float CostOfClassification(CNomAttr& predictedClass, CNomAttr& trueClass)const;

	/**
	 * Funkcja zwraca koszt atrybutu o indeksie currentID. Uwzglednia wektor
	 * podwojnego indeksowania m_vMapVector.
	 */
	virtual float GetCostOfAttribute(size_t currentID)const;
		
	bool IsCostOfClassification()const{
		return m_bCostOfClassification;
	}
	
	bool IsCostOfAttrs()const{
		return m_bCostOfAttrs;
	}
	
	/**
	 * Funkcja uwzglednia podwojne indeksowanie. Atrybuty wylaczone nie sa uwzgleniane w tym
	 * sprawdzaniu.
	 * @see m_vMapVector
	 * @see GetAttrDesc
	 * @return - funkcja zwraca true kiedy zbior danych ma tylko tylko rzeczywiste atrybuty.
	 */
	bool OnlyRealAttributes()const;
	
	/**
	 * Atrybuty o numerach przekazanych w parametrze sa oznaczane
	 * jako niewidoczne. Kazde wywolanie tej funkcji dziala na
	 * nowym wektorze, tzn. nie jest uwzgledniany wynik poprzedniego
	 * dzialania (jest ono najpierw anulowane).
	 * @param attr2rem - wektor numerew atrybutew do usunięcia
	 * @see m_vMapVector
	 * @see m_uCurrentAttrCount
	 */
	virtual void RemoveAttr(vector<unsigned> attr2rem);

	/**
	 * Podaje aktualna liczbe atrybutow. Nie uzglednia klasy.
	 * Klasa jest ostatnim atrybutem i nie jest uwzgledniana w tej
	 * liczbie.
	 * @see m_uCurrentAttrCount
	 * @see m_vMapVector
	 */
	virtual unsigned GetCurrentAttrCount()const{
		return m_uCurrentAttrCount;
	}		

	virtual unsigned GetCurrentRealAttrCount()const{
		return m_inRealAttrs;
	}
	
	/**
	 * Zwraca obiekt opisu dla atrybutu o podanym indeksie uIndex.
	 * Uwzglednia fakt, ze niektore aktrybuty nalezy pominac.
	 * @param uIndex - indeks atrybutu
	 * @return NULL - kiedy uIndex poza zakrasem.
	 * @see m_vMapVector
	 */
	virtual CAttrDesc* GetAttrDesc(unsigned uIndex)const;

	/**
	 * Zwraca opis klasy.
	 * @return NULL - kiedy dataset nie ma okreslonej klasy.
	 * @see m_pCAttrDescClass
	 */	
	virtual CAttrDesc* GetClassDesc()const{
		if(!m_bClass) return NULL;
		return m_pCAttrDescClass;
	}

	/**
	 * Wczytuje dane.
	 * Znaczenie parametrow zalezy od konkretnej implementacji.
	 * @return Zero kiedy funkcja konczy sie bez bledu.
	 */
	virtual int ReadData(char *str1, char *str2)=0;

	/**
	 * Wczytuje dane dotyczace kosztow.
	 * Znaczenie parametrow zalezy od konkretnej implementacji.
	 * @return Zero kiedy funkcja konczy sie bez bledu.
	 */
	virtual int ReadCostData(char *str1, char *str2)=0;
	
	/**
	 * Drukuje dane w formacie C4.5.
	 * @param ostr1 - wskaznik do strumienia, do ktorego maja byc wyslane dane
	 * @param ostr2 - wsaznik do sturmienia, do ktorego maja byc zapisane metadane,
	 * jak w przypadku formatu C4.5. Kiedy ostr2 jest NULL metadane nie sa
	 * zapisywane.
	 */
	virtual void PrintDSInC45Format(ostream *ostr1, ostream *ostr2=NULL)const;

	/**
	 * Drukuje dane w formacie czytelnym do przegladania. 
	 */
	virtual void PrintDSInHumanReadableFormat(ostream *ostr)const;

	/**
	 * Ustawienie klasy dla zbioru danych.
	 * @param uIndex - indeks atrybutu, który ma byc klasa.
	 * @see m_pCAttrDescClass
	 */
	virtual void SetClass(unsigned uIndex);

	/**
	 * Zwraca pozycje atrybutu rzeczywistego pod ktora znajduje sie on
	 * w danych indeksowanych przez m_vMapVector.
	 */
	int GetRealAttrPos(unsigned uIndex)const;
	
	/**
	 * Zwraca ilosc przykladow w zbiorze danych.
	 * @see m_uCurrentAttrCount
	 */
	virtual unsigned GetSize()const{
		return (unsigned)(m_dCDataObject.size());
	}

	CDataObject *operator[](unsigned uIndex)const;

	/**
	 * Do sprawdzenia czy zbior danych ma okreslona klase.
	 * @see m_bClass
	 */
	virtual bool IsClass()const{
		return m_bClass;
	}
	
	/**
	 * Zwraca kolejke sturktur, zawierajaca wartosci atrybutu o indeksie 
	 * uIndex wraz z klasa przykladu, z ktorej pochodzi dana wartosc.
	 *
	 * ATRYBUTY NIEOKRESLONE SA POMIJANE!
	 * @param uIndex - indeks atrybutu do pobrania
	 */
	virtual deque<sColAndClass> GetColAndClass(unsigned uIndex)const;

	/**
	 * Zwraca aktualna liczbe klas.
	 */
	virtual unsigned GetCurrentClassCount() const;

	/**
	 * Zwraca liczbe klas WSZYSTKICH klas.
	 * @return -1 kiedy zbior nie ma okreslonej klasy.
	 */
	virtual int GetClassCount() const;
	
	/**
	 * Zaraca wektor mozliwych wartosci atrybutu bedacego klasa.
	 * Wektor ten jest posortowany nierosnaco. Kiedy klasa nie jest
	 * zdefiniowana w danym zbiorze danych zwracany jest wektor pusty.
	 */
	virtual vector<IAttrValue*> GetClassValues()const;

	/**
	 * Metoda dokonuje podzialu zbioru danych. Zwracana jest kolejka zawierajaca
	 * n IDataSet'ow. Zwrócone IDataSet'y zawieraja odwolania do tych samych danych
	 * co wyjsciowy IDataSet. Dlatego nie mozna modyfikowac danych w IDataSet'ach
	 * zwroconych ta metoda.
	 * @see CTempDataSet
	 */
	virtual deque<const IDataSet*> SplitDataSet(unsigned n, bool stratified=true)const;

	/**
	 * Metoda dokonuje podzialu zbioru danych na czesc uczaca (2/3) i czesc testowa (1/3).
	 * Zwracana jest kolejka zawierajaca
	 * 2 IDataSet'y. Zwrocone IDataSet'y zawieraja odwolania do tych samych danych
	 * co wyjsciowy IDataSet. Dlatego nie mozna modyfikowac danych w IDataSet'ach
	 * zwroconych ta metoda.
	 * @return pierwszy element tablicy jest zbiorem uczacym
	 * @return drugi element tablicy jest zbiorem testowym
	 * @see CTempDataSet
	 */
	virtual deque<const IDataSet*> SplitDataSet23(bool stratified=true)const;

	/** Ustawia nazwe dla zbioru danych.
	 * @param name - nazwa zbioru danych
	 */
	virtual void SetDataSetName(string name);

	/**
	 * Ustawia wektor potrojnego indeksowania dla atrybutow rzeczywistych.
	 * @see m_vRealAttrsMapVector
	 */
	void SetRealAttrsMap();

	/**
	 * Konstruktor ustawiajacy nazwe zbioru danych.
	 */
	IDataSet(char *name);
	
	virtual ~IDataSet(void);
};

#endif
