#ifndef __CTEMPDATASET__
#define __CTEMPDATASET__

#include "IDataSet.h"

/** Klasa do wykorzystnia, kiedy potrzebny jest dataset wskazujacy na dane 
 * z innego datasetu. Realizuje sie to w celu unikniecia kopiowania danych.
 * Przy usuwaniu obiektow typu CTempDataSet nie mozna usuwac danych, gdyż sa
 * one wspoldziekone z wlaaciwym datasetem. Fakt ten jest sprawdzany w destruktorze
 * IDataSet.
 * @author Marek Grzes'
 * @see IDataSet#m_bTempDataSet
 */
class CTempDataSet : public IDataSet
{
public:
	/**
	 * Pusta implementacja.
	 * @return Zwraca 1 wskazujac blad.
	 */
	virtual int ReadData(char *str1, char *str2);
	
	/**
	 * Pusta implementacja.
	 * @return Zwraca 1 wskazujac blad.
	 */
	virtual int ReadCostData(char *str1, char *str2);
	
	/** Operator dodania przykladow z innego zbioru danych. Poniewaz oba zbiory
	 * danych sa zbiorami tymczasowamy typu CTempDataSet, dodawane sa tylko
	 * wskazniki. Danych tych nie wolno usuwac. Usuniecie moze nastapic tylko
	 * w glownym zbiorze danych.
	 */
	virtual IDataSet& operator+=(const IDataSet &ids);

	/**
	 * Konstruktor kopujacy, ktory kopiuje wszystkie pola oprócz odwolan do danych.
	 * Odwołania do danych musza byc ustawnione po stworzeniu obiektu. 
	 * Wykorzytywane jest to przy dzieleniu zbioru danych, np. do kroswalidacji.
	 * @see IDataSet#SplitDataSet
	 */
	CTempDataSet(const IDataSet &templ);

	// DORDU BEGIN
	/**
	 * Konstruktor utworzy zbior danych na podstawie templ.
	 * W nowym zbiorze danych znajda sie wylacznie obiekty,
	 * dla ktorych indeks klasy jest rowny ktoremus z indeksow
	 * podanych w vClassIndexes
	 */
	CTempDataSet(const IDataSet &templ, const vector<int> &vClassIndexes);
	// DORDU END

	CTempDataSet(char* name);
	virtual ~CTempDataSet(void);

};

#endif
