#include "DatasetDescription.h"
#include "../../mlpcore/MLPDataset/DataSetC45.h"

#include <sstream>

CDatasetDescription::CDatasetDescription(void){
	m_ptrainDS = 0;
	m_ptestDS = 0;
	m_bStratified = true;
	m_iFolds = 10;
	m_bCostOfClassification = false;
	m_bCostOfAttrs = false;
	m_bmakeSplit = false;
	m_iExperimentType = 0;
	m_sDSFormat = "C4.5";
}

CDatasetDescription::~CDatasetDescription(void){
	if(m_ptrainDS) delete m_ptrainDS;
	if(m_ptestDS) delete m_ptestDS;
}

DSType CDatasetDescription::GetDSType(){
	if(m_iExperimentType == 0){
		if(m_bmakeSplit){
			return e_CVMAKESPLIT;
		}else{
			return e_CV;
		}
	}
	return e_TEST;
}

bool CDatasetDescription::LoadData(){
	// Tu mozna bedzie dopisac obsluge innych formatow danych.
	if(m_sDSFormat=="C4.5"){
		// Test albo dzielenie danych do kroswalidacji.
		if(m_iExperimentType == 1 || m_bmakeSplit){
			CDataSetC45 *trainds = new CDataSetC45("");
			trainds->SetDataSetName(m_sName);
			string data = m_sPath2Stem+".data";
			string names = m_sPath2Stem+".names";
			trainds->ReadData((char*)(data.c_str()),(char*)(names.c_str()));
			if(!m_vsCols2Rem.empty()){
				trainds->RemoveAttr(m_vsCols2Rem);
			}
			if(m_bCostOfClassification){
				string cost = m_sPath2Stem+".costs";
				trainds->ReadCostData((char*)(cost.c_str()),NULL);
			}
			if(m_bCostOfAttrs){
				string testscost = m_sPath2Stem+".cost";
				trainds->ReadCostData(NULL,(char*)(testscost.c_str()));
			}
			m_ptrainDS = trainds;
			// Test
			if(m_iExperimentType == 1){
				CDataSetC45 *testds = new CDataSetC45("");
				testds->SetDataSetName(m_sName);
				string testStem = m_sPath2Stem+".test";
				testds->ReadData((char*)(testStem.c_str()),(char*)(names.c_str()));
				if(!m_vsCols2Rem.empty()){
					testds->RemoveAttr(m_vsCols2Rem);
				}
				if(m_bCostOfClassification){
					string cost = m_sPath2Stem+".costs";
					testds->ReadCostData((char*)(cost.c_str()),NULL);
				}
				if(m_bCostOfAttrs){
					string testscost = m_sPath2Stem+".cost";
					testds->ReadCostData(NULL,(char*)(testscost.c_str()));
				}
				m_ptestDS=testds;
			}
		}else{
			// Wczytanie zadanego podzialu danych do kroswalidacji.
			string names = m_sPath2Stem+".names";
			for(int t=0; t<m_iFolds; t++){
				CDataSetC45 *ds = new CDataSetC45("");
				ds->SetDataSetName(m_sName);
				stringstream ss;
				ss<<t;
				string data = m_sPath2Stem+".data."+ss.str();
				ds->ReadData((char*)(data.c_str()),(char*)(names.c_str()));
				if(!m_vsCols2Rem.empty()){
					ds->RemoveAttr(m_vsCols2Rem);
				}
				if(m_bCostOfClassification){
					string cost = m_sPath2Stem+".costs";
					ds->ReadCostData((char*)(cost.c_str()),NULL);
				}
				if(m_bCostOfAttrs){
					string testscost = m_sPath2Stem+".cost";
					ds->ReadCostData(NULL,(char*)(testscost.c_str()));
				}
				
				m_deqDatasets.push_back(ds);
			}
		}
	}

	return true;
}

void CDatasetDescription::ReleaseData(){
	if(m_ptrainDS){
		delete m_ptrainDS;
		m_ptrainDS = 0;
	}
	if(m_ptestDS){
		m_ptestDS = 0;
		delete m_ptestDS;
	}
	if(m_deqDatasets.size() > 0){
		for(size_t t=0; t<m_deqDatasets.size(); t++){
			delete m_deqDatasets[t];
		}
		m_deqDatasets.clear();
	}
}

