#include "EAIDTIndividual.h"
#include "../../mlpcore/MLPDT/ObliqueTest.h"

CEAIDTIndividual::CEAIDTIndividual(void):IIndividual(){
	m_CDTIndRoot = NULL;
}

CEAIDTIndividual::CEAIDTIndividual(const CEAIDTIndividual &templ):IIndividual((const IIndividual&)templ){
	m_CDTIndRoot = new CDTreeNode(*templ.m_CDTIndRoot);
}

IIndividual* CEAIDTIndividual::Clone() const{
	IIndividual* res = new CEAIDTIndividual(*this);
	return res;
}

CEAIDTIndividual::~CEAIDTIndividual(void){
	if(m_CDTIndRoot) delete m_CDTIndRoot;
}

double CEAIDTIndividual::GetNodeCount_APOB(double pr){
	return GetNodeCount_APOB(pr,m_CDTIndRoot);
}

double CEAIDTIndividual::GetNodeCount_APOB(double pr, CDTreeNode* node){
	double n;
	CObliqueTest* oblTest = dynamic_cast<CObliqueTest*>(node->m_pITest);
	// Wliczenie wezla biezacego
	if(oblTest==NULL){
		n = 1.0;
	}else{
		unsigned nz=0;
		unsigned N=(unsigned)(oblTest->m_vdWReal.size())-1;
		for(size_t t=0; t<N; t++){
			if(oblTest->m_vdWReal[t] != 0){
				nz++;
			}
		}
		if(nz<N){
			// Jak pr=0 to liczymy jak w AP a jak 1, to kazda checha jest liczona
			// jako wezel
			n = 1 + (nz - 1) * pr;
		}else{
			n = 1.0;
		}
	}
	for(unsigned i=0; i<node->m_vBranch.size(); i++){
		n += GetNodeCount_FS(pr,node->m_vBranch[i]);
	}
	return n;	
}

double CEAIDTIndividual::GetNodeCount_FS(double beta){
	return GetNodeCount_FS(beta,m_CDTIndRoot);
}

double CEAIDTIndividual::GetNodeCount_FS(double beta, CDTreeNode* node){
	double n;
	CObliqueTest* oblTest = dynamic_cast<CObliqueTest*>(node->m_pITest);
	// Wliczenie wezla biezacego
	if(oblTest==NULL){
		n = 1.0;
	}else{
		unsigned nz=0;
		unsigned N=(unsigned)(oblTest->m_vdWReal.size())-1;
		for(size_t t=0; t<N; t++){
			if(oblTest->m_vdWReal[t] != 0){
				nz++;
			}
		}
		if(nz<N){
			n = ( 1 - beta) + beta * (double)nz/(double)N;
		}else{
			n = 1.0;
		}
	}
	for(unsigned i=0; i<node->m_vBranch.size(); i++){
		n += GetNodeCount_FS(beta,node->m_vBranch[i]);
	}
	return n;	
}
