/**
 * $Revision: 1.107 $
 * $Date: 2006/07/13 09:25:37 $
 */
 
#include "EAIDTInducer.h"
#include <sstream>
#include <algorithm>
#include <ctime>
#include <cmath>
#include <string>
#include "../../mlpcore/MLPDataset/TempDataSet.h"
#include "../../mlpcore/MLPDT/DTClassifier.h"
#include "../../mlpcore/MLPCommon/MLPStatistics.h"
#include "../../mlpcore/MLPDataset/NomAttr.h"
#include "../../mlpcore/MLPDataset/RealAttr.h"
#include <iomanip>
#include <typeinfo>

// g++ -E -dM test.cpp | grep -i linux

bool CDTreeNodeLess(CDTreeNode *n1, CDTreeNode *n2){
	if(n1->m_dE < n2->m_dE){
		return true;
	}
	return false;
}


bool CDTreeNodePure(CDTreeNode *n){
	if(n->m_dE==0){
		return true;
	}
	return false;
}


IEAIDTInducer::IEAIDTInducer(void){
	m_uMaxTreeSize = 0;
	m_uAverageTreeSize = 0;
	m_bUseMaxTreeSize = false;
	m_bUseAverageTreeSize = false;
	m_dFitnessAlfa = 0.05;
	m_bMaximizeFitness = true;
	m_iMaxFitnessProbability = 0;
	m_dGamma = 0.1;
	m_sFitnessCalculation = "sum";
	m_fMaxCostOfClassification = 1;
	m_fMaxCostOfAttributes = 1;
	m_iMutationType = 1;
	// Dla InitMaxDepth = 7, drzewa z populacji poczatkowej beda nie wieksze niz 255 wezlow
	m_iInitMaxDepth = 7;

	m_bInnerCrossValidation = true;
	m_iICVFolds	= 2;
	m_iICVRe_runs	= 5;
	m_sICVParam	= "alfa";
	m_dICVlvalue	= 0.001;
	m_dICVrvalue	= 0.1;
	m_iICVsteps	= 10;
}

IEAIDTInducer::~IEAIDTInducer(void){
}

void IEAIDTInducer::CheckParams(){
	int p=0;
	for(size_t t=0; t<m_viPc.size(); t++){
		p+=m_viPc[t];
	}
	if(p!=100){
		CLogger::LogEvent("error in pc values",__FILE__,__LINE__);
	}
	p=0;
	for(size_t t=0; t<m_viPm.size(); t++){
		p+=m_viPm[t];
	}
	if(p!=100){
		CLogger::LogEvent("error in pm values",__FILE__,__LINE__);
	}
	p=0;
	for(size_t t=0; t<m_viPmL.size(); t++){
		p+=m_viPmL[t];
	}
	if(p!=100){
		CLogger::LogEvent("error in pml values",__FILE__,__LINE__);
	}
	if(!m_bCostOfClassification && m_sFitnessCalculation=="division"){
		CLogger::LogEvent("fitnesscalculasion=division is used with costofclassification=no",__FILE__,__LINE__);
	}
}

void IEAIDTInducer::EstimateMaxTreeSize(){
	m_uMaxTreeSize = 0;
	for(unsigned i=0; i<m_dqPopulation.size(); i++){
		if(((CEAIDTIndividual*)m_dqPopulation[i])->m_CDTIndRoot->GetNodeCount() > m_uMaxTreeSize)
			m_uMaxTreeSize = ((CEAIDTIndividual*)m_dqPopulation[i])->m_CDTIndRoot->GetNodeCount();
	}
	if(m_iLogLevel>=1){
		stringstream ss;
		ss<<"Max tree size = "<<m_uMaxTreeSize;
		CLogger::LogEvent(ss.str().c_str(),"",__LINE__,m_ofsLogFile);
	}
}

void IEAIDTInducer::EstimateAverageTreeSize(){
	vector<unsigned> vu;
	for(unsigned i=0; i<m_dqPopulation.size(); i++){
		vu.push_back(((CEAIDTIndividual*)m_dqPopulation[i])->m_CDTIndRoot->GetNodeCount());
	}
	CMLPStatistics stat;
	m_uAverageTreeSize = stat.MLPAverage<unsigned>(vu);
	if(m_iLogLevel>=1){
		stringstream ss;
		ss<<"Average tree size = "<<m_uAverageTreeSize;
		CLogger::LogEvent(ss.str().c_str(),"",__LINE__,m_ofsLogFile);
	}
}

void IEAIDTInducer::GetNodeClassification(CDTreeNode* node){
	node->m_dN = node->m_deqCI.size();
	node->m_dE=0;
	node->m_dEstE=0;
	node->m_iSubtreeSize = 1;
	// W lisciach koszt atrybutow jest zero
	node->m_fCostOfAttrs = 0;

	if(node->m_dN > 0){
		// Obliczenie liczby przykladow niepoprawnie klasyfikowanych.
		if(!m_bCostOfClassification){
			for(unsigned i=0; i<node->m_deqCI.size(); i++){
				if( *(node->m_deqCI[i]->GetClass()) != *(node->m_pIAClass) )	node->m_dE++;
			}
		}else{
			node->m_fCostOfClassification = 0.0;
			node->m_uNExamples = node->m_deqCI.size();
			for(unsigned i=0; i<node->m_uNExamples; i++){
				if( *(node->m_deqCI[i]->GetClass()) != *(node->m_pIAClass) )	node->m_dE++;
				try{
					CNomAttr& predC = dynamic_cast<CNomAttr&>(*node->m_pIAClass);
					CNomAttr& trueC = dynamic_cast<CNomAttr&>(*(node->m_deqCI[i]->GetClass()));
					node->m_fCostOfClassification += m_pDataset->CostOfClassification(predC,trueC);
				}
				catch(bad_cast){
				}
			}
			//REM assert(m_deqCI.size()>0);
			//REM m_fCostOfClassification /= (float)(m_deqCI.size());
		}
		// Pobranie gornej granicy dla przedzialu ufnosci wskaznika struktury
		node->m_dEstE = node->m_dN*CMLPStatistics::UCF(node->m_dE,node->m_dN);
	}
}

void IEAIDTInducer::EstimateErrors(IIndividual* individual){
	vector<bool> vb(m_pDataset->GetCurrentAttrCount(),false);
	EstimateErrors(EAIND(individual)->m_CDTIndRoot,vb);
}

void IEAIDTInducer::EstimateErrors(CDTreeNode* node, vector<bool> vbUsed){
	if(node->m_bLeaf){
		GetNodeClassification(node);
		return;
	}

	// Liczenie kosztu atrybutow
	if(m_bCostOfAttrs){
		// Metoda GetCostOfAttributes uaktualni wektor ubUsed o uzyte w tym wezle cechy
		node->m_fCostOfAttrs = node->m_deqCI.size() * node->m_pITest->GetCostOfAttributes(m_pDataset,vbUsed);
	}


	node->m_dN = 0;
	node->m_dE = 0;
	node->m_dEstE = 0;
	node->m_iSubtreeSize = 1;
	node->m_fCostOfClassification = 0;
	node->m_uNExamples = 0;
	
	// Rekurencyjne schodzenie w dol drzewa i obliczanie na biezaco parametrow.
	for(unsigned i=0; i<node->m_vBranch.size(); i++){
		// Tutaj trzeci parametr wywolania tej funkcji musi byc zsumowany przed petla
		EstimateErrors(node->m_vBranch[i],vbUsed);
		node->m_dN += node->m_vBranch[i]->m_dN;
		node->m_dE += node->m_vBranch[i]->m_dE;
		node->m_dEstE += node->m_vBranch[i]->m_dEstE;
		node->m_iSubtreeSize += node->m_vBranch[i]->m_iSubtreeSize;
		node->m_fCostOfClassification += node->m_vBranch[i]->m_fCostOfClassification;
		node->m_uNExamples += node->m_vBranch[i]->m_uNExamples;
		node->m_fCostOfAttrs += node->m_vBranch[i]->m_fCostOfAttrs;
	}

	// Tutaj sprawdzamy czy wezel nie ma takich synow-lisci, ktore maja ta sama klase.
	// jesli tak, to biezacy wezel staje sie od razu lisciem.
	if(!node->m_bLeaf){
		CheckRepeatedClasses(node);
		// Kiedy wezel byl przyciety to trzeba wyznaczyc jego parametry klasyfikacji.
		if(node->m_bLeaf){
			GetNodeClassification(node);
			return;
		}
	}
}

double IEAIDTInducer::FindOptimalValue(const IDataSet* test){
	vector<pair<double, double> > vpq(m_iICVsteps+1);
	stringstream ss;
	ss<<m_dICVlvalue;
	SetParam(m_sICVParam,ss.str());
	vpq[0].first = m_dICVlvalue;
	vpq[0].second = SingleRun(test);
	// Dla ostatniego punktu wynik wstawiony bedzie na drugiej pozycji (ale to nie ma znaczenia)
	ss.str("");ss<<m_dICVrvalue;
	SetParam(m_sICVParam,ss.str());
	vpq[1].first = m_dICVrvalue;
	vpq[1].second = SingleRun(test);

	// Krok
	double krok = (m_dICVrvalue - m_dICVlvalue) / m_iICVsteps;

	// Petla idzie tyle razy ile jest krokow n
	for(int i=0; i<m_iICVsteps-1; i++){
		ss.str("");ss<<m_dICVlvalue + (i+1) * krok;
		SetParam(m_sICVParam,ss.str());
		vpq[i+2].first = m_dICVlvalue + (i+1) * krok;
		vpq[i+2].second = SingleRun(test);
	}

	double max = vpq[0].second;
	int maxindex = 0;
	for(int i=1; i<m_iICVsteps+1; i++){
		// Tutaj zwracamy UWAGE na znak nierownosci
		if(vpq[i].second > max){
			max = vpq[i].second;
			maxindex = i;
		}
	}
/*	for(int i=0; i<m_iICVsteps+1; i++){
		cerr<<vpq[i].first<<" "<<vpq[i].second<<endl;
	}	
	cerr<<vpq[maxindex].first<<" "<<vpq[maxindex].second<<endl; */
	return vpq[maxindex].first;
}

double IEAIDTInducer::FitnessSizeCCost(double dSize, double dCCost) const {
	if(m_sFitnessCalculation=="sum"){
		return m_dFitnessAlfa * (dSize - 1) + dCCost;
	}
	if(m_sFitnessCalculation=="division"){
		// (1)
		// Dla funkcji fitness ponizej nie bylo mozliwosci dobrania wartosci alfa
		// Dobrana dla jednego zbioru byla zla dla innych (alfa =[0.05;0.06])
		// Byc moze trzeba zastosowac dzielenie przed liczbe altrybutow (jak
		// w pracy Kretwoski&Kwedlo o regulach).
		// double f = pow(dSize + 1,m_dFitnessAlfa) / (1.0 - dCCost/m_fMaxCostOfClassification);
		
		// (2)
		// W tej funkcji alfa=0 wylacza wplyw rozmiaru. Wzrost wartosci alfa powoduje
		// zwiekszanie wplywu rozmiaru drzewa na jakosc
		double f = (dSize * m_dFitnessAlfa + 1) / (1.0 - dCCost/m_fMaxCostOfClassification);
		
		// cout<<"alfa="<<m_dFitnessAlfa<<" dSize="<<dSize<<" dCCost="<<dCCost<<" maxcost="<<m_fMaxCostOfClassification<<" fitness="<<f<<endl;
		return f;
	}
	CLogger::LogError("unknown fitness calculation",__FILE__,__LINE__);
	return 0;
}


int IEAIDTInducer::GetCrossoverOperatorID(){
	int r = rand()%100 + 1;
	int p = 0;
	for(size_t t=0; t<m_viPc.size(); t++){
		p += m_viPc[t];
		if (r<=p){
			return (int)t;
		}
		
	}
	// CLogger::LogEvent("GetCrossoverOperatorID",__FILE__,__LINE__);
	return -1;
}

void IEAIDTInducer::GetNodesLeaves(CDTreeNode* node, deque<CDTreeNode*> &nodes, deque<CDTreeNode*> &leaves){
	deque<CDTreeNode*> q;
	q.push_back(node);
	size_t nnodes = 0;
	size_t nleaves = 0;
	while(!q.empty()){
		CDTreeNode* n = q.front();
		q.pop_front();
		if(n->m_bLeaf){
			if(nleaves >= leaves.size()){
				leaves.resize(leaves.size()+100);
			}
			leaves[nleaves++] = n;
		}else{
			for(size_t t=0; t<n->m_vBranch.size(); t++){
				q.push_back(n->m_vBranch[t]);
			}
			if(nnodes >= nodes.size()){
				nodes.resize(nodes.size()+100);
			}
			nodes[nnodes++] =n;
		}
	}
	// Obciece kolejek
	leaves.resize(nleaves);
	nodes.resize(nnodes);
}

int IEAIDTInducer::GetMutationOperatorID(CDTreeNode* pNode){
	if(!pNode->m_bLeaf){
		// Dla wezlow wewnerznych wywolywana jest funkcja, ktory zwoci indeksy operatorow, ktore
		// mozna zastosowac do biezacego wezla
		vector<size_t> IDs = SetMutationOperators(pNode);
		if(!IDs.empty()){
			vector<int> distr(IDs.size(),0);
			distr[0] = m_viPm[IDs[0]];
			for(size_t i=1; i<IDs.size(); i++){
				distr[i] += (distr[i-1] + m_viPm[IDs[i]]);
			}
			// Losujuemy z przedzialu [1 ; suma p-w dopuszczalnych operatorow]
			int r = rand()%distr.back() + 1;
			for(size_t t=0; t<distr.size(); t++){
				if (r<=distr[t]){
					return (int)IDs[t];
				}
				
			}
		}
	}else{
		// Dla lisci nie ma ograniczen, losowanie sposrod wszystkich operatorow zgodnie z
		// zadanym wektorem p-w.
		int r = rand()%100 + 1;
		int p = 0;
		for(size_t t=0; t<m_viPmL.size(); t++){
			p += m_viPmL[t];
			if (r<=p){
				// Indeksy operatorow dla lisci sa kontynuacja numeracji operatorow na wezlow wewnetrznych.
				return (int)(t + m_viPm.size());
			}
			
		}
	}
	// Jak p-a nie sumuja sie do 100 (np. wszystkie sa rowne 0) to zwracane -1 co powoduje
	// ze zaden operator z mozliwych nie zostanie wylosowany
	return -1;
}

void IEAIDTInducer::InnerCrossValidation(){
	stringstream ss;
	ss<<m_dICVlvalue;
	if(!SetParam(m_sICVParam,ss.str())){
		ss.str("");
		ss<<"Param "<<m_sICVParam<<" can not be used with value "<<m_dICVlvalue<<" in inner cross validation";
		CLogger::LogError(ss.str(),__FILE__,__LINE__);
	}

	// Zapamietujemy caly zbior uczacy jaki zostal przekazany do algorytmu.
	const IDataSet* originalTrainingData = m_pDataset;

	// Jako wartosc koncowa zwracana jest srednia wartosci z kazdego
	// powtorzenia.
	double m = 0;

	// Liczba powtorzen
	for(int t=0; t<m_iICVRe_runs; t++){
		double mc = 0;
		// Podzial zbioru uczacego.
		deque<const IDataSet*> ds = originalTrainingData->SplitDataSet(m_iICVFolds);
		// Petla idzie m_iICVFolds razy
		for(size_t t=0; t<ds.size(); t++){
			const IDataSet* test = ds.front();
			ds.pop_front();
			CTempDataSet* train = new CTempDataSet(*(ds.front()));
			for(size_t i=0; i<ds.size(); i++){
				*train += *ds[i];
			}
			m_pDataset = train;
			mc += FindOptimalValue(test);
		}
		// Liczymy srednia z danego podzialu na m_iICVFolds czesci.
		mc /= m_iICVFolds;
		m += mc;
	}
	
	// Obliczenie sredniej wartosci szukanego parametru
	m /= m_iICVRe_runs;
	ss.str(""); ss<<m;
	// Ustawienie znalezionej sredniej wartosci (wartosc znaleziona
	// wedlug wewnetrznej krosswalidacji) jako wartosci obowiazujacej.
	SetParam(m_sICVParam,ss.str());

	// Ustawienie zbioru uczacego na caly zbior uczacy.
	m_pDataset = originalTrainingData;
}

void IEAIDTInducer::MaximizeFitness(IIndividual* individual){
	vector<bool> vb(m_pDataset->GetCurrentAttrCount(),false);
	MaximizeFitness(EAIND(individual)->m_CDTIndRoot,vb);
}

void IEAIDTInducer::MaximizeFitness(CDTreeNode* node, vector<bool> vbUsed){

	if(node->m_bLeaf){
		GetNodeClassification(node);
		return;
	}

	// Liczenie kosztu atrybutow
	if(m_bCostOfAttrs){
		// Metoda GetCostOfAttributes uaktualni wektor ubUsed o uzyte w tym wezle cechy
		node->m_fCostOfAttrs = node->m_deqCI.size() * node->m_pITest->GetCostOfAttributes(m_pDataset,vbUsed);
	}


	node->m_dN = 0;
	node->m_dE = 0;
	node->m_dEstE = 0;
	node->m_iSubtreeSize = 1;
	node->m_fCostOfClassification = 0;
	node->m_uNExamples = 0;
	
	// Rekurencyjne schodzenie w dol drzewa i obliczanie na biezaco parametrow.
	for(unsigned i=0; i<node->m_vBranch.size(); i++){
		// Tutaj trzeci parametr wywolania tej funkcji musi byc zsumowany przed petla
		MaximizeFitness(node->m_vBranch[i],vbUsed);
		node->m_dN += node->m_vBranch[i]->m_dN;
		node->m_dE += node->m_vBranch[i]->m_dE;
		node->m_dEstE += node->m_vBranch[i]->m_dEstE;
		node->m_iSubtreeSize += node->m_vBranch[i]->m_iSubtreeSize;
		node->m_fCostOfClassification += node->m_vBranch[i]->m_fCostOfClassification;
		node->m_uNExamples += node->m_vBranch[i]->m_uNExamples;
		node->m_fCostOfAttrs += node->m_vBranch[i]->m_fCostOfAttrs;
	}

	// Tutaj sprawdzamy czy wezel nie ma takich synow-lisci, ktore maja ta sama klase.
	// jesli tak, to biezacy wezel staje sie od razu lisciem.
	if(!node->m_bLeaf){
		CheckRepeatedClasses(node);
		// Kiedy wezel byl przyciety to trzeba wyznaczyc jego parametry klasyfikacji.
		if(node->m_bLeaf){
			GetNodeClassification(node);
			return;
		}
	}


	// To co jest ponizej to jest to co rozni funkcje MaximizeFitness od EstimateErrors.


	// Tutaj maksymalizowanie funkcji fitness przez ewentualne ciachniecie poddrzewa
	// zaczynajacego sie w biezacym wezle
		//________________________ Maximize fitness _______________________________
		// Teraz dla danego wezla jestesmy juz po przejciu jego poddrzewa, parametry uaktualnione
	
		// Sprawdzamy czy oplaca sie przyciac biezacy wezel
		assert(node->m_dN>0);
		// reklasyfikacja dla poddrzewa (blad)
		double rOld;
		if(m_iError == 0){
			rOld = node->GetEstimateError() / 100.0;
		}else{
			rOld = node->GetErrorOnTestDS() / 100.0;
		}
		// Tutaj nie korzystamy z dzielenia przez sredni rozmiar albo maksymalny
		// bo mamy poddrzewa i takie dzielenie w prostej postaci nie mialo by sensu
		int sOld = node->m_iSubtreeSize;
		int iE = 0;
		float cCOld=0;
		if(m_bCostOfClassification){
			cCOld = node->GetCostOfClassificationOnTrainDS();
		}
		// Tutaj koszt cAOLD to jest koszt koszt we biezacym wezle i jego poddrzewie
		float cAOld = node->m_fCostOfAttrs / node->m_uNExamples;
		float cCNewCumulative = 0;
		float cCNew = 0;
		// koszt po przycieciu czyli zero
		float cANew = 0;
		// Obliczenie liczby przykladow niepoprawnie klasyfikowanych,
		// przez dany wezel (czyli traktujemy go jakby byl lisciem).
		if(m_bCostOfClassification){
			for(unsigned i=0; i<node->m_deqCI.size(); i++){
				if( *(node->m_deqCI[i]->GetClass()) != *(node->m_pIAClass) ){
					iE++;
				}
				try{
					CNomAttr& predC = dynamic_cast<CNomAttr&>(*(node->m_pIAClass));
					CNomAttr& trueC = dynamic_cast<CNomAttr&>(*(node->m_deqCI[i]->GetClass()));
					cCNewCumulative += m_pDataset->CostOfClassification(predC,trueC);
				}
				catch(bad_cast){
				}
			}
			// Tu bedzie srednia OK, bo jest tylko jeden wezel (lisc)
			cCNew = cCNewCumulative / (float)(node->m_deqCI.size());
		}else{
			for(unsigned i=0; i<node->m_deqCI.size(); i++){
				if( *(node->m_deqCI[i]->GetClass()) != *(node->m_pIAClass) ){
					iE++;
				}
			}
		}

		double rNew;
		if(m_iError == 0){
			rNew = CMLPStatistics::UCF((double)iE,node->m_dN);
		}else{
			rNew = (double)iE / node->m_dN;
		}
		int sNew = 1;
		// Obliczenie roznicy wartosci funkcji dopasowania po i przed przycieciem
		double dFOld;
		double dFNew;
		if(m_bCostOfClassification){
			if(m_bCostOfAttrs){
				dFOld = - FitnessSizeCCostACost(sOld, cCOld , cAOld);
				dFNew = - FitnessSizeCCostACost(sNew, cCNew, cANew);
			}else{
				dFOld = - FitnessSizeCCost(sOld, cCOld);
				dFNew = - FitnessSizeCCost(sNew, cCNew);
			}
		}else{
			if(m_bCostOfAttrs){
				dFOld = - FitnessReclSizeACost(rOld ,sOld, cAOld);
				dFNew = - FitnessReclSizeACost(rNew, sNew, cANew);
			}else{
				dFOld = - FitnessReclSize(rOld , sOld);
				dFNew = - FitnessReclSize(rNew , sNew);
			}
		}
	
		//cout<<"size "<<sOld<<" -> "<<sNew<<endl;
		//cout<<"reclassification "<<rOld<<" -> "<<rNew<<endl;
		
		// Czy przyciecie daje poprawe wartosci fitness
		if(dFNew >= dFOld){
			//cerr<<"prunning >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>"<<endl;
			// Przyciecie drzewa. Biezacy wezel staje sie lisciem
			for(size_t t=0; t<node->m_vBranch.size(); t++){
				delete node->m_vBranch[t];
			}
			node->m_vBranch.clear();
			delete node->m_pITest;
			node->m_pITest = NULL;
			node->m_bLeaf = true;
			node->m_dE = iE;
			node->m_iSubtreeSize = 1;
			// node->m_dN pozostaje bez zmian
			node->m_dEstE = node->m_dN*CMLPStatistics::UCF(node->m_dE,node->m_dN);
			node->m_fCostOfClassification = cCNewCumulative;
			node->m_fCostOfAttrs = cANew;
			node->m_uNExamples = node->m_deqCI.size();
		}
		// Bez wzgledu czy wejdziemy to if'a czy nie, wartosci parametrow w wezle beda OK

		//________________________ Maximize fitness koniec ________________________
}


void IEAIDTInducer::PreTrain(const IDataSet *trainingDataSet){
	IInducer::PreTrain(trainingDataSet);
	// Sprawdzenie parametrow przed uruchomieniem alg. ewolucyjnego
	CheckParams();
}

IClassifier *IEAIDTInducer::ActualTrain(const IDataSet *trainingDataSet){

	if(m_iLogLevel>=1){
		CLogger::TimeStamp("Genetic Tree Inducer Started: ",m_ofsLogFile);
		m_ofsLogFile<<"Data set:\t"<<trainingDataSet->m_sDataSetName<<endl;
	}

	// Uruchomienie wewnetrznej kroswalidacji. Po zakonczeniu InnerCrossValidation
	// wartosc dobieranego parametru jest ustawiona na wartsc najlepsza jaka
	// zostala znaleziona.
	if(m_bInnerCrossValidation){
		InnerCrossValidation();
	}

	// ---------------- Uruchomienie algorytmu ewolucyjnego -------------------
	IIndividual* bestIndividual = Run(m_sResultsFileStem,m_iLogLevel,m_ofsLogFile);
	
	if(bestIndividual){
		CDTClassifier *pClass = new CDTClassifier(((CEAIDTIndividual*)bestIndividual)->m_CDTIndRoot,m_bCostOfClassification,m_bCostOfAttrs,trainingDataSet);
		((CEAIDTIndividual*)bestIndividual)->m_CDTIndRoot = NULL;
		delete bestIndividual;
		if(m_iLogLevel>=1) CLogger::TimeStamp("Genetic Tree Inducer Stopped: ",m_ofsLogFile);
		if(m_iLogLevel>=1) pClass->PrintGnuplotHiperplanes(m_ofsLogFile);
		return pClass;
	}
	
	return 0;
}

int IEAIDTInducer::GetProbabilityOfMutation(CDTreeNode* node) const{
	if(node->m_bLeaf){
		double q = 1.0 - node->m_dE / node->m_dN;
		double k = m_pDataset->GetClassCount();
		assert(k>0);
		double c = 2 * k* ((double)m_iProbabilityOfMutation) / (k - 1);
		return (int)( c * (1.0 - q) );
	}else{
		// Dla wezla mozna z node pobrac q i rozmiar jego poddrzewa.
		return m_iProbabilityOfMutation;
	}
}


void IEAIDTInducer::MutationOperator(int iterNum){
#ifdef MLPEAOPERATORSTEST
	int improved = 0;
#endif
	int nMutated =0;
	for(unsigned i=0; i<m_dqPopulation.size(); i++){
		switch(m_iMutationType){
			case 0:{
#ifdef MLPEAOPERATORSTEST
				double prevF = m_dqPopulation[i]->m_dFitness;
#endif
				// p-o mutacji dla pojedynczego wezla
				if( MutateTree( ((CEAIDTIndividual*)m_dqPopulation[i])->m_CDTIndRoot,m_pDataset,((CEAIDTIndividual&)(*(m_dqPopulation[i]))) ) ){
					EstimateErrors(m_dqPopulation[i]);
					EvaluateIndividualsFitness(m_dqPopulation[i]);
					nMutated++;
#ifdef MLPEAOPERATORSTEST
				if(m_dqPopulation[i]->m_dFitness > prevF){
					improved++;	
				}
#endif
				}
				break;
			}
			case 1:{
#ifdef MLPEAOPERATORSTEST
				double prevF = m_dqPopulation[i]->m_dFitness;
#endif
				//TOTH: tu mozna zrobic tak, zeby dla duzych drzew mutowac wiecej niz jeden wezel.
				// tylko jakie proporcje wzgledem rozmiaru drzewa.
				// p-o mutacji dla calego drzewa
				if( MutateTreeGradeNodes( ((CEAIDTIndividual*)m_dqPopulation[i])->m_CDTIndRoot,m_pDataset,((CEAIDTIndividual&)(*(m_dqPopulation[i]))) ) ){
					EstimateErrors(m_dqPopulation[i]);
					EvaluateIndividualsFitness(m_dqPopulation[i]);
					nMutated++;
#ifdef MLPEAOPERATORSTEST
				if(m_dqPopulation[i]->m_dFitness > prevF){
					improved++;	
				}
#endif
				}
				break;
			}
		}
	}
#ifdef MLPEAOPERATORSTEST
	m_ofsMutationCounting<<iterNum<<" "<<nMutated<<" "<<improved<<endl;
#endif
	if(m_iLogLevel>=2){
		stringstream ss;
		if(m_iMutationType == 1){
			ss<<"pop_size="<<m_dqPopulation.size()<<"\tmutated="<<nMutated<<"\t"<<100.0*(double)nMutated/(double)m_dqPopulation.size()<<"% (expected "<<m_iProbabilityOfMutation<<"%)";
		}else{
			ss<<"pop_size="<<m_dqPopulation.size()<<"\tmutated="<<nMutated<<"\t"<<100.0*(double)nMutated/(double)m_dqPopulation.size()<<"%";
		}
		CLogger::LogEvent(ss.str(),"",__LINE__,m_ofsLogFile);
	}
}

bool IEAIDTInducer::MutateTree(CDTreeNode* node, const IDataSet* dataset, CEAIDTIndividual &indv){
bool mutated = false;
//CLogger::LogEvent("wejscie",__FILE__,__LINE__);
	if(node == NULL) return false;
	
	// if( rand()%100 + 1 <= m_iProbabilityOfMutation ){
	if( rand()%100 + 1 <= GetProbabilityOfMutation(node)){
		if(!(node->m_bLeaf && node->m_dE==0)){
			if(MutateNode(node,dataset,indv)){
				mutated = true;
			}
		}
	}
	
	for(size_t t=0; t<node->m_vBranch.size(); t++){
		if( !(node->m_vBranch[t]->m_bLeaf && node->m_vBranch[t]->m_dE==0) ){
			if(MutateTree(node->m_vBranch[t],dataset,indv)) mutated = true;
		}
	}
	//CLogger::LogEvent("wyjscie",__FILE__,__LINE__);
	return mutated;
}

bool IEAIDTInducer::MutateTreeGradeNodes(CDTreeNode* node, const IDataSet* dataset, CEAIDTIndividual &indv){
	// Sprawdzamy czy mutowac dane drzewo o korzeniu w node
	if( rand()%100 + 1 <= m_iProbabilityOfMutation ){
		deque<CDTreeNode*> nodes;
		deque<CDTreeNode*> leaves;
		GetNodesLeaves(node,nodes,leaves);
		if( nodes.size()>0 && rand()%2){
			// mutujemy wezel wewnetrzny
			size_t pop_size = nodes.size();
			// To jest bardzo wazny parametr np. dla wartosci 2 jak sa 4 wezly to wtedy
			// pierwszy ma prawie zerowe p-o wylosowania
			double SelectivePressure = 1.5;
			double q = SelectivePressure * 1.0/(double)pop_size;
			double r = 2.0* q / ((double)pop_size-1.0) - 2.0/((double)pop_size*((double)pop_size-1.0));
			vector<double> vdDist(pop_size);
			vdDist[0] = q - (pop_size /*- 0*/ - 1)*r;
			for(size_t i=1; i<pop_size; i++){
				vdDist[i] = (vdDist[i-1] + q - (pop_size - i - 1) * r);
			}
			double k = CMLPStatistics::Rand_01();
			int i = 0;
			while(k>vdDist[i]){
				i++;
			}
			// Wezel do mutacji.
			CDTreeNode *m = nodes[i];
			return MutateNode(m,dataset,indv);
		}else{
			// Z listy wezlow wyrzucamy od razu te, ktore sa czyste
			deque<CDTreeNode*>::iterator newend = remove_if(leaves.begin(),leaves.end(),CDTreeNodePure);
			leaves.erase(newend,leaves.end());
			if(leaves.empty()){
				return false;
			}
			// mutujemy liscia
			sort(leaves.begin(),leaves.end(),CDTreeNodeLess);
			/*cout<<endl;
			for(size_t t=0; t<leaves.size(); t++){
				cout<<leaves[t]->m_dE<<" ";
			}
			cout<<endl;*/
			size_t pop_size = leaves.size();
			double SelectivePressure = 1.5;
			double q = SelectivePressure * 1.0/(double)pop_size;
			double r = 2.0* q / ((double)pop_size-1.0) - 2.0/((double)pop_size*((double)pop_size-1.0));
			vector<double> vdDist(pop_size);
			vdDist[0] = q - (pop_size /*- 0*/ - 1)*r;
			for(size_t i=1; i<pop_size; i++){
				vdDist[i] = (vdDist[i-1] + q - (pop_size - i - 1) * r);
			}
			double k = CMLPStatistics::Rand_01();
			int i = 0;
			while(k>vdDist[i]){
				i++;
			}
			// Lisc do mutacji.
			// cout<<endl<<pop_size<<" "<<i<<endl;
			CDTreeNode *m = leaves[i];
			return MutateNode(m,dataset,indv);
		}
	}
	return false;
}

void IEAIDTInducer::PostCreateInitialPopulation(){
	if(m_bCostOfClassification || m_bCostOfAttrs){
		// m.in.: m_sFitnessCalculation=="division"
		// Wyznaczenie maksymalnego kosztu blednej klasyfikacji na zbiorze uczacym
		vector<IAttrValue*> vClassVal = m_pDataset->GetClassValues();
		pair<float,int> p(0,0);
		vector<pair<float,int> > vpCostN(vClassVal.size(),p);
		for(size_t i=0; i<vClassVal.size(); i++){
			for(size_t j=0; j<vClassVal.size(); j++){
				try{
					CNomAttr& predC = dynamic_cast<CNomAttr&>(*vClassVal[j]);
					CNomAttr& trueC = dynamic_cast<CNomAttr&>(*vClassVal[i]);
					float c = m_pDataset->CostOfClassification(predC,trueC);
					if(vpCostN[i].first < c){
						vpCostN[i].first = c;
					}
				}
				catch(bad_cast){
				}
			}
		}
		for(size_t d=0; d<m_pDataset->GetSize(); d++){
			CDataObject *dOb = (*m_pDataset)[d];
			for(size_t c=0; c<vClassVal.size(); c++){
				if( *(dOb->GetClass()) == *(vClassVal[c])){
					vpCostN[c].second++;
				}
			}
		}
		m_fMaxCostOfClassification = 0;
		int sum = 0;
		for(size_t i=0; i<vpCostN.size(); i++){
			m_fMaxCostOfClassification += vpCostN[i].first * vpCostN[i].second;
			sum += vpCostN[i].second;
		}
		assert(sum>0);
		m_fMaxCostOfClassification /= sum;
		assert(m_fMaxCostOfClassification>0);

		// Wyznaczenie maksymalnego kosztu atrybutow na zbiorze uczacym
		m_fMaxCostOfAttributes = 0;
		for(size_t t=0; t<m_pDataset->GetCurrentAttrCount(); t++){
			m_fMaxCostOfAttributes += m_pDataset->GetAttrDesc(t)->m_fCost;
		}
		if(m_fMaxCostOfAttributes==0){
			// Jak jest zero to ustawiamy na 1 zeby mozna bylo wykonac dzielenie.
			m_fMaxCostOfAttributes = 1;
		}
	}
	
	// Po utworzeniu populacji poczatkowej ta funcja ustawia w wezlach sume wartosci z 
	// synow (chodzi tu o ustawienie wartosci rozmiaru poddrzewa dla kazdego wezla).
	for(size_t t=0; t<m_dqPopulation.size();t++){
		CEAIDTIndividual* eaindividual = (CEAIDTIndividual*)m_dqPopulation[t];
		EstimateErrors(eaindividual);
	}
}

bool IEAIDTInducer::SetParam(string sParamName, string sParamValue){
	string::iterator i = sParamName.begin();
	while(i != sParamName.end()){
		*i = tolower(*i);
		++i;
	}
	i = sParamValue.begin();
	while(i != sParamValue.end()){
		*i = tolower(*i);
		++i;
	}

	if(sParamName=="fitnesscalculation"){
		if(sParamValue=="sum"){
			m_sFitnessCalculation = "sum";
			return true;
		}
		if(sParamValue=="division"){
			m_sFitnessCalculation = "division";
			return true;
		}
		return false;
	}
	if(sParamName=="max tree size" || sParamName=="maxtreesize"){
		if(sParamValue=="yes" && !m_bUseAverageTreeSize){
			m_bUseMaxTreeSize = true;
			return true;
		}
		if(sParamValue=="no"){
			m_bUseMaxTreeSize = false;
			return true;
		}
		return false;
	}
	if(sParamName=="average tree size" || sParamName=="averagetreesize"){
		if(sParamValue=="yes" && !m_bUseMaxTreeSize){
			m_bUseAverageTreeSize = true;
			return true;
		}
		if(sParamValue=="no"){
			m_bUseAverageTreeSize = false;
			return true;
		}
		return false;
	}
	if(sParamName=="alfa"){
		m_dFitnessAlfa = atof(sParamValue.c_str());
		return true;
	}
	for(size_t t=0; t<m_viPc.size(); t++){
		stringstream ss;
		ss<<"pc"<<(unsigned)t;
		if(sParamName==ss.str()){
			m_viPc[t] = (int)atoi(sParamValue.c_str());
			if(m_viPc[t] < 0 || m_viPc[t] > 100){
				m_viPc[t] = 0;
				return false;
			}else{
				return true;
			}
		}
	}
	for(size_t t=0; t<m_viPm.size(); t++){
		stringstream ss;
		ss<<"pm"<<(unsigned)t;
		if(sParamName==ss.str()){
			m_viPm[t] = (int)atoi(sParamValue.c_str());
			if(m_viPm[t] < 0 || m_viPm[t] > 100){
				m_viPm[t] = 0;
				return false;
			}else{
				return true;
			}
		}
	}
	for(size_t t=0; t<m_viPmL.size(); t++){
		stringstream ss;
		ss<<"pml"<<(unsigned)t;
		if(sParamName==ss.str()){
			m_viPmL[t] = (int)atoi(sParamValue.c_str());
			if(m_viPmL[t] < 0 || m_viPmL[t] > 100){
				m_viPmL[t] = 0;
				return false;
			}else{
				return true;
			}
		}
	}
	if(sParamName=="maximize fitness" || sParamName=="maximizefitness"){
		if(sParamValue=="yes"){
			m_bMaximizeFitness = true;
			return true;
		}
		if(sParamValue=="no"){
			m_bMaximizeFitness = false;
			return true;
		}
		return false;
	}
	if(sParamName=="mfprobability"){
		m_iMaxFitnessProbability = atoi(sParamValue.c_str());
		if(m_iMaxFitnessProbability < 0 || m_iMaxFitnessProbability > 100){
			return false;
		}
		return true;
	}
	if(sParamName=="gamma"){
		m_dGamma = atof(sParamValue.c_str());
		return true;
	}
	if(sParamName=="mutationtype"){
		m_iMutationType = atoi(sParamValue.c_str());
		if(m_iMutationType < 0 || m_iMutationType > 1){
			return false;
		}
		return true;
	}
	if(sParamName=="initmaxdepth"){
		m_iInitMaxDepth = atoi(sParamValue.c_str());
		if(m_iInitMaxDepth<-1){
			m_iInitMaxDepth = -1;
			return false;
		}
		return true;
	}
	if(sParamName=="innercrossvalidation"){
		if(sParamValue=="yes"){
			m_bInnerCrossValidation = true;
			return true;
		}
		if(sParamValue=="no"){
			m_bInnerCrossValidation = false;
			return true;
		}
		return false;
	}
	if(IEvolutionaryAlg::SetParam(sParamName,sParamValue)) return true;
	if(IDTInducer::SetParam(sParamName,sParamValue)) return true;
	return false;
}

double IEAIDTInducer::SingleRun(const IDataSet* test){
	double q = 0;
	IIndividual* bestIndividual = Run(m_sResultsFileStem,m_iLogLevel,m_ofsLogFile);
	if(bestIndividual){
		CDTClassifier *pClass = new CDTClassifier(((CEAIDTIndividual*)bestIndividual)->m_CDTIndRoot,m_bCostOfClassification,m_bCostOfAttrs,m_pDataset);
		((CEAIDTIndividual*)bestIndividual)->m_CDTIndRoot = NULL;
		delete bestIndividual;
		for(size_t j=0; j<test->GetSize(); j++){
			CDataObject *inst = (*test)[j];
			IAttrValue* pIA;
			pair<double,double> costs;
			if(IsCostOfClassification() || IsCostOfAttrs()){
				pIA=pClass->Classify(inst,costs);
				// Dla kosztow bierzemy wartosci ze znakiem przeciwnym
				// bo zwrocona wartosc bedzie maksymalizowana
				if(IsCostOfClassification()){
					q -= costs.first;
				}
				if(IsCostOfAttrs()){
					q -=costs.second;
				}
			}else{
				pIA=pClass->Classify(inst);
				IAttrValue *ia=inst->GetClass();
				if((*pIA)==(*ia)){
					// dla jakosci sumujemy po prostu liczbe
					// poprawnie klasyfikowanych przykladow
					q++;
				}
			}
		}
		q /= test->GetSize();
	}
	return q;
}

bool IEAIDTInducer::SetSubParam(string sParentParamName, string sSubParamName, string sSubParamValue){
	string::iterator i = sSubParamName.begin();
	while(i != sSubParamName.end()){
		*i = tolower(*i);
		++i;
	}
	i = sSubParamValue.begin();
	while(i != sSubParamValue.end()){
		*i = tolower(*i);
		++i;
	}
	i = sParentParamName.begin();
	while(i != sParentParamName.end()){
		*i = tolower(*i);
		++i;
	}

	if(sParentParamName=="innercrossvalidation"){
		if(sSubParamName=="folds"){
			m_iICVFolds = atoi(sSubParamValue.c_str());
			return true;
		}
		if(sSubParamName=="re-runs"){
			m_iICVRe_runs = atoi(sSubParamValue.c_str());
			return true;
		}
		if(sSubParamName=="steps"){
			m_iICVsteps = atoi(sSubParamValue.c_str());
			return true;
		}
		if(sSubParamName=="lvalue"){
			m_dICVlvalue = atof(sSubParamValue.c_str());
			return true;
		}
		if(sSubParamName=="rvalue"){
			m_dICVrvalue = atof(sSubParamValue.c_str());
			return true;
		}
		if(sSubParamName=="param"){
			m_sICVParam = sSubParamValue;
			return true;
		}
	}
	
	return false;
}

void IEAIDTInducer::PrintParamsXML(ostream &ostr, int nsp){
	IEvolutionaryAlg::PrintParamsXML(ostr,nsp);
	IDTInducer::PrintParamsXML(ostr,nsp);
	ostr.fill(' ');
	ostr<<setw(nsp)<<" "<<"<!-- [ sum | division ] -->"<<endl;
	ostr<<setw(nsp)<<" "<<"<MLPParam Name=\"fitnesscalculation\" Value=\"sum\" />"<<endl;
	ostr<<setw(nsp)<<" "<<"<MLPParam Name=\"maxtreesize\" Value=\"no\" />"<<endl;
	ostr<<setw(nsp)<<" "<<"<MLPParam Name=\"averagetreesize\" Value=\"no\" />"<<endl;
	ostr<<setw(nsp)<<" "<<"<!-- alfa * size -->"<<endl;
	ostr<<setw(nsp)<<" "<<"<MLPParam Name=\"alfa\" Value=\"0.005\" />"<<endl;
	ostr<<setw(nsp)<<" "<<"<!-- maximizing fitness by prunning -->"<<endl;
	ostr<<setw(nsp)<<" "<<"<MLPParam Name=\"maximizefitness\" Value=\"yes\" />"<<endl;
	ostr<<setw(nsp)<<" "<<"<MLPParam Name=\"mfprobability\" Value=\"0\" />"<<endl;
	ostr<<setw(nsp)<<" "<<"<!-- gamma * cost of attributes -->"<<endl;
	ostr<<setw(nsp)<<" "<<"<MLPParam Name=\"gamma\" Value=\"0.1\" />"<<endl;
	ostr<<setw(nsp)<<" "<<"<!-- 0 - prob. for a single node; 1 - prob. for the whole tree -->"<<endl;
	ostr<<setw(nsp)<<" "<<"<MLPParam Name=\"mutationtype\" Value=\"1\" />"<<endl;
	ostr<<setw(nsp)<<" "<<"<MLPParam Name=\"initmaxdepth\" Value=\"-1\" />"<<endl;
	ostr<<setw(nsp)<<" "<<"<MLPParam Name=\"innercrossvalidation\" Value=\"no\" >"<<endl;
	ostr<<setw(nsp+2)<<" "<<"<MLPSubParam Name=\"folds\" Value=\"2\" />"<<endl;
	ostr<<setw(nsp+2)<<" "<<"<MLPSubParam Name=\"re-runs\" Value=\"5\" />"<<endl;
	ostr<<setw(nsp+2)<<" "<<"<MLPSubParam Name=\"param\" Value=\"alfa\" />"<<endl;
	ostr<<setw(nsp+2)<<" "<<"<MLPSubParam Name=\"lvalue\" Value=\"0.001\" />"<<endl;
	ostr<<setw(nsp+2)<<" "<<"<MLPSubParam Name=\"rvalue\" Value=\"0.1\" />"<<endl;
	ostr<<setw(nsp+2)<<" "<<"<MLPSubParam Name=\"steps\" Value=\"10\" />"<<endl;
	ostr<<setw(nsp)<<" "<<"</MLPParam>"<<endl;
	ostr.fill(' ');
}

vector<string> IEAIDTInducer::GetParams() const{
vector<string> res;
stringstream ss;

	res = IEvolutionaryAlg::GetParams();
	vector<string> temp = IDTInducer::GetParams();
	for(size_t t=0; t<temp.size(); t++){
		res.push_back(temp[t]);
	}

	res.push_back("fitnesscalculation");
	res.push_back(m_sFitnessCalculation);
	res.push_back("maxtreesize");
	if(m_bUseMaxTreeSize) res.push_back("yes");
		else res.push_back("no");
	res.push_back("averagetreesize");
	if(m_bUseAverageTreeSize) res.push_back("yes");
		else res.push_back("no");
	res.push_back("alfa");
	ss.str(""); ss<<m_dFitnessAlfa;
	res.push_back(ss.str());
	for(size_t t=0; t<m_viPc.size(); t++){
		ss.str(""); ss<<"pc"<<(int)t;
		res.push_back(ss.str());
		ss.str(""); ss<<m_viPc[t];
		res.push_back(ss.str());
	}
	for(size_t t=0; t<m_viPm.size(); t++){
		ss.str(""); ss<<"pm"<<(int)t;
		res.push_back(ss.str());
		ss.str(""); ss<<m_viPm[t];
		res.push_back(ss.str());
	}
	for(size_t t=0; t<m_viPmL.size(); t++){
		ss.str(""); ss<<"pml"<<(int)t;
		res.push_back(ss.str());
		ss.str(""); ss<<m_viPmL[t];
		res.push_back(ss.str());
	}
	res.push_back("maximizefitness");
	if(m_bMaximizeFitness) res.push_back("yes");
		else res.push_back("no");
	res.push_back("mfprobability");
	ss.str(""); ss<<m_iMaxFitnessProbability;
	res.push_back(ss.str());
	res.push_back("gamma");
	ss.str(""); ss<<m_dGamma;
	res.push_back(ss.str());
	res.push_back("mutationtype");
	ss.str(""); ss<<m_iMutationType;
	res.push_back(ss.str());
	res.push_back("initmaxdepth");
	ss.str(""); ss<<m_iInitMaxDepth;
	res.push_back(ss.str());
	
	res.push_back("innercrossvalidation");
	if(m_bInnerCrossValidation) res.push_back("yes");
		else res.push_back("no");
	res.push_back("param");
	res.push_back(m_sICVParam);
	res.push_back("lvalue");
	ss.str(""); ss<<m_dICVlvalue;
	res.push_back(ss.str());
	res.push_back("rvalue");
	ss.str(""); ss<<m_dICVrvalue;
	res.push_back(ss.str());
	res.push_back("folds");
	ss.str(""); ss<<m_iICVFolds;
	res.push_back(ss.str());
	res.push_back("re-runs");
	ss.str(""); ss<<m_iICVRe_runs;
	res.push_back(ss.str());
	res.push_back("steps");
	ss.str(""); ss<<m_iICVsteps;
	res.push_back(ss.str());


	return res;
}

void IEAIDTInducer::PrintFitnessLogHeader(ofstream &ofFlog){
	CLogger::TimeStamp("# This file is generated by Machine Learning Platform ",ofFlog);
	if(m_bCostOfAttrs || m_bCostOfClassification){
		ofFlog<<"# Iter.\tMin. fitness\tAverage fitness\tMax. fitness\tMin. Q\tAvrg. Q\tMax. Q\tMin. T\tAvrg. T\tMax. T\tMin. C\tAvrg. C\tMax. C\tBest fitness\tBest quality\tBest Tree size"<<endl;
	}else{
		ofFlog<<"# Iter.\tMin. fitness\tAverage fitness\tMax. fitness\tMin. Q\tAvrg. Q\tMax. Q\tMin. T\tAvrg. T\tMax. T\t\tBest fitness\tBest quality\tBest Tree size"<<endl;
	}
}

void IEAIDTInducer::PrintIndividualToLog(string prefix, IIndividual* ind, bool bPrintStructure){
	stringstream msg;
	if(m_bCostOfClassification || m_bCostOfAttrs){
		msg<<prefix<<": F="<<ind->m_dFitness<<" (q="<<ind->m_dQuality<<", s="<<ind->m_dSize<<", c="<<ind->m_dCost<<")";
	}else{
		msg<<prefix<<": F="<<ind->m_dFitness<<" (q="<<ind->m_dQuality<<", s="<<ind->m_dSize<<")";
	}
	CLogger::LogEvent(msg.str(),"",__LINE__,m_ofsLogFile);
	// Drukowanie struktury osobnika
	if(bPrintStructure){
		//cerr<<endl<<((CEAIDTIndividual*)ind)->m_CDTIndRoot->m_fCostOfAttrs<<endl;
		//cerr<<": F="<<ind->m_dFitness<<" (q="<<ind->m_dQuality<<", s="<<ind->m_dSize<<", c="<<ind->m_dCost<<")";

		CDTreeNode *node = new CDTreeNode(*((CEAIDTIndividual*)ind)->m_CDTIndRoot);
		CDTClassifier *pClass = new CDTClassifier(node,m_bCostOfClassification,m_bCostOfAttrs,m_pDataset);
		pClass->DisplayStruct(m_ofsLogFile,0);
		delete pClass;
		m_ofsLogFile<<"\n";
	}
}

/**
 * TOTH:
 * to wypisanie mozna zrealizowac tak zeby osobnik sam wypisywal informacje o sobie
 * tak aby ta funkcja nie byla zalezna od algorytmu
 */
void IEAIDTInducer::PrintFitnessLog(deque<IIndividual*> &p,IIndividual* bestind, unsigned iternum, ostream &ostr){
vector<double> fitness;
vector<double> quality;
vector<double> size;
vector<double> cost;
CMLPStatistics stat;

	for(size_t t=0; t<p.size(); t++){
		fitness.push_back(p[t]->m_dFitness);
		quality.push_back(p[t]->m_dQuality);
		size.push_back(p[t]->m_dSize);
		cost.push_back(p[t]->m_dCost);
	}
	vector<double>::iterator maxf = max_element(fitness.begin(),fitness.end());
	vector<double>::iterator minf = min_element(fitness.begin(),fitness.end());
	double avrgf = stat.MLPAverage<double>(fitness);

	vector<double>::iterator maxq = max_element(quality.begin(),quality.end());
	vector<double>::iterator minq = min_element(quality.begin(),quality.end());
	double avrgq = stat.MLPAverage<double>(quality);

	vector<double>::iterator maxt = max_element(size.begin(),size.end());
	vector<double>::iterator mint = min_element(size.begin(),size.end());
	double avrgt = stat.MLPAverage<double>(size);

	vector<double>::iterator maxc = max_element(cost.begin(),cost.end());
	vector<double>::iterator minc = min_element(cost.begin(),cost.end());
	double avrgc = stat.MLPAverage<double>(cost);
	
	if(m_bCostOfAttrs || m_bCostOfClassification){
		ostr<<iternum<<"\t"<<*minf<<"\t"<<avrgf<<"\t"<<*maxf<<"\t"<<*minq<<"\t"<<avrgq<<"\t"<<*maxq<<"\t"<<*mint<<"\t"<<avrgt<<"\t"<<*maxt<<"\t"<<*minc<<"\t"<<avrgc<<"\t"<<*maxc<<"\t"<<bestind->m_dFitness<<"\t"<<bestind->m_dQuality<<"\t"<<bestind->m_dSize<<"\t"<<bestind->m_dCost<<endl;
	}else{
		ostr<<iternum<<"\t"<<*minf<<"\t"<<avrgf<<"\t"<<*maxf<<"\t"<<*minq<<"\t"<<avrgq<<"\t"<<*maxq<<"\t"<<*mint<<"\t"<<avrgt<<"\t"<<*maxt<<"\t"<<bestind->m_dFitness<<"\t"<<bestind->m_dQuality<<"\t"<<bestind->m_dSize<<"\t"<<bestind->m_dCost<<endl;
	}
}

void IEAIDTInducer::CreateGnuplotProjects(string datafilename, string outfilenamestem){
string datasetname = m_pDataset->m_sDataSetName;
ofstream	outfile;
string outfilenamef = outfilenamestem+"_FL_fitness.plt";
string outfilenameq = outfilenamestem+"_FL_quality.plt";
string outfilenamet = outfilenamestem+"_FL_treesize.plt";
string outfilenamec = outfilenamestem+"_FL_totalcost.plt";
string outfilenameb = outfilenamestem+"_FL_bestindividual.plt";
// dane zaczynaja sie od drugiej kolumny
int iCol = 2;

	vector<string> params = GetParams();
	stringstream ssParams;
	for(size_t i=0; i<params.size()-1; i+=2){
		ssParams<<"["<<params[i]<<"="<<params[i+1]<<"]";
		if(i%3==0){
			ssParams<<"\\n";
		}
	}

	outfile.open(outfilenamef.c_str(),ofstream::out);
	if(outfile.is_open()){
		outfile<<"# Gnuplot script file for plotting data in file "<<datafilename<<endl;
		CLogger::TimeStamp("# This file is generated by Machine Learning Platform ",outfile);
		outfile<<endl;
		outfile<<"set title \""<<datasetname<<" - "<<"fitness "<<ssParams.str()<<"\""<<endl;
		outfile<<"set terminal png"<<endl;
		outfile<<"set output '"<<outfilenamestem<<"_FL_fitness.png"<<"'"<<endl;
		outfile<<"set xlabel \"iteration number\""<<endl;
		outfile<<"set ylabel \"fitness\""<<endl;
		outfile<<"set y2label \" \""<<endl;
		outfile<<"#set key 2.5,2.5"<<endl;
		outfile<<"#set xr [0.0:5.0]"<<endl;
		outfile<<"#set yr [0:2]"<<endl;
		outfile<<"set ytics mirror"<<endl;
		outfile<<"set noy2tics"<<endl;
		outfile<<"plot \""<<datafilename<<"\" using 1:"<<iCol++<<" title 'min. fitness' with lines , \\"<<endl;
		outfile<<"\""<<datafilename<<"\" using 1:"<<iCol++<<" title 'avrg. fitness' with lines, \\"<<endl;
		outfile<<"\""<<datafilename<<"\" using 1:"<<iCol++<<" title 'max. fitness' with lines"<<endl;
		outfile.close();
	}
	outfile.open(outfilenameq.c_str(),ofstream::out);
	if(outfile.is_open()){
		outfile<<"# Gnuplot script file for plotting data in file "<<datafilename<<endl;
		CLogger::TimeStamp("# This file is generated by Machine Learning Platform ",outfile);
		outfile<<endl;
		outfile<<"set title \""<<datasetname<<" - "<<"quality "<<ssParams.str()<<"\""<<endl;
		outfile<<"set terminal png"<<endl;
		outfile<<"set output '"<<outfilenamestem<<"_FL_quality.png"<<"'"<<endl;
		outfile<<"set xlabel \"iteration number\""<<endl;
		outfile<<"set ylabel \"quality\""<<endl;
		outfile<<"set y2label \" \""<<endl;
		outfile<<"#set key 2.5,2.5"<<endl;
		outfile<<"#set xr [0.0:5.0]"<<endl;
		outfile<<"#set yr [0:2]"<<endl;
		outfile<<"set ytics mirror"<<endl;
		outfile<<"set noy2tics"<<endl;
		outfile<<"plot \""<<datafilename<<"\" using 1:"<<iCol++<<" title 'min. quality' with lines , \\"<<endl;
		outfile<<"\""<<datafilename<<"\" using 1:"<<iCol++<<" title 'avrg. quality' with lines, \\"<<endl;
		outfile<<"\""<<datafilename<<"\" using 1:"<<iCol++<<" title 'max. quality' with lines"<<endl;
		outfile.close();
	}
	outfile.open(outfilenamet.c_str(),ofstream::out);
	if(outfile.is_open()){
		outfile<<"# Gnuplot script file for plotting data in file "<<datafilename<<endl;
		CLogger::TimeStamp("# This file is generated by Machine Learning Platform ",outfile);
		outfile<<endl;
		outfile<<"set title \""<<datasetname<<" - "<<"treesize "<<ssParams.str()<<"\""<<endl;
		outfile<<"set terminal png"<<endl;
		outfile<<"set output '"<<outfilenamestem<<"_FL_treesize.png"<<"'"<<endl;
		outfile<<"set xlabel \"iteration number\""<<endl;
		outfile<<"set ylabel \"tree size\""<<endl;
		outfile<<"set y2label \" \""<<endl;
		outfile<<"#set key 2.5,2.5"<<endl;
		outfile<<"#set xr [0.0:5.0]"<<endl;
		outfile<<"#set yr [0:2]"<<endl;
		outfile<<"set ytics mirror"<<endl;
		outfile<<"set noy2tics"<<endl;
		outfile<<"plot \""<<datafilename<<"\" using 1:"<<iCol++<<" title 'min. tree size' with lines , \\"<<endl;
		outfile<<"\""<<datafilename<<"\" using 1:"<<iCol++<<" title 'avrg. tree size' with lines, \\"<<endl;
		outfile<<"\""<<datafilename<<"\" using 1:"<<iCol++<<" title 'max. tree size' with lines"<<endl;
		outfile.close();
	}
	if(m_bCostOfAttrs || m_bCostOfClassification){
		outfile.open(outfilenamec.c_str(),ofstream::out);
		if(outfile.is_open()){
			outfile<<"# Gnuplot script file for plotting data in file "<<datafilename<<endl;
			CLogger::TimeStamp("# This file is generated by Machine Learning Platform ",outfile);
			outfile<<endl;
			outfile<<"set title \""<<datasetname<<" - "<<"totalcost "<<ssParams.str()<<"\""<<endl;
			outfile<<"set terminal png"<<endl;
			outfile<<"set output '"<<outfilenamestem<<"_FL_totalcost.png"<<"'"<<endl;
			outfile<<"set xlabel \"iteration number\""<<endl;
			outfile<<"set ylabel \"total cost\""<<endl;
			outfile<<"set y2label \" \""<<endl;
			outfile<<"#set key 2.5,2.5"<<endl;
			outfile<<"#set xr [0.0:5.0]"<<endl;
			outfile<<"#set yr [0:2]"<<endl;
			outfile<<"set ytics mirror"<<endl;
			outfile<<"set noy2tics"<<endl;
			outfile<<"plot \""<<datafilename<<"\" using 1:"<<iCol++<<" title 'min. total cost' with lines , \\"<<endl;
			outfile<<"\""<<datafilename<<"\" using 1:"<<iCol++<<" title 'avrg. total cost' with lines, \\"<<endl;
			outfile<<"\""<<datafilename<<"\" using 1:"<<iCol++<<" title 'max. total cost' with lines"<<endl;
			outfile.close();
		}
	}
	outfile.open(outfilenameb.c_str(),ofstream::out);
	if(outfile.is_open()){
		outfile<<"# Gnuplot script file for plotting data in file "<<datafilename<<endl;
		CLogger::TimeStamp("# This file is generated by Machine Learning Platform ",outfile);
		outfile<<endl;
		outfile<<"set title \""<<datasetname<<" - "<<"the best individual "<<ssParams.str()<<"\""<<endl;
		outfile<<"set terminal png"<<endl;
		outfile<<"set output '"<<outfilenamestem<<"_FL_bestindividual.png"<<"'"<<endl;
		outfile<<"set xlabel \"iteration number\""<<endl;
		outfile<<"set ylabel \"tree size\""<<endl;
		if(m_bCostOfAttrs || m_bCostOfClassification){
			outfile<<"set y2label \"fitness, quality, total cost\""<<endl;
		}else{
			outfile<<"set y2label \"fitness, quality\""<<endl;
		}
		outfile<<"#set key 2.5,2.5"<<endl;
		outfile<<"#set xr [0.0:5.0]"<<endl;
		outfile<<"#set yr [0:2]"<<endl;
		outfile<<"set ytics nomirror"<<endl;
		outfile<<"set y2tics 0.5"<<endl;
		outfile<<"plot \""<<datafilename<<"\" using 1:"<<iCol++<<" axes x1y2 title 'fitness' with lines , \\"<<endl;
		outfile<<"\""<<datafilename<<"\" using 1:"<<iCol++<<" axes x1y2 title 'quality' with lines, \\"<<endl;
		outfile<<"\""<<datafilename<<"\" using 1:"<<iCol++<<" title 'tree size' with lines";
		if(m_bCostOfAttrs || m_bCostOfClassification){
			outfile<<", \\"<<endl;
			outfile<<"\""<<datafilename<<"\" using 1:17 axes x1y2 title 'total cost' with lines";
		}
		outfile<<endl;
		outfile.close();
	}
}

void IEAIDTInducer::ApplyTestChange(const IDataSet* dataset, CDTreeNode* node, CDTreeNode* parentNode){
	assert(node);
	assert(!node->m_deqCI.empty());

	// Znalezienie klasy wiekszosciowej dla wezla. Jak parentnode jest NULL, to
	// nie trzeba szukac, bo to jest wezel, w ktorym klasa wiekszosciowa sie nie 
	// zmienila, bo to z tego wezla startuje wywolanie rekurencyjne.
	IAttrValue* temp=NULL;
	if(parentNode){
		temp = Category(dataset, node->m_deqCI,parentNode->m_pIAClass);
	}
	// NULL, moglby byc wtedy, kiedy node->m_deqCI.empty() && parentNode->m_pIAClass==NULL
	// Zamiana klasy wiekszosciowej.
	if(temp){
		delete node->m_pIAClass;
		if(temp->GetAttrType() == e_NominalAttr){
			node->m_pIAClass = new CNomAttr((CNomAttr&)*temp);
		}
		if(temp->GetAttrType() == e_RealAttr){
			node->m_pIAClass = new CRealAttr((CRealAttr&)*temp);
		}
	}

	if(node->m_bLeaf) return;

	// Dane nalezy podzielic na tyle pozbiorow ile poddrzew wychodzi z biezacego
	// wezla.
	deque<deque<CDataObject*> > ddeqCI(node->m_vBranch.size());
	// Dzielenie danych.
	assert(node->m_pITest);
	for(unsigned i=0; i<node->m_deqCI.size(); i++){
		unsigned id = node->m_pITest->GetBranchID(node->m_deqCI[i]);
		assert(ddeqCI.size()>id);
		ddeqCI[id].push_back(node->m_deqCI[i]);
	}
	// Sprawdzenie czy nie ma pustego podzialu.
	for(size_t t=0; t<ddeqCI.size(); t++){
		if(ddeqCI[t].empty()){
			if(parentNode){
				// Jak, chociaz jeden podzial jest pusty i parentNode nie jest NULL,
				// to biezacy wezel jest zastepowany losowo wybranym synem.
				if(m_iLogLevel>=3) CLogger::LogEvent("A2 Node replaced by its son - an empty branch","",__LINE__,m_ofsLogFile);
				CDTreeNode *tempNode = node->m_vBranch[rand()%node->m_vBranch.size()];
				for(size_t s=0; s<node->m_vBranch.size(); s++){
					if(node->m_vBranch[s] != tempNode){
						delete node->m_vBranch[s];
					}
				}
				node->m_vBranch.clear();
				for(size_t s=0; s<parentNode->m_vBranch.size(); s++){
					if(parentNode->m_vBranch[s]==node){
						tempNode->m_deqCI=node->m_deqCI;
						parentNode->m_vBranch[s]=tempNode;
						delete node;
						ApplyTestChange(dataset,parentNode->m_vBranch[s],parentNode);
						break;
					}
				}
				return;
			}else{
				// Jak, chociaz jeden podzial jest pusty i parentNode jest NULL,
				// to biezacy wezel staje sie lisciem.
				if(m_iLogLevel>=3) CLogger::LogEvent("A1 Tree prunned - an empty branch","",__LINE__,m_ofsLogFile);
				node->m_bLeaf = true;
				while(!node->m_vBranch.empty()){
					delete node->m_vBranch.back();
					node->m_vBranch.pop_back();
				}
				delete node->m_pITest;
				node->m_pITest = NULL;
				return;
			}
		}
	}

	// Rekurencyjne schodzenie w dol drzewa i obliczanie na biezaco parametrow.
	for(unsigned i=0; i<node->m_vBranch.size(); i++){
		node->m_vBranch[i]->m_deqCI = ddeqCI[i];
		ApplyTestChange(dataset,node->m_vBranch[i],node);
	}
}

void IEAIDTInducer::CheckRepeatedClasses(CDTreeNode* node){
	if(!node->m_vBranch[0]->m_bLeaf){
		return;
	}
	CDTreeNode* first = node->m_vBranch[0];
	for(unsigned i=1; i<node->m_vBranch.size(); i++){
		if(node->m_vBranch[i]->m_bLeaf){
			if( *(first->m_pIAClass) != *(node->m_vBranch[i]->m_pIAClass) ){
				// Jesli sa chociaz dwa liscie z rozna klasa to konczymy
				return;
			}
		}else{
			// Kiedy chociaz jeden potomek nie jest lisciem to konczymy sprawdzanie
			return;
		}
	}
	// Jesli wszystkie potomki sa liscmi majacymi taka sama decyzje to przycinamy wezel
	if(m_iLogLevel>=3) CLogger::LogEvent("A1 Tree prunned - the same classes","",__LINE__,m_ofsLogFile);
	node->m_bLeaf = true;
	while(!node->m_vBranch.empty()){
		delete node->m_vBranch.back();
		node->m_vBranch.pop_back();
	}
	delete node->m_pITest;
	node->m_pITest = NULL;
}

#ifdef MLPEAOPERATORSTEST
void IEAIDTInducer::OPTEST_OperatorsTestLog(IIndividual &indv, string sOpName){
	CEAIDTIndividual &eaIndv = (CEAIDTIndividual&)indv;
	double prevFitness = indv.m_dFitness;
	EstimateErrors(&eaIndv);
	EvaluateIndividualsFitness(&indv);

	map<string,pair<int,int> >::iterator iter = m_mOpEfficiency.find(sOpName);
	if(iter==m_mOpEfficiency.end()){
		CLogger::LogError("operator not found",__FILE__,__LINE__);
	}else{
		if(indv.m_dFitness > prevFitness){		
			iter->second.first += 1;
		}
		iter->second.second += 1;
	}
	//--------------------------------------------------------------------------------------------------
	map<string,pair<int,int> >::iterator iterBestIndv = m_mOpEfficiencyBestIndv.find(sOpName);
	if(iterBestIndv==m_mOpEfficiencyBestIndv.end()){
		CLogger::LogError("operator not found",__FILE__,__LINE__);
	}else{
		if(indv.m_dFitness > m_dMaxFitness){
			// Wykluczenie sytuacji, ze to dzialanie poprzeniego operatora mogla juz poprawic jakosc.
			if(prevFitness <= m_dMaxFitness){
				iterBestIndv->second.first += 1;
			}
		}
		iterBestIndv->second.second += 1;
	}
}
#endif

