/**
 * $Revision: 1.62 $
 * $Date: 2006/07/13 09:25:37 $
 */

#ifndef __IEAIDTINDUCERH__
#define __IEAIDTINDUCERH__

#include <fstream>
#include <deque>
#include <map>
#include "../../mlpcore/MLPCommon/IEvolutionaryAlg.h"
#include "../../mlpcore/MLPDT/IDTInducer.h"
#include "../MLPDT_EA/EAIDTIndividual.h"

/**
 * Metoda do porownywania lisci drzewa przy ocenie do mutacji.
 * @see MutateTreeGradeNodes
 */
bool CDTreeNodeLess(CDTreeNode *n1, CDTreeNode *n2);

/**
 * Abstrakcyjna klasa ucznia generujacego drzewo decyzyjne z wykorzystaniem algorytmu
 * ewolucyjnego.
 * @author Marek Grzes'
 */
class IEAIDTInducer: public IDTInducer, public IEvolutionaryAlg{

public:
	virtual vector<string> GetParams() const;
	virtual bool SetParam(string sParamName, string sParamValue);
	virtual bool SetSubParam(string sParentParamName, string sSubParamName, string sSubParamValue);
	virtual void PrintParamsXML(ostream &ostr, int nsp);

	IEAIDTInducer(void);
	virtual ~IEAIDTInducer(void);
	
protected:

	/**
	 * Parametry dla wewenetrznej kroswalidacji.
	 */
	bool	m_bInnerCrossValidation;
	int	m_iICVFolds;
	int	m_iICVRe_runs;
	string	m_sICVParam;
	/**
	 * Parametr, ktory jest podkrecany za pomoca wewnetrznej kroswalidacji musi
	 * byc typu liczbowego.
	 */
	double	m_dICVlvalue;
	double	m_dICVrvalue;
	int	m_iICVsteps;

	/**
	 * Wplyw kosztu na funkcje dopasowania.
	 */
	double m_dGamma;
	
	/**
	 * Zmienna wskazujaca czy wykorzystywac operator, ktory probuje
	 * maksymalizowac fitness przycinajac drzewo.
	 * @see MaximizeFitness
	 */
	bool	m_bMaximizeFitness;

	/**
	 * Wartosc wskazujaca maksymalna glebokosc drzew w generowanej populacji poczatkowej.
	 * -1 - to parametr nie jest uwzgledniany
	 */
	int m_iInitMaxDepth;

	
	/**
	 * P-o z jakim dany osobnik z kazdej populacji (poczawszy od drugiej)
	 * bedzie poddany operatorowi MaximizeFitness
	 */
	int	m_iMaxFitnessProbability;

	/**
	 * P-a wywolan poszczegolnych operatorow krzyzowania. Rozmiar i wartosci tego wektora jest
	 * ustawiany w konstruktorach klas potomnych i jest zawsze rowny liczbie dostepnych
	 * operatorow krzyzowania.
	 */
	vector<int>		m_viPc;

	/**
	 * P-a wywolan poszczegolnych operatorow mutacji. Rozmiar i wartosci tego wektora jest
	 * ustawiany w konstruktorach klas potomnych i jest zawszy rowny liczbie dostepnych
	 * operatorow mutacji na wezlach wewnetrznych.
	 */
	vector<int>		m_viPm;
	
	/**
	 * P-a wywolan poszczegolnych operatorow mutacji. Rozmiar i wartosci tego wektora jest
	 * ustawiany w konstruktorach klas potomnych i jest zawszy rowny liczbie dostepnych
	 * operatorow mutacji na lisciach.
	 */
	vector<int>		m_viPmL;

	/**
	 * Zmienna wykorzystywana przy liczeniu funckji fitness.
	 */
	double	m_dFitnessAlfa;

	/**
	 * Sposob liczenia funkcji fitness (rodzaj funkcji fitness).
	 * sum - fitness z suma
	 * division - fitness z dzieleniem
	 */
	string	m_sFitnessCalculation;
	
	/**
	 * Najwiekszy znaleziony w populacji poczatkowej rozmiar drzewa.
	 */
	unsigned	m_uMaxTreeSize;
	
	/**
	 * Maksymalny koszt klasyfikacji wg. zbioru uczacego.
	 */
	float	m_fMaxCostOfClassification;

	/**
	 * Maksymalny koszt atrybutow wg. zbioru uczacego.
	 */
	float	m_fMaxCostOfAttributes;
	
	/**
	 * Sredni rozmiar drzewa z populacji poczatkowej.
	 */
	unsigned	m_uAverageTreeSize;
	
	bool		m_bUseMaxTreeSize;
	bool		m_bUseAverageTreeSize;
	
	/**
	 * Parametr okresla rodzaj mutacji:
	 * 0 - p-o mutacji oznacza p-o mutowania kazdego wezla z osobna
	 * 1 - p-o mutacji oznacza p-o mutowania calego drzewa
	 */
	int		m_iMutationType;
	
	/**
	 * Metoda do sprawdzenia poprawnosci parametrow algorytmu podanych przez uzytkownika.
	 * Jest wywolywana na poczatku metody train. Jesli metoda wykryje blad wartosci
	 * parametrow to zakonczy dzialanie programu i wypisze na cerr komunikat o 
	 * przyczynie bledu.
	 * Sprawdza czy p-a Pc, Pm i PmL dla operatorow krzyzowania i mutacji sa dobrze okreslnone.
	 */
	virtual void CheckParams();
	
	virtual void CreateGnuplotProjects(string datafilename, string outfilenamestem);
	virtual void PrintFitnessLogHeader(ofstream &ofFlog);
	virtual void PrintFitnessLog(deque<IIndividual*> &p,IIndividual* bestind, unsigned iternum, ostream &ostr);
	virtual void PrintIndividualToLog(string prefix, IIndividual* ind, bool bPrintStructure=false);
	
	/**
	 * Oblicza funkcje fitness na podstawie kosztu blednej klasyfikacji i rozmiaru
	 * drzewa.
	 * <b>Korzystajac z tych funkcji trzeba zwracac uwage na to jak jest liczona
	 * roznica dopasowania w funkcji MaximizeFitness. Dla "sum" mozemy badac 
	 * ronice, ale dla "division" nie.</b>
	 */
	virtual double FitnessSizeCCost(double dSize, double dCCost) const;

	/**
	 * Oblicza funkcje fitness na podstawie kosztu blednej klasyfikacji,
	 * kosztu atrybutow i rozmiaru drzewa.
	 */
	virtual double FitnessSizeCCostACost(double iSize, double dCCost, double dACost) const {
		//cerr<<"ccost="<<dCCost<<"\tacost="<<dACost<<endl;
		//cerr<<m_fMaxCostOfClassification<<" "<<m_fMaxCostOfAttributes<<endl;
		//cerr<<dCCost/m_fMaxCostOfClassification<<" "<<dACost/m_fMaxCostOfAttributes<<endl;
		//return m_dFitnessAlfa*(iSize-1) + dCCost/m_fMaxCostOfClassification + dACost/m_fMaxCostOfAttributes;
		//return m_dFitnessAlfa*(iSize-1) + dCCost + dACost;
		return m_dFitnessAlfa*(iSize-1) + (dCCost + dACost)/(m_fMaxCostOfClassification + m_fMaxCostOfAttributes);
	}
	
	/**
	 * Oblicza funcje fitness na podstawie bledu reklasyfikacji i rozmiaru
	 * drzewa.
	 */
	virtual double FitnessReclSize(double dRecl, double iSize) const {
		return dRecl + m_dFitnessAlfa * (iSize - 1);
	}
	
	/**
	 * Oblicza funcje fitness na podstawie bledu reklasyfikacji, rozmiaru
	 * drzewa oraz kosztu atrybutow.
	 */
	virtual double FitnessReclSizeACost(double dRecl, double iSize, double dACost) const {
		return dRecl + m_dFitnessAlfa * (iSize - 1) + m_dGamma * dACost;
	}
	
	/**
	 * Metoda wywolywana po utworzeniu populacji poczatkowej.
	 */
	virtual void PostCreateInitialPopulation();
	
	/**
	 * Operator rozwija lisc drzewa w poddrzewo (zamienia lisc na poddrzewo).
	 */
	virtual bool ExpandLeaf(CDTreeNode* node, unsigned attrCount, double dStop, const IDataSet *dataset)=0;

	/**
	 * Wywolanie tej funkcji powoduje przypisane do zmiennej m_uAverageTreeSize,
	 * sredniego rozmiaru drzewa sposrod wszystkich drzew znajdujacych sie
	 * w m_dqPopulation.
	 */
	virtual void EstimateAverageTreeSize();
		
	/**
	 * Wywolanie tej funkcji powoduje przypisane do zmiennej m_uMaxTreeSize,
	 * maksymalnego rozmiaru drzewa sposrod wszystkich drzew znajdujacych sie 
	 * w m_dqPopulation.
	 */
	virtual void EstimateMaxTreeSize();

	/**
	 * Zwraca numer operatora krzyzowania do zastosowania. Uwzglednia rozklad p-a operatorw
	 * zadany wektorem m_viPc. Jesli p-a w tym wektorze nie sumuja sie do 100, to moze byc
	 * zwrocony numer operatora -1, co bedzie oznaczalo, ze zaden operator sie nie wykona.
	 * @return -1 kiedy p-a nie sumuja sie do 100 (np. wszystkie maja wartosc 0).
	 * @see m_viPc
	 */
	virtual int GetCrossoverOperatorID();

	/**
	 * Funkcja zwraca w wektrach nodes i leaves odpowiednio wszystkie wezly weznetrzne
	 * uporzadkowanie tak jak przechodzenie wszerz oraz wszystkie liscie, poczynajac od
	 * wezla node.
	 */
	virtual void GetNodesLeaves(CDTreeNode* node, deque<CDTreeNode*> &nodes, deque<CDTreeNode*> &leaves);

	/**
	 * Zwraca numer operatora mutacji do zastosowania. Uwzglednia rozklad p-a operatorw
	 * zadany wektorami m_viPm (dla wezlow wewnetrznych) i m_viPmL (dla lisci).
	 * Jesli p-a w tych wektorach nie sumuja sie wzajemnie do 100, to moze byc
	 * zwrocony numer operatora -1, co bedzie oznaczalo, ze zaden operator sie nie wykona.
	 * @return -1 kiedy p-a nie sumuja sie do 100 (np. wszystkie maja wartosc 0).
	 * @param bLeaf - true kiedy wezel, dla ktorego losowany jest operator jest lisciem
	 * @see m_viPm
	 * @see m_viPmL
	 * @see SetMutationOperators
	 */
	virtual int GetMutationOperatorID(CDTreeNode* pNodef);
	
	/**
	 * Medota zwraca wektor indeksow tych operatorow, ktore mozna zastosowac na 
	 * przekazanym z parametrze wezle drzewa. Indeksy te musze byc zgodne z indeksami w
	 * switch'u odpowiedniej funkcji MutateTree.
	 * @see MutateTree
	 */
	virtual vector<size_t> SetMutationOperators(CDTreeNode* pNode)=0;

	/**
	 * Estymuje nowe wartosci bledow w drzewie oraz koryguje wartosc
	 * dopasowania. Dlatego po wykonaniu mutacji wartosci te sa aktualne
	 * i odpowiadaja wprowadzonym zmianom.
	 */
	virtual void MutationOperator(int iterNum);

	/**
	 * Mutacja wezlow drzewa. Rekurencyjne przechodzenie drzewa i przeprowadzanie
	 * mutacji z prawdopodobienstwem m_dProbabilityOfMutation.
	 * @return true - zwracane kiedy co najmniej jeden wezel w poddrzewie byl
	 * poddany mutacji
	 */
	virtual bool MutateTree(CDTreeNode* node, const IDataSet* dataset, CEAIDTIndividual &indv);

	virtual bool MutateNode(CDTreeNode* node, const IDataSet* dataset, CEAIDTIndividual &indv)=0;
	
	/**
	 * Mutacja drzewa. Drzewo jest mutowane z p-em m_dProbabilityOfMutation.
	 * @return true - zwracane kiedy drzewo bylo mutowane
	 */
	virtual bool MutateTreeGradeNodes(CDTreeNode* node, const IDataSet* dataset, CEAIDTIndividual &indv);
	
	/**
	 * Zwraca prawdopodobienstwo mutacji dla podanego wezla. Dla wezlow wewnetrznych zwraca 
	 * m_iProbabilityOfMutation a dla lisci zwraca wartosc przeskalowana wzgledem jakosci
	 * wezla.
	 */
	virtual int GetProbabilityOfMutation(CDTreeNode* node) const;

	/**
	 * Fukcja rekurencyjna, ktora uaktualnia wartosci m_deqCI pamietane w 
	 * wezlach po tym jak w wezle nadrzednym zmieniony zostal test.
	 * Funkcja nie uaktualnia wartosci bledow. 
	 * Funkcja przeprowadza korekcje klas wiekszosciowych.
	 *
	 * <p><b>UWAGA:</b> Funkcja przycina drzewo kiedy w podziale danych jest co najmniej
	 * jeden pusty zbior.</p>
	 */
	virtual void ApplyTestChange(const IDataSet* dataset, CDTreeNode* node, CDTreeNode* parentNode=NULL);

	/**
	 * Funkcja wewnetrznej dyzjunkcji. Po jej zakonczeniu wartosc dobieranego
	 * parametru jest ustawiona na wartosc znaleziona przez ta funkcje.
	 */
	void InnerCrossValidation();

	/**
	 * Metoda redukcji rozmiaru drzewa, ktora maksymalizuje dopasowanie
	 * osobnika.
	 * Przed wywolaniem tej funkcji dla danego osobnika nie musi byc najpierw
	 * wywolana fukcja EstimateErrors.
	 * Funkcja po zakonczeniu dzialania zostawia drzewo w takim stanie jak
	 * funkcja EsitmateErrors. Tzn. po zakonczeniu dzialania MaximizeFitness mozna
	 * wywolac EvaluateIndividualsFitness.
	 * <p><b>UWAGA!</b> Funkcja ta ma uwzgledniac tez koszty.</p>
	 * @see CDTreeNode#EstimateErrors
	 * @see m_bMaximizeFitness
	 * <strong>UWAGA: cialo tej funkcji jest identyczne jak EstimateErrors z tym ze na koncu
	 * funkcji EstimateErrors trzeba dodac kod maksymalizujacy dopasowanie czyli przycinajacy
	 * drzewo oraz zmianic nazwe funkcji w wywolaniu rekurencyjnym.</strong>
	 */
	 virtual void MaximizeFitness(CDTreeNode* node, vector<bool> vbUsed);
	 virtual void MaximizeFitness(IIndividual* individual);
	
	virtual void PreTrain(const IDataSet *trainingDataSet);
	
	/**
	 * Metoda sprawdza czy w wezlach potomnych bedacych liscmi nie ma tej samej klasy.
	 * Jesli dla co najmniej jednej pary lisci jest ta sama decyzja to podany jako parametr
	 * wezel jest przycinany do liscia.
	 */
	void CheckRepeatedClasses(CDTreeNode* node);

	/**
	 * Funkcja ta jest wywolywana aby przygotowac informacje o bledach w drzewie
	 * do liczenia fitness. Musi byc wywolana na calym drzewie dlatego jako
	 * parametr pobiera wskaznik do osobnika.
	 */
	virtual void EstimateErrors(IIndividual* individual);
	
	/**
	 * Wylicza parametry klasyfikacji dla podanego wezla (liscia).
	 */
	virtual void GetNodeClassification(CDTreeNode* node);
	
#ifdef MLPEAOPERATORSTEST
	void OPTEST_OperatorsTestLog(IIndividual &indv, string sOpName);
#endif

private:
	virtual IClassifier *ActualTrain(const IDataSet *trainingDataSet);

	/**
	 * Funkcja oblicza blad na zbiorze uczacym na podstawie rozlozenia przykladow
	 * pamietanego w polu m_deqCI. W lisciach wstawia liczbe
	 * pokrywanych przykladow m_dN, liczbe niepoprawnie pokrywanych 
	 * przykladow m_dE oraz wylicza blad pesymistyczny na podstawie
	 * przedzialu ufnosci dla wskaznika strukury. W wezlach wstawiane sa
	 * wartosci bedace suma wartosci w popdrzewach.
	 * Funkcja oblicza takze koszt klasyfikacji oraz testow i obliczone wartosci
	 * zapamietuje w polach obiektu.
	 * @param node - wezel startowy
	 * @param vbUsed - wektor logiczny wskazujacy, ktore atrybuty byly juz wliczone powyzej danego
	 * wezla. Jesli vbUsed[i]=true to atrybut i wystapil juz <b>na sciezce</b> powyzej biezacego wezla.
	 * @see CDTreeNode#m_fCostOfClassification
	 * @see CDTreeNode#m_fCostOfAttrs
	 */
	virtual void EstimateErrors(CDTreeNode* node, vector<bool> vbUsed);

	/**
	 * Znajduje "optymalna" wartosc parametru, dla ktorego przeprowadzana jest
	 * wewnetrzna kroswalidacja.
	 * @param test - testowy zbior danych, na ktorym bedzie wyznaczana wartosc parametru
	 * @return - nalepsza wartosc parametru dla zadanego zbioru uczacego i testowego
	 * wynikajacego z podzialu wlasciwego zbioru uczacego na podstawie wewnetrznej
	 * kroswalidacji.
	 */
	double FindOptimalValue(const IDataSet* test);

	/**
	 * Pojedyncze uruchomienie algorytmu. Wywolywane przez funkcje
	 * FindOptimalValue(). Zwraca jakosc na zbiorze test.
	 * Algorytm naucza sie na zbiorze danych z m_pIDataset.
	 * @param test - zbior danych, na ktorym sprawdza jest jakosc uzyskanego
	 * klasyfikatora
	 * @return - jakosc na zbiorze test. <strong>UWAGA!</strong> dla jakosci
	 * klasyfikacji jest to w naturalny sposob liczba poprawnie klasyfikowanych
	 * przykladow podzielona przez liczbe wszystkich obserwacji. Dla kosztow
	 * blednej klasyfikacji (i kosztu atrybutuow) jest to koszt ze znakiem minus.
	 * Bo zwrocona wartosc trzeba maksymalizowac.
	 */
	double SingleRun(const IDataSet* test);

};

#endif
