/**
 * $Revision: 1.84 $
 * $Date: 2006/05/25 13:02:42 $
 */

#include "EAIDTInducerAP.h"
#include <cstdlib>
#include <algorithm>
#include "../../mlpcore/MLPDT/DTClassifier.h"
#include "../../mlpcore/MLPDT/NominalTest.h"
#include "../../mlpcore/MLPDT/RealTest.h"
#include "../../mlpcore/MLPDT/TestContainer2.h"
#include "../../mlpcore/MLPCommon/MLPStatistics.h"
#include <iostream>
#include <cmath>
#include "../../mlpcore/MLPDataset/NomAttr.h"
#include "../../mlpcore/MLPDataset/RealAttr.h"
#include <sstream>
#include <typeinfo>
#include <iomanip>

#define __PM5__ "[add branche]"
#define __PM6__ "[merge branches]"
#define __PM7__ "[move value between branches]"
#define __PM8__ "[Mixed dipol cut]"
#define __PM9__ "[Pure dipol joined]"

using namespace std;

bool ITestGreater(ITest* it1, ITest* it2){
	return it1->GetSubSetCount() > it2->GetSubSetCount();
}

CEAIDTInducerAP::CEAIDTInducerAP(void){
	m_bEnlargeTheMargin = false;
	m_pITestContainer = NULL;
	
	//if(!m_DTIDTInducer.SetParam("selectionofatest","gain ratio criterion")){
	if(!m_DTIDTInducer.SetParam("selectionofatest","dipolar")){
		CLogger::LogEvent("SetParam",__FILE__,__LINE__);
	}

	m_sThreshold	 = "midpoint";
	//m_ofsLogFile.open("MLP_EATreeInducerAP.log",ofstream::out | ofstream::app);
	m_dFitnessAlfa = 0.5;

	// Ustawienie domyslnych wartosci i inicjalizacja rozmiaru wektorow. Kazda klasa
	// potomna klasy EAIDTInducer musi ustawiac te parametry we wlasnym konstruktorze.
	vector<int> temppc(3,33);
	temppc[2] = 34;
	vector<int> temppm(10,10);
	vector<int> temppml(2,50);
	m_viPc = temppc;
	m_viPm = temppm;
	m_viPmL = temppml;
	
	// Kazda implementacja klasy EAIDTInducer musi miec przypisanie nazw operatow w swoim konstrukorze
#ifdef MLPEAOPERATORSTEST
	// Operatory krzyzowania
	m_vOpNames.push_back("[branches swapped in random order]"); //ok
	m_vOpNames.push_back("[subtrees swapped]"); //ok
	m_vOpNames.push_back("[tests swapped]"); //ok
	
	// Operatory mutacji
	m_vOpNames.push_back("[tests between father and son exchanged]");
	m_vOpNames.push_back("[symmetric mutation]");
	m_vOpNames.push_back("[Test in node changed]");
	m_vOpNames.push_back("[Node changed into leaf]");
	m_vOpNames.push_back("[threshold in real outcome test changed]");
	m_vOpNames.push_back("[Leaf changed into node]");
	m_vOpNames.push_back("[Leaf expanded into subtree]");

	// Operatory mutacji dla wewnetrznej dyzjunkcji
	m_vOpNames.push_back(__PM5__);
	m_vOpNames.push_back(__PM6__);
	m_vOpNames.push_back(__PM7__);

	// Operatory dipolowe dla drzew AP
	m_vOpNames.push_back(__PM8__);
	m_vOpNames.push_back(__PM9__);

	pair<string,pair<int,int> > p;
	pair<int,int> zero(0,0);
	p.second = zero;
	for(size_t t=0; t<m_vOpNames.size(); t++){
		p.first = m_vOpNames[t];
		m_mOpEfficiency.insert(p);
	}
	// Kopia mapy dla najlepszego osobnika.
	m_mOpEfficiencyBestIndv = m_mOpEfficiency;
#endif
}

CEAIDTInducerAP::~CEAIDTInducerAP(void){
	if(m_pITestContainer){
		delete m_pITestContainer;
	}
	//m_ofsLogFile.close();
}

void CEAIDTInducerAP::ActualCreateInitialPopulation(){
stringstream sp;

	// choose initial population
	
	IInducer *inducer = new CTDIDTInducer();
	
	sp<<m_iInitMaxDepth;
	if(!inducer->SetParam("maxdepth",sp.str())){
		CLogger::LogError("SetParam - maxdepth",__FILE__,__LINE__);
	}
	if(m_bInnerDisjunction){
		if(!inducer->SetParam("innerdisjunction","yes")){
			CLogger::LogError("SetParam - innerdisjunction",__FILE__,__LINE__);
		}
	}
	if(m_bCostOfAttrs){
		if(!inducer->SetParam("costofattributes","yes")){
			CLogger::LogError("SetParam - costofattributes",__FILE__,__LINE__);
		}
	}
	if(m_bCostOfClassification){
		if(!inducer->SetParam("costofclassification","yes")){
			CLogger::LogError("SetParam - costofclassification",__FILE__,__LINE__);
		}
	}
	if(!inducer->SetParam("selection of a test","dipolar")){
		CLogger::LogError("SetParam - selection of a test",__FILE__,__LINE__);
	}
	if(!inducer->SetParam("pruning","no")){
		CLogger::LogError("SetParam - pruning",__FILE__,__LINE__);
	}
	if(m_bMaximizeFitness){
		if(!inducer->SetParam("random stop","0.0")){
			CLogger::LogError("SetParam - random stop",__FILE__,__LINE__);
		}
	}else{
		if(!inducer->SetParam("random stop","0.3")){
			CLogger::LogError("SetParam - random stop",__FILE__,__LINE__);
		}
	}
	if(!inducer->SetParam("minimum items","5")){
		CLogger::LogError("SetParam - minimum items",__FILE__,__LINE__);
	}
	if(!inducer->SetParam("threshold","midpoint")){
		CLogger::LogError("SetParam - threshold",__FILE__,__LINE__);
	}
	if(!inducer->SetParam("clear coverd","no")){
		CLogger::LogError("SetParam - clear coverd",__FILE__,__LINE__);
	}
	for(unsigned i=0; i<m_uSizeOfPop; i++){
		CDTClassifier *DTClassifier = (CDTClassifier*)inducer->Train(m_pDataset);
		CEAIDTIndividual* temp = new CEAIDTIndividual();
		temp->m_CDTIndRoot = new CDTreeNode( *(DTClassifier->GetRoot()) );
		delete DTClassifier;
		m_dqPopulation.push_back(temp);
	}
	delete inducer;
}

void CEAIDTInducerAP::EvaluateIndividualsFitness(IIndividual* individual){
double a,b,tSize;
CEAIDTIndividual* eaindividual = (CEAIDTIndividual*)individual;
	
	if(m_iError == 0){
		a = eaindividual->m_CDTIndRoot->GetEstimateError() / 100.0;
	}else{
		a = eaindividual->m_CDTIndRoot->GetErrorOnTestDS() / 100.0;
	}
	tSize = (double)(eaindividual->m_CDTIndRoot->GetNodeCount());
	assert(!m_bUseAverageTreeSize || !m_bUseMaxTreeSize);	
	if(m_bUseAverageTreeSize){
		assert(m_uAverageTreeSize>0);
		b = tSize/(double)m_uAverageTreeSize;
	}else{
		if(m_bUseMaxTreeSize){
			assert(m_uMaxTreeSize>0);
			b = tSize/(double)m_uMaxTreeSize;
		}else{
			b = tSize;
		}
	}

	if(m_bCostOfClassification){
		if(m_bCostOfAttrs){
			eaindividual->m_dFitness = - FitnessSizeCCostACost(b, eaindividual->m_CDTIndRoot->GetCostOfClassificationOnTrainDS(), eaindividual->m_CDTIndRoot->GetCostOfAttributesOnTrainDS());
		}else{
			eaindividual->m_dFitness = - FitnessSizeCCost(b, eaindividual->m_CDTIndRoot->GetCostOfClassificationOnTrainDS());
		}
	}else{
		if(m_bCostOfAttrs){
			eaindividual->m_dFitness = - FitnessReclSizeACost(a,b,eaindividual->m_CDTIndRoot->GetCostOfAttributesOnTrainDS());
		}else{
			eaindividual->m_dFitness = - FitnessReclSize(a,b);
		}
	}
	
	eaindividual->m_dQuality = 1.0 - a;
	eaindividual->m_dSize = tSize;
	eaindividual->m_dCost = 0;
	if(m_bCostOfClassification){
		// to przypisanie bo pole m_dCost jest z klasy nadrzednej a sa miejsca gdzie
		// korzystamy z polimorfizmu
		eaindividual->m_dCost = eaindividual->m_CDTIndRoot->GetCostOfClassificationOnTrainDS();
	}
	if(m_bCostOfAttrs){
		eaindividual->m_dCost += eaindividual->m_CDTIndRoot->GetCostOfAttributesOnTrainDS();
	}
}

bool CEAIDTInducerAP::ExpandLeaf(CDTreeNode* node, unsigned attrCount, double dStop, const IDataSet *dataset){

	if(!node->m_bLeaf){
		return false;
	}
	
	// Random stop, dla procesu rozwijania drzewa.
	if( CMLPStatistics::Event((int)(100.0*dStop)) ){
		return false;
	}
	
	// Test wybierany zgodnie z metoda podana dla m_DTIDTInducer
	ITest* test=m_DTIDTInducer.PickOutTest(node->m_deqCI,*m_pITestContainer,dataset);
	if(test){
		// Utworz tablice podzialow
		vector<vector<unsigned> > vvDiv = test->GetSubSets(node->m_deqCI);

		// Sprawdzamy czy sa chociaz 2 niepuste podzialy
		int div=0;
		for(unsigned i=0; i<vvDiv.size(); i++){
			if(vvDiv[i].size()>0){
				div++;
			}
			if(div==2){
				break;
			}
		}
		// Jesli nie ma co namjmniej dwoch nie pustych podzialow to tworzymy lisc w tym miejscu
		if(div<2){
			delete test;
			node->m_bLeaf = true;
			return false;
		}

		node->m_bLeaf = false;
		//memcheck node->m_pITest = test->CloneTest();
		node->m_pITest = test;
		// Dla kazego podzbioru wynikajacego z podzialu: utworzenie potomka
		// bedacego lisciem, a nastepnie proba jego rozwiniecia.
		for(unsigned i=0; i<vvDiv.size(); i++){
			deque<CDataObject*> subset;
			vector<unsigned> subIDs = vvDiv[i];
			for(unsigned j=0; j<subIDs.size(); j++){
				subset.insert(subset.end(),node->m_deqCI[subIDs[j]]);
			}
			CDTreeNode* newnode = MakeTreeLeaf(dataset,subset,node->m_pIAClass);
			node->m_vBranch.push_back(newnode);
			ExpandLeaf(newnode,attrCount,dStop,dataset);
		}
		return true;
	}else{
		return false;
	}
}

vector<size_t> CEAIDTInducerAP::SetMutationOperators(CDTreeNode* pNode){
vector<size_t> res(4);
	if(!pNode->m_bLeaf){
		res[0] = 0;
		res[1] = 1;
		res[2] = 2;
		res[3] = 3;
		CRealTest* test = dynamic_cast<CRealTest*>(pNode->m_pITest);
		if(test!=0){
			res.push_back(4);
			// mixed dipol cut
			res.push_back(8);
			// pure dipol joined
			res.push_back(9);
		}else{
			CNominalTestInnerDisj* test2 = dynamic_cast<CNominalTestInnerDisj*>(pNode->m_pITest);
			if(test2!=0){
				res.push_back(5);
				res.push_back(6);
				res.push_back(7);
			}
		}
	}
	return res;
}

bool CEAIDTInducerAP::MutateNode(CDTreeNode* node, const IDataSet* dataset, CEAIDTIndividual &indv){
bool mutated = false;
		int _op_ = GetMutationOperatorID(node);
		switch(_op_){
		// *********************** operacje na wezlach wewn. ***************************
			case 0:{//0
				assert(node->m_pITest);
				assert(node->m_deqCI.size() > 0);
				// ************* Zamiana testu w danym wezle na test z jego losowo wybranego syna *****
				size_t son = rand()%node->m_vBranch.size();
				CDTreeNode* node2 = node->m_vBranch[son];
				// Tu trzeba sprawdzic liczba galezi wychodzocych z obu wezlow
				if(!node2->m_bLeaf && node->m_vBranch.size()==node2->m_vBranch.size()){
						ITest* test = node->m_pITest;
						node->m_pITest = node2->m_pITest;
						node2->m_pITest = test;
					if(m_iLogLevel>=2) CLogger::LogEvent("M0 tests between father and son exchanged","",__LINE__,m_ofsLogFile);
					mutated = true;
					ApplyTestChange(dataset,node);
#ifdef MLPEAOPERATORSTEST
					OPTEST_OperatorsTestLog(indv,"[tests between father and son exchanged]");
#endif
				}
				break;
			}//0
			case 1:{//1
				assert(node->m_pITest);
				assert(node->m_deqCI.size() > 0);
				// ************* Mutowanie symetryczne **************
				// Wylosowanie poddrzewa, ktore zostanie powielone i zastapi pozostale.
				size_t sbtree = rand()%node->m_vBranch.size();
				CDTreeNode* node2 = node->m_vBranch[sbtree];
				if(!node2->m_bLeaf){
					for(size_t t=0; t<node->m_vBranch.size(); t++){
						if(t!=sbtree){
							delete node->m_vBranch[t];
							node->m_vBranch[t] = new CDTreeNode(*node2);
						}
					}
					if(m_iLogLevel>=2) CLogger::LogEvent("M1 symmetric mutation","",__LINE__,m_ofsLogFile);
					mutated = true;
					ApplyTestChange(dataset,node);
#ifdef MLPEAOPERATORSTEST
					OPTEST_OperatorsTestLog(indv,"[symmetric mutation]");
#endif
				}
				break;
			}//1
			case 2:{//2
				assert(node->m_pITest);
				assert(node->m_deqCI.size() > 0);
				// *********** Zamiana testu na inny test ***************
				size_t n = node->m_pITest->GetSubSetCount();
				assert(n>0);
				ITest* test = FindTest(n,node->m_pITest);
				if(test){
					mutated = true;
					// Usuniecie istniejacego testu.
					delete node->m_pITest;
					// Tu nie trzeba skopiowac testu !!!, bo find zwraca kopie
					node->m_pITest = test;
					//TAK TEZ W EAIDTInducerOB => node->m_pITest = test;
					// Korekcja wartosci m_deqCI we wszystkich wezlach
					// poddrzewa o korzeniu w wezle node. Korekcja
					// klas wiekszosciowych.
					ApplyTestChange(dataset,node);
					if(m_iLogLevel>=2) CLogger::LogEvent("M2 Test in node changed","",__LINE__,m_ofsLogFile);
#ifdef MLPEAOPERATORSTEST
					OPTEST_OperatorsTestLog(indv,"[Test in node changed]");
#endif
				}
				break;
			}//2
			case 3:{//3
				// ************ Przyciecie - zamiana wezla na lisc. ****************************
				mutated = true;
				node->m_bLeaf = true;
				while(!node->m_vBranch.empty()){
					delete node->m_vBranch.back();
					node->m_vBranch.pop_back();
				}
				node->m_vBranch.clear();
				delete node->m_pITest;
				node->m_pITest = NULL;
				if(m_iLogLevel>=2) CLogger::LogEvent("M3 Node changed into leaf","",__LINE__,m_ofsLogFile);
#ifdef MLPEAOPERATORSTEST
				OPTEST_OperatorsTestLog(indv,"[Node changed into leaf]");
#endif
				break;
			}//3
			case 4:{//4
				// ************ Zamiana progu w tescie na atrybucie ciaglnym ********************
				CRealTest* pITest = dynamic_cast<CRealTest*>(node->m_pITest);
				if(pITest!=0){
						pITest->m_dThreshold = m_pITestContainer->GetRandomThreshold(pITest->GetAttrNum());
						mutated = true;
						ApplyTestChange(dataset,node);
						if(m_iLogLevel>=2) CLogger::LogEvent("M4 Threshold in real outcome test changed","",__LINE__,m_ofsLogFile);
#ifdef MLPEAOPERATORSTEST
						OPTEST_OperatorsTestLog(indv,"[threshold in real outcome test changed]");
#endif
				}
				break;
			}//4
			// *********** trzy operatory dla wewnetrznej dyzjunkcji
			case 5:{//5
				if(!m_bInnerDisjunction){
					break;
				}
				// ************ dodanie galezi do testu ********************
				CNominalTestInnerDisj* pITest = dynamic_cast<CNominalTestInnerDisj*>(node->m_pITest);
				if(pITest!=0){
					if(AddBranch(pITest,node)){
						mutated = true;
						ApplyTestChange(dataset,node);
						if(m_iLogLevel>=2) CLogger::LogEvent(string("M5 ")+__PM5__,"",__LINE__,m_ofsLogFile);
#ifdef MLPEAOPERATORSTEST
						OPTEST_OperatorsTestLog(indv,__PM5__);
#endif
					}
				}
				break;
			}//5
			case 6:{//6
				if(!m_bInnerDisjunction){
					break;
				}
				// ************ zlaczenie galezi w tescie ********************
				CNominalTestInnerDisj* pITest = dynamic_cast<CNominalTestInnerDisj*>(node->m_pITest);
				// Operator tylko dla testow z wewnetrzna dyzjunckja
				if(pITest!=0){
					// Zwrocony wezel musi miec wstanione pokrywane przyklady, i musi byc lisciem
					CDTreeNode* newNode = JoinBranches(this,pITest,node);
					if(newNode!=0){
						if(m_bMaximizeFitness){
							mutated = ExpandLeaf(newNode,dataset->GetCurrentAttrCount(),0.0,dataset);
						}else{
							mutated = ExpandLeaf(newNode,dataset->GetCurrentAttrCount(),0.3,dataset);
						}
						if(m_bMaximizeFitness){
							// Nie jest to maksymalizacja dokladna, bo tu nie uwzgledniamy kosztu
							// cech, ktore sa wyzej na sciezce (ale nie mamy tu do nich dostepu).
							vector<bool> vb(m_pDataset->GetCurrentAttrCount(),false);
							MaximizeFitness(node,vb);
						}
						mutated = true;
						// Uaktualnienie do wezla biezacego, czyli node, w dol
						ApplyTestChange(dataset,node);
						if(m_iLogLevel>=2) CLogger::LogEvent(string("M6 ")+__PM6__,"",__LINE__,m_ofsLogFile);
#ifdef MLPEAOPERATORSTEST
						OPTEST_OperatorsTestLog(indv,__PM6__);
#endif
					}
				}
				break;
			}//6
			case 7:{//7
				if(!m_bInnerDisjunction){
					break;
				}
				// ******* wartosc nominalna zmienia przyporzadkowanie do galezi **********
				CNominalTestInnerDisj* pITest = dynamic_cast<CNominalTestInnerDisj*>(node->m_pITest);
				if(pITest!=0){
					if(MoveValueBetweenBranches(pITest)){
						mutated = true;
						ApplyTestChange(dataset,node);
						if(m_iLogLevel>=2) CLogger::LogEvent(string("M7 ")+__PM7__,"",__LINE__,m_ofsLogFile);
#ifdef MLPEAOPERATORSTEST
						OPTEST_OperatorsTestLog(indv,__PM7__);
#endif
					}
				}
				break;
			}//7

			// ++++++++++ Operatory dipolowe dla drzew AP +++++++++
			case 8:{//8
				assert(node->m_pITest);
				assert(node->m_deqCI.size() > 0);
				// ***** Przeciecie dipola mieszanego *****
				CDataObject* ob1=NULL;
				CDataObject* ob2=NULL;
				if(FindMixedDipolAP(node->m_deqCI,ob1,ob2,node->m_pITest)){
					assert(ob1);
					assert(ob2);
					if(CutMixedDipolAP(ob1,ob2,node->m_pITest)){
						mutated = true;
						ApplyTestChange(dataset,node);
						if(m_iLogLevel>=2) CLogger::LogEvent("M8 Mixed dipol cut","",__LINE__,m_ofsLogFile);
#ifdef MLPEAOPERATORSTEST
						OPTEST_OperatorsTestLog(indv,__PM8__);
#endif
					}else{
						if(m_iLogLevel>=4) CLogger::LogEvent("NOT M8 Mixed dipol cut","",__LINE__,m_ofsLogFile);
					}
				}else{
					if(m_iLogLevel>=4) CLogger::LogEvent("NOT M8 Mixed dipol cut","",__LINE__,m_ofsLogFile);
				}
				break;
			}//8
			case 9:{//9
				assert(node->m_pITest);
				assert(node->m_deqCI.size() > 0);
				CDataObject* ob1=NULL;
				CDataObject* ob2=NULL;
				// ***** Usuniecie przeciecia dipola czystego *****
				if(FindPureDipolAP(node->m_deqCI,ob1,ob2,node->m_pITest)){
					assert(ob1);
					assert(ob2);
					if(JoinPureDipolAP(ob1,ob2,node->m_pITest)){
						mutated = true;
						ApplyTestChange(dataset,node);
						if(m_iLogLevel>=2) CLogger::LogEvent("M9 Pure dipol joined","",__LINE__,m_ofsLogFile);
#ifdef MLPEAOPERATORSTEST
						OPTEST_OperatorsTestLog(indv,__PM9__);
#endif
					}else{
						if(m_iLogLevel>=4) CLogger::LogEvent("NOT M9 Pure dipol joined","",__LINE__,m_ofsLogFile);
					}
				}else{
					if(m_iLogLevel>=4) CLogger::LogEvent("NOT M9 Pure dipol joined","",__LINE__,m_ofsLogFile);
				}
				break;
			}//9
			// ++++++++++ Operatory dipolowe dla drzew AP - KONIEC +++++++++

			// ******************************** operacje na lisciu ******************************
			case 10:{//10
//CLogger::LogEvent("	wejscie",__FILE__,__LINE__);
				// *************************** Dodanie testu w lisciu. **************************
				// Sprawdzenie czy jest sens robic wezel w tym miejscu.
				if(node->m_deqCI.size() > 1 && m_pITestContainer->size() > 0){
					ITest* test=0;
					// Test wybierany zgodnie z metoda podana dla m_DTIDTInducer
					test = m_DTIDTInducer.PickOutTest(node->m_deqCI,*m_pITestContainer,dataset);
					if(test!=0){
						// Utworz tablice podzialow
						vector<vector<unsigned> > vvDiv = test->GetSubSets(node->m_deqCI);
						mutated = true;
						node->m_bLeaf = false;
						// Dla kazego podzbioru wynikajacego z podzialu: utworzenie potomka
						// bedacego lisciem.
						for(unsigned i=0; i<vvDiv.size(); i++){
							deque<CDataObject*> subset;
							vector<unsigned> subIDs = vvDiv[i];
							for(unsigned j=0; j<subIDs.size(); j++){
								subset.insert(subset.end(),node->m_deqCI[subIDs[j]]);
							}
							CDTreeNode* newnode = MakeTreeLeaf(dataset,subset,node->m_pIAClass);
							node->m_vBranch.push_back(newnode);
						}
						node->m_pITest = test;
						if(m_iLogLevel>=2) CLogger::LogEvent("M10 Leaf changed into node","",__LINE__,m_ofsLogFile);
#ifdef MLPEAOPERATORSTEST
						OPTEST_OperatorsTestLog(indv,"[Leaf changed into node]");
#endif
					}
				}
//CLogger::LogEvent("	wyjscie",__FILE__,__LINE__);
				break;
			}//10
			case 11:{//11
//CLogger::LogEvent("		wejscie",__FILE__,__LINE__);
				// *************************** Rozwiniecie liscia w poddrzewo. **************************
				if(m_bMaximizeFitness){
//CLogger::LogEvent("		w1",__FILE__,__LINE__);
					mutated = ExpandLeaf(node,dataset->GetCurrentAttrCount(),0.0,dataset);
//CLogger::LogEvent("		w2",__FILE__,__LINE__);					
				}else{
					mutated = ExpandLeaf(node,dataset->GetCurrentAttrCount(),0.3,dataset);
				}
				if(m_bMaximizeFitness){
					// Nie jest to maksymalizacja dokladna, bo tu nie uwzgledniamy kosztu
					// cech, ktore sa wyzej na sciezce (ale nie mamy tu do nich dostepu).
					vector<bool> vb(m_pDataset->GetCurrentAttrCount(),false);
					MaximizeFitness(node,vb);
				}
				if(mutated){
					if(m_iLogLevel>=2) CLogger::LogEvent("M11 Leaf expanded into subtree","",__LINE__,m_ofsLogFile);
#ifdef MLPEAOPERATORSTEST
					OPTEST_OperatorsTestLog(indv,"[Leaf expanded into subtree]");
#endif
				}
				// tego poddrzewa nie mutujemy, od razu powrot.
//CLogger::LogEvent("		wyjscie",__FILE__,__LINE__);
				return mutated;
			}//11
		}//switch
	return mutated;
}

void CEAIDTInducerAP::CrossoverOperator(){
	size_t i;
	size_t j;
	size_t k;
	vector<unsigned> used;
	vector<size_t> vindv;
	
	for(size_t t=0; t<m_dqPopulation.size(); t++){
		if( rand()%100 + 1 <= m_iProbabilityOfCrossOver ){
			vindv.push_back(t);
		}
	}
	while(vindv.size() > 1){
		size_t ind1 = rand()%vindv.size();
		size_t ind2;
		do{
			 ind2 = rand()%vindv.size();
		}while(ind1==ind2);
		i = vindv[ind1];
		j = vindv[ind2];
		k=0;
		for(vector<size_t>::iterator it=vindv.begin(); it!=vindv.end(); it++){
			if(k==ind1){
				vindv.erase(it);
				break;
			}
			k++;
		}
		k=0;
		for(vector<size_t>::iterator it=vindv.begin(); it!=vindv.end(); it++){
			if(k==ind2){
				vindv.erase(it);
				break;
			}
			k++;
		}
// _____________ W tym miejscu mamy indeksy osobnikow do krzyzowania: i,j ______________
// _____________________________________________________________________________________
		int m = GetCrossoverOperatorID();
		switch(m){//009
			// ********** Zamiana galezi w kolejnosci losowej *********
			case 0:{
				//CDTreeNode* node1 = p[i]->m_CDTIndRoot->FindRandomNode();
				CDTreeNode* node1 = EAIND(m_dqPopulation[i])->m_CDTIndRoot->GetNodeEvenProb();
				if(node1==NULL || node1->m_deqCI.empty()){
					continue;
				}
				//CDTreeNode* node2 = p[j]->m_CDTIndRoot->FindRandomNode();
				CDTreeNode* node2 = EAIND(m_dqPopulation[j])->m_CDTIndRoot->GetNodeEvenProb();
				if(node2==NULL || node2->m_deqCI.empty()){
					continue;
				}
				used.clear();
				for(unsigned i1=0; i1<node1->m_vBranch.size(); i1++){
					if(used.size()>=node1->m_vBranch.size() || used.size()>=node2->m_vBranch.size()){
						break;
					}
					unsigned i2 = (unsigned)CMLPStatistics::RandVal((int)(node2->m_vBranch.size()));
					if(find(used.begin(),used.end(),i2) != used.end()){
						i2 = 0;
						while(find(used.begin(),used.end(),i2) != used.end()){
							i2++;
						}
					}
					if(i2>=node2->m_vBranch.size()){
						CLogger::LogError("Index error",__FILE__,__LINE__);
						continue;
					}
					CDTreeNode * temp = node1->m_vBranch[i1];
					node1->m_vBranch[i1] = node2->m_vBranch[i2];
					node2->m_vBranch[i2] = temp;
					used.push_back(i2);
				}
				ApplyTestChange(m_pDataset,node1);
				ApplyTestChange(m_pDataset,node2);
				if(m_iLogLevel>=2) CLogger::LogEvent("C1 Branches swapped in random order","",__LINE__,m_ofsLogFile);
#ifdef MLPEAOPERATORSTEST
				OPTEST_OperatorsTestLog(*(EAIND(m_dqPopulation[i])),"[branches swapped in random order]");
				OPTEST_OperatorsTestLog(*(EAIND(m_dqPopulation[j])),"[branches swapped in random order]");
#endif
				break;
			}
			// *************************************************************************

			// ********** Zamiana calych poddrzew ********************
			case 1:{
				//CDTreeNode* node1 = p[i]->m_CDTIndRoot->FindRandomNodeOrLeaf();
				CDTreeNode* node1 = EAIND(m_dqPopulation[i])->m_CDTIndRoot->GetNodeOrLeafEvenProb();
				if(node1==NULL || node1->m_deqCI.empty()){
					continue;
				}
				//CDTreeNode* node2 = p[j]->m_CDTIndRoot->FindRandomNodeOrLeaf();
				CDTreeNode* node2 = EAIND(m_dqPopulation[j])->m_CDTIndRoot->GetNodeOrLeafEvenProb();
				if(node2==NULL || node2->m_deqCI.empty()) {
					continue;
				}
				bool changed = false;
				if(!node1->m_bLeaf && !node2->m_bLeaf){
					ITest* pITest;
					vector<CDTreeNode*> branch;
					pITest=node1->m_pITest;
					branch=node1->m_vBranch;
					node1->m_pITest=node2->m_pITest;
					node1->m_vBranch=node2->m_vBranch;
					node2->m_pITest=pITest;
					node2->m_vBranch=branch;
					changed = true;
				}else{
					if(node1->m_bLeaf && !node2->m_bLeaf){
						node1->m_pITest=node2->m_pITest;
						node1->m_vBranch=node2->m_vBranch;
						node1->m_bLeaf=false;
						node2->m_pITest=NULL;
						node2->m_vBranch.clear();
						node2->m_bLeaf=true;
						changed = true;
					}else{
						if(!node1->m_bLeaf && node2->m_bLeaf){
							node2->m_pITest=node1->m_pITest;
							node2->m_vBranch=node1->m_vBranch;
							node2->m_bLeaf=false;
							node1->m_pITest=NULL;
							node1->m_vBranch.clear();
							node1->m_bLeaf=true;
							changed = true;
						}
					}
				}
				if(changed){
					ApplyTestChange(m_pDataset,node1);
					ApplyTestChange(m_pDataset,node2);
					if(m_iLogLevel>=2) CLogger::LogEvent("C2 Subtrees swapped","",__LINE__,m_ofsLogFile);
#ifdef MLPEAOPERATORSTEST
					OPTEST_OperatorsTestLog(*(EAIND(m_dqPopulation[i])),"[subtrees swapped]");
					OPTEST_OperatorsTestLog(*(EAIND(m_dqPopulation[j])),"[subtrees swapped]");
#endif
				}
				break;
			}
			// *************************************************************************
			
			// ********** Zamiana testow w wezlach **********************************
			case 2:{
				//CDTreeNode* node1 = p[i]->m_CDTIndRoot->FindRandomNode();
				CDTreeNode* node1 = EAIND(m_dqPopulation[i])->m_CDTIndRoot->GetNodeEvenProb();
				if(node1==NULL || node1->m_deqCI.empty()){
					continue;
				}
				CDTreeNode* node2 = EAIND(m_dqPopulation[j])->m_CDTIndRoot->GetNodeEvenProb((unsigned)(node1->m_vBranch.size()));
				if(node2==NULL || node2->m_deqCI.empty()) {
					continue;
				}
				assert(node1->m_vBranch.size()==node2->m_vBranch.size());
				ITest* pITest;
				pITest=node1->m_pITest;
				node1->m_pITest=node2->m_pITest;
				node2->m_pITest=pITest;
				ApplyTestChange(m_pDataset,node1);
				ApplyTestChange(m_pDataset,node2);
				if(m_iLogLevel>=2) CLogger::LogEvent("C3 Tests swapped","",__LINE__,m_ofsLogFile);
#ifdef MLPEAOPERATORSTEST
				OPTEST_OperatorsTestLog(*(EAIND(m_dqPopulation[i])),"[tests swapped]");
				OPTEST_OperatorsTestLog(*(EAIND(m_dqPopulation[j])),"[tests swapped]");
#endif
				break;
			}
			// *************************************************************************
		}//009
//______________________________________________________________________________________
		EstimateErrors(m_dqPopulation[i]);
		EstimateErrors(m_dqPopulation[j]);
		EvaluateIndividualsFitness(m_dqPopulation[i]);
		EvaluateIndividualsFitness(m_dqPopulation[j]);
	}
}

void CEAIDTInducerAP::EnlargeTheMargin(CDTreeNode *node){
	assert(node);
	CRealTest* rTest = dynamic_cast<CRealTest*>(node->m_pITest);
	// Przesuwanie marginesow tylko dla testow na trybutach rzeczywistych
	if(rTest!=0){
			deque<CDataObject*> &branch1 = node->m_vBranch[0]->m_deqCI;
			deque<CDataObject*> &branch2 = node->m_vBranch[1]->m_deqCI;
			// w galezi 1 sa wartosci <= threshold, czyli szukamy max
			double maxVal = (*branch1[0])[rTest->GetAttrNum()]->m_AttrValue.u_dReal;
			for(size_t i=1; i<branch1.size(); i++){
				if((*branch1[i])[rTest->GetAttrNum()]->m_AttrValue.u_dReal > maxVal){
					maxVal = (*branch1[i])[rTest->GetAttrNum()]->m_AttrValue.u_dReal;
				}
			}
			// w galezi 2 sa wartosci > threshold, czyli szukamy min
			double minVal = (*branch2[0])[rTest->GetAttrNum()]->m_AttrValue.u_dReal;
			for(size_t i=1; i<branch2.size(); i++){
				if((*branch2[i])[rTest->GetAttrNum()]->m_AttrValue.u_dReal < minVal){
					minVal = (*branch2[i])[rTest->GetAttrNum()]->m_AttrValue.u_dReal;
				}
			}
			// Margines dla atrybutu, na ktorym jest biezacy test.
			// jest tu (min - max) bo min jest po prawej stronie ciecia
			double margin = minVal-maxVal;
			// Teraz sprawdzamy, czy nie atrybutu, ktory rodziela tak samo przyklady, ale
			// z wiekszym marginesem.
			size_t nAttrs = m_pDataset->GetCurrentAttrCount();
			double margin2 = 0;
			size_t attr2 = rTest->GetAttrNum();
			double threshold2 = rTest->m_dThreshold;
			for(size_t t=0; t<nAttrs; t++){
				// dla roznych atrybutow niz w tescie biezacym oraz tylko dla ciaglych
				if((int)t!=rTest->GetAttrNum() && m_pDataset->GetAttrDesc(t)->m_iContinuous == 1){
					// Dla przykladow w galezi 1
					double maxVal1 = (*branch1[0])[t]->m_AttrValue.u_dReal;
					double minVal1 = (*branch1[0])[t]->m_AttrValue.u_dReal;
					for(size_t i=1; i<branch1.size(); i++){
						if((*branch1[i])[t]->m_AttrValue.u_dReal > maxVal1){
							maxVal1 = (*branch1[i])[t]->m_AttrValue.u_dReal;
						}
						if((*branch1[i])[t]->m_AttrValue.u_dReal < minVal1){
							minVal1 = (*branch1[i])[t]->m_AttrValue.u_dReal;
						}
					}
					// Dla przykladow w galezi 2
					double maxVal2 = (*branch2[0])[t]->m_AttrValue.u_dReal;
					double minVal2 = (*branch2[0])[t]->m_AttrValue.u_dReal;
					if(maxVal2 < maxVal1 && maxVal2 > minVal1){
						// Juz tu warunek niespelniony, nie da sie rozdzielic przykladow tak samo
						continue;
					}
					bool notFound = false;
					for(size_t i=1; i<branch2.size(); i++){
						if((*branch2[i])[t]->m_AttrValue.u_dReal > maxVal2){
							maxVal2 = (*branch2[i])[t]->m_AttrValue.u_dReal;
						}
						if((*branch2[i])[t]->m_AttrValue.u_dReal < minVal2){
							minVal2 = (*branch2[i])[t]->m_AttrValue.u_dReal;
						}
						if( ! (maxVal1 > minVal2 || maxVal2 > minVal1) ){
							notFound = true;
							break;
						}
					}
					if(notFound){
						continue;
					}else{
						if(maxVal1 < minVal2){
							double tmargin = minVal2-maxVal1;
							if(margin2<tmargin){
								margin2 = tmargin;
								threshold2 = maxVal2 + tmargin/2.0;
								attr2 = t;
							}
						}
						if(maxVal2 < minVal1){
							double tmargin = minVal1-maxVal2;
							if(margin2<tmargin){
								margin2 = tmargin;
								threshold2 = maxVal2 + tmargin/2.0;
								attr2 = t;
							}
						}
					}	

				}
			}
			// sprawdzamy czy mamy jakis atrybut z lepszym marginesem
			if(margin2 > margin){
				cerr<<"attribute "<<rTest->GetAttrNum()<<" -> "<<attr2<<endl;
				CRealTest *t = new CRealTest(threshold2,attr2);
				delete node->m_pITest;
				node->m_pITest = t;
				rTest = NULL;
			}else{
				// jak nie to tylko srodkujemy punkt ciecia
				double m = maxVal + margin/2.0;
				cerr<<maxVal<<" "<<rTest->m_dThreshold<<" "<<minVal<<" -> "<<maxVal<<" "<<m<<" "<<minVal<<endl;
				rTest->m_dThreshold = m;
			}
	}
	if(!node->m_bLeaf){
		// Rekurencyjne schodzenie w dol drzewa.
		for(size_t t=0; t<node->m_vBranch.size(); t++){
			EnlargeTheMargin(node->m_vBranch[t]);
		}
	}
}

bool CEAIDTInducerAP::SetParam(string sParamName, string sParamValue){
	string::iterator i = sParamName.begin();
	while(i != sParamName.end()){
		*i = tolower(*i);
		++i;
	}
	i = sParamValue.begin();
	while(i != sParamValue.end()){
		*i = tolower(*i);
		++i;
	}

	if(sParamName=="threshold"){
		if(sParamValue=="midpoint" ||
			sParamValue=="largest value"){
				m_sThreshold=sParamValue;
				return true;
		}
		else
			return false;
	}
	if(sParamName=="enlargethemargin"){
		if(sParamValue=="yes"){
			m_bEnlargeTheMargin = true;
			return true;
		}
		if(sParamValue=="no"){
			m_bEnlargeTheMargin = false;
			return true;
		}
		return false;
	}
	
	if(IDTInducerAP::SetParam(sParamName,sParamValue)) return true;
	if(IEAIDTInducer::SetParam(sParamName,sParamValue)) return true;
	return false;
}

void CEAIDTInducerAP::PrintParamsXML(ostream &ostr, int nsp){
	IDTInducerAP::PrintParamsXML(ostr,nsp);
	IEAIDTInducer::PrintParamsXML(ostr,nsp);
	ostr.fill(' ');
	ostr<<setw(nsp)<<" "<<"<MLPParam Name=\"threshold\" Value=\"midpoint\" />"<<endl;
	ostr<<setw(nsp)<<" "<<"<!-- enlarging the margin of the best tree -->"<<endl;
	ostr<<setw(nsp)<<" "<<"<MLPParam Name=\"enlargethemargin\" Value=\"yes\" />"<<endl;
	ostr<<setw(nsp)<<" "<<"<!-- branches swapped in random order -->"<<endl;
	ostr<<setw(nsp)<<" "<<"<MLPParam Name=\"pc0\" Value=\"33\" />"<<endl;
	ostr<<setw(nsp)<<" "<<"<!-- subtrees swapped -->"<<endl;
	ostr<<setw(nsp)<<" "<<"<MLPParam Name=\"pc1\" Value=\"33\" />"<<endl;
	ostr<<setw(nsp)<<" "<<"<!-- tests swapped -->"<<endl;
	ostr<<setw(nsp)<<" "<<"<MLPParam Name=\"pc2\" Value=\"34\" />"<<endl;
	ostr<<setw(nsp)<<" "<<"<!-- tests between father and son exchanged -->"<<endl;
	ostr<<setw(nsp)<<" "<<"<MLPParam Name=\"pm0\" Value=\"10\" />"<<endl;
	ostr<<setw(nsp)<<" "<<"<!-- symmetric mutation -->"<<endl;
	ostr<<setw(nsp)<<" "<<"<MLPParam Name=\"pm1\" Value=\"10\" />"<<endl;
	ostr<<setw(nsp)<<" "<<"<!-- Test in node changed -->"<<endl;
	ostr<<setw(nsp)<<" "<<"<MLPParam Name=\"pm2\" Value=\"10\" />"<<endl;
	ostr<<setw(nsp)<<" "<<"<!-- Node changed into leaf -->"<<endl;
	ostr<<setw(nsp)<<" "<<"<MLPParam Name=\"pm3\" Value=\"10\" />"<<endl;
	ostr<<setw(nsp)<<" "<<"<!-- threshold in real outcome test changed -->"<<endl;
	ostr<<setw(nsp)<<" "<<"<MLPParam Name=\"pm4\" Value=\"10\" />"<<endl;
	ostr<<setw(nsp)<<" "<<"<!-- "<<__PM5__<<" -->"<<endl;
	ostr<<setw(nsp)<<" "<<"<MLPParam Name=\"pm5\" Value=\"10\" />"<<endl;
	ostr<<setw(nsp)<<" "<<"<!-- "<<__PM6__<<" -->"<<endl;
	ostr<<setw(nsp)<<" "<<"<MLPParam Name=\"pm6\" Value=\"10\" />"<<endl;
	ostr<<setw(nsp)<<" "<<"<!-- "<<__PM7__<<" -->"<<endl;
	ostr<<setw(nsp)<<" "<<"<MLPParam Name=\"pm7\" Value=\"10\" />"<<endl;
	ostr<<setw(nsp)<<" "<<"<!-- "<<__PM8__<<" -->"<<endl;
	ostr<<setw(nsp)<<" "<<"<MLPParam Name=\"pm8\" Value=\"10\" />"<<endl;
	ostr<<setw(nsp)<<" "<<"<!-- "<<__PM9__<<" -->"<<endl;
	ostr<<setw(nsp)<<" "<<"<MLPParam Name=\"pm9\" Value=\"10\" />"<<endl;
	ostr<<setw(nsp)<<" "<<"<!-- Leaf changed into node -->"<<endl;
	ostr<<setw(nsp)<<" "<<"<MLPParam Name=\"pml0\" Value=\"50\" />"<<endl;
	ostr<<setw(nsp)<<" "<<"<!-- Leaf expanded into subtree -->"<<endl;
	ostr<<setw(nsp)<<" "<<"<MLPParam Name=\"pml1\" Value=\"50\" />"<<endl;
	ostr.fill(' ');
}

vector<string> CEAIDTInducerAP::GetParams() const{
vector<string> res;
stringstream ss;
	
	res = IEAIDTInducer::GetParams();
	
	vector<string> temp = IDTInducerAP::GetParams();
	for(size_t t=0; t<temp.size(); t++){
		res.push_back(temp[t]);
	}
	
	res.push_back("threshold");
	res.push_back(m_sThreshold);
	res.push_back("enlargethemargin");
	if(m_bEnlargeTheMargin){
		res.push_back("yes");
	}else{
		res.push_back("no");
	}
	
	return res;
}

bool CEAIDTInducerAP::MergeLeavesWithRepeatedClasses(CDTreeNode* node){
bool changed = false;
		// kiedy lisc to konczymy bo nie ma czego scalac
		if(node->m_bLeaf){
			return false;
		}
		for(size_t i=0; i<node->m_vBranch.size(); i++){
			if(MergeLeavesWithRepeatedClasses(node->m_vBranch[i])){
				changed = true;
			}
		}

		CNominalTestInnerDisj* pTest = dynamic_cast<CNominalTestInnerDisj*>(node->m_pITest);
		// Tylko dla testow z wewnetrzna dyzjunkcja robimy skladanie lisci
		if(pTest==0){
			return false;
		}
		vector<size_t> vt(pTest->m_vtInnerDisIndex.size());
		size_t t=0;
		size_t T=node->m_vBranch.size();
		while(t<T){
			if(node->m_vBranch[t]->m_bLeaf){
				// zerujemy liczne znalezionych lisci z ta sama klasa
				size_t vtL = 0;
				for(size_t u=t+1; u<node->m_vBranch.size(); u++){
					if(node->m_vBranch[u]->m_bLeaf){
						if( *(node->m_vBranch[t]->m_pIAClass) == *(node->m_vBranch[u]->m_pIAClass) ){
							vt[vtL++] = u;
						}
					}
				}
				if(vtL>0){
					changed = true;
					// w miennej t jest indeks do ktorego dolaczamy reszte

					// Uaktualnie wektora podwojnego indeksowania dla wewn. dyzjunkcji
					for(size_t w=0; w<pTest->m_vtInnerDisIndex.size(); w++){
						for(size_t x=0; x<vtL; x++){
							if(pTest->m_vtInnerDisIndex[w] == vt[x]){
								pTest->m_vtInnerDisIndex[w] = t;
							}
						}
					}
					// Zmniejszenie wartosci indeksow ktore sa wieksze od t oraz
					// wartosc ktorych jest wieksza od jakiegokolwiek elementu z vt
					// Zmniejszamy o tyle, ile jest elementow w wektroze vt mniejszych od 
					// rozwazanego elementu
					for(size_t w=0; w<pTest->m_vtInnerDisIndex.size(); w++){
						// Rozpatrujemy tylko te, ktore sa wieksze od elementu do ktorego scalilismy
						if(pTest->m_vtInnerDisIndex[w]>t){
							// liczymy ile jest elementow w vt mniejszych od rozwazanego
							int n=0;
							for(size_t x=0; x<vtL; x++){
								if(vt[x]<pTest->m_vtInnerDisIndex[w]){
									n++;
								}
							}
							pTest->m_vtInnerDisIndex[w] -= n;
						}
					}
					
					// usuwanie lisci, ktore sa na pozycjach z vt
					for(size_t w=t; w<node->m_vBranch.size(); w++){
						for(size_t x=0; x<vtL; x++){
							if(vt[x] == w){
								delete node->m_vBranch[w];
								node->m_vBranch[w] = 0;
							}
						}
					}
					// usuniecie z wektora node->m_vBranch elementow ktore sa NULL
					vector<CDTreeNode*>::iterator newend = remove(node->m_vBranch.begin(),node->m_vBranch.end(),(CDTreeNode*)0);
					node->m_vBranch.erase(newend,node->m_vBranch.end());
				}
			}
			// Uaktualnie wartosci dla sprawdzenia warunkow petli
			t++;
			T=node->m_vBranch.size();
		}
	return changed;
}

void CEAIDTInducerAP::ImproveTheBestIndividual(IIndividual *pBest){
	CEAIDTIndividual* peaBest = (CEAIDTIndividual*)pBest;
	if(m_bInnerDisjunction){
		if(MergeLeavesWithRepeatedClasses(peaBest->m_CDTIndRoot)){
			// uaktualnienie rozlazenie przykladow
			ApplyTestChange(m_pDataset,peaBest->m_CDTIndRoot);
			EstimateErrors(peaBest);
			EvaluateIndividualsFitness(pBest);
		}
	}
	if(m_bEnlargeTheMargin){
		EnlargeTheMargin(peaBest->m_CDTIndRoot);
	}
}

void CEAIDTInducerAP::MakeTrainTests(const IDataSet* dataset){
	if(m_pITestContainer){
		delete m_pITestContainer;
	}
	m_pITestContainer = new CTestContainer2(dataset->GetCurrentAttrCount());
	// Tworzenie zbioru testow.	
	for(unsigned i=0; i<dataset->GetCurrentAttrCount(); i++){
		CAttrDesc *pCAD = dataset->GetAttrDesc(i);
		// Test dla atrybutu nominalnego.
		if(pCAD->m_iContinuous == 0){
			CNominalTest* pnomtest = 0;
			if(m_bInnerDisjunction){
				pnomtest = new CNominalTestInnerDisj(pCAD,i);
			}else{
				pnomtest = new CNominalTest(pCAD,i);
			}
			m_pITestContainer->push(i,(ITest*)pnomtest);
		}
		// Test dla atrybutu ciaglego.
		else if(pCAD->m_iContinuous == 1){
			deque<sColAndClass> sCAC = dataset->GetColAndClass(i);
			CTestContainer2::MakeRealTests(sCAC,*m_pITestContainer,i,dataset,m_sThreshold);
		}
	}
}

ITest* CEAIDTInducerAP::FindTest(size_t n, ITest* it){

	// Jak n jest minimalne i jest wewnetrzna dyzjunkcja to sprawa jest banalna
	if(n==2 && m_bInnerDisjunction){
		// Losujemy dowolny test z kontenera
		ITest* test = (*m_pITestContainer)[rand()%m_pITestContainer->size()];
		CNominalTestInnerDisj* pITest = dynamic_cast<CNominalTestInnerDisj*>(test);
		// Jesli jest to test nominalny z wieksza liczba wyjsc to stosujemy losowa wewnetrzna dyzjunkcje.
		if(pITest!=0 && pITest->GetSubSetCount() > n){
			pITest->SetSubSets(n);
		}
		return test->CloneTest();
	}

vector<ITest*> temp;
// Drugi wektor dlatego, ze tu powstaja kopie testow, ktore trzeba usunac przed wyjsciem z metody
// Dla testow z innerDisj trzeba tworzyc kopie, bo inaczej zmodyfikuje sie test oryginalny
vector<ITest*> innerDisjTests;

	// Wstawinie to temp tych testow, ktore daja n wyjsc
	for(size_t iter=0; iter < m_pITestContainer->size(); iter++){
		unsigned nBranches = (*m_pITestContainer)[iter]->GetSubSetCount();
		if( nBranches == n){
			temp.push_back((*m_pITestContainer)[iter]);
		}else{
			// Kiedy jest wewnetrzna dyzjunkcja to test nominalny o wiecej niz n wyjsciach mozna
			// sprowadzic do testu o n wyjsciach.
			if(m_bInnerDisjunction && nBranches > n){
				CNominalTestInnerDisj* pITest = dynamic_cast<CNominalTestInnerDisj*>((*m_pITestContainer)[iter]);
				if(pITest!=0){
					ITest* tCopy = pITest->CloneTest();
					if( ((CNominalTestInnerDisj*)tCopy)->SetSubSets(n)){
						temp.push_back(tCopy);
						innerDisjTests.push_back(tCopy);
					}else{
						delete tCopy;
					}
				}
			}
		}
	}
	
	if(temp.size() < 2){
		for(size_t t=0; t<innerDisjTests.size(); t++){
			delete innerDisjTests[t];
		}
		return 0;
	}

	// Losujemy kilka razy
	for(size_t i=0; i < temp.size(); i++){
		// W temp sa tylko takie testy, ktore daja podzial na n podzbiorow.
		int k = rand()%temp.size();
		// Kiedy to nie jest ten sam test to ok.
		if( !(*temp[k] == *it) ){
			// Najpierw kopia zeby nie skasowac w petli nizej
			ITest* res = temp[k]->CloneTest();
			// Teraz zwolnienie testow z innerDisj z pomocniczego wektora wskaznikow
			for(size_t t=0; t<innerDisjTests.size(); t++){
				delete innerDisjTests[t];
			}
			return res;
		}
	}

	for(size_t t=0; t<innerDisjTests.size(); t++){
		delete innerDisjTests[t];
	}
	return 0;
}

string CEAIDTInducerAP::GetName(){
	return "DT_EA_AP";
}

void CEAIDTInducerAP::Check(deque<CEAIDTIndividual*> &p, const IDataSet* ds){
	for(size_t t=0; t<p.size(); t++){
		CDTreeNode* temp = new CDTreeNode(*(p[t]->m_CDTIndRoot));
		CDTClassifier *pClassifier = new CDTClassifier(temp,m_bCostOfClassification,m_bCostOfAttrs,ds);
		int all=0;
		int ok=0;
		for(unsigned int i=0; i<ds->GetSize(); i++){
			CDataObject *inst = (*ds)[i];
			IAttrValue* pIA=pClassifier->Classify(inst);
			all++;
			IAttrValue *ia=inst->GetClass();
			if((*pIA)==(*ia))ok++;
		}
		double a=p[t]->m_CDTIndRoot->GetErrorOnTestDS();
		a = 100.0 - a;
		double b=(double)ok/(double)all * 100.0;
		if((int)a!=(int)b){
			cerr.setf(ios::showpoint);
			cerr.precision(6);
			cerr<<a<<"\t"<<b<<endl;
		}
		delete pClassifier;
	}
	cout<<"__________________________________________\n";
}

void CEAIDTInducerAP::PostCreateInitialPopulation(){
	IEAIDTInducer::PostCreateInitialPopulation();
	
	if(m_bUseAverageTreeSize){
		EstimateAverageTreeSize();
	}else{
		if(m_bUseMaxTreeSize){
			EstimateMaxTreeSize();
		}
	}
	// To wywolanie dla uczenia z kosztem klasyfikacji po to zeby policzyc
	// koszt klasyfikacji dla kazdego osobnika
	if(m_bCostOfClassification || m_bCostOfAttrs){
		for(size_t t=0; t<m_dqPopulation.size(); t++){
			CEAIDTIndividual* eaindividual = (CEAIDTIndividual*)m_dqPopulation[t];
			EstimateErrors(eaindividual);
		}
	}

	if(m_bMaximizeFitness){
		for(size_t t=0; t<m_dqPopulation.size(); t++){
			CEAIDTIndividual* eaindividual = (CEAIDTIndividual*)m_dqPopulation[t];
			// Maksymalizowanie funkcji fitness poprzez przycinanie drzewa.
			MaximizeFitness(eaindividual);
		}
	}

}

void CEAIDTInducerAP::PostEAOperations(deque<IIndividual*> &p){

	if(m_bMaximizeFitness){
		// Maksymalizowanie funkcji fitness
		for(size_t t=0; t<p.size(); t++){
			if( rand()%100 + 1 <= m_iMaxFitnessProbability){
				CEAIDTIndividual *dtInd=(CEAIDTIndividual*)p[t];
				MaximizeFitness(dtInd);
				// Fitness trzeba przliczyc, bo po wywolaniu tej funkcji
				// fitness nie jest juz uaktualniany
				EvaluateIndividualsFitness(p[t]);
			}
		}
	}

}

void CEAIDTInducerAP::PreCreateInitialPopulation(){
	MakeTrainTests(m_pDataset);
}

void CEAIDTInducerAP::PostTrain(const IDataSet *trainingDataSet){
	if(m_pITestContainer){
		// w destruktarze zostana skasowane wszystkie testy.
		delete m_pITestContainer;
	}
	m_pITestContainer = 0;
}

void CEAIDTInducerAP::CheckParams(){
	IEAIDTInducer::CheckParams();
	if(!m_bInnerDisjunction){
		if(m_viPm[5]!=0 || m_viPm[6]!=0 || m_viPm[7]!=0){
			//m_viPm[5] = 0;
			//m_viPm[6] = 0;
			//m_viPm[7] = 0;
			CLogger::LogEvent("error in pm values - non zero probabilities for innter disjuntion mutation params",__FILE__,__LINE__);
		}
	}
}
