#ifndef __EAIDTINDUCERAPH__
#define __EAIDTINDUCERAPH__

#include "../../mlpcore/MLPDT/IDTInducerAP.h"
#include "../../mlpcore/MLPDT/NominalTestInnerDisj.h"
#include "../../mlpcore/MLPDT_TD/TDIDTInducer.h"
#include "../MLPDT_EA/EAIDTInducer.h"
#include "../MLPDT_EA/EAIDTIndividual.h"
#include <deque>
#include "../../mlpcore/MLPDT/ITestContainer.h"
#include <fstream>

/**
 * Klasa ucznia generujacego drzewo decyzyjne z wykorzystaniem algorytmu
 * ewolucyjnego.
 * @author Marek Grzes'
 */
class CEAIDTInducerAP: public IEAIDTInducer, public IDTInducerAP{
public:
	virtual vector<string> GetParams() const;
	virtual void PrintParamsXML(ostream &ostr, int nsp);
	virtual bool SetParam(string sParamName, string sParamValue);
	virtual string GetName();

	CEAIDTInducerAP(void);
	virtual ~CEAIDTInducerAP(void);

protected:
	
	/**
	 * Zwiekszanie marginesu. Wykonuje sie ono dla najlepszego osobnika tuz przed jego
	 * zwroceniem jako wyniku indukcji.
	 * @see EnlargeTheMargin
	 */
	bool m_bEnlargeTheMargin;

	/**
	 * Parametr wskazujacy jak maja byc tworzone punkty ciecia, przy
	 * tworzeniu testow dla atrybutow ciaglych.
	 * midpoint - ciecie w poowie odleglosci
	 */
	string		m_sThreshold;

	/**
	 * Kolekcja mozliwych testw.
	 */
	ITestContainer *m_pITestContainer;

	/**
	 * Tworzy populacje poczatkowa i zapisuje ja w m_dqPopulation.
	 * @see m_dqPopulation
	 */
	virtual void ActualCreateInitialPopulation();

	virtual void CheckParams();

	/**
	 * Funkcja zwieksza margines testow rzeczywistych w calym drzewie. Sprawdza tez
	 * czy nie ma innego atrybutu, ktory z wiekszym marginesem pozwala osiagnac takie
	 * samo ciecie.
	 */
	virtual void EnlargeTheMargin(CDTreeNode *node);

	/**
	 * Dla podanego w parametrze osobnika oblicza dopasowanie.
	 * @param individual - osobnik, dla ktrego ma byc wyznaczone dopasowanie
	 */
	virtual void EvaluateIndividualsFitness(IIndividual* individual);
	
	virtual bool ExpandLeaf(CDTreeNode* node, unsigned attrCount, double dStop, const IDataSet *dataset);

	virtual bool MutateNode(CDTreeNode* node, const IDataSet* dataset, CEAIDTIndividual &indv);
	
	/**
	 * Operator krzyzowania populacji podanej w parametrze p. Pary stanowi kolejno
	 * wystepujace dwjki elementw. Funkcja estymuje nowe wartosci bedw w drzewie
	 * oraz koryguje wartosc dopasowania. Dlatego po wykonaniu krzyzowania wartosci
	 * te sa aktualne i odpowiadaja wprowadzonym zmianom.
	 */
	virtual void CrossoverOperator();
	
	/**
	 * Tworzy wektor testw i zapisuje go w m_vITests.
	 * @see m_vITests
	 */
	virtual void MakeTrainTests(const IDataSet* dataset);

	/**
	 * Znajduje test na liscie testow, ktory daje podzial na n podzbiorow.
	 * Zwraca wskaznik do znalezionego testu lub NULL, kiedy nie udalo sie
	 * znalesc testu o podanej liczbie podzialow.
	 * Kiedy jest wlaczona wewnetrzna dyzjunkcja to w losowaniu biora udzial rowniez
	 * testy, ktore daja wiecej niz n wyjsc (wczesniej sa losowo sprowadzane do testow 
	 * o n wyjsciach).
	 * @see CNominalTestInnerDisj#SetSubSets
	 * @return kopia testu
	 * @warning Uwaga: zwracana jest kopia testu dlatego zwrocony test trzeba
	 * usunac jak nie jest potrzebny.
	 */
	virtual ITest* FindTest(size_t n, ITest* it);
	
	virtual void Check(deque<CEAIDTIndividual*> &p, const IDataSet* ds);

	/**
	 * Metoda wywolywana przed rozpoczeciem generowania populacji poczatkowej.
	 */
	virtual void PreCreateInitialPopulation();

	/**
	 * Metoda wywolywana po utworzeniu populacji poczatkowej.
	 */
	virtual void PostCreateInitialPopulation();

	virtual void PostEAOperations(deque<IIndividual*> &p);
	virtual void PostTrain(const IDataSet *trainingDataSet);

	virtual vector<size_t> SetMutationOperators(CDTreeNode* pNode);

	virtual void ImproveTheBestIndividual(IIndividual *pBest);	

private:
	/**
	 * Inducer top down dla drzew prostych. Jego metody sa wykorzystywane dlatego warto nie tworzyc go
	 * za kazdym razem.
	 * @see ActualCreateInitialPopulation
	 */
	CTDIDTInducer m_DTIDTInducer;

	/**
	 * Zlacza liscie z ta sama klasa.
	 * @return true - kiedy byla jakas zmiana w drzewie
	 */
	virtual bool MergeLeavesWithRepeatedClasses(CDTreeNode* node);
};

#endif

