/**
 * $Revision: 1.7 $
 * $Date: 2006/04/03 22:18:56 $
 */

#ifndef __EAIDTINDUCERAPOBH__
#define __EAIDTINDUCERAPOBH__

#include "../MLPDT_EA_AP/EAIDTInducerAP.h"
#include "../MLPDT_EA_OB/EAIDTInducerOB.h"

/**
 * Klasa ucznia generujacego drzewo decyzyjne z wykorzystaniem algorytmu
 * ewolucyjnego. Generowane drzewo moze miec testy proste jak i skosne.
 * @author Marek Grzes'
 */
class CEAIDTInducerAPOB: public IEAIDTInducer, public IDTInducerAP{
public:
	CEAIDTInducerAPOB();
	void PrintParamsXML(ostream &ostr, int nsp);

protected:

	/**
	 * Zwiekszanie marginesu. Wykonuje sie ono dla najlepszego osobnika tuz przed jego
	 * zwroceniem jako wyniku indukcji.
	 * @see EnlargeTheMargin
	 */
	bool m_bEnlargeTheMargin;

	/**
	 * Parametr wykorzystywany przy liczeniu rozmiaru drzewa.
	 */
	double	m_dFitnessBeta2;

	bool m_bCentreHiperplane;
	bool m_bDipolOperator;

	/**
	 * Kolekcja mozliwych testw.
	 */
	ITestContainer *m_pITestContainer;

	/**
	 * Parametr wskazujacy jak maja byc tworzone punkty ciecia, przy
	 * tworzeniu testow dla atrybutow ciaglych.
	 * midpoint - ciecie w poowie odleglosci
	 */
	string		m_sThreshold;

	/**
	 * prevention underfitting factor
	 */
	unsigned	m_uPUF;

	CEAIDTInducerOB m_inducerOB;

	void ActualCreateInitialPopulation();
	void CrossoverOperator();
	void EnlargeTheMargin(CDTreeNode *node);
	void EvaluateIndividualsFitness(IIndividual* individual);
	bool ExpandLeaf(CDTreeNode* node, unsigned attrCount, double dStop, const IDataSet *dataset);

	/**
	 * Znajduje test na liscie testow, ktory daje podzial na n podzbiorow.
	 * Zwraca wskaznik do znalezionego testu lub NULL, kiedy nie udalo sie
	 * znalesc testu o podanej liczbie podzialow.
	 * Kiedy jest wlaczona wewnetrzna dyzjunkcja to w losowaniu biora udzial rowniez
	 * testy, ktore daja wiecej niz n wyjsc (wczesniej sa losowo sprowadzane do testow 
	 * o n wyjsciach).
	 * @see CNominalTestInnerDisj#SetSubSets
	 * @return kopia testu
	 * @warning Uwaga: zwracana jest kopia testu dlatego zwrocony test trzeba
	 * usunac jak nie jest potrzebny.
	 */
	virtual ITest* FindTest(size_t n, ITest* it);

	string GetName(){ return "DT_EA_APOB"; }
	vector<string> GetParams() const;
	void ImproveTheBestIndividual(IIndividual *pBest);
	void MakeTrainTests(const IDataSet* dataset);
	bool MergeLeavesWithRepeatedClasses(CDTreeNode* node);
	bool MutateNode(CDTreeNode* node, const IDataSet* dataset, CEAIDTIndividual &indv);
	void PostCreateInitialPopulation();
	void PostEAOperations(deque<IIndividual*> &p);
	void PostTrain(const IDataSet *trainingDataSet);
	void PreCreateInitialPopulation();
	vector<size_t> SetMutationOperators(CDTreeNode* pNode);
	bool SetParam(string sParamName, string sParamValue);

private:
	/**
	 * Inducer top down dla drzew prostych. Jego metody sa wykorzystywane dlatego warto nie tworzyc go
	 * za kazdym razem.
	 * @see ActualCreateInitialPopulation
	 */
	CTDIDTInducer m_DTIDTInducer;

};

#endif
