/**
 * $Revision: 1.92 $
 * $Date: 2006/04/04 23:15:31 $
 */

#include "EAIDTInducerOB.h"
#include "../MLPDT_TD_OB/TDIDTInducerOB.h"
#include "../MLPDT_TD_OB/ObliqueTestCreator.h"
#include "../../mlpcore/MLPDT/DTClassifier.h"
#include "../../mlpcore/MLPCommon/MLPStatistics.h"
#include "../../mlpcore/MLPDataset/NomAttr.h"
#include "../../mlpcore/MLPDataset/RealAttr.h"
#include <algorithm>
#include <sstream>
#include <iomanip>

CEAIDTInducerOB::CEAIDTInducerOB(void){
	//m_ofsLogFile.open("MLP_EATreeInducerOB.log",ofstream::out | ofstream::app);
	m_bCentreHiperplane = false;
	m_dFitnessAlfa = 0.05;
	m_dFitnessBeta = 0.2;
	m_bFeatureSelection = false;
	m_uPUF = 5;
	m_bDipolOperator = true;
	
	// Ustawienie domyslnych wartosci i inicjalizacja rozmiaru wektorow. Kazda klasa
	// potomna klasy EAIDTInducer musi ustawiac te parametry we wlasnym konstruktorze.
	vector<int> temppc(5,20);
	vector<int> temppm(8,12);
	vector<int> temppml(2,50);
	temppm[7]=13;
	temppm[6]=13;
	temppm[5]=13;
	temppm[4]=13;	
	//temppc[2]=34;
	m_viPc = temppc;
	m_viPm = temppm;
	m_viPmL = temppml;
	
	// Kazda implementacja klasy EAIDTInducer musi miec przypisanie nazw operatow w swoim konstrukorze
#ifdef MLPEAOPERATORSTEST
	// Operatory krzyzowania
	m_vOpNames.push_back("[branches swapped in random order]");
	m_vOpNames.push_back("[subtrees swapped]");
	m_vOpNames.push_back("[hiperplane crossed]");
	m_vOpNames.push_back("[Dipolar crossover - subtrees]");
	m_vOpNames.push_back("[Dipolar crossover - tests]");
	
	// Operatory mutacji
	m_vOpNames.push_back("[random zero]");
	m_vOpNames.push_back("[Single factor mutated]");
	m_vOpNames.push_back("[tests between father and son exchanged]");
	m_vOpNames.push_back("[symmetric mutation]");
	m_vOpNames.push_back("[Test in node changed]");
	m_vOpNames.push_back("[Mixed dipol cut]");
	m_vOpNames.push_back("[Pure dipol joined]");
	m_vOpNames.push_back("[Node changed into leaf]");
	m_vOpNames.push_back("[Leaf changed into node]");
	m_vOpNames.push_back("[Leaf expanded into subtree]");
	pair<string,pair<int,int> > p;
	pair<int,int> zero(0,0);
	p.second = zero;
	for(size_t t=0; t<m_vOpNames.size(); t++){
		p.first = m_vOpNames[t];
		m_mOpEfficiency.insert(p);
	}
	// Kopia mapy dla najlepszego osobnika.
	m_mOpEfficiencyBestIndv = m_mOpEfficiency;
#endif
}

CEAIDTInducerOB::~CEAIDTInducerOB(void){
	//m_ofsLogFile.close();
}

bool CEAIDTInducerOB::ExpandLeaf(CDTreeNode* node, unsigned attrCount, double dStop, const IDataSet *dataset){
static CObliqueTestCreator obliqueTestCreator;

	if(!node->m_bLeaf){
		return false;
	}
	
	// Random stop, dla procesu rozwijania drzewa.
	if( CMLPStatistics::Event((int)(100.0*dStop)) ){
		return false;
	}
	
	
	ITest* test=NULL;
	if(m_bDipolOperator){
		test = obliqueTestCreator.CreateObliqueTest(node->m_deqCI,m_pDataset->GetCurrentRealAttrCount(),m_uPUF);
	}else{
		test = obliqueTestCreator.CreateRandObliqueTest(node->m_deqCI,m_pDataset->GetCurrentRealAttrCount(),m_uPUF);
	}
	if(test){
		// Utworz tablice podzialow
		vector<vector<unsigned> > vvDiv = test->GetSubSets(node->m_deqCI);

		node->m_bLeaf = false;
		//memcheck node->m_pITest = test->CloneTest();
		node->m_pITest = test;
		// Dla kazego podzbioru wynikajacego z podzialu: utworzenie potomka
		// bedacego lisciem, a nastepnie proba jego rozwiniecia.
		for(unsigned i=0; i<vvDiv.size(); i++){
			deque<CDataObject*> subset;
			vector<unsigned> subIDs = vvDiv[i];
			for(unsigned j=0; j<subIDs.size(); j++){
				subset.insert(subset.end(),node->m_deqCI[subIDs[j]]);
			}
			CDTreeNode* newnode = MakeTreeLeaf(dataset,subset,node->m_pIAClass);
			node->m_vBranch.push_back(newnode);
			ExpandLeaf(newnode,attrCount,dStop,dataset);
		}
		return true;
	}else{
		return false;
	}
}

vector<size_t> CEAIDTInducerOB::SetMutationOperators(CDTreeNode* pNode){
vector<size_t> res(m_viPm.size());
	
	/*
	for(size_t t=0; t<res.size(); t++){
		res[t] = t;
	}
	*/
	
	size_t i=0;
	
	// (0) Losowe zero
	res[i++] = 0;
	// (1) Mutacja jednego wsp.
	res[i++] = 1;
	for(size_t t=0; t<pNode->m_vBranch.size(); t++){
		if(!pNode->m_vBranch[t]->m_bLeaf){
			// (2) zamiana testu na test z syna
			res[i++] = 2;
			// (3) mutowanie symetryczne
			res[i++] = 3;
			break;
		}
	}
	// (4) Zamiana testu na inny test
	res[i++] = 4;
	// (5) Przyciecie dipola mieszanego
	res[i++] = 5;
	// (6) Usuniecie przeciecia dipola czystego
	res[i++] = 6;
	// (7) Przyciecie, wezel staje sie lisciem
	res[i++] = 7;
	res.resize(i);
	return res;
}

bool CEAIDTInducerOB::MutateNode(CDTreeNode* node, const IDataSet* dataset, CEAIDTIndividual &indv){
static CObliqueTestCreator obliqueTestCreator;
bool mutated = false;
		int _op_ = GetMutationOperatorID(node);
		switch(_op_){
		// *********************** operacje na wezlach wewn. ***************************
			case 0:{
				// ***************** Losowe zero w jednym wsp. hip. ********************
				assert(node->m_pITest);
				assert(node->m_deqCI.size() > 0);
				// ************* Wyzerowanie losowego wsp. hiperplaszczyzny ************
				obliqueTestCreator.SetRandomZero(node->m_pITest);
				if(m_iLogLevel>=2) CLogger::LogEvent("M1 Random zero","",__LINE__,m_ofsLogFile);
				mutated = true;
				ApplyTestChange(dataset,node);
#ifdef MLPEAOPERATORSTEST
				OPTEST_OperatorsTestLog(indv,"[random zero]");
#endif
				break;
			}//0
			case 1:{//1
				assert(node->m_pITest);
				assert(node->m_deqCI.size() > 0);
				// ************* Mutacja jednego losowo wybranego wspolczynnika *****
				obliqueTestCreator.MutateTest(node->m_pITest);
				if(m_iLogLevel>=2) CLogger::LogEvent("M2 Single factor mutated","",__LINE__,m_ofsLogFile);
				mutated = true;
				ApplyTestChange(dataset,node);
#ifdef MLPEAOPERATORSTEST
				OPTEST_OperatorsTestLog(indv,"[Single factor mutated]");
#endif
				break;
			}//1
			case 2:{//2
				assert(node->m_pITest);
				assert(node->m_deqCI.size() > 0);
				// ************* Zamiana testu w danym wezle na test z jego losowo wybranego syna *****
				vector<size_t> nNodes(node->m_vBranch.size());
				size_t nN=0;
				// zapamietujenow synow ktorzy nie sa liscmi
				for(size_t i=0; i<node->m_vBranch.size(); i++){
					if(!node->m_vBranch[i]->m_bLeaf){
						nNodes[nN++]=i;
					}
				}
				// losujemy syna (nie bedacego lisciem)
				size_t son = nNodes[rand()%nN];
				CDTreeNode* node2 = node->m_vBranch[son];
				if(!node2->m_bLeaf){
					assert(node->m_vBranch.size()==node2->m_vBranch.size());
						ITest* test = node->m_pITest;
						node->m_pITest = node2->m_pITest;
						node2->m_pITest = test;
					if(m_iLogLevel>=2) CLogger::LogEvent("M3 tests between father and son exchanged","",__LINE__,m_ofsLogFile);
					mutated = true;
					ApplyTestChange(dataset,node);
#ifdef MLPEAOPERATORSTEST
					OPTEST_OperatorsTestLog(indv,"[tests between father and son exchanged]");
#endif
				}else{
					if(m_iLogLevel>=4) CLogger::LogEvent("NOT M3 tests between father and son exchanged","",__LINE__,m_ofsLogFile);
				}
				break;
			}//2
			case 3:{//3
				assert(node->m_pITest);
				assert(node->m_deqCI.size() > 0);
				// ************* Mutowanie symetryczne **************
				vector<size_t> nNodes(node->m_vBranch.size());
				size_t nN=0;
				// zapamietujenow synow ktorzy nie sa liscmi
				for(size_t i=0; i<node->m_vBranch.size(); i++){
					if(!node->m_vBranch[i]->m_bLeaf){
						nNodes[nN++]=i;
					}
				}
				// Wylosowanie poddrzewa, ktore zostanie powielone i zastapi pozostale.
				// losujemy tylko sposrod wezlow wewnetrznych
				size_t sbtree = nNodes[rand()%nN];
				CDTreeNode* node2 = node->m_vBranch[sbtree];
				if(!node2->m_bLeaf){
					for(size_t t=0; t<node->m_vBranch.size(); t++){
						if(t!=sbtree){
							delete node->m_vBranch[t];
							node->m_vBranch[t] = new CDTreeNode(*node2);
						}
					}
					if(m_iLogLevel>=2) CLogger::LogEvent("M4 symmetric mutation","",__LINE__,m_ofsLogFile);
					mutated = true;
					ApplyTestChange(dataset,node);
#ifdef MLPEAOPERATORSTEST
					OPTEST_OperatorsTestLog(indv,"[symmetric mutation]");
#endif
				}else{
					if(m_iLogLevel>=4) CLogger::LogEvent("NOT M4 symmetric mutation","",__LINE__,m_ofsLogFile); 
				}
				break;
			}//3
			case 4:{//4
				assert(node->m_pITest);
				assert(node->m_deqCI.size() > 0);
				// *********** Zamiana testu na inny test ***************
				ITest* test=NULL;
				if(m_bDipolOperator){
					test = obliqueTestCreator.CreateObliqueTest(node->m_deqCI,dataset->GetCurrentRealAttrCount(),m_uPUF);
				}else{
					test = obliqueTestCreator.CreateRandObliqueTest(node->m_deqCI,dataset->GetCurrentRealAttrCount(),m_uPUF);
				}
				if(test){
					mutated = true;
					// Usuniecie istniejacego testu.
					delete node->m_pITest;
					//memcheck node->m_pITest = test->CloneTest();
					node->m_pITest = test;
					// Korekcja wartosci m_deqCI we wszystkich wezlach
					// poddrzewa o korzeniu w wezle node. Korekcja
					// klas wiekszosciowych.
					ApplyTestChange(dataset,node);
					if(m_iLogLevel>=2) CLogger::LogEvent("M5 Test in node changed","",__LINE__,m_ofsLogFile);
#ifdef MLPEAOPERATORSTEST
					OPTEST_OperatorsTestLog(indv,"[Test in node changed]");
#endif
				}else{
					if(m_iLogLevel>=4) CLogger::LogEvent("NOT M5 Test in node changed","",__LINE__,m_ofsLogFile);
				}
				break;
			}//4
			case 5:{//5
				assert(node->m_pITest);
				assert(node->m_deqCI.size() > 0);
				// ***** Przeciecie dipola mieszanego *****
				CDataObject* ob1=NULL;
				CDataObject* ob2=NULL;
				if(obliqueTestCreator.FindMixedDipol(node->m_deqCI,ob1,ob2,node->m_pITest)){
					assert(ob1);
					assert(ob2);
					if(obliqueTestCreator.CutMixedDipol(ob1,ob2,node->m_pITest)){
						mutated = true;
						ApplyTestChange(dataset,node);
						if(m_iLogLevel>=2) CLogger::LogEvent("M6 Mixed dipol cut","",__LINE__,m_ofsLogFile);
#ifdef MLPEAOPERATORSTEST
						OPTEST_OperatorsTestLog(indv,"[Mixed dipol cut]");
#endif
					}else{
						if(m_iLogLevel>=4) CLogger::LogEvent("NOT M6 Mixed dipol cut","",__LINE__,m_ofsLogFile);
					}
				}else{
					if(m_iLogLevel>=4) CLogger::LogEvent("NOT M6 Mixed dipol cut","",__LINE__,m_ofsLogFile);
				}
				break;
			}//5
			case 6:{//6
				assert(node->m_pITest);
				assert(node->m_deqCI.size() > 0);
				CDataObject* ob1=NULL;
				CDataObject* ob2=NULL;
				// ***** Usuniecie przeciecia dipola czystego *****
				if(obliqueTestCreator.FindPureDipol(node->m_deqCI,ob1,ob2,node->m_pITest)){
					assert(ob1);
					assert(ob2);
					if(obliqueTestCreator.JoinPureDipol(ob1,ob2,node->m_pITest)){
						mutated = true;
						ApplyTestChange(dataset,node);
						if(m_iLogLevel>=2) CLogger::LogEvent("M7 Pure dipol joined","",__LINE__,m_ofsLogFile);
#ifdef MLPEAOPERATORSTEST
						OPTEST_OperatorsTestLog(indv,"[Pure dipol joined]");
#endif
					}else{
						if(m_iLogLevel>=4) CLogger::LogEvent("NOT M7 Pure dipol joined","",__LINE__,m_ofsLogFile);
					}
				}else{
					if(m_iLogLevel>=4) CLogger::LogEvent("NOT M7 Pure dipol joined","",__LINE__,m_ofsLogFile);
				}
				break;
			}//6
			case 7:{//7
				// ************ Przyciecie - zamiana wezla na lisc. *************
				mutated = true;
				node->m_bLeaf = true;
				while(!node->m_vBranch.empty()){
					delete node->m_vBranch.back();
					node->m_vBranch.pop_back();
				}
				node->m_vBranch.clear();
				delete node->m_pITest;
				node->m_pITest = NULL;
				if(m_iLogLevel>=2) CLogger::LogEvent("M8 Node changed into leaf","",__LINE__,m_ofsLogFile);
#ifdef MLPEAOPERATORSTEST
				OPTEST_OperatorsTestLog(indv,"[Node changed into leaf]");
#endif
				break;
			}//7
			// ******************************** operacje na lisciu *************************
			case 8:{
				// *************************** Dodanie testu w lisciu. **************************
				ITest* test=NULL;
				if(m_bDipolOperator){
					test = obliqueTestCreator.CreateObliqueTest(node->m_deqCI,dataset->GetCurrentRealAttrCount(),m_uPUF);
				}else{
					test = obliqueTestCreator.CreateRandObliqueTest(node->m_deqCI,dataset->GetCurrentRealAttrCount(),m_uPUF);
				}
				if(test){
					// Utworz tablice podzialow
					vector<vector<unsigned> > vvDiv = test->GetSubSets(node->m_deqCI);
					mutated = true;
					node->m_bLeaf = false;
					// Dla kazego podzbioru wynikajacego z podzialu: utworzenie potomka
					// bedacego lisciem.
					for(unsigned i=0; i<vvDiv.size(); i++){
						deque<CDataObject*> subset;
						vector<unsigned> subIDs = vvDiv[i];
						for(unsigned j=0; j<subIDs.size(); j++){
							subset.insert(subset.end(),node->m_deqCI[subIDs[j]]);
						}
						CDTreeNode* newnode = MakeTreeLeaf(dataset,subset,node->m_pIAClass);
						node->m_vBranch.push_back(newnode);
					}
					//memcheck node->m_pITest = test->CloneTest();
					node->m_pITest = test;
					if(m_iLogLevel>=2) CLogger::LogEvent("M9 Leaf changed into node","",__LINE__,m_ofsLogFile);
#ifdef MLPEAOPERATORSTEST
					OPTEST_OperatorsTestLog(indv,"[Leaf changed into node]");
#endif
				}else{
					if(m_iLogLevel>=4) CLogger::LogEvent("NOT M9 Leaf changed into node","",__LINE__,m_ofsLogFile);
				}
				break;
			}
			case 9:{
				// *************************** Rozwiniecie liscia w poddrzewo. **************************
				if(m_bMaximizeFitness){
					mutated = ExpandLeaf(node,dataset->GetCurrentAttrCount(),0,dataset);
				}else{
					mutated = ExpandLeaf(node,dataset->GetCurrentAttrCount(),0.3,dataset);
				}
				if(m_bMaximizeFitness){
					// Nie jest to maksymalizacja dokladna, bo tu nie uwzgledniamy kosztu
					// cech, ktore sa wyzej na sciezce (ale nie mamy tu do nich dostepu).
					vector<bool> vb(m_pDataset->GetCurrentAttrCount(),false);
					MaximizeFitness(node,vb);
				}
				if(mutated){
					if(m_iLogLevel>=2) CLogger::LogEvent("M10 Leaf expanded into subtree","",__LINE__,m_ofsLogFile);
#ifdef MLPEAOPERATORSTEST
					OPTEST_OperatorsTestLog(indv,"[Leaf expanded into subtree]");
#endif
				}else{
					if(m_iLogLevel>=2) CLogger::LogEvent("NOT M10 Leaf expanded into subtree","",__LINE__,m_ofsLogFile);
				}
				// tego poddrzewa nie mutujemy, od razu powrot.
				return mutated;
			}
		}//switch

	return mutated;
}

void CEAIDTInducerOB::CrossoverOperator(){
	static CObliqueTestCreator obliqueTestCreator;
	size_t i;
	size_t j;
	size_t k;
	vector<unsigned> used;
	vector<size_t> vindv;
	
	for(size_t t=0; t<m_dqPopulation.size(); t++){
		if( rand()%100 + 1 <= m_iProbabilityOfCrossOver ){
			vindv.push_back(t);
		}
	}
	while(vindv.size() > 1){
		size_t ind1 = rand()%vindv.size();
		size_t ind2;
		do{
			 ind2 = rand()%vindv.size();
		}while(ind1==ind2);
		i = vindv[ind1];
		j = vindv[ind2];
		k=0;
		for(vector<size_t>::iterator it=vindv.begin(); it!=vindv.end(); it++){
			if(k==ind1){
				vindv.erase(it);
				break;
			}
			k++;
		}
		k=0;
		for(vector<size_t>::iterator it=vindv.begin(); it!=vindv.end(); it++){
			if(k==ind2){
				vindv.erase(it);
				break;
			}
			k++;
		}
// _____________ W tym miejscu mamy indeksy osobnikow do krzyzowania: i,j ______________
// _____________________________________________________________________________________
		int m = GetCrossoverOperatorID();
		switch(m){//009
			// ********** Zamiana galezi w kolejnosci losowej *********
			case 0:{
				//CDTreeNode* node1 = p[i]->m_CDTIndRoot->FindRandomNode();
				CDTreeNode* node1 = EAIND(m_dqPopulation[i])->m_CDTIndRoot->GetNodeEvenProb();
				if(node1==NULL || node1->m_deqCI.empty()){
					continue;
				}
				//CDTreeNode* node2 = p[j]->m_CDTIndRoot->FindRandomNode();
				CDTreeNode* node2 = EAIND(m_dqPopulation[j])->m_CDTIndRoot->GetNodeEvenProb();
				if(node2==NULL || node2->m_deqCI.empty()){
					continue;
				}
				used.clear();
				for(unsigned i1=0; i1<node1->m_vBranch.size(); i1++){
					if(used.size()>=node1->m_vBranch.size() || used.size()>=node2->m_vBranch.size()){
						break;
					}
					unsigned i2 = (unsigned)CMLPStatistics::RandVal((int)(node2->m_vBranch.size()));
					if(find(used.begin(),used.end(),i2) != used.end()){
						i2 = 0;
						while(find(used.begin(),used.end(),i2) != used.end()){
							i2++;
						}
					}
					if(i2>=node2->m_vBranch.size()){
						CLogger::LogError("Index error",__FILE__,__LINE__);
						continue;
					}
					CDTreeNode * temp = node1->m_vBranch[i1];
					node1->m_vBranch[i1] = node2->m_vBranch[i2];
					node2->m_vBranch[i2] = temp;
					used.push_back(i2);
				}
				ApplyTestChange(m_pDataset,node1);
				ApplyTestChange(m_pDataset,node2);
				if(m_iLogLevel>=2) CLogger::LogEvent("C1 Branches swapped in random order","",__LINE__,m_ofsLogFile);
#ifdef MLPEAOPERATORSTEST
				OPTEST_OperatorsTestLog(*(EAIND(m_dqPopulation[i])),"[branches swapped in random order]");
				OPTEST_OperatorsTestLog(*(EAIND(m_dqPopulation[j])),"[branches swapped in random order]");
#endif
				break;
			}
			// *************************************************************************

			// ********** Zamiana calych poddrzew ********************
			case 1:{
				//CDTreeNode* node1 = p[i]->m_CDTIndRoot->FindRandomNodeOrLeaf();
				CDTreeNode* node1 = EAIND(m_dqPopulation[i])->m_CDTIndRoot->GetNodeOrLeafEvenProb();
				if(node1==NULL || node1->m_deqCI.empty()){
					continue;
				}
				//CDTreeNode* node2 = p[j]->m_CDTIndRoot->FindRandomNodeOrLeaf();
				CDTreeNode* node2 = EAIND(m_dqPopulation[j])->m_CDTIndRoot->GetNodeOrLeafEvenProb();
				if(node2==NULL || node2->m_deqCI.empty()) {
					continue;
				}
				bool changed = false;
				if(!node1->m_bLeaf && !node2->m_bLeaf){
					ITest* pITest;
					vector<CDTreeNode*> branch;
					pITest=node1->m_pITest;
					branch=node1->m_vBranch;
					node1->m_pITest=node2->m_pITest;
					node1->m_vBranch=node2->m_vBranch;
					node2->m_pITest=pITest;
					node2->m_vBranch=branch;
					changed = true;
				}else{
					if(node1->m_bLeaf && !node2->m_bLeaf){
						node1->m_pITest=node2->m_pITest;
						node1->m_vBranch=node2->m_vBranch;
						node1->m_bLeaf=false;
						node2->m_pITest=NULL;
						node2->m_vBranch.clear();
						node2->m_bLeaf=true;
						changed = true;
					}else{
						if(!node1->m_bLeaf && node2->m_bLeaf){
							node2->m_pITest=node1->m_pITest;
							node2->m_vBranch=node1->m_vBranch;
							node2->m_bLeaf=false;
							node1->m_pITest=NULL;
							node1->m_vBranch.clear();
							node1->m_bLeaf=true;
							changed = true;
						}
					}
				}
				if(changed){
					ApplyTestChange(m_pDataset,node1);
					ApplyTestChange(m_pDataset,node2);
					if(m_iLogLevel>=2) CLogger::LogEvent("C2 Subtrees swapped","",__LINE__,m_ofsLogFile);
#ifdef MLPEAOPERATORSTEST
					OPTEST_OperatorsTestLog(*(EAIND(m_dqPopulation[i])),"[subtrees swapped]");
					OPTEST_OperatorsTestLog(*(EAIND(m_dqPopulation[j])),"[subtrees swapped]");
#endif
				}
				break;
			}
			// *************************************************************************

			// ********** Krzyzowanie hiperplaszczyzn z dwoch testow *******************
			case 2:{
				//CDTreeNode* node1 = p[i]->m_CDTIndRoot->FindRandomNode();
				CDTreeNode* node1 = EAIND(m_dqPopulation[i])->m_CDTIndRoot->GetNodeEvenProb();
				if(node1==NULL || node1->m_deqCI.empty()){
					continue;
				}
				//CDTreeNode* node2 = p[j]->m_CDTIndRoot->FindRandomNode((unsigned)(node1->m_vBranch.size()));
				CDTreeNode* node2 = EAIND(m_dqPopulation[j])->m_CDTIndRoot->GetNodeEvenProb((unsigned)(node1->m_vBranch.size()));
				if(node2==NULL || node2->m_deqCI.empty()) {
					continue;
				}
				assert(node1->m_vBranch.size()==node2->m_vBranch.size());
				/*ITest* pITest;
				pITest=node1->m_pITest;
				node1->m_pITest=node2->m_pITest;
				node2->m_pITest=pITest;*/
				if(obliqueTestCreator.CrossoverHiperplanes(node1->m_pITest,node2->m_pITest)){
					ApplyTestChange(m_pDataset,node1);
					ApplyTestChange(m_pDataset,node2);
					if(m_iLogLevel>=2) CLogger::LogEvent("C3 Hiperplane crossed","",__LINE__,m_ofsLogFile);
#ifdef MLPEAOPERATORSTEST
					OPTEST_OperatorsTestLog(*(EAIND(m_dqPopulation[i])),"[hiperplane crossed]");
					OPTEST_OperatorsTestLog(*(EAIND(m_dqPopulation[j])),"[hiperplane crossed]");
#endif
				}
				break;
			}
			// ******************* Krzyzowanie dipolowe - poddrzewa *********************
			case 3:{
				// 1. wybranie pierwszego wezla.
				//CDTreeNode* node1 = p[i]->m_CDTIndRoot->FindRandomNode();
				CDTreeNode* node1 = EAIND(m_dqPopulation[i])->m_CDTIndRoot->GetNodeEvenProb();
				if(node1==NULL){
					continue;
				}
				// 2. wyszukanie przecietego dipola mieszanego.
				CDataObject* ob1=NULL;
				CDataObject* ob2=NULL;
				if(obliqueTestCreator.FindMixedDipolCut(node1->m_deqCI,ob1,ob2,node1->m_pITest)){
					assert(ob1);
					assert(ob2);
					pair<CDataObject*, CDataObject*> pr(ob1,ob2);
					// 3. wyszukanie w drugim drzewie wezla przecinajacego badany dipol
					CDTreeNode* node2 = EAIND(m_dqPopulation[j])->m_CDTIndRoot->FindNodeCuttingDipol(pr);
					if(node2==NULL){
						continue;
					}
					// 4. zamiana poddrzew 
					ITest* pITest;
					vector<CDTreeNode*> branch;
					pITest=node1->m_pITest;
					branch=node1->m_vBranch;
					node1->m_pITest=node2->m_pITest;
					node1->m_vBranch=node2->m_vBranch;
					node2->m_pITest=pITest;
					node2->m_vBranch=branch;
					ApplyTestChange(m_pDataset,node1);
					ApplyTestChange(m_pDataset,node2);
					if(m_iLogLevel>=2) CLogger::LogEvent("C4 Dipolar crossover - subtrees","",__LINE__,m_ofsLogFile);
#ifdef MLPEAOPERATORSTEST
					OPTEST_OperatorsTestLog(*(m_dqPopulation[i]),"[Dipolar crossover - subtrees]");
					OPTEST_OperatorsTestLog(*(m_dqPopulation[j]),"[Dipolar crossover - subtrees]");
#endif
				}
				break;
			}
			// ******************* Krzyzowanie dipolowe - testy *********************
			case 4:{
				// 1. wybranie pierwszego wezla.
				//CDTreeNode* node1 = p[i]->m_CDTIndRoot->FindRandomNode();
				CDTreeNode* node1 = EAIND(m_dqPopulation[i])->m_CDTIndRoot->GetNodeEvenProb();
				if(node1==NULL){
					continue;
				}
				// 2. wyszukanie przecietego dipola mieszanego.
				CDataObject* ob1=NULL;
				CDataObject* ob2=NULL;
				if(obliqueTestCreator.FindMixedDipolCut(node1->m_deqCI,ob1,ob2,node1->m_pITest)){
					assert(ob1);
					assert(ob2);
					pair<CDataObject*, CDataObject*> pr(ob1,ob2);
					// 3. wyszukanie w drugim drzewie wezla przecinajacego badany dipol
					CDTreeNode* node2 = EAIND(m_dqPopulation[j])->m_CDTIndRoot->FindNodeCuttingDipol(pr);
					if(node2==NULL){
						continue;
					}
					// 4. zamiana testow
					ITest* pITest=node1->m_pITest;
					node1->m_pITest=node2->m_pITest;
					node2->m_pITest=pITest;
					ApplyTestChange(m_pDataset,node1);
					ApplyTestChange(m_pDataset,node2);
					if(m_iLogLevel>=2) CLogger::LogEvent("C5 Dipolar crossover - tests","",__LINE__,m_ofsLogFile);
#ifdef MLPEAOPERATORSTEST
					OPTEST_OperatorsTestLog(*(m_dqPopulation[i]),"[Dipolar crossover - tests]");
					OPTEST_OperatorsTestLog(*(m_dqPopulation[j]),"[Dipolar crossover - tests]");
#endif
				}
				break;
			}
			// *************************************************************************
		}//009
//______________________________________________________________________________________
		EstimateErrors(m_dqPopulation[i]);
		EstimateErrors(m_dqPopulation[j]);
		EvaluateIndividualsFitness(EAIND(m_dqPopulation[i]));
		EvaluateIndividualsFitness(EAIND(m_dqPopulation[j]));
	}
}

void CEAIDTInducerOB::ActualCreateInitialPopulation(){
stringstream sp;

	if(!m_pDataset->OnlyRealAttributes()){
		stringstream ss;
		ss<<"Dataset "<<m_pDataset->m_sDataSetName<<" with ";
		ss<<m_pDataset->GetCurrentAttrCount() - m_pDataset->GetCurrentRealAttrCount();
		ss<<"/"<<m_pDataset->GetCurrentAttrCount();
		ss<<" nominal attributes in CEAIDTInducerOB";
		CLogger::LogEvent(ss.str(),__FILE__,__LINE__);
	}

	/*
	 * choose initial population
	 */  
	IInducer *inducer = new CTDIDTInducerOB();
	sp<<m_iInitMaxDepth;
	if(!inducer->SetParam("maxdepth",sp.str())){
		CLogger::LogError("SetParam - maxdepth",__FILE__,__LINE__);
	}
	if(!m_bDipolOperator){
		if(!inducer->SetParam("split","random")){
			CLogger::LogError("SetParam - split",__FILE__,__LINE__);
		}
	}
	if(!inducer->SetParam("pruning","no")){
		CLogger::LogError("SetParam - pruning",__FILE__,__LINE__);
	}
	if(m_bMaximizeFitness){
		if(!inducer->SetParam("random stop","0")){
			CLogger::LogError("SetParam - random stop",__FILE__,__LINE__);
		}
	}else{
		if(!inducer->SetParam("random stop","0.3")){
			CLogger::LogError("SetParam - random stop",__FILE__,__LINE__);
		}
	}
	if(!inducer->SetParam("minimum items","5")){
		CLogger::LogError("SetParam - minimum items",__FILE__,__LINE__);
	}
	if(!inducer->SetParam("clear coverd","no")){
		CLogger::LogError("SetParam - clear coverd",__FILE__,__LINE__);
	}
	stringstream ss;
	ss<<m_uPUF;
	if(!inducer->SetParam("prevention underfitting factor",ss.str())){
		CLogger::LogError("SetParam - prevention underfitting factor",__FILE__,__LINE__);
	}
	for(unsigned i=0; i<m_uSizeOfPop; i++){
		CDTClassifier *DTClassifier = (CDTClassifier*)inducer->Train(m_pDataset);
		CEAIDTIndividual* temp = new CEAIDTIndividual();
		temp->m_CDTIndRoot = new CDTreeNode( *(DTClassifier->GetRoot()) );
		//DTClassifier->DisplayStruct(m_ofsLogFile,0);
		delete DTClassifier;
		m_dqPopulation.push_back(temp);
		//cerr<<"Individual :"<<i<<" created ("<<temp->m_CDTIndRoot->GetNodeCount()<<")"<<endl;
	}
	delete inducer;
}

/**
 * Przed wywolaniem tej funkcji dla danego osobnika musi byc najpierw
 * wywolana fukcja CDTreeNode::EstimateErrors, ktora wyliczy liczbe
 * zle pokrywanych przykladow w zbiorze uczacym.
 */
void CEAIDTInducerOB::EvaluateIndividualsFitness(IIndividual* individual){
double a,b,c,tSize;
CEAIDTIndividual* eaindividual = (CEAIDTIndividual*)individual;

	if(m_iError == 0){
		a = (100.0 - eaindividual->m_CDTIndRoot->GetEstimateError()) / 100.0;
	}else{
		a = (100.0 - eaindividual->m_CDTIndRoot->GetErrorOnTestDS()) / 100.0;
	}
	if(m_bFeatureSelection){
		tSize = eaindividual->GetNodeCount_FS(m_dFitnessBeta);
	}else{
		tSize = (double)(eaindividual->m_CDTIndRoot->GetNodeCount());
	}
	assert(!m_bUseAverageTreeSize || !m_bUseMaxTreeSize);
	if(m_bUseAverageTreeSize){
		assert(m_uAverageTreeSize>0);
		b = m_dFitnessAlfa * tSize/(double)m_uAverageTreeSize;
	}else{
		if(m_bUseMaxTreeSize){
			assert(m_uMaxTreeSize>0);
			b = m_dFitnessAlfa * tSize/(double)m_uMaxTreeSize;
		}else{
			b = m_dFitnessAlfa * tSize;
		}
	}
	c = 0.0;
	if(m_bCostOfClassification){
		c = m_dGamma * eaindividual->m_CDTIndRoot->GetCostOfClassificationOnTrainDS();
	}
	/*if(m_bCostOfAttrs){
		c+= m_dGamma * eaindividual->m_CDTIndRoot->m_fCostOfAttrs;
	}*/
	
	eaindividual->m_dFitness = a - b - c;
	eaindividual->m_dQuality = a;
	eaindividual->m_dSize = tSize;
	if(m_bCostOfClassification){
		// to przypisanie bo pole m_dCost jest z klasy nadrzednej a sa miejsca gdzie
		// korzystamy z polimorfizmu
		eaindividual->m_dCost = eaindividual->m_CDTIndRoot->GetCostOfClassificationOnTrainDS();
	}
}

bool CEAIDTInducerOB::SetParam(string sParamName, string sParamValue){
	string::iterator i = sParamName.begin();
	while(i != sParamName.end()){
		*i = tolower(*i);
		++i;
	}
	i = sParamValue.begin();
	while(i != sParamValue.end()){
		*i = tolower(*i);
		++i;
	}

	if(sParamName=="centre hiperplane" || sParamName=="centrehiperplane"){
		if(sParamValue=="yes"){
			m_bCentreHiperplane = true;
			return true;
		}
		if(sParamValue=="no"){
			m_bCentreHiperplane = false;
			return true;
		}
		return false;
	}
	if(sParamName=="feature selection" || sParamName=="featureselection"){
		if(sParamValue=="yes"){
			m_bFeatureSelection = true;
			return true;
		}
		if(sParamValue=="no"){
			m_bFeatureSelection = false;
			return true;
		}
		return false;
	}
	if(sParamName=="dipol operator" || sParamName=="dipoloperator"){
		if(sParamValue=="yes"){
			m_bDipolOperator = true;
			return true;
		}
		if(sParamValue=="no"){
			m_bDipolOperator = false;
			return true;
		}
		return false;
	}
	if(sParamName=="beta"){
		m_dFitnessBeta = atof(sParamValue.c_str());
		return true;
	}
	if(sParamName=="prevention underfitting factor" || sParamName=="preventionunderfittingfactor"){
		m_uPUF = atoi(sParamValue.c_str());
		return true;
	}

	if(IEAIDTInducer::SetParam(sParamName,sParamValue)) return true;
	return false;
}

void CEAIDTInducerOB::PostTrain(const IDataSet *trainingDataSet){
}

void CEAIDTInducerOB::PrintParamsXML(ostream &ostr, int nsp){
	IEAIDTInducer::PrintParamsXML(ostr,nsp);
	ostr.fill(' ');
	ostr<<setw(nsp)<<" "<<"<MLPParam Name=\"centrehiperplane\" Value=\"yes\" />"<<endl;
	ostr<<setw(nsp)<<" "<<"<MLPParam Name=\"featureselection\" Value=\"no\" />"<<endl;
	ostr<<setw(nsp)<<" "<<"<MLPParam Name=\"dipoloperator\" Value=\"yes\" />"<<endl;
	ostr<<setw(nsp)<<" "<<"<MLPParam Name=\"beta\" Value=\"0.2\" />"<<endl;
	ostr<<setw(nsp)<<" "<<"<MLPParam Name=\"preventionunderfittingfactor\" Value=\"5\" />"<<endl;
	ostr<<setw(nsp)<<" "<<"<!-- branches swapped in random order -->"<<endl;
	ostr<<setw(nsp)<<" "<<"<MLPParam Name=\"pc0\" Value=\"20\" />"<<endl;
	ostr<<setw(nsp)<<" "<<"<!-- subtrees swapped -->"<<endl;
	ostr<<setw(nsp)<<" "<<"<MLPParam Name=\"pc1\" Value=\"20\" />"<<endl;
	ostr<<setw(nsp)<<" "<<"<!-- hiperplane crossed -->"<<endl;
	ostr<<setw(nsp)<<" "<<"<MLPParam Name=\"pc2\" Value=\"20\" />"<<endl;
	ostr<<setw(nsp)<<" "<<"<!-- Dipolar crossover - subtrees -->"<<endl;
	ostr<<setw(nsp)<<" "<<"<MLPParam Name=\"pc3\" Value=\"20\" />"<<endl;
	ostr<<setw(nsp)<<" "<<"<!-- Dipolar crossover - tests -->"<<endl;
	ostr<<setw(nsp)<<" "<<"<MLPParam Name=\"pc4\" Value=\"20\" />"<<endl;		
	ostr<<setw(nsp)<<" "<<"<!-- random zero -->"<<endl;
	ostr<<setw(nsp)<<" "<<"<MLPParam Name=\"pm0\" Value=\"12\" />"<<endl;
	ostr<<setw(nsp)<<" "<<"<!-- single factor mutated -->"<<endl;
	ostr<<setw(nsp)<<" "<<"<MLPParam Name=\"pm1\" Value=\"12\" />"<<endl;
	ostr<<setw(nsp)<<" "<<"<!-- tests between father and son exchanged -->"<<endl;
	ostr<<setw(nsp)<<" "<<"<MLPParam Name=\"pm2\" Value=\"12\" />"<<endl;
	ostr<<setw(nsp)<<" "<<"<!-- symmetric mutation -->"<<endl;
	ostr<<setw(nsp)<<" "<<"<MLPParam Name=\"pm3\" Value=\"12\" />"<<endl;
	ostr<<setw(nsp)<<" "<<"<!-- test in node changed -->"<<endl;
	ostr<<setw(nsp)<<" "<<"<MLPParam Name=\"pm4\" Value=\"13\" />"<<endl;
	ostr<<setw(nsp)<<" "<<"<!-- mixed dipol cut -->"<<endl;
	ostr<<setw(nsp)<<" "<<"<MLPParam Name=\"pm5\" Value=\"13\" />"<<endl;
	ostr<<setw(nsp)<<" "<<"<!-- pure dipol joined -->"<<endl;
	ostr<<setw(nsp)<<" "<<"<MLPParam Name=\"pm6\" Value=\"13\" />"<<endl;
	ostr<<setw(nsp)<<" "<<"<!-- node changed into leaf -->"<<endl;
	ostr<<setw(nsp)<<" "<<"<MLPParam Name=\"pm7\" Value=\"13\" />"<<endl;
	ostr<<setw(nsp)<<" "<<"<!-- leaf changed into node -->"<<endl;
	ostr<<setw(nsp)<<" "<<"<MLPParam Name=\"pml0\" Value=\"50\" />"<<endl;
	ostr<<setw(nsp)<<" "<<"<!-- Leaf expanded into subtree -->"<<endl;
	ostr<<setw(nsp)<<" "<<"<MLPParam Name=\"pml1\" Value=\"50\" />"<<endl;
	ostr.fill(' ');
}

vector<string> CEAIDTInducerOB::GetParams() const{
vector<string> res;
stringstream ss;

	res = IEAIDTInducer::GetParams();

	res.push_back("centrehiperplane");
	if(m_bCentreHiperplane) res.push_back("yes");
		else res.push_back("no");
	res.push_back("featureselection");
	if(m_bFeatureSelection) res.push_back("yes");
		else res.push_back("no");
	res.push_back("dipoloperator");
	if(m_bDipolOperator) res.push_back("yes");
		else res.push_back("no");
	res.push_back("beta");
	ss.str(""); ss<<m_dFitnessBeta;
	res.push_back(ss.str());
	res.push_back("preventionunderfittingfactor");
	ss.str(""); ss<<m_uPUF;
	res.push_back(ss.str());
	return res;
}


string CEAIDTInducerOB::GetName(){
	return "DT_EA_OB";
}

void CEAIDTInducerOB::PreCreateInitialPopulation(){
}

void CEAIDTInducerOB::PostCreateInitialPopulation(){
	IEAIDTInducer::PostCreateInitialPopulation();

	if(m_bUseAverageTreeSize){
		EstimateAverageTreeSize();
	}else{
		if(m_bUseMaxTreeSize){
			EstimateMaxTreeSize();
		}
	} 
	// To wywolanie dla uczenia z kosztem klasyfikacji po to zeby policzyc
	// koszt klasyfikacji dla kazdego osobnika
	if(m_bCostOfClassification || m_bCostOfAttrs){
		for(size_t t=0; t<m_dqPopulation.size(); t++){
			CEAIDTIndividual* eaindividual = (CEAIDTIndividual*)m_dqPopulation[t];
			EstimateErrors(eaindividual);
		}
	}

	if(m_bMaximizeFitness){
		for(size_t t=0; t<m_dqPopulation.size(); t++){
			CEAIDTIndividual* eaindividual = (CEAIDTIndividual*)m_dqPopulation[t];
			// Maksymalizowanie funkcji fitness poprzez przycinanie drzewa.
			MaximizeFitness(eaindividual);
		}
	}
}

void CEAIDTInducerOB::PostEAOperations(deque<IIndividual*> &p){
	if(m_bCentreHiperplane && m_bDipolOperator){
		for(size_t t=0; t<p.size(); t++){
			// Tu nie trzeba przeliczac bledow, bo hiperplaszczyzna jest przesuwana
			// o tyle, ze nie zmienia sie pokrywanie zbioru danych
			CEAIDTIndividual *dtInd=(CEAIDTIndividual*)p[t];
			CentreHiperplanes(dtInd->m_CDTIndRoot);
		}
	}

	if(m_bMaximizeFitness){
		// Maksymalizowanie funkcji fitness
		for(size_t t=0; t<p.size(); t++){
			if( rand()%100 + 1 <= m_iMaxFitnessProbability){
				CEAIDTIndividual *dtInd=(CEAIDTIndividual*)p[t];
				MaximizeFitness(dtInd);
				// Fitness trzeba przliczyc, bo po wywolaniu tej funkcji
				// fitness nie jest juz uaktualniany
				EvaluateIndividualsFitness(dtInd);
			}
		}
	}
}
	
void CEAIDTInducerOB::CentreHiperplanes(CDTreeNode* node){
static CObliqueTestCreator obliqueTestCreator;
	assert(node);
	obliqueTestCreator.ShiftHiperplane(node->m_pITest,node->m_deqCI);
	if(!node->m_bLeaf){
		// Rekurencyjne schodzenie w dol drzewa.
		for(size_t t=0; t<node->m_vBranch.size(); t++){
			CentreHiperplanes(node->m_vBranch[t]);
		}
	}
}
