#ifndef __CEAIDTINDUCEROB__
#define __CEAIDTINDUCEROB__

#include "../MLPDT_EA/EAIDTInducer.h"

/**
 * Klasa ucznia generujacego drzewo decyzyjne z wykorzystaniem algorytmu
 * ewolucyjnego i skosnych testow.
 *
 * @author Marek Grzes'
 */
class CEAIDTInducerOB: public IEAIDTInducer{
friend class CEAIDTInducerAPOB;

public:
	virtual vector<string> GetParams() const;
	virtual bool SetParam(string sParamName, string sParamValue);
	virtual void PrintParamsXML(ostream &ostr, int nsp);
	virtual string GetName();

	CEAIDTInducerOB(void);
	virtual ~CEAIDTInducerOB(void);

protected:
	bool	m_bDipolOperator;

	/**
	 * prevention underfitting factor
	 */
	unsigned	m_uPUF;

	bool m_bFeatureSelection;

	/**
	 * Parametr wykorzystywany przy selekcji cech. Jest uwzgledniany przy liczeniu
	 * rozmiaru drzewa.
	 */
	double	m_dFitnessBeta;

	bool	m_bCentreHiperplane;

	/**
	 * Operator mutacji skosnego drzewa decyzyjnego, ktory zastepuje dany lisc poddrzewem wygenerowanym dla
	 * przykladow znajdujacych sie w tym lisciu. Wezel przekazany w parametrze node
	 * musi byc lisciem. W przeciwnym przypadku funkcja konczy dzialanie i zwraca false.
	 * @param node - wezel w drzewie, dla ktorego przeprowadzana jest mutacja
	 * @param attrCount - liczba atrybutow zbioru danych
	 * @param dStop - p-o zatrzymania rekurencyjnego rozbudowywania drzewa
	 * @return true - kiedy nastapila mutacja / false - kiedy mutacji nie bylo (np. przyklady tylko z jednej klasy)
	 */
	virtual bool ExpandLeaf(CDTreeNode* node, unsigned attrCount, double dStop, const IDataSet *dataset);
	
	/**
	 * Tworzy populacje poczatkowa i zapisuje ja w m_dqPopulation.
	 * @see m_dqPopulation
	 */
	virtual void ActualCreateInitialPopulation();
	
	/**
	 * Dla podanego w parametrze osobnika oblicza dopasowanie.
	 * @param individual - osobnik, dla ktorego ma byc wyznaczone dopasowanie
	 */
	virtual void EvaluateIndividualsFitness(IIndividual* individual);

	virtual bool MutateNode(CDTreeNode* node, const IDataSet* dataset, CEAIDTIndividual &indv);
	virtual void CrossoverOperator();
	virtual void PreCreateInitialPopulation();
	virtual void PostCreateInitialPopulation();
	virtual void PostEAOperations(deque<IIndividual*> &p);
	virtual void CentreHiperplanes(CDTreeNode* node);
	virtual void PostTrain(const IDataSet *trainingDataSet);

	virtual vector<size_t> SetMutationOperators(CDTreeNode* pNode);
};
#endif

