#include "MTTest.h"
#include <typeinfo>

CMTTest::CMTTest(vector<ITest*> tests):ITest(-1){
	m_vTests = tests;
}

CMTTest::~CMTTest(void){
	for(size_t t=0; t<m_vTests.size(); t++){
		delete m_vTests[t];
	}
	m_vTests.clear();
}

ITest* CMTTest::CloneTest()const{
	vector<ITest*> v(m_vTests.size());
	for(size_t t=0; t<m_vTests.size(); t++){
		v[t] = m_vTests[t]->CloneTest();
	}
	return new CMTTest(v);
}

unsigned CMTTest::GetBranchID(CDataObject *pCInst)const{
	assert(pCInst);
	vector<pair<unsigned,unsigned> > vote;
	for(size_t t=0; t<m_vTests.size(); t++){
		unsigned b = m_vTests[t]->GetBranchID(pCInst);
		bool found = false;
		for(size_t u=0; u<vote.size(); u++){
			if(vote[u].first == b){
				found = true;
				vote[u].second++;
			}
		}
		if(!found){
			pair<unsigned, unsigned> p(b,1);
			vote.push_back(p);
		}
	}
	size_t maxU = 0;
	unsigned max = 0;
	for(size_t u=0; u<vote.size(); u++){
		if(vote[u].second > max){
			maxU = u;
		}
	}
	return maxU;
}

float CMTTest::GetCostOfAttributes(const IDataSet* dataset)const{
	float res = 0;
	for(size_t t=0; t<m_vTests.size(); t++){
		res += m_vTests[t]->GetCostOfAttributes(dataset);
	}
	return res;
}

float CMTTest::GetCostOfAttributes(const IDataSet* dataset, vector<bool> &vUsed)const{
	float res = 0;
	for(size_t t=0; t<m_vTests.size(); t++){
		res += m_vTests[t]->GetCostOfAttributes(dataset,vUsed);
	}
	return res;
}

unsigned CMTTest::GetSubSetCount()const{
	if(!m_vTests.empty()){
		return m_vTests[0]->GetSubSetCount();
	}
	return 0;
}

vector<vector<unsigned> > CMTTest::GetSubSets(deque<CDataObject*> deqCI)const{
vector<vector<unsigned> > res(this->GetSubSetCount());
	for(unsigned i=0; i<deqCI.size(); i++){
		unsigned uSubSetId = GetBranchID(deqCI[i]);
		res[uSubSetId].push_back(i);
	}
	return res;
}

bool CMTTest::operator==(const ITest& it)const{
   try {
		CMTTest& ref = dynamic_cast<CMTTest&>((ITest&)it);
		// Jak jest inna liczba testow to testy sa od razu rozne.
		if(ref.m_vTests.size() != m_vTests.size()){
			return false;
		}
		for(size_t t=0; t<m_vTests.size(); t++){
			// Jak chociaz jeden skladnik na danej pozycji jest rozny to testy sa rozne
			if(! m_vTests[t]->operator==(*(ref.m_vTests[t])) ){
				return false;
			}
		}
	}
	catch (bad_cast) {
		return false;
	}
	return true;
}

void CMTTest::PrintTest(unsigned branchid, const IDataSet* dataset, ostream &ostr)const{
	assert(m_pAttrDesc);
	assert(m_pAttrDesc->m_vAttrDomain.size()>branchid);
	for(size_t t=0; t<m_vTests.size(); t++){
		ostr<<"[";
		m_vTests[t]->PrintTest(branchid,dataset,ostr);
		ostr<<"]";
	}
}
