/**
 * $Revision: 1.2 $
 * $Date: 2006/09/20 22:07:16 $
 */

#include "TDIDTMTNAPInducer.h"
#include "../../mlpcore/MLPDT/TestCreator.h"
#include "../../mlpcore/MLPDT/RealTest.h"
#include "../../mlpcore/MLPDT/Entropy.h"
#include "MTTest.h"
#include <iomanip>
#include <algorithm>

// Author: Marek Grzes <marekg@ii.pb.bialystok.pl>, (C) 2006

bool vpless(pair<size_t,long double> arg1, pair<size_t,long double> arg2){
	if(arg1.second < arg2.second){
		return true;
	}
	return false;
}

CTDIDTMTNAPInducer::CTDIDTMTNAPInducer(){
	m_uNTests = 3;
}

CTDIDTMTNAPInducer::~CTDIDTMTNAPInducer(){

}

ITest* CTDIDTMTNAPInducer::BuildMultiTest(deque<CDataObject*> &deqCI, ITestContainer &tests, const IDataSet *trainingDataSet){
long double maxgain;
size_t maxidx;

	size_t testsSize = tests.size();

	if(testsSize<0){
		return 0;
	}

	if(m_sSelectionOfATest=="surrogate"){
			maxidx = 0;
			maxgain = CEntropy::GainRatio(deqCI,tests[0]);
			for(size_t t=1; t<testsSize; t++){
				long double gain = CEntropy::GainRatio(deqCI,tests[t]);
				if(gain >= maxgain){
					maxgain = gain;
					maxidx = t;
				}
			}
			// Jan nie ma przyrostu informacji (maxgain == 0) to warunek stopu -
			// zwrocenie NULL;
			if(maxgain == 0) return NULL;
			ITest* res = tests[maxidx]->CloneTest();
			
			// numertestu, odleglosc
			vector<pair<size_t, long double> > cl(testsSize);
			unsigned nChecked = 0;
			for(size_t t=0; t<testsSize; t++){
				// Biezemy pod uwage tylko testy na innych atrybutach
				if( res->GetAttrNum() != tests[t]->GetAttrNum() ){
					size_t ds = deqCI.size();
					unsigned distance = 0;
					for(size_t i=0; i<ds; i++){
						if( tests[t]->GetBranchID(deqCI[i]) != res->GetBranchID(deqCI[i])){
							distance++;
						}
					}
					cl[nChecked].first = t;
					cl[nChecked].second = distance;
					nChecked++;
				}
			}
			cl.resize(nChecked);
			sort(cl.begin(),cl.end(),vpless);
			cerr<<cl.front().second<<" "<<cl.back().second<<endl;
			// __________________ Budowanie wektora testow __________________________
			vector<ITest*> v(m_uNTests);
			// Test najlepszy od razu idzie do wektora
			v[0] = res;
			// Indeks testu do dodania z listy cl
			unsigned i=0;
			// Liczba testow dodanych do wektora v
			unsigned a=1;
			while(a<m_uNTests){
				// Ten if po to aby przeskoczyc testy na atrybutach z ktorych testy sa juz dodane
				// do wektora v
				if(a>1){
					bool found;
					do{
						found = false;
						// Ta petna idzie do "a" bo na razie pozstale elementy nie sa zdefiniowane
						for(size_t k=0; k<a; k++){
							cerr<<v[k]->GetAttrNum()<<"\t"<<tests[cl[i].first]->GetAttrNum()<<endl;
							if( v[k]->GetAttrNum() == tests[cl[i].first]->GetAttrNum() ){
								found = true;
								i++;
								break;
							}
						}
					}while(found && i<cl.size());
					cerr<<"i="<<i<<endl;
				}
				if(i>=cl.size()){
					v.resize(a);
					break;
				}
				ITest* res2 = tests[cl[i].first]->CloneTest();
				v[a]=res2;
				a++;
				i++;
			}
			// __________________ Budowanie wektora testow KONIEC ___________________
		return (ITest*) CheckMultiTest( new CMTTest(v) , deqCI, trainingDataSet );
	}

	if(m_sSelectionOfATest == "random"){
		vector<ITest*> v(3);
		size_t testsSize = tests.size();
		for(int i=0; i<3; i++){
			ITest* res = tests[rand()%testsSize]->CloneTest();
			v[i]=res;
		}
		return (ITest*)(new CMTTest(v));
	}

	if(m_sSelectionOfATest=="gain ratio criterion"){
		vector<pair<size_t, long double> > g(testsSize);
		for(size_t t=0; t<testsSize; t++){
			g[t].first = t;
			g[t].second = CEntropy::GainRatio(deqCI,tests[t]);
		}
		sort(g.begin(),g.end(),vpless);
		vector<ITest*> v(m_uNTests);
		for(unsigned i=0; i<m_uNTests; i++){
			ITest* res = tests[g[g.size()-1-i].first]->CloneTest();
			v[i]=res;
		}
		return (ITest*) CheckMultiTest( new CMTTest(v) , deqCI, trainingDataSet );
	}
	return 0;
}

ITest* CTDIDTMTNAPInducer::CheckMultiTest(CMTTest* test, deque<CDataObject*> &deqCI, const IDataSet *trainingDataSet){
	bool removed;
	do{
		removed=false;
		// Utworz tablice podzialow
		vector<vector<unsigned> > vvDiv = test->GetSubSets(deqCI);
	
		// Sprawdzamy czy sa chociaz 2 niepuste podzialy
		int div=0;
		for(unsigned i=0; i<vvDiv.size(); i++){
			if(vvDiv[i].size()>0){
				div++;
			}
			if(div==2){
				break;
			}
		}
		// Jesli nie ma co namjmniej dwoch nie pustych podzialow to wyczucamy dzwie cechy (o ile jest z czego wyrzucac)
		if(div<2){
			size_t size = test->m_vTests.size();
			if(size>=3){
				test->m_vTests.resize(size-2);
				cerr<<"New size is "<<size-2<<endl;
				removed = true;
			}else{
				cerr<<"do d"<<endl;
			}
		}
	}while(removed);

	return test;
}

string CTDIDTMTNAPInducer::GetName(){
	return "MLPDT_MTN";
}

vector<string> CTDIDTMTNAPInducer::GetParams() const{
vector<string> res;
stringstream ss;

	res = CTDIDTInducer::GetParams();

	res.push_back("ntests");
	ss.str("");ss<<m_uNTests;
	res.push_back(ss.str());
		
	return res;	
}

/**
 * Tu mozna wykorzystac dipole do znajdowania kilku testow podobnych. Np. zbior testow
 * moze byc zdefiniowany jako zbior takich testow, ktore przecinaja ten sam dipol.
 * Mozna by to zapewne wykorzystac w alg. ewolucyjnym dla takich drzew.
 */
ITest* CTDIDTMTNAPInducer::PickOutTest(deque<CDataObject*> &deqCI, ITestContainer &tests, const IDataSet *trainingDataSet){
//long double maxgain;
//size_t maxidx;

	// Jak jest osiagnieta maksymalna glebokosc drzewa to zwracamy NULL i robudowywanie
	// drzewa zostanie zatrzymane.
	if(m_iMaxDepth > -1 && m_iCurrentDepth >= m_iMaxDepth){
		return 0;
	}

	// Zwracamy test zbudowany metoda budujaca testy
	return BuildMultiTest(deqCI,tests,trainingDataSet);

/*
	size_t testsSize = tests.size();

	if(m_sSelectionOfATest=="ganzrandom"){
		if(testsSize > 0){
			ITest* res = tests[rand()%testsSize]->CloneTest();
			if(m_bInnerDisjunction){
				RandomInnerDisjuntion(res);
			}
			return res;
		}else{
			return NULL;
		}
	}
	
	if(m_sSelectionOfATest=="dipolar"){
		if(testsSize > 0){
			static CTestCreator tc;
			CDataObject* pob1;
			CDataObject* pob2;
			if(tc.FindMixedDipol(deqCI,pob1,pob2)){
				// zeby nie dodawac przyrostowo do wektora to od razo tworze o maksymalnym rozmiarze
				vector<int> vAttrs(trainingDataSet->GetCurrentAttrCount(),-1);
				// liczba elementow w wektorze atrybutow
				size_t vmax = 0;
				for(size_t t=0; t<trainingDataSet->GetCurrentAttrCount(); t++){
					if( ! ( *((*pob1)[t]) == *((*pob2)[t]) ) ){
						vAttrs[vmax] = (int)t;
						vmax++;
					}
				}
				// attrId jest to losowany index atrybutu do uzycia przy wyszukiwaniu testu
				if(vmax == 0){
					return NULL;
				}
				size_t attrId = vAttrs[rand()%vmax];
				if(tests.size(attrId) == 0){
					return NULL;
				}
				
				if((*(pob1->operator [ ](attrId))).GetAttrType() == e_NominalAttr){
					// Jezeli to jest atrybut nomilany to losujemy jeden test, tu na razie
					// jest wylko jeden test to losujemy jeden z jednego
					ITest* res = tests.GetAttrsTest(attrId, rand()%tests.size(attrId))->CloneTest();
					if(m_bInnerDisjunction){
						RandomInnerDisjuntion(res);
					}
					return res;
				}
				
				if((*(pob1->operator[ ](attrId))).GetAttrType() == e_RealAttr){
					double val1 = pob1->operator [ ](attrId)->m_AttrValue.u_dReal;
					double val2 = pob2->operator [ ](attrId)->m_AttrValue.u_dReal;
					if(val1 > val2){
						double temp = val1;
						val1 = val2;
						val2 = temp;
					}
					vector<CRealTest*> vRealTest(tests.size(attrId));
					int nTests = 0;
					for(size_t i=0; i<tests.size(attrId); i++){
        				CRealTest* pTest = dynamic_cast<CRealTest*>(tests.GetAttrsTest(attrId,i));
						if(pTest && val1 < pTest->m_dThreshold && val2 > pTest->m_dThreshold){
							vRealTest[nTests] = pTest;
							nTests++;
						}
					}
					if(nTests == 0){
						return NULL;
					}
					return vRealTest[rand()%nTests]->CloneTest();
					
				}
				
				return tests[rand()%testsSize]->CloneTest();
			}else{
				return NULL;
			}
		}else{
			return NULL;
		}
	}

	if(m_sSelectionOfATest=="gain criterion"){
		if(testsSize>0){
			maxidx = 0;
			maxgain = CEntropy::Gain(deqCI,tests[0]);
			for(size_t t=1; t<testsSize; t++){
				long double gain = CEntropy::Gain(deqCI,tests[t]);
				if(gain > maxgain){
					maxgain = gain;
					maxidx = t;
				}
			}
			// Jan nie ma przyrostu informacji (maxgain == 0) to warunek stopu -
			// zwrocenie NULL;
			if(maxgain == 0) return NULL;
			ITest* res = tests[maxidx]->CloneTest();
			if(m_bInnerDisjunction){
				RandomInnerDisjuntion(res);
			}
			return res;
		}else{
			return NULL;
		}
	}

	if(m_sSelectionOfATest=="gain ratio criterion"){
		if(testsSize>0){
			maxidx = 0;
			maxgain = CEntropy::GainRatio(deqCI,tests[0]);
//- cout<<maxgain<<" "<<(*maxiter)->GetDescription()<<" "<<((CRealTrainTest*)*maxiter)->m_dThreshold<<endl;
			for(size_t t=1; t<testsSize; t++){
				long double gain = CEntropy::GainRatio(deqCI,tests[t]);
//- cout<<gain<<" "<<(*triter)->GetDescription()<<" "<<((CRealTrainTest*)*triter)->m_dThreshold<<endl;
				if(gain >= maxgain){
					maxgain = gain;
					maxidx = t;
				}
			}
			// Jan nie ma przyrostu informacji (maxgain == 0) to warunek stopu -
			// zwrocenie NULL;
			if(maxgain == 0) return NULL;
			ITest* res = tests[maxidx]->CloneTest();
			if(m_bInnerDisjunction){
				RandomInnerDisjuntion(res);
			}
			return res;
		}else{
			return 0;
		}
	}
	
*/
	/**
	 * Test wybierany jest w sposob losowy. Dla wybranego testu spawdzane jest
	 * gain ratio. Jesli wynosi ono 0 to nastepuje ponowne losowanie. Liczba
	 * losowan jest rowna liczbie testow. Kiedy w zadnym z losowan nie uzyska
	 * sie testu dajacego gain ratio wieksze od zera to zwracany jest NULL
	 * co stanowi warunek stopu.
	 */
/*	if(m_sSelectionOfATest=="random"){
		if(testsSize>0){
			size_t repeat=testsSize;
			while(repeat-->0){
				size_t n = rand()%testsSize;
				if(CEntropy::GainRatio(deqCI,tests[n])>0){
					ITest* res = tests[n]->CloneTest();
					if(m_bInnerDisjunction){
						RandomInnerDisjuntion(res);
					}
					return res;
				}
			}
		}else{
			return NULL;
		}
	}
*/
	/**
	 * Kiedy parametr m_sSelectionOfATest nie jest ustawiony
	 * zwracany jest test zupelnie losowy.
	 */
/*	if(testsSize>0){
		ITest* res = tests[rand()%testsSize]->CloneTest();
		if(m_bInnerDisjunction){
			RandomInnerDisjuntion(res);
		}
		return res;
	}
	return 0;*/
}

void CTDIDTMTNAPInducer::PrintParamsXML(ostream &ostr, int nsp){
	CTDIDTInducer::PrintParamsXML(ostr,nsp);
	ostr.fill(' ');
	ostr<<setw(nsp)<<" "<<"<MLPParam Name=\"ntests\" Value=\"3\" />"<<endl;
	ostr.fill(' ');
}

bool CTDIDTMTNAPInducer::SetParam(string sParamName, string sParamValue){
	string::iterator i = sParamName.begin();
	while(i != sParamName.end()){
		*i = tolower(*i);
		++i;
	}
	i = sParamValue.begin();
	while(i != sParamValue.end()){
		*i = tolower(*i);
		++i;
	}

	if(sParamName=="ntests"){
		m_uNTests = atoi(sParamValue.c_str());
		return true;
	}

	if(sParamName=="selection of a test" || sParamName=="selectionofatest"){
		if(sParamValue=="surrogate"){
			m_sSelectionOfATest=sParamValue;
			return true;
		}else{
			if(CTDIDTInducer::SetParam(sParamName,sParamValue)){
				return true;
			}else{
				return false;
			}
		}
	}

	if(CTDIDTInducer::SetParam(sParamName,sParamValue)) return true;

	return false;
}
