#include "EAObliqueTestCreator.h"
#include "../../mlpcore/MLPDT/ObliqueTest.h"
#include "../../mlpcore/MLPDataset/DataObject.h"
#include "../../mlpcore/MLPDT/Entropy.h"
#include "../../mlpcore/MLPDataset/RealAttr.h"
#include "../../mlpcore/MLPCommon/MLPStatistics.h"
#include "ObliqueTestIndividual.h"
#include <cmath>
#include <numeric>
#include <sstream>
#include <iomanip>
#include <cstdlib>
#include <algorithm>

CEAObliqueTestCreator::CEAObliqueTestCreator(void){
	m_dAlfa = 0.01;
	m_dBeta = 0.2;
	m_sFitnessFunction = "dipolar";
	m_bFeatureSelection = false;
}

CEAObliqueTestCreator::~CEAObliqueTestCreator(void){
}


ITest* CEAObliqueTestCreator::CreateObliqueTest(deque<CDataObject*> &deqCI, unsigned attrCount, unsigned uPuf){
	m_uAttrCount = attrCount;
	m_uPUF = uPuf;
	m_pdeqCI = &deqCI;
	return (ITest*) Evaluate();
}

void* CEAObliqueTestCreator::Evaluate(void* ep1){

	ActualCreateInitialPopulation();
	EvaluateFitness(m_dqPopulation);
	deque<IIndividual*>::iterator maxind = max_element(m_dqPopulation.begin(),m_dqPopulation.end(),IIndividualLess);
	m_BestIndividual = (*maxind)->Clone();
	
	unsigned currentIt = 0;
	unsigned idleIt = 0;
	do{
		SelectIndividualsToReproduce();
		CrossoverOperator((deque<CObliqueTestIndividual*>&)m_dqPopulation,*m_pdeqCI);
		MutationOperator((deque<CObliqueTestIndividual*>&)m_dqPopulation,*m_pdeqCI);
		PreventUnderfitting((deque<CObliqueTestIndividual*>&)m_dqPopulation,*m_pdeqCI);
		EvaluateFitness(m_dqPopulation);
		if(!m_dqPopulation.empty()){
			deque<IIndividual*>::iterator maxind = max_element(m_dqPopulation.begin(),m_dqPopulation.end(),IIndividualLess);
			if(m_BestIndividual->m_dFitness < (*maxind)->m_dFitness){
				delete m_BestIndividual;
				m_BestIndividual = (*maxind)->Clone();
				idleIt=0;
			}
		}
		currentIt++;
	}while(currentIt<=m_uMinIterations || (currentIt<=m_uMaxIterations && ++idleIt<=m_uMaxIdleIterations));
	
	
	if(m_BestIndividual){
		CObliqueTest *res = new CObliqueTest(*(((CObliqueTestIndividual*)m_BestIndividual)->m_pOblTest));
		for(unsigned i=0; i<m_dqPopulation.size(); i++)
			delete m_dqPopulation[i];
		m_dqPopulation.clear();
		delete m_BestIndividual;
		m_BestIndividual=0;
		ShiftHiperplane(res,*m_pdeqCI);
		return res;
	}
	return NULL;
}

void CEAObliqueTestCreator::PreventUnderfitting(deque<CObliqueTestIndividual*> &dqPopulation, deque<CDataObject*> &deqCI){
	for(size_t i=0; i<dqPopulation.size(); i++){
		vector<double> &w = dqPopulation[i]->m_pOblTest->m_vdWReal;
		int removedAttrCount=0;
		for(size_t t=0; t<w.size()-1; t++){
			if(w[t]==0){
				removedAttrCount++;
			}
		}
		
		if( ((size_t)removedAttrCount) == w.size()){
			CLogger::LogEvent("all features equal in mixed dipol",__FILE__,__LINE__);
		}
		
		// Usuwanie atrybutow do momentu spelnienia warunku.Mozna zastanowic sie
		// nad tym, aby wyrzucac w pierwszej kolejnosci te atrybuty, 
		// ktore nie wplywaja na separowalnosc przykladow tworzacych dipol.
		while( (m_uAttrCount - removedAttrCount) * m_uPUF > deqCI.size()){
			unsigned k = rand()%m_uAttrCount;
			unsigned ind = k;
			while(w[ind]==0){
				ind = ++k%m_uAttrCount;
			}
			removedAttrCount++;
			w[ind] = 0;
			//cerr<<ind<<" removed "<<ind<<" in "<<i<<endl;
		}
	}
}

void CEAObliqueTestCreator::EvaluateIndividualsFitness(IIndividual* individual){
	CObliqueTestIndividual *ind = (CObliqueTestIndividual*)individual;
	if(ind->m_bChanged){
		//Wlaczenie funkcji ponizej wydluza dwukrotnie czas ewolucji.
		//ShiftHiperplane(ind->m_pOblTest,*m_pdeqCI);
		if(m_sFitnessFunction=="dipolar"){
			ind->m_dFitness = DipolarFitness(*m_pdeqCI,(ITest*)(ind->m_pOblTest));
		}else{
			ind->m_dFitness = CEntropy::GainRatio(*m_pdeqCI,(ITest*)(ind->m_pOblTest));
		}
		ind->m_bChanged = false;
	}
}

double CEAObliqueTestCreator::DipolarFitness(deque<CDataObject*> &deqCI, ITest *pIT){
	unsigned size = (unsigned)deqCI.size();
	assert(size>0);
	CObliqueTest* oblTest = dynamic_cast<CObliqueTest*>(pIT);
	if(oblTest){
		// Wszystkie przyklady domyslne na lewo.
		vector<char> sides(size,'L');
		for(size_t t=0; t<size; t++){
			// Dla tych na prawo ustawienie R.
			if(oblTest->GetInnerProduct(deqCI[t]) > 0){
				sides[t] = 'R';
			}
		}
		double fitness;
		int fmixed = 0;
		int fpure = 0;
		vector<IAttrValue*> classes = deqCI[0]->GetClassValues();
		for(size_t c=0; c<classes.size(); c++){
			int nLClass1 = 0;
			int nLClass2 = 0;
			int nRClass1 = 0;
			int nRClass2 = 0;
			for(size_t i=0; i<size; i++){
				if( *(deqCI[i]->GetClass()) == *(classes[c]) ){
					if(sides[i] == 'R') nRClass1++;
					else nLClass1++;
				}
				if( *(deqCI[i]->GetClass()) != *(classes[c]) ){
					if(sides[i] == 'R') nRClass2++;
					else nLClass2++;
				}
			}
			fpure += nRClass1*nLClass1;
			fmixed += nRClass1*nLClass2 + nLClass1*nRClass2;
		}
		// Trzeba podzielic przez 2, bo algorym powyzej kazdy mieszany przeciety
		// dipol zlicza dwukrotnie.
		fmixed /= 2;
		double pureFraction = (double)fpure / ((double)(size*(size-1))/2.0);
		double mixedFraction = (double)fmixed / ((double)(size*(size-1))/2.0);
		fitness = mixedFraction + m_dAlfa*(1.0-pureFraction);
		if(m_bFeatureSelection){
			unsigned N = (unsigned)(oblTest->m_vdWReal.size());
			unsigned n = 0;
			for(size_t w=0; w<N; w++){
				if(oblTest->m_vdWReal[w]!=0){
					n++;
				}
			}
			fitness *= ((1.0-m_dBeta) + m_dBeta*((double)N+1.0)/((double)n+1.0));
		}
		return fitness;
	}
	return -1;
}
	
void CEAObliqueTestCreator::ActualCreateInitialPopulation(){
	m_dqPopulation.resize(m_uSizeOfPop);
	for(unsigned i=0; i<m_uSizeOfPop; i++){
		// Metoda z klasy nadrzednej tworzone sa testy, ktore przecinaja
		// losowy wybrany dipol.
		CObliqueTestIndividual *ind = new CObliqueTestIndividual();
		ind->m_pOblTest = (CObliqueTest*)CObliqueTestCreator::CreateObliqueTest(*m_pdeqCI,m_uAttrCount,m_uPUF);
		if(ind->m_pOblTest==0){
			ind->m_pOblTest = (CObliqueTest*)CObliqueTestCreator::CreateRandObliqueTest(*m_pdeqCI,m_uAttrCount,m_uPUF);
			if(ind->m_pOblTest==0){
				// Jak 2 razy nie znajdziemy testu to wstawiamy zupelnie losowy test
				vector<double> v(m_uAttrCount+1);
				for(size_t i=0;i<v.size();i++){
					v[i]=rand();
				}
				ind->m_pOblTest = new CObliqueTest(v);
			}
		}
		m_dqPopulation[i] = ind;
	}
}

void CEAObliqueTestCreator::PreCreateInitialPopulation(){
}

void CEAObliqueTestCreator::PostCreateInitialPopulation(){
}

bool CEAObliqueTestCreator::MutateTest(ITest* pITest){
	bool mutated = false;
	CObliqueTest* oblTest = dynamic_cast<CObliqueTest*>(pITest);
	if(oblTest){
		for(size_t mPoint=0; mPoint<oblTest->m_vdWReal.size(); mPoint++){
			if( rand()%100 + 1 <= m_iProbabilityOfMutation ){
				if(oblTest->m_vdWReal[mPoint]==0){
					if(rand()%2==0){
						oblTest->m_vdWReal[mPoint] = CMLPStatistics::Rand_01();
					}else{
						oblTest->m_vdWReal[mPoint] = -CMLPStatistics::Rand_01();
					}
					mutated = true;
				}else{
					int m = rand()%4;
					CDataObject* ob1=0;
					CDataObject* ob2=0;
					switch(m){
						case 0:{
							SetRandomZero(oblTest);
							mutated = true;
							break;
						}
						case 1:{
							if(fabs(oblTest->m_vdWReal[mPoint])<=1){
								if(rand()%2==0){
									oblTest->m_vdWReal[mPoint] += CMLPStatistics::Rand_01();
								}else{
									oblTest->m_vdWReal[mPoint] -= CMLPStatistics::Rand_01();
								}
							}else{
								double diff = oblTest->m_vdWReal[mPoint] * CMLPStatistics::Rand_01();
								if(rand()%2==0){
									oblTest->m_vdWReal[mPoint] += diff;
								}else{
									oblTest->m_vdWReal[mPoint] -= diff;
								}
							}
							mutated = true;
							break;
						}
						case 2:{
							// ***** Usuniecie przeciecia dipola czystego *****
							if(FindPureDipol(*m_pdeqCI,ob1,ob2,oblTest)){
								assert(ob1);
								assert(ob2);
								if(JoinPureDipol(ob1,ob2,oblTest)){
									mutated = true;
								}
							}
							break;
						}
						case 3:{
							// ***** Przeciecie dipola mieszanego *****
							if(FindMixedDipol(*m_pdeqCI,ob1,ob2,oblTest)){
								assert(ob1);
								assert(ob2);
								if(CutMixedDipol(ob1,ob2,oblTest)){
									mutated = true;
								}
							}
							break;
						}
					}
				}
			}
		}
		return mutated;
	}
	return false;
}

void CEAObliqueTestCreator::MutationOperator(deque<CObliqueTestIndividual*> &dqPopulation, deque<CDataObject*> &deqCI){
	for(size_t t=0; t<dqPopulation.size(); t++){
		if(MutateTest(dqPopulation[t]->m_pOblTest)){
			dqPopulation[t]->m_bChanged = true;
		}
	}
}

void CEAObliqueTestCreator::CrossoverOperator(deque<CObliqueTestIndividual*> &p,deque<CDataObject*> &deqCI){
	unsigned i=0;
	unsigned j;
	vector<unsigned> used;

	while(i+1<p.size()){
		if( !( rand()%100 + 1 <= m_iProbabilityOfCrossOver ) ){
			i++;
			continue;
		}else{
			j=i+1;
			while(j<p.size()){
				if( !( rand()%100 + 1 <= m_iProbabilityOfCrossOver ) )  j++;
				else break;
			}
		}
		if(j>=p.size()) break;

// W tym miejscu mamy indeksy osobnikow do krzyzowania: i,j
//______________________________________________________________________________________
		if(TwoPointCrossoverHiperplanes(p[i]->m_pOblTest,p[j]->m_pOblTest)){
			p[i]->m_bChanged = true;
			p[j]->m_bChanged = true;
		}
//______________________________________________________________________________________
		i=j+1;
	}

}

bool CEAObliqueTestCreator::SetParam(string sParamName, string sParamValue){
	string::iterator i = sParamName.begin();
	while(i != sParamName.end()){
		*i = tolower(*i);
		++i;
	}
	i = sParamValue.begin();
	while(i != sParamValue.end()){
		*i = tolower(*i);
		++i;
	}
	
	if(sParamName=="alfa"){
		m_dAlfa = (double)atof(sParamValue.c_str());
		if(m_dAlfa < 0 || m_dAlfa > 1){
			m_dAlfa = 0.01;
			return false;
		}else{
			return true;
		}
	}
	if(sParamName=="beta"){
		m_dBeta = (double)atof(sParamValue.c_str());
		if(m_dBeta < 0 || m_dBeta > 1){
			m_dBeta = 0.2;
			return false;
		}else{
			return true;
		}
	}
	if(sParamName=="fitness function" || sParamName=="fitnessfunction"){
		if(sParamValue=="dipolar"){
			m_sFitnessFunction = sParamValue;
			return true;
		}
		if(sParamValue=="gain ratio criterion"){
			m_sFitnessFunction = sParamValue;
			return true;
		}
		return false;
	}
	if(sParamName=="feature selection" || sParamName=="featureselection"){
		if(sParamValue=="yes"){
			m_bFeatureSelection = true;
			return true;
		}
		if(sParamValue=="no"){
			m_bFeatureSelection = false;
			return true;
		}
		return false;
	}
	
	if(IEvolutionaryAlg::SetParam(sParamName,sParamValue)) return true;
	return false;
}

vector<string> CEAObliqueTestCreator::GetParams() const{
vector<string> res;
stringstream ss;

	res = IEvolutionaryAlg::GetParams();
	ss.str(""); ss<<m_dAlfa;
	res.push_back("alfa");
	res.push_back(ss.str());
	res.push_back("beta");
	ss.str(""); ss<<m_dBeta;
	res.push_back(ss.str());
	res.push_back("fitnessfunction");
	res.push_back(m_sFitnessFunction);
	res.push_back("featureselection");	
	if(m_bFeatureSelection) res.push_back("yes");
		else res.push_back("no");
	return res;
}

void CEAObliqueTestCreator::PrintXML(ostream &ostr, int nsp, string sParamName){
	ostr.fill(' ');
	ostr<<setw(nsp)<<" "<<"<"<<sParamName<<" Name=\"alfa\" Value=\"0.01\" />"<<endl;
	ostr<<setw(nsp)<<" "<<"<"<<sParamName<<" Name=\"beta\" Value=\"0.2\" />"<<endl;
	ostr<<setw(nsp)<<" "<<"<"<<sParamName<<" Name=\"fitnessfunction\" Value=\"dipolar\" />"<<endl;
	ostr<<setw(nsp)<<" "<<"<"<<sParamName<<" Name=\"featureselection\" Value=\"yes\" />"<<endl;
	ostr.fill(' ');
}

void CEAObliqueTestCreator::PrintParamsXML(ostream &ostr, int nsp){
	IEvolutionaryAlg::PrintParamsXML(ostr,nsp);
	PrintXML(ostr,nsp,"MLPParam");
}

void CEAObliqueTestCreator::PrintSubParamsXML(ostream &ostr, int nsp){
	IEvolutionaryAlg::PrintSubParamsXML(ostr,nsp);
	PrintXML(ostr,nsp,"MLPSubParam");
}

