#ifndef __CEAOBLIQUETESTCREATOR__
#define __CEAOBLIQUETESTCREATOR__ 

#include "ObliqueTestCreator.h"
#include "ObliqueTestIndividual.h"
#include "../../mlpcore/MLPCommon/IEvolutionaryAlg.h"
#include "../../mlpcore/MLPDT/ITest.h"

/**
 * @author Marek Grzes'
 */ 
class CEAObliqueTestCreator: public CObliqueTestCreator, public IEvolutionaryAlg{
public:
	/**
	 * Na podstawie podanego zbioru przykadow tworzy test skosny. Kiedy przekazany
	 * jako parametr zbior przykladow jest z jednej klasy to zwracany jest NULL.
	 *
	 * @param deqCI - kolejka przykladow
	 * @param attrCount - liczba atrybutow branych pod uwage
	 * @param puf - prevention underfitting factor
	 */
	virtual ITest* CreateObliqueTest(deque<CDataObject*> &deqCI, unsigned attrCount, unsigned uPuf);

	virtual vector<string> GetParams() const;
	virtual bool SetParam(string sParamName, string sParamValue);
	virtual void PrintParamsXML(ostream &ostr, int nsp);
	virtual void PrintSubParamsXML(ostream &ostr, int nsp);
	
	CEAObliqueTestCreator(void);
	virtual ~CEAObliqueTestCreator(void);

protected:
	/**
	 * Rodzaj funkcji dopasowania:
	 * <pre>
	 *	"dipolar" - dipolowa funkcja dopasowania
	 * 	"gain ratio criterion" - dopasowanie na podstawie wspolczynika przyrostu
	 *	informacji.
	 * </pre>
	 */
	string m_sFitnessFunction;

	bool m_bFeatureSelection;
		
	/**
	 * Parametry dipolowej funkcji dopasowania.
	 */
	double	m_dAlfa;
	double	m_dBeta;

	/**
	 * Parametry pomocnicze.
	 */
	unsigned		m_uAttrCount;
	unsigned		m_uPUF;
	deque<CDataObject*>	*m_pdeqCI;
	
	/**
	 * Realizuje przebieg algorytmu ewolucyjnego.
	 * @return wskaznik do testu skosnego
	 * @return 0 kiedy nie ma juz podzialow
	 */
	virtual void* Evaluate(void* ep1=0);
	
	virtual void EvaluateIndividualsFitness(IIndividual* individual);
	virtual void PreCreateInitialPopulation();
	virtual void ActualCreateInitialPopulation();
	virtual void PostCreateInitialPopulation();
	
	/**
	 * Mutacja testu do wykorzystania w algorytnie ewolucyjnym szukania testu
	 * skosnego. Kazdy wspolczynik jest mutowany z p-em m_dProbabilityOfMutation.
	 * @return true - kiedy test zostal zmodyfikowany
	 */
	virtual bool MutateTest(ITest* pITest);	
	
	virtual void MutationOperator(deque<CObliqueTestIndividual*> &dqPopulation,deque<CDataObject*> &deqCI);
	virtual void CrossoverOperator(deque<CObliqueTestIndividual*> &p,deque<CDataObject*> &deqCI);
	virtual void MutationOperator(int iterNum){
	}
	virtual void CrossoverOperator(){
	}
	
	virtual void PrintXML(ostream &ostr, int nsp, string sParamName);

	/**
	 * Obicza dopasowanie hiperplaszczyzny wzgledem dipolowej kryterialnej.
	 */
	virtual double DipolarFitness(deque<CDataObject*> &deqCI, ITest *pIT);
	
	/**
	 * Metoda do wywolania po operatorach ewolucyjnych. Losowo zeruje 
	 * wspolczyniki do momentu az bedzie spelniony warunek unikania
	 * underfittingu.
	 */
	virtual void PreventUnderfitting(deque<CObliqueTestIndividual*> &dqPopulation,deque<CDataObject*> &deqCI);
	
	virtual void CreateGnuplotProjects(string datafilename, string outfilenamestem){
	}
	virtual void PrintFitnessLogHeader(ofstream &ofFlog){
	}
	virtual void PrintFitnessLog(deque<IIndividual*> &p,IIndividual* bestind, unsigned iternum, ostream &ostr){
	}
	virtual void PrintIndividualToLog(string prefix, IIndividual* ind, bool bPrintStructure=false){
	}

	virtual void PostEAOperations(deque<IIndividual*> &p){
	}

#ifdef MLPEAOPERATORSTEST
	void OPTEST_PrintOperatorsTestLog(unsigned iterNum){
	}
	virtual void OPTEST_OperatorsTestLog(IIndividual &indv, string sOpName){
	}
	virtual void OPTEST_PrintOperatorsNames(){
	}
#endif

};

#endif

