#include "ObliqueTestCreator.h"
#include "../../mlpcore/MLPDT/ObliqueTest.h"
#include "../../mlpcore/MLPDataset/DataObject.h"
#include "../../mlpcore/MLPDT/Entropy.h"
#include "../../mlpcore/MLPDataset/RealAttr.h"
#include "../../mlpcore/MLPCommon/MLPStatistics.h"
#include <cmath>
#include <numeric>

CObliqueTestCreator::CObliqueTestCreator(void){
}

CObliqueTestCreator::~CObliqueTestCreator(void){
}

bool CObliqueTestCreator::FindPureDipol(deque<CDataObject*> &deqCI, CDataObject* &ob1, CDataObject* &ob2){
	bool found=false;
	if(deqCI.size()>0){
		IAttrValue *pIA1 = deqCI[0]->GetClass();
		for(unsigned i=1; i<deqCI.size(); i++){
			IAttrValue *pIA2 = deqCI[i]->GetClass();
			if( ! ((*pIA1) == (*pIA2)) ){
				found=true;
				break;
			}
		}
	}else{
		return false;
	}
	if(!found){
		return false;
	}
	vector<vector<CDataObject*> > splitCI(1);
	splitCI[0].push_back(deqCI[0]);
	for(size_t t=1; t<deqCI.size(); t++){
		bool found = false;
		for(size_t k=0; k<splitCI.size(); k++){
			if( *(splitCI[k].front()->GetClass()) == *(deqCI[t]->GetClass()) ){
				splitCI[k].push_back(deqCI[t]);
				found=true;
				break;
			}
		}
		if(!found){
			splitCI.resize(splitCI.size()+1);
			splitCI[splitCI.size()-1].push_back(deqCI[t]);
		}
	}
	if(!(splitCI.size()>(size_t)1)){
		return false;
	}
	size_t atrind = rand()%splitCI.size();
	if(splitCI[atrind].size()<2){
		return false;
	}
	size_t ind1 = rand()%splitCI[atrind].size();
	size_t ind2 = rand()%splitCI[atrind].size();
	while(ind1 == ind2){
		ind2 = rand()%splitCI[atrind].size();
	}
	ob1 = splitCI[atrind][ind1];
	ob2 = splitCI[atrind][ind2];
	return true;
}

ITest* CObliqueTestCreator::CreateObliqueTest(deque<CDataObject*> &deqCI, unsigned attrCount, unsigned uPuf){
	assert(attrCount>0);
	if(deqCI.size()<2 || uPuf>deqCI.size()){
		return NULL;
	}

	// ob1=(x1,y1,z1)
	CDataObject* ob1;
	// ob2=(x2,y2,z2)
	CDataObject* ob2;
	vector<bool> vbFS(attrCount,true);

	if(!FindMixedDipol(deqCI,ob1,ob2)){
		return NULL;
	}

	vector<double> vdN(attrCount);
	vector<double> vdX(attrCount);
	for(unsigned j=0; j<attrCount; j++){
		IAttrValue* pIA1 = ob1->GetRealAttrValue(j);
		IAttrValue* pIA2 = ob2->GetRealAttrValue(j);
		assert(pIA1->GetAttrType()==e_RealAttr);
		assert(pIA2->GetAttrType()==e_RealAttr);
		if( *pIA1 != *pIA2){
			// vdN=(A,B,C)=(x2-x1,y2-y1,z2-z1)
			//gprof vdN.push_back(pIA2->m_AttrValue.u_dReal - pIA1->m_AttrValue.u_dReal);
			vdN[j]=pIA2->m_AttrValue.u_dReal - pIA1->m_AttrValue.u_dReal;
			// x3=(x2-x1)/2 + x1 itd.
			//gprof vdX.push_back( (pIA2->m_AttrValue.u_dReal - pIA1->m_AttrValue.u_dReal) /2 + pIA1->m_AttrValue.u_dReal);
			vdX[j]=(pIA2->m_AttrValue.u_dReal - pIA1->m_AttrValue.u_dReal) /2 + pIA1->m_AttrValue.u_dReal;
		}else{
			// Cechy dla ktorych wartosci w dipolu sa rowne, nie sa brane pod uwage
			// przy tworzeniu hiperplaszczyzny.
			vbFS[j] = false;
			//gprof vdN.push_back(0.0);
			vdN[j]=0.0;
			//gprof vdX.push_back(0.0);
			vdX[j]=0.0;
		}
	}
	size_t removedAttrCount=0;
	for(size_t t=0; t<vbFS.size(); t++){
		if(!vbFS[t]){
			removedAttrCount++;
		}
	}
	
	// Kiedy dipol mieszany skladal sie z dwoch przykladach o takich samych
	// wartosciach wszystkich atrybutow za wyjatkiem klasy, to zwracamy od razu NULL.
	if(removedAttrCount == attrCount){
		return 0;
	}
	
	// Usuwanie atrybutow do momentu spelnienia warunku. Atrybuty usuwane sa w
	// sposob losowy. Mozna zastanowic sie nad tym, aby wyrzucac w pierwszej kolejnosci
	// te atrybuty, ktore nie wplywaja na separowalnosc przykladow tworzacych dipol.
	while( (attrCount - removedAttrCount) * uPuf > deqCI.size()){
		unsigned k = rand()%attrCount;
		unsigned ind = k;
		while(!vbFS[ind]){
			ind = ++k%attrCount;
		}
		removedAttrCount++;
		vbFS[ind] = false;
		vdN[ind] = 0;
		vdX[ind] = 0;
	}

	// Ax + By + Cz
	vector<double> vdW=vdN;
	// (-Ax3-By3-Cz3)
	double w0=0;
	for(unsigned j=0; j<attrCount; j++){
		w0 -= vdN[j]*vdX[j];
	}
	vdW.push_back(w0);
	// W=(Ax + By + Cz + (-Ax3-By3-Cz3))
	return (ITest*) new CObliqueTest(vdW);
}

ITest* CObliqueTestCreator::CreateRandObliqueTest(deque<CDataObject*> &deqCI, unsigned attrCount, unsigned uPuf){
	assert(attrCount>0);
	if(deqCI.size()<2 || uPuf>deqCI.size()){
		return NULL;
	}

	// ob1=(x1,y1,z1)
	CDataObject* ob1;
	// ob2=(x2,y2,z2)
	CDataObject* ob2;
	vector<bool> vbFS(attrCount,true);

	if(!FindMixedDipol(deqCI,ob1,ob2)){
		return NULL;
	}

	vector<double> vdN(attrCount);
	vector<double> vdX(attrCount);
	// ********************* Wsp. zeby nie ciac dokladnie posrodku dipola. **********
	double factor=CMLPStatistics::Rand_01();
	for(unsigned j=0; j<attrCount; j++){
		IAttrValue* pIA1 = ob1->GetRealAttrValue(j);
		IAttrValue* pIA2 = ob2->GetRealAttrValue(j);
		assert(pIA1->GetAttrType()==e_RealAttr);
		assert(pIA2->GetAttrType()==e_RealAttr);
		if( *pIA1 != *pIA2){
			// vdN=(A,B,C)=(x2-x1,y2-y1,z2-z1)
			//gprof vdN.push_back(pIA2->m_AttrValue.u_dReal - pIA1->m_AttrValue.u_dReal);
			vdN[j]=pIA2->m_AttrValue.u_dReal - pIA1->m_AttrValue.u_dReal;
			// **************************** Losowe kiwanie wektorem. ****************
			if(rand()%2){
				//gprof vdN.back() += vdN.back()*0.1*CMLPStatistics::Rand_01();
				vdN[j] += vdN[j]*0.1*CMLPStatistics::Rand_01();
			}
			// x3=(x2-x1)*[0;1] + x1 itd.
			//gprof vdX.push_back( (pIA2->m_AttrValue.u_dReal - pIA1->m_AttrValue.u_dReal)*factor + pIA1->m_AttrValue.u_dReal);
			vdX[j]=(pIA2->m_AttrValue.u_dReal - pIA1->m_AttrValue.u_dReal)*factor + pIA1->m_AttrValue.u_dReal;
		}else{
			// Cechy dla ktorych wartosci w dipolu sa rowne, nie sa brane pod uwage
			// przy tworzeniu hiperplaszczyzny.
			vbFS[j] = false;
			//gprof vdN.push_back(0.0);
			vdN[j]=0.0;
			//gprof vdX.push_back(0.0);
			vdX[j]=0.0;
		}
	}
	size_t removedAttrCount=0;
	for(size_t t=0; t<vbFS.size(); t++){
		if(!vbFS[t]){
			removedAttrCount++;
		}
	}
	
	// Kiedy dipol mieszany skladal sie z dwoch przykladach o takich samych
	// wartosciach wszystkich atrybutow za wyjatkiem klasy, to zwracamy od razu NULL.
	if(removedAttrCount == attrCount){
		return 0;
	}

	// Usuwanie atrybutow do momentu spelnienia warunku. Atrybuty usuwane sa w
	// sposob losowy. Mozna zastanowic sie nad tym, aby wyrzucac w pierwszej kolejnosci
	// te atrybuty, ktore nie wplywaja na separowalnosc przykladow tworzacych dipol.
	while( (attrCount - removedAttrCount) * uPuf > deqCI.size()){
		unsigned ind = rand()%attrCount;
		while(!vbFS[ind]){
			ind = rand()%attrCount;
		}
		removedAttrCount++;
		vbFS[ind] = false;
		vdN[ind] = 0;
		vdX[ind] = 0;
	}

	// Ax + By + Cz
	vector<double> vdW=vdN;
	// (-Ax3-By3-Cz3)
	double w0=0;
	for(unsigned j=0; j<attrCount; j++){
		w0 -= vdN[j]*vdX[j];
	}
	vdW.push_back(w0);
	// W=(Ax + By + Cz + (-Ax3-By3-Cz3))
	return (ITest*) new CObliqueTest(vdW);
}

ITest* CObliqueTestCreator::CreateObliqueTestWP(deque<CDataObject*> &deqCI, unsigned attrCount){
	assert(attrCount>0);
	if(deqCI.size()<2){
		return NULL;
	}

	// ob1=(x1,y1,z1)
	CDataObject* ob1;
	// ob2=(x2,y2,z2)
	CDataObject* ob2;

	if(!FindMixedDipol(deqCI,ob1,ob2)){
		return NULL;
	}

	vector<double> vdN(attrCount);
	vector<double> vdX(attrCount);
	for(unsigned j=0; j<attrCount; j++){
		IAttrValue* pIA1 = ob1->GetRealAttrValue(j);
		IAttrValue* pIA2 = ob2->GetRealAttrValue(j);
		assert(pIA1->GetAttrType()==e_RealAttr);
		assert(pIA2->GetAttrType()==e_RealAttr);
		// vdN=(A,B,C)=(x2-x1,y2-y1,z2-z1)
		//gprof vdN.push_back(pIA2->m_AttrValue.u_dReal - pIA1->m_AttrValue.u_dReal);
		vdN[j]=pIA2->m_AttrValue.u_dReal - pIA1->m_AttrValue.u_dReal;
		// x3=(x2-x1)/2 + x1 itd.
		//gprof vdX.push_back( (pIA2->m_AttrValue.u_dReal - pIA1->m_AttrValue.u_dReal) /2 + pIA1->m_AttrValue.u_dReal);
		vdX[j]=(pIA2->m_AttrValue.u_dReal - pIA1->m_AttrValue.u_dReal) /2 + pIA1->m_AttrValue.u_dReal;
	}
	// Ax + By + Cz
	vector<double> vdW=vdN;
	// (-Ax3-By3-Cz3)
	double w0=0;
	for(unsigned j=0; j<attrCount; j++){
		w0 -= vdN[j]*vdX[j];
	}
	vdW.push_back(w0);
	// W=(Ax + By + Cz + (-Ax3-By3-Cz3))
	return (ITest*) new CObliqueTest(vdW);
}


bool CObliqueTestCreator::FindMixedDipolCut(deque<CDataObject*> &deqCI, CDataObject* &ob1, CDataObject* &ob2, ITest* pITest){
	CObliqueTest* oblTest = dynamic_cast<CObliqueTest*>(pITest);
	assert(oblTest);

	bool found=false;
	if(deqCI.size()>0){
		IAttrValue *pIA1 = deqCI[0]->GetClass();
		for(unsigned i=1; i<deqCI.size(); i++){
			IAttrValue *pIA2 = deqCI[i]->GetClass();
			if( ! ((*pIA1) == (*pIA2)) ){
				found=true;
				break;
			}
		}
	}else{
		return false;
	}
	if(!found){
		return false;
	}
	vector<vector<CDataObject*> > splitCI(1);
	splitCI[0].push_back(deqCI[0]);
	for(size_t t=1; t<deqCI.size(); t++){
		bool found = false;
		for(size_t k=0; k<splitCI.size(); k++){
			if( *(splitCI[k].front()->GetClass()) == *(deqCI[t]->GetClass()) ){
				splitCI[k].push_back(deqCI[t]);
				found=true;
				break;
			}
		}
		if(!found){
			splitCI.resize(splitCI.size()+1);
			splitCI[splitCI.size()-1].push_back(deqCI[t]);
		}
	}
	if(!(splitCI.size()>(size_t)1)){
		return false;
	}
	size_t atrind1 = rand()%splitCI.size();
	size_t atrind2 = rand()%splitCI.size();
	while(atrind1 == atrind2){
		atrind2 = rand()%splitCI.size();
	}
	// Pierwszy losujemy
	size_t ind1 = rand()%splitCI[atrind1].size();
	double ip1 = oblTest->GetInnerProduct(splitCI[atrind1][ind1]);
	// wyrzucamy z splitCI te, ktore leza po tej samej stronie hip. zeby
	// z nich nie losowac.
	vector<CDataObject*>::iterator iter = splitCI[atrind2].begin();
	while(iter<splitCI[atrind2].end()){
		if(ip1*oblTest->GetInnerProduct(*iter)>0){
			splitCI[atrind2].erase(iter);
		}
		iter++;
	}
	if(splitCI[atrind2].empty()){
		return false;
	}

	// I tu juz losowanie z innej klasy, ale lezacego po drugiej stronie hip.
	size_t ind2 = rand()%splitCI[atrind2].size();

	ob1 = splitCI[atrind1][ind1];
	ob2 = splitCI[atrind2][ind2];
	return true;
}

bool CObliqueTestCreator::FindPureDipol(deque<CDataObject*> &deqCI, CDataObject* &ob1, CDataObject* &ob2, ITest* pITest){
	CObliqueTest* oblTest = dynamic_cast<CObliqueTest*>(pITest);
	assert(oblTest);

	bool found=false;
	if(deqCI.size()>0){
		IAttrValue *pIA1 = deqCI[0]->GetClass();
		for(unsigned i=1; i<deqCI.size(); i++){
			IAttrValue *pIA2 = deqCI[i]->GetClass();
			if( ! ((*pIA1) == (*pIA2)) ){
				found=true;
				break;
			}
		}
	}else{
		return false;
	}
	if(!found){
		return false;
	}
	vector<vector<CDataObject*> > splitCI(1);
	splitCI[0].push_back(deqCI[0]);
	for(size_t t=1; t<deqCI.size(); t++){
		bool found = false;
		for(size_t k=0; k<splitCI.size(); k++){
			if( *(splitCI[k].front()->GetClass()) == *(deqCI[t]->GetClass()) ){
				splitCI[k].push_back(deqCI[t]);
				found=true;
				break;
			}
		}
		if(!found){
			splitCI.resize(splitCI.size()+1);
			splitCI[splitCI.size()-1].push_back(deqCI[t]);
		}
	}
	if(!(splitCI.size()>(size_t)1)){
		return false;
	}
	// Losowanie indeksu klasy.
	size_t atrind = rand()%splitCI.size();
	if(splitCI[atrind].size()<2){
		return false;
	}
	// Pierwszy losujemy
	size_t ind1 = rand()%splitCI[atrind].size();
	double ip1 = oblTest->GetInnerProduct(splitCI[atrind][ind1]);
	// wyrzucamy z splitCI te, ktore leza po tej samej stronie hip. zeby
	// z nich nie losowac.
	vector<CDataObject*>::iterator iter = splitCI[atrind].begin();
	while(iter<splitCI[atrind].end()){
		if(ip1*oblTest->GetInnerProduct(*iter)>=0){
			splitCI[atrind].erase(iter);
		}
		iter++;
	}
	if(splitCI[atrind].empty()){
		return false;
	}
	// I tu juz losowanie z tej samej klasy lezacego po tej samej stronie hip.
	size_t ind2 = rand()%splitCI[atrind].size();

	ob1 = splitCI[atrind][ind1];
	ob2 = splitCI[atrind][ind2];
	return true;
}

bool CObliqueTestCreator::JoinPureDipol(CDataObject* ob1, CDataObject* ob2, ITest* pITest){
	CObliqueTest* oblTest = dynamic_cast<CObliqueTest*>(pITest);
	assert(oblTest);

	// Obliczenie wsp. punktu przez, ktory trzeba poprowadzic hiperplane.
	int r=rand()%2;
	unsigned size = (unsigned)(oblTest->m_vdWReal.size());
	vector<double> vdX(size);
	for(unsigned t=0; t<size-1; t++){
		IAttrValue* pIA1 = ob1->GetRealAttrValue(t);
		IAttrValue* pIA2 = ob2->GetRealAttrValue(t);
		assert(pIA1);
		assert(pIA2);
		CRealAttr* ref_ra1 = dynamic_cast<CRealAttr*>(pIA1);
		CRealAttr* ref_ra2 = dynamic_cast<CRealAttr*>(pIA2);
		assert(ref_ra1!=NULL);
		assert(ref_ra2!=NULL);
		if(r==0){
			//gprof vdX.push_back(ref_ra1->m_AttrValue.u_dReal - (ref_ra2->m_AttrValue.u_dReal-ref_ra1->m_AttrValue.u_dReal)/2.0);
			vdX[t]=ref_ra1->m_AttrValue.u_dReal - (ref_ra2->m_AttrValue.u_dReal-ref_ra1->m_AttrValue.u_dReal)/2.0;
		}else{
			//gprof vdX.push_back((ref_ra2->m_AttrValue.u_dReal-ref_ra1->m_AttrValue.u_dReal)/2.0+ref_ra2->m_AttrValue.u_dReal);
			vdX[t]=(ref_ra2->m_AttrValue.u_dReal-ref_ra1->m_AttrValue.u_dReal)/2.0+ref_ra2->m_AttrValue.u_dReal;
		}
	}
	//gprof vdX.push_back(1.0);
	vdX[size-1]=1.0;
	// Indeks wspolrzednej.
	size_t ind = rand()%oblTest->m_vdWReal.size();
	int i=0;
	while(vdX[ind] == 0 && i++ < 5){
		ind = rand()%oblTest->m_vdWReal.size();
	}
	if(vdX[ind]==0){
		return false;
	}
	double sum = 0;
	for(unsigned t=0; t<oblTest->m_vdWReal.size(); t++){
		if(t!=ind){
			sum -= vdX[t]*oblTest->m_vdWReal[t];
		}
	}
	oblTest->m_vdWReal[ind] = sum/vdX[ind];
	return true;
}

bool CObliqueTestCreator::ShiftHiperplane(ITest* pITest, deque<CDataObject*> &deqCI){
	CObliqueTest* oblTest = dynamic_cast<CObliqueTest*>(pITest);
	if(oblTest && deqCI.size()>2){
		bool lfound=false;
		bool rfound=false;
		size_t lminID=0;
		size_t rminID=0;
		double lminval=0;
		double rminval=0;
		double mi = oblTest->GetMI();
		for(size_t t=0; t<deqCI.size(); t++){//loop
			if(oblTest->GetInnerProduct(deqCI[t]) > 0){
				if(!rfound){
					rfound=true;
					rminID=t;
					rminval=fabs(oblTest->GetDistance(deqCI[t],mi));
				}else{
					double aktval = fabs(oblTest->GetDistance(deqCI[t],mi));
					if(aktval<rminval){
						rminID=t;
						rminval=aktval;
					}
				}
			}else{
				if(!lfound){
					lfound=true;
					lminID=t;
					lminval=fabs(oblTest->GetDistance(deqCI[t],mi));
				}else{
					double aktval = fabs(oblTest->GetDistance(deqCI[t],mi));
					if(aktval<lminval){
						lminID=t;
						lminval=aktval;
					}
				}
			}
		}//loop

		// Tu trzeba by sprawdzac czy po tej samej stronie co rfound nie lezy wiecej przykladow
		// z tej klasy, do ktorej nalezy lfound. I odwrotnie czy po stronie lfound nie 
		// znajduje sie wiecej przykladow z rfound

		if(rfound && lfound){//found
			/*
				double temp1 = oblTest->GetInnerProduct(deqCI[rminID]);
				double temp2 = oblTest->GetInnerProduct(deqCI[lminID]);
			*/

			// Wyznaczenie punktu lezacego miedzy rmin i lmin.
			vector<double> vdX(oblTest->m_vdWReal.size());
			CDataObject* lobj = deqCI[lminID];
			CDataObject* robj = deqCI[rminID];

			// w tej chwili dosuwanie do srodka ma sens tylko wtedy kiedy dipol jest mieszany
			// TOTH: mozna tu zapamietywac kolejne nie najblizsze elementy i robic jakies bardziej 
			// inteligentne przesuwanie kiedy dipol jest czysty.
			if( (*lobj->GetClass())!=(*robj->GetClass()) ){ //mixed dipol
				for(unsigned t=0; t<oblTest->m_vdWReal.size()-1; t++){
					IAttrValue* pIA1 = lobj->GetRealAttrValue(t);
					IAttrValue* pIA2 = robj->GetRealAttrValue(t);
					assert(pIA1);
					assert(pIA2);
					CRealAttr* ref_ra1 = dynamic_cast<CRealAttr*>(pIA1);
					CRealAttr* ref_ra2 = dynamic_cast<CRealAttr*>(pIA2);
					assert(ref_ra1!=NULL);
					assert(ref_ra2!=NULL);
					//gprof vdX.push_back((ref_ra2->m_AttrValue.u_dReal-ref_ra1->m_AttrValue.u_dReal)/2.0+ref_ra1->m_AttrValue.u_dReal);
					vdX[t]=(ref_ra2->m_AttrValue.u_dReal-ref_ra1->m_AttrValue.u_dReal)/2.0+ref_ra1->m_AttrValue.u_dReal;
				}
				//gprof vdX.push_back(1.0);
				vdX[oblTest->m_vdWReal.size()-1]=1.0;
				// Przesuniecie rownolegle hiperlaszczyzny.
				double w0=0;
				for(unsigned t=0; t<oblTest->m_vdWReal.size()-1; t++){
					w0 -= oblTest->m_vdWReal[t]*vdX[t];
				}
				oblTest->m_vdWReal[oblTest->m_vdWReal.size()-1] = w0;
				assert(fabs(0.0 - inner_product(vdX.begin(),vdX.end(),oblTest->m_vdW.begin(),0.0)) < 1e-3);
				assert(oblTest->GetInnerProduct(deqCI[rminID]) * oblTest->GetInnerProduct(deqCI[lminID]) <= 1e-3);
	
				/*
					double temp = inner_product(vdX.begin(),vdX.end(),oblTest->m_vdW.begin(),0.0);
					double temp3 = oblTest->GetInnerProduct(deqCI[rminID]);
					double temp4 = oblTest->GetInnerProduct(deqCI[lminID]);
				*/
	
				return true;
			}// dipol mixed

			return false;
		}//found
		else{
			return false;
		}
	}else{
		return false;
	}
}

void CObliqueTestCreator::SetRandomZero(ITest* pITest){
	CObliqueTest* oblTest = dynamic_cast<CObliqueTest*>(pITest);
	assert(oblTest);
	if(oblTest){
		oblTest->m_vdWReal[rand()%oblTest->m_vdWReal.size()] = 0.0;
	}
}

bool CObliqueTestCreator::MutateTest(ITest* pITest){
        CObliqueTest* oblTest = dynamic_cast<CObliqueTest*>(pITest);
        if(oblTest){
                size_t mPoint = rand()%oblTest->m_vdWReal.size();
                if(oblTest->m_vdWReal[mPoint]==0){
                        if(rand()%2==0){
                                oblTest->m_vdWReal[mPoint] = CMLPStatistics::Rand_01();
                        }else{
                                oblTest->m_vdWReal[mPoint] = -CMLPStatistics::Rand_01();
                        }
                }else{
                        if(fabs(oblTest->m_vdWReal[mPoint])<=1){
                                if(rand()%2==0){
                                        oblTest->m_vdWReal[mPoint] += CMLPStatistics::Rand_01();
                                }else{
                                        oblTest->m_vdWReal[mPoint] -= CMLPStatistics::Rand_01();
                                }
                        }else{
                                double diff = oblTest->m_vdWReal[mPoint] * CMLPStatistics::Rand_01();
                                if(rand()%2==0){
                                        oblTest->m_vdWReal[mPoint] += diff;
                                }else{
                                        oblTest->m_vdWReal[mPoint] -= diff;
                                }
                        }
                }
		return true;
        }
	return false;
}

bool CObliqueTestCreator::CrossoverHiperplanes(ITest* pITest1, ITest* pITest2){
        CObliqueTest* oblTest1 = dynamic_cast<CObliqueTest*>(pITest1);
        CObliqueTest* oblTest2 = dynamic_cast<CObliqueTest*>(pITest2);
        double temp;
        if(oblTest1 && oblTest2){
                assert(oblTest1->m_vdWReal.size()==oblTest2->m_vdWReal.size());
                size_t cutPoint = rand()%oblTest1->m_vdWReal.size();
                for(size_t  t=cutPoint; t<oblTest1->m_vdWReal.size()-1; t++){
                        temp=oblTest1->m_vdWReal[t];
                        oblTest1->m_vdWReal[t]=oblTest2->m_vdWReal[t];
                        oblTest2->m_vdWReal[t]=temp;
                }
                return true;
        }else{
                return false;
        }
        return false;
}

bool CObliqueTestCreator::TwoPointCrossoverHiperplanes(ITest* pITest1, ITest* pITest2){
        CObliqueTest* oblTest1 = dynamic_cast<CObliqueTest*>(pITest1);
        CObliqueTest* oblTest2 = dynamic_cast<CObliqueTest*>(pITest2);
        double temp;
        if(oblTest1 && oblTest2){
            assert(oblTest1->m_vdW.size()==oblTest2->m_vdW.size());
            size_t cutPoint1 = rand()%oblTest1->m_vdWReal.size();
            size_t cutPoint2 = rand()%oblTest1->m_vdWReal.size();
			if(cutPoint1 == cutPoint2) return false;
			if(cutPoint1 > cutPoint2){
				size_t temp = cutPoint1;
				cutPoint1 = cutPoint2;
				cutPoint2 = temp;
			}
            for(size_t  t=cutPoint1; t<cutPoint2; t++){
            	temp=oblTest1->m_vdWReal[t];
            	oblTest1->m_vdWReal[t]=oblTest2->m_vdWReal[t];
            	oblTest2->m_vdWReal[t]=temp;
            }
            return true;
        }else{
                return false;
        }
	return false;
}

bool CObliqueTestCreator::CutMixedDipol(CDataObject* ob1, CDataObject* ob2, ITest* pITest){
	CObliqueTest* oblTest = dynamic_cast<CObliqueTest*>(pITest);
	assert(oblTest);

	// Obliczenie wsp. punktu przez, ktory trzeba poprowadzic hiperplane.
	unsigned size = (unsigned)(oblTest->m_vdWReal.size());
	vector<double> vdX(size);
	for(unsigned t=0; t<size-1; t++){
		IAttrValue* pIA1 = ob1->GetRealAttrValue(t);
		IAttrValue* pIA2 = ob2->GetRealAttrValue(t);
		assert(pIA1);
		assert(pIA2);
		CRealAttr* ref_ra1 = dynamic_cast<CRealAttr*>(pIA1);
		CRealAttr* ref_ra2 = dynamic_cast<CRealAttr*>(pIA2);
		assert(ref_ra1!=NULL);
		assert(ref_ra2!=NULL);
		//vdX.push_back((ref_ra2->m_AttrValue.u_dReal-ref_ra1->m_AttrValue.u_dReal)/2.0+ref_ra1->m_AttrValue.u_dReal);
		vdX[t]=(ref_ra2->m_AttrValue.u_dReal-ref_ra1->m_AttrValue.u_dReal)/2.0+ref_ra1->m_AttrValue.u_dReal;
	}
	//vdX.push_back(1.0);
	vdX[size-1]=1.0;
	// Indeks wspolrzednej.
	size_t ind = rand()%oblTest->m_vdWReal.size();
	int i=0;
	while(vdX[ind] == 0 && i++ < 5){
		ind = rand()%oblTest->m_vdWReal.size();
	}
	if(vdX[ind]==0){
		return false;
	}
	double sum = 0;
	for(unsigned t=0; t<oblTest->m_vdWReal.size(); t++){
		if(t!=ind){
			sum -= vdX[t]*oblTest->m_vdWReal[t];
		}
	}
	oblTest->m_vdWReal[ind] = sum/vdX[ind];
	return true;
}

