#ifndef __COBLIQUETESTCREATOR__
#define __COBLIQUETESTCREATOR__ 

#include "../../mlpcore/MLPDT/ITest.h"
#include "../../mlpcore/MLPDT/TestCreator.h"

/**
 * @author Marek Grzes'
 */ 
class CObliqueTestCreator : public CTestCreator {
public:
	/** Na podstawie podanego zbioru przykadow tworzy test skosny. Kiedy przekazany
	 * jako parametr zbior przykladow jest z jednej klasy to zwracany jest NULL.
	 *
	 * @param deqCI - kolejka przykladow
	 * @param attrCount - liczba atrybutow branych pod uwage
	 */
	virtual ITest* CreateObliqueTestWP(deque<CDataObject*> &deqCI, unsigned attrCount);
	virtual ITest* CreateObliqueTest(deque<CDataObject*> &deqCI, unsigned attrCount, unsigned uPuf);
	virtual ITest* CreateRandObliqueTest(deque<CDataObject*> &deqCI, unsigned attrCount, unsigned uPuf);

	/** Znajduje czysty dipol.
	 * @return false kiedy dane sa tylko z jednej klasy.
	 */
	virtual bool FindPureDipol(deque<CDataObject*> &deqCI, CDataObject* &ob1, CDataObject* &ob2);
	
	/** Znajduje czysty dipol, ktory jest przecinany przez hiperplaszczyzne
	 * testu.
	 * @return false kiedy dane sa tylko z jednej klasy.
	 */
	virtual bool FindPureDipol(deque<CDataObject*> &deqCI, CDataObject* &ob1, CDataObject* &ob2, ITest* pITest);

	/**
	 * Znajduje dipol mieszany, ktory jest przecinany przez hiperplaszczyzne testu.
	 * @return false kiedy dane sa tylko z jednej klasy luz zaden dipol mieszany nie jest
	 * przeciety.
	 */
	virtual bool FindMixedDipolCut(deque<CDataObject*> &deqCI, CDataObject* &ob1, CDataObject* &ob2, ITest* pITest);
	
	/** Przecina mieszany dipol.
	 */
	virtual bool CutMixedDipol(CDataObject* ob1, CDataObject* ob2, ITest* pITest);

	/** Laczy przeciety czysty dipol.
	 */
	virtual bool JoinPureDipol(CDataObject* ob1, CDataObject* ob2, ITest* pITest);

	/** Przesuwa hiperplaszyszyzne tak, aby przeciecie bylo w polowie odleglosci.
	 *
	 * @return false - kiedy nie przesuwano hiperplszczyzny
	 * @retrun true - kiedy hiperplaszczyzna zostala przesunieta
	 */
	virtual bool ShiftHiperplane(ITest* pITest, deque<CDataObject*> &deqCI);
	
	virtual void SetRandomZero(ITest* pITest);
	
	/**
	 * Mutacja, w ktorej jeden losowo wybrany wspolnczynik jest mutowany. Metoda
	 * do wykorzystania przy globalnej indukcji drzewa.
	 */
	virtual bool MutateTest(ITest* pITest);
	
	/**
	 * Krzyzowanie jednopunktowe dwoch hiperplaszczyzn.
	 */
	virtual bool CrossoverHiperplanes(ITest* pITest1, ITest* pITest2);
	
	/**
	 * Krzyzowanie dwupunktowe dwoch hiperplaszczyzn.
	 */
	virtual bool TwoPointCrossoverHiperplanes(ITest* pITest1, ITest* pITest2);
	
	CObliqueTestCreator(void);
	virtual ~CObliqueTestCreator(void);

protected:
};

#endif

