#include "Chromosome.h"
#include <cmath>
#include <cstdlib>

CChromosome::CChromosome(CChromosomeDesc *pChrDesc){
	m_pChrDesc=pChrDesc;
	m_ldFitness=0;
	Randomize();
}


void CChromosome::Randomize(){
	for(int i=0;i<m_pChrDesc->m_iChromosomeSize;i++){
		if (((double)rand()/(double)RAND_MAX)>0.5){
			m_vString.push_back(false);
		}
		else{
			m_vString.push_back(true);
		}
	}
}


void CChromosome::ToString(ostream *ostr){
	for(int i=0;i<(int)m_vString.size();i++){
		if (m_vString.at(i)){
			*ostr<<"1";
		}
		else{
			*ostr<<"0";
		}
	}
}

CChromosome::~CChromosome(void){
	m_vString.clear();
}


CChromosome::CChromosome(CChromosome *c){
	m_pChrDesc=c->m_pChrDesc;
	m_ldFitness=c->m_ldFitness;

	for (int i=0;i<m_pChrDesc->m_iChromosomeSize;i++){
		if (c->m_vString[i]){
			m_vString.push_back(true);
		}
		else{
			m_vString.push_back(false);
		}
	}
}

void CChromosome::Mutate(double mutrate){
	for (int i=0;i<(int)m_vString.size();i++){
		if ((double)rand()/(double)RAND_MAX<=mutrate){
		m_vString[i]=!m_vString[i];
		}
	}
}


vector<long double> CChromosome::GetFonotype(){
	vector<long double> result;
	int k=0; //wykladnik potgi 2
	int value=0;
	int maxvalue=(int)pow((long double)2,(long double)m_pChrDesc->m_iNBits);
	for (int i=0;i<m_pChrDesc->m_iChromosomeSize;i++){
		if (m_vString[i]){
			value+=(int)pow((long double)2,(long double)k);
		}
		k++;
		if (((i+1)%m_pChrDesc->m_iNBits)==0){
			value=GreyToBinary(value);
			result.push_back((long double)value/maxvalue);
			k=0;
			value=0;
		}
	}
	return result;
}

int CChromosome::BinaryToGrey(int value){
	return (value ^ (value>>1));
}

int CChromosome::GreyToBinary(int value){
	value ^= value>>1; 
	value ^= value>>2; 
	value ^= value>>4; 
	value ^= value>>8; 
	value ^= value>>16; 
	return value;
}

void CChromosome::Change(CChromosome* pChr, int pos){
	bool tmp=m_vString[pos];
	
	if (pChr->m_vString[pos]){
		m_vString[pos]=true;
	}
	else{
		m_vString[pos]=false;
	}
	
	if (tmp){
		pChr->m_vString[pos]=true;
	}
	else{
		pChr->m_vString[pos]=false;
	}
}

void CChromosome::Crossover(CChromosome* pChr, int end){
	for (int i=0;i<=end;i++){
		Change(pChr,i);
	}
}

void CChromosome::Crossover(CChromosome* pChr, int start, int end){
	for (int i=start;i<=end;i++){
		Change(pChr,i);
	}
}
