#include "ICET.h"

CICET::CICET():CGeneticAlgorithm(){
}

IClassifier* CICET::ActualTrain(const IDataSet* pDS){
	m_pDataset=pDS;
	IClassifier* pIC=NULL;
	IInducer *itInducer = new CTDIDTCSInducer();
	vector<double> vOldCost;
	
	//stoworzenie inducera i pobranie satrych kosztw 
	itInducer->SetParam("selection of a test","eg2");
	for (unsigned int i=0;i<m_pDataset->GetCurrentAttrCount();i++){
		vOldCost.push_back(m_pDataset->GetAttrDesc(i)->m_fCost);
	}
	//ustawienie paramerow algorytmu
	SetParams();
	//stworzenie populacji
	CreatePopulation();
	//odpalenie algorytmu genetycznego
	Run();
	//pobranie najlepszego 
	vector<long double> vFenotype=m_pPopulation->GetFittest()->GetFonotype();
	
	//ustawienie omega i kosztow w datasecie
	CICF::omega=(double)vFenotype.back();
	for (unsigned int j=0;j<vFenotype.size()-1;j++){
			m_pDataset->GetAttrDesc(j)->m_fCost=(double)vFenotype[j];
	}
	
	//stworzenie klasyfikatora
	pIC=itInducer->Train(m_pDataset);
	
	//podstawienie starych kosztw
	for (unsigned int i=0;i<m_pDataset->GetCurrentAttrCount();i++){
		m_pDataset->GetAttrDesc(i)->m_fCost=vOldCost[i];
	}
	delete itInducer;
	//vOldCost.clear();
	return pIC;
	
}

void CICET::SetParams(){
	m_pChrDesc->m_iNBits=10;
	m_pChrDesc->m_iNFeatures=(int)m_pDataset->GetCurrentAttrCount()+1;
	m_pChrDesc->m_iChromosomeSize=m_pChrDesc->m_iNBits*m_pChrDesc->m_iNFeatures;
	CICF::omega=1;
	m_dMaxCost=1000;
	m_iLoop=1000;
	m_iPopulationSize=2;
}

void CICET::SetFitness(CPopulation *pPop){
	vector<long double> vFenotype;
	IClassifier* pIC=NULL;
	deque<const IDataSet*> deqIDS;
	//wyliczeniemaksymalnego kosztu
	IInducer *itInducer = new CTDIDTCSInducer();

	for (unsigned int i=0;i<pPop->m_deqPopulation.size();i++){
		//pobranie wektora cech z algorytmu genetycznego
		vFenotype=pPop->m_deqPopulation[i]->GetFonotype();
		//ustawienie parametru omega
		CICF::omega=(double)vFenotype.back();
		for (unsigned int j=0;j<vFenotype.size()-1;j++){
			//wstawienie go do datasetu
			m_pDataset->GetAttrDesc(j)->m_fCost=(double)(vFenotype[j]);
		}
		//podzia zbioru
		deqIDS=m_pDataset->SplitDataSet(2);
		//stworzenie drzewa z pierwszej czesci
		CDTClassifier* dt=NULL;
		pIC=itInducer->Train(deqIDS[0]);	
		
		//wstawienie - kosztu klasyfikacji jako dostosowania z minusem 
		m_pPopulation->m_deqPopulation[i]->m_ldFitness=-GetCostOfClassification(pIC,deqIDS[1]);
		//usuniecie drzewa
		
		delete deqIDS[0];
		delete deqIDS[1];
		deqIDS.clear();
		dt=(CDTClassifier*)pIC;
		delete dt;
	}
	delete itInducer;
}

CICET::~CICET(){
}

double CICET::GetCostOfClassification(IClassifier *pIC,const IDataSet *pIDS){
	if (pIDS->GetSize()>0){
		double result=0;
		
		//pobranie kosztow klasyfikacji
		for (unsigned int i=0;i<pIDS->GetSize();i++){
			CDataObject *inst = (*pIDS)[i];
			//klasa otrzymana
			pair<double,double> pCost;
			pIC->Classify(inst,pCost);
			result+=pCost.first;
			result+=pCost.second;
		}
		return result/pIDS->GetSize();
	}
	else return 0;
}
