#include "Population.h"
#include <cstdlib>

CPopulation::CPopulation(CChromosomeDesc* pChrDesc,CPopDesc* pPopDesc,int popSize){
	m_pChrDesc=pChrDesc;
	m_pPopDesc=pPopDesc;

	for(int i=0;i<popSize;i++){
		m_deqPopulation.push_back(new CChromosome(m_pChrDesc));
	}
}


CPopulation::~CPopulation(){
	for (unsigned i=0;i<m_deqPopulation.size();i++){
		delete m_deqPopulation[i];
	}
	m_deqPopulation.clear();
}

CChromosome* CPopulation::GetFittest(){
	int maxindex=0;
	long double maxfitness=m_deqPopulation[0]->m_ldFitness;
	for(unsigned int i=1;i<m_deqPopulation.size();i++){
		if (m_deqPopulation[i]->m_ldFitness>maxfitness){
			maxindex=i;
			maxfitness=m_deqPopulation[i]->m_ldFitness;
		}
	}
	return m_deqPopulation[maxindex];
}

void CPopulation::Mutate(){
	for (int i=0;i<(int)m_deqPopulation.size();i++){
		m_deqPopulation[i]->Mutate(m_pPopDesc->m_dMutationRate);
	}
}

void CPopulation::Replace(CChromosome* pChr){
	delete m_deqPopulation.front();
	m_deqPopulation.pop_front();
	m_deqPopulation.push_back(pChr);
	
}

vector<CChromosome*> CPopulation::GetParents(void)
{
	vector<CChromosome*> vResult;
	int nChromosome=(int)(m_deqPopulation.size()*((double)rand()/RAND_MAX))-1;
	CChromosome *parent1=NULL, *parent2=NULL;
	int ind1=0;
	int ind2=0;

	while(ind1==m_deqPopulation.size()){
		ind1=(int)(m_deqPopulation.size()*((double)rand()/RAND_MAX));
	}
	while(ind2==m_deqPopulation.size()){
		ind2=(int)(m_deqPopulation.size()*((double)rand()/RAND_MAX));
	}
	//bez powtorzen
	while(ind1==ind2 || ind2==m_deqPopulation.size()){
		ind2=(int)(m_deqPopulation.size()*((double)rand()/RAND_MAX));
	}
	
	if (nChromosome>1){
		//indeksy wylosowanych 
		vector<int> vIndexes;
		long double fitness1,fitness2;
		//czy losowa kolejny
		bool next=true;

		vIndexes.push_back(ind1);
		vIndexes.push_back(ind2);
		fitness1=m_deqPopulation[ind1]->m_ldFitness;
		fitness2=m_deqPopulation[ind2]->m_ldFitness;
		//fitnes1>fitness2
		if(fitness1<fitness2){
			long double ldtmp;
			int itmp;
			ldtmp=fitness1;
			fitness1=fitness2;
			fitness2=ldtmp;
			
			itmp=ind1;
			ind1=ind2;
			ind2=itmp;
		}

		for (int i=2;i<nChromosome;i++){
			int ind;
			while(next){
				next = false;
				ind=(int)(m_deqPopulation.size()*((double)rand()/RAND_MAX));
				for (unsigned int j=0;j<vIndexes.size();j++){
					if (vIndexes[j]==ind || ind==m_deqPopulation.size()){
						next=true;
						break;
					}
				}
			}
			vIndexes.push_back(ind);
			if(fitness1<m_deqPopulation[ind]->m_ldFitness){
				fitness1=m_deqPopulation[ind]->m_ldFitness;
				ind1=ind;
			}
			else if(fitness2<m_deqPopulation[ind]->m_ldFitness){
				fitness2=m_deqPopulation[ind]->m_ldFitness;
				ind2=ind;
			}
			next=true;
		}
		vIndexes.clear();
	}
	vResult.push_back(m_deqPopulation[ind1]);
	vResult.push_back(m_deqPopulation[ind2]);
	return vResult;
}

vector<CChromosome*> CPopulation::Crossover(vector<CChromosome*> vParents)
{
	vector<CChromosome*> vOffspring;
	vOffspring.push_back(new CChromosome(vParents[0]));
	vOffspring.push_back(new CChromosome(vParents[1]));
	if (m_pPopDesc->m_sCrossoverType=="one point"){
		int end = (int)(m_pChrDesc->m_iChromosomeSize*((double)rand()/(double)RAND_MAX));
		vOffspring[0]->Crossover(vOffspring[1],end);
	}
	else{
		int end = (int)(m_pChrDesc->m_iChromosomeSize*((double)rand()/(double)RAND_MAX));
		int start = (int)(m_pChrDesc->m_iChromosomeSize*((double)rand()/(double)RAND_MAX));
		if (end>start){
			vOffspring[0]->Crossover(vOffspring[1],start,end);
		}
		else{
			vOffspring[0]->Crossover(vOffspring[1],end,start);
		}
	}
	return vOffspring;
}
